/*
Address.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* Object that represents a SIP address. 
*
* The {@link Address} is an opaque object to represents SIP addresses, ie the
* content of SIP's 'from' and 'to' headers. A SIP address is made of display
* name, username, domain name, port, and various uri headers (such as tags). It
* looks like 'Alice <sip:alice@example.net>'. The {@link Address} has methods to
* extract and manipulate all parts of the address. When some part of the address
* (for example the username) is empty, the accessor methods return null. 
*/
public interface Address {
    /**
    * Returns the display name. 
    *
    */
    public String getDisplayName();

    /**
    * Sets the display name. 
    *
    */
    public int setDisplayName(String displayName);

    /**
    * Returns the domain name. 
    *
    */
    public String getDomain();

    /**
    * Sets the domain. 
    *
    */
    public int setDomain(String domain);

    /**
    * returns true if address is a routable sip address 
    *
    */
    public boolean isSip();

    /**
    * Get the value of the method parameter. 
    *
    */
    public String getMethodParam();

    /**
    * Set the value of the method parameter. 
    *
    */
    public void setMethodParam(String methodParam);

    /**
    * Get the password encoded in the address. 
    *
    * It is used for basic authentication (not recommended). 
    * 
    * @return the password, if any, null otherwise. 
    */
    public String getPassword();

    /**
    * Set the password encoded in the address. 
    *
    * It is used for basic authentication (not recommended). 
    * 
    * @param password the password to set. 
    */
    public void setPassword(String password);

    /**
    * Get port number as an integer value, 0 if not present. 
    *
    */
    public int getPort();

    /**
    * Sets the port number. 
    *
    */
    public int setPort(int port);

    /**
    * Returns the address scheme, normally "sip". 
    *
    */
    public String getScheme();

    /**
    * Returns true if address refers to a secure location (sips) 
    *
    */
    public boolean getSecure();

    /**
    * Make the address refer to a secure location (sips scheme) 
    *
    * @param enabled  if address is requested to be secure. 
    */
    public void setSecure(boolean enabled);

    /**
    * Get the transport. 
    *
    */
    public TransportType getTransport();

    /**
    * Set a transport. 
    *
    */
    public int setTransport(TransportType transport);

    /**
    * Returns the username. 
    *
    */
    public String getUsername();

    /**
    * Sets the username. 
    *
    */
    public int setUsername(String username);

    /**
    * Returns the address as a string. 
    *
    * The returned char * must be freed by the application. Use ms_free(). 
    */
    public String asString();

    /**
    * Returns the SIP uri only as a string, that is display name is removed. 
    *
    * The returned char * must be freed by the application. Use ms_free(). 
    */
    public String asStringUriOnly();

    /**
    * Removes address's tags and uri headers so that it is displayable to the user. 
    *
    */
    public void clean();

    /**
    * Clones a {@link Address} object. 
    *
    */
    public Address clone();

    /**
    * Compare two {@link Address} taking the tags and headers into account. 
    *
    * @param address2 {@link Address} object 
    * 
    * @return Boolean value telling if the {@link Address} objects are equal. 
    * 
    * See also: {@link Address#weakEqual} 
    */
    public boolean equal(Address address2);

    /**
    * Get the header encoded in the address. 
    *
    * @param headerName the header name 
    */
    public String getHeader(String headerName);

    /**
    * Get the value of a parameter of the address. 
    *
    * @param paramName The name of the parameter 
    * 
    * @return The value of the parameter 
    */
    public String getParam(String paramName);

    /**
    * Get the value of a parameter of the URI of the address. 
    *
    * @param uriParamName The name of the parameter 
    * 
    * @return The value of the parameter 
    */
    public String getUriParam(String uriParamName);

    /**
    * Tell whether a parameter is present in the address. 
    *
    * @param paramName The name of the parameter 
    * 
    * @return A boolean value telling whether the parameter is present in the address 
    */
    public boolean hasParam(String paramName);

    /**
    * Tell whether a parameter is present in the URI of the address. 
    *
    * @param uriParamName The name of the parameter 
    * 
    * @return A boolean value telling whether the parameter is present in the URI of
    * the address 
    */
    public boolean hasUriParam(String uriParamName);

    /**
    * Removes the value of a parameter of the URI of the address. 
    *
    * @param uriParamName The name of the parameter 
    */
    public void removeUriParam(String uriParamName);

    /**
    * Set a header into the address. 
    *
    * Headers appear in the URI with '?', such as
    * <sip:test@linphone.org?SomeHeader=SomeValue>. 
    * 
    * @param headerValue the header value 
    */
    public void setHeader(String headerName, String headerValue);

    /**
    * Set the value of a parameter of the address. 
    *
    * @param paramValue The new value of the parameter 
    */
    public void setParam(String paramName, String paramValue);

    /**
    * Set the value of a parameter of the URI of the address. 
    *
    * @param uriParamValue The new value of the parameter 
    */
    public void setUriParam(String uriParamName, String uriParamValue);

    /**
    * Compare two {@link Address} ignoring tags and headers, basically just domain,
    * username, and port. 
    *
    * @param address2 {@link Address} object 
    * 
    * @return Boolean value telling if the {@link Address} objects are equal. 
    * 
    * See also: {@link Address#equal} 
    */
    public boolean weakEqual(Address address2);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class AddressImpl implements Address {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected AddressImpl(long ptr) {
        nativePtr = ptr;
    }


    private native String getDisplayName(long nativePtr);
    @Override
    synchronized public String getDisplayName()  {
        
        return getDisplayName(nativePtr);
    }

    private native int setDisplayName(long nativePtr, String displayName);
    @Override
    synchronized public int setDisplayName(String displayName)  {
        
        return setDisplayName(nativePtr, displayName);
    }

    private native String getDomain(long nativePtr);
    @Override
    synchronized public String getDomain()  {
        
        return getDomain(nativePtr);
    }

    private native int setDomain(long nativePtr, String domain);
    @Override
    synchronized public int setDomain(String domain)  {
        
        return setDomain(nativePtr, domain);
    }

    private native boolean isSip(long nativePtr);
    @Override
    synchronized public boolean isSip()  {
        
        return isSip(nativePtr);
    }

    private native String getMethodParam(long nativePtr);
    @Override
    synchronized public String getMethodParam()  {
        
        return getMethodParam(nativePtr);
    }

    private native void setMethodParam(long nativePtr, String methodParam);
    @Override
    synchronized public void setMethodParam(String methodParam)  {
        
        setMethodParam(nativePtr, methodParam);
    }

    private native String getPassword(long nativePtr);
    @Override
    synchronized public String getPassword()  {
        
        return getPassword(nativePtr);
    }

    private native void setPassword(long nativePtr, String password);
    @Override
    synchronized public void setPassword(String password)  {
        
        setPassword(nativePtr, password);
    }

    private native int getPort(long nativePtr);
    @Override
    synchronized public int getPort()  {
        
        return getPort(nativePtr);
    }

    private native int setPort(long nativePtr, int port);
    @Override
    synchronized public int setPort(int port)  {
        
        return setPort(nativePtr, port);
    }

    private native String getScheme(long nativePtr);
    @Override
    synchronized public String getScheme()  {
        
        return getScheme(nativePtr);
    }

    private native boolean getSecure(long nativePtr);
    @Override
    synchronized public boolean getSecure()  {
        
        return getSecure(nativePtr);
    }

    private native void setSecure(long nativePtr, boolean enabled);
    @Override
    synchronized public void setSecure(boolean enabled)  {
        
        setSecure(nativePtr, enabled);
    }

    private native int getTransport(long nativePtr);
    @Override
    synchronized public TransportType getTransport()  {
        
        return TransportType.fromInt(getTransport(nativePtr));
    }

    private native int setTransport(long nativePtr, int transport);
    @Override
    synchronized public int setTransport(TransportType transport)  {
        
        return setTransport(nativePtr, transport.toInt());
    }

    private native String getUsername(long nativePtr);
    @Override
    synchronized public String getUsername()  {
        
        return getUsername(nativePtr);
    }

    private native int setUsername(long nativePtr, String username);
    @Override
    synchronized public int setUsername(String username)  {
        
        return setUsername(nativePtr, username);
    }

    private native String asString(long nativePtr);
    @Override
    synchronized public String asString()  {
        
        return asString(nativePtr);
    }

    private native String asStringUriOnly(long nativePtr);
    @Override
    synchronized public String asStringUriOnly()  {
        
        return asStringUriOnly(nativePtr);
    }

    private native void clean(long nativePtr);
    @Override
    synchronized public void clean()  {
        
        clean(nativePtr);
    }

    private native Address clone(long nativePtr);
    @Override
    synchronized public Address clone()  {
        
        return (Address)clone(nativePtr);
    }

    private native boolean equal(long nativePtr, Address address2);
    @Override
    synchronized public boolean equal(Address address2)  {
        
        return equal(nativePtr, address2);
    }

    private native String getHeader(long nativePtr, String headerName);
    @Override
    synchronized public String getHeader(String headerName)  {
        
        return getHeader(nativePtr, headerName);
    }

    private native String getParam(long nativePtr, String paramName);
    @Override
    synchronized public String getParam(String paramName)  {
        
        return getParam(nativePtr, paramName);
    }

    private native String getUriParam(long nativePtr, String uriParamName);
    @Override
    synchronized public String getUriParam(String uriParamName)  {
        
        return getUriParam(nativePtr, uriParamName);
    }

    private native boolean hasParam(long nativePtr, String paramName);
    @Override
    synchronized public boolean hasParam(String paramName)  {
        
        return hasParam(nativePtr, paramName);
    }

    private native boolean hasUriParam(long nativePtr, String uriParamName);
    @Override
    synchronized public boolean hasUriParam(String uriParamName)  {
        
        return hasUriParam(nativePtr, uriParamName);
    }

    private native void removeUriParam(long nativePtr, String uriParamName);
    @Override
    synchronized public void removeUriParam(String uriParamName)  {
        
        removeUriParam(nativePtr, uriParamName);
    }

    private native void setHeader(long nativePtr, String headerName, String headerValue);
    @Override
    synchronized public void setHeader(String headerName, String headerValue)  {
        
        setHeader(nativePtr, headerName, headerValue);
    }

    private native void setParam(long nativePtr, String paramName, String paramValue);
    @Override
    synchronized public void setParam(String paramName, String paramValue)  {
        
        setParam(nativePtr, paramName, paramValue);
    }

    private native void setUriParam(long nativePtr, String uriParamName, String uriParamValue);
    @Override
    synchronized public void setUriParam(String uriParamName, String uriParamValue)  {
        
        setUriParam(nativePtr, uriParamName, uriParamValue);
    }

    private native boolean weakEqual(long nativePtr, Address address2);
    @Override
    synchronized public boolean weakEqual(Address address2)  {
        
        return weakEqual(nativePtr, address2);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
