/*
AuthInfo.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* Object holding authentication information. 
*
* Note: The object's fields should not be accessed directly. Prefer using the
* accessor methods.
* 
* In most case, authentication information consists of a username and password.
* Sometimes, a userid is required by proxy, and realm can be useful to
* discriminate different SIP domains.
* 
* Once created and filled, a {@link AuthInfo} must be added to the {@link Core}
* in order to become known and used automatically when needed. Use {@link
* Core#addAuthInfo} for that purpose.
* 
* The {@link Core} object can take the initiative to request authentication
* information when needed to the application through the auth_info_requested
* callback of the LinphoneCoreVTable structure.
* 
* The application can respond to this information request later using {@link
* Core#addAuthInfo}. This will unblock all pending authentication transactions
* and retry them with authentication headers. 
*/
public interface AuthInfo {
    /**
    * Gets the algorithm. 
    *
    * @return The algorithm. 
    */
    public String getAlgorithm();

    /**
    * Sets the algorithm. 
    *
    * @param algorithm The algorithm. 
    */
    public void setAlgorithm(String algorithm);

    /**
    * Gets the domain. 
    *
    * @return The domain. 
    */
    public String getDomain();

    /**
    * Sets the domain for which this authentication is valid. 
    *
    * @param domain The domain. This should not be necessary because realm is
    * supposed to be unique and sufficient. However, many SIP servers don't set realm
    * correctly, then domain has to be used to distinguish between several SIP
    * account bearing the same username. 
    */
    public void setDomain(String domain);

    /**
    * Gets the ha1. 
    *
    * @return The ha1. 
    */
    public String getHa1();

    /**
    * Sets the ha1. 
    *
    * @param ha1 The ha1. 
    */
    public void setHa1(String ha1);

    /**
    * Gets the password. 
    *
    * @return The password. 
    * 
    * @deprecated , use linphone_auth_info_get_password instead 
    */
    @Deprecated
    public String getPasswd();

    /**
    * Sets the password. 
    *
    * @param passwd The password. 
    * 
    * @deprecated , use linphone_auth_info_set_password instead 
    */
    @Deprecated
    public void setPasswd(String passwd);

    /**
    * Gets the password. 
    *
    * @return The password. 
    */
    public String getPassword();

    /**
    * Sets the password. 
    *
    * @param passwd The password. 
    */
    public void setPassword(String passwd);

    /**
    * Gets the realm. 
    *
    * @return The realm. 
    */
    public String getRealm();

    /**
    * Sets the realm. 
    *
    * @param realm The realm. 
    */
    public void setRealm(String realm);

    /**
    * Gets the TLS certificate. 
    *
    * @return The TLS certificate. 
    */
    public String getTlsCert();

    /**
    * Sets the TLS certificate. 
    *
    * @param tlsCert The TLS certificate. 
    */
    public void setTlsCert(String tlsCert);

    /**
    * Gets the TLS certificate path. 
    *
    * @return The TLS certificate path. 
    */
    public String getTlsCertPath();

    /**
    * Sets the TLS certificate path. 
    *
    * @param tlsCertPath The TLS certificate path. 
    */
    public void setTlsCertPath(String tlsCertPath);

    /**
    * Gets the TLS key. 
    *
    * @return The TLS key. 
    */
    public String getTlsKey();

    /**
    * Sets the TLS key. 
    *
    * @param tlsKey The TLS key. 
    */
    public void setTlsKey(String tlsKey);

    /**
    * Gets the TLS key path. 
    *
    * @return The TLS key path. 
    */
    public String getTlsKeyPath();

    /**
    * Sets the TLS key path. 
    *
    * @param tlsKeyPath The TLS key path. 
    */
    public void setTlsKeyPath(String tlsKeyPath);

    /**
    * Gets the userid. 
    *
    * @return The userid. 
    */
    public String getUserid();

    /**
    * Sets the userid. 
    *
    * @param userid The userid. 
    */
    public void setUserid(String userid);

    /**
    * Gets the username. 
    *
    * @return The username. 
    */
    public String getUsername();

    /**
    * Sets the username. 
    *
    * @param username The username. 
    */
    public void setUsername(String username);

    /**
    * Instantiates a new auth info with values from source. 
    *
    * @return The newly created {@link AuthInfo} object. 
    */
    public AuthInfo clone();

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class AuthInfoImpl implements AuthInfo {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected AuthInfoImpl(long ptr) {
        nativePtr = ptr;
    }


    private native String getAlgorithm(long nativePtr);
    @Override
    synchronized public String getAlgorithm()  {
        
        return getAlgorithm(nativePtr);
    }

    private native void setAlgorithm(long nativePtr, String algorithm);
    @Override
    synchronized public void setAlgorithm(String algorithm)  {
        
        setAlgorithm(nativePtr, algorithm);
    }

    private native String getDomain(long nativePtr);
    @Override
    synchronized public String getDomain()  {
        
        return getDomain(nativePtr);
    }

    private native void setDomain(long nativePtr, String domain);
    @Override
    synchronized public void setDomain(String domain)  {
        
        setDomain(nativePtr, domain);
    }

    private native String getHa11(long nativePtr);
    @Override
    synchronized public String getHa1()  {
        
        return getHa11(nativePtr);
    }

    private native void setHa11(long nativePtr, String ha1);
    @Override
    synchronized public void setHa1(String ha1)  {
        
        setHa11(nativePtr, ha1);
    }

    private native String getPasswd(long nativePtr);
    @Override
    synchronized public String getPasswd()  {
        
        return getPasswd(nativePtr);
    }

    private native void setPasswd(long nativePtr, String passwd);
    @Override
    synchronized public void setPasswd(String passwd)  {
        
        setPasswd(nativePtr, passwd);
    }

    private native String getPassword(long nativePtr);
    @Override
    synchronized public String getPassword()  {
        
        return getPassword(nativePtr);
    }

    private native void setPassword(long nativePtr, String passwd);
    @Override
    synchronized public void setPassword(String passwd)  {
        
        setPassword(nativePtr, passwd);
    }

    private native String getRealm(long nativePtr);
    @Override
    synchronized public String getRealm()  {
        
        return getRealm(nativePtr);
    }

    private native void setRealm(long nativePtr, String realm);
    @Override
    synchronized public void setRealm(String realm)  {
        
        setRealm(nativePtr, realm);
    }

    private native String getTlsCert(long nativePtr);
    @Override
    synchronized public String getTlsCert()  {
        
        return getTlsCert(nativePtr);
    }

    private native void setTlsCert(long nativePtr, String tlsCert);
    @Override
    synchronized public void setTlsCert(String tlsCert)  {
        
        setTlsCert(nativePtr, tlsCert);
    }

    private native String getTlsCertPath(long nativePtr);
    @Override
    synchronized public String getTlsCertPath()  {
        
        return getTlsCertPath(nativePtr);
    }

    private native void setTlsCertPath(long nativePtr, String tlsCertPath);
    @Override
    synchronized public void setTlsCertPath(String tlsCertPath)  {
        
        setTlsCertPath(nativePtr, tlsCertPath);
    }

    private native String getTlsKey(long nativePtr);
    @Override
    synchronized public String getTlsKey()  {
        
        return getTlsKey(nativePtr);
    }

    private native void setTlsKey(long nativePtr, String tlsKey);
    @Override
    synchronized public void setTlsKey(String tlsKey)  {
        
        setTlsKey(nativePtr, tlsKey);
    }

    private native String getTlsKeyPath(long nativePtr);
    @Override
    synchronized public String getTlsKeyPath()  {
        
        return getTlsKeyPath(nativePtr);
    }

    private native void setTlsKeyPath(long nativePtr, String tlsKeyPath);
    @Override
    synchronized public void setTlsKeyPath(String tlsKeyPath)  {
        
        setTlsKeyPath(nativePtr, tlsKeyPath);
    }

    private native String getUserid(long nativePtr);
    @Override
    synchronized public String getUserid()  {
        
        return getUserid(nativePtr);
    }

    private native void setUserid(long nativePtr, String userid);
    @Override
    synchronized public void setUserid(String userid)  {
        
        setUserid(nativePtr, userid);
    }

    private native String getUsername(long nativePtr);
    @Override
    synchronized public String getUsername()  {
        
        return getUsername(nativePtr);
    }

    private native void setUsername(long nativePtr, String username);
    @Override
    synchronized public void setUsername(String username)  {
        
        setUsername(nativePtr, username);
    }

    private native AuthInfo clone(long nativePtr);
    @Override
    synchronized public AuthInfo clone()  {
        
        return (AuthInfo)clone(nativePtr);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
