/*
Call.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* The {@link Call} object represents a call issued or received by the {@link
* Core}. 
*
*/
public interface Call {
    public enum Status {
        /**
        * The call was sucessful. 
        *
        */
        Success(0),

        /**
        * The call was aborted. 
        *
        */
        Aborted(1),

        /**
        * The call was missed (unanswered) 
        *
        */
        Missed(2),

        /**
        * The call was declined, either locally or by remote end. 
        *
        */
        Declined(3),

        /**
        * The call was aborted before being advertised to the application - for protocol
        * reasons. 
        *
        */
        EarlyAborted(4),

        /**
        * The call was answered on another device. 
        *
        */
        AcceptedElsewhere(5),

        /**
        * The call was declined on another device. 
        *
        */
        DeclinedElsewhere(6);

        protected final int mValue;

        private Status (int value) {
            mValue = value;
        }

        static public Status fromInt(int value) throws RuntimeException {
            switch(value) {
            case 0: return Success;
            case 1: return Aborted;
            case 2: return Missed;
            case 3: return Declined;
            case 4: return EarlyAborted;
            case 5: return AcceptedElsewhere;
            case 6: return DeclinedElsewhere;
            default:
                throw new RuntimeException("Unhandled enum value " + value + " for Status");
            }
        }

        public int toInt() {
            return mValue;
        }
    };

    public enum Dir {
        /**
        * outgoing calls 
        *
        */
        Outgoing(0),

        /**
        * incoming calls 
        *
        */
        Incoming(1);

        protected final int mValue;

        private Dir (int value) {
            mValue = value;
        }

        static public Dir fromInt(int value) throws RuntimeException {
            switch(value) {
            case 0: return Outgoing;
            case 1: return Incoming;
            default:
                throw new RuntimeException("Unhandled enum value " + value + " for Dir");
            }
        }

        public int toInt() {
            return mValue;
        }
    };

    public enum State {
        /**
        * Initial state. 
        *
        */
        Idle(0),

        /**
        * Incoming call received. 
        *
        */
        IncomingReceived(1),

        /**
        * Outgoing call initialized. 
        *
        */
        OutgoingInit(2),

        /**
        * Outgoing call in progress. 
        *
        */
        OutgoingProgress(3),

        /**
        * Outgoing call ringing. 
        *
        */
        OutgoingRinging(4),

        /**
        * Outgoing call early media. 
        *
        */
        OutgoingEarlyMedia(5),

        /**
        * Connected. 
        *
        */
        Connected(6),

        /**
        * Streams running. 
        *
        */
        StreamsRunning(7),

        /**
        * Pausing. 
        *
        */
        Pausing(8),

        /**
        * Paused. 
        *
        */
        Paused(9),

        /**
        * Resuming. 
        *
        */
        Resuming(10),

        /**
        * Referred. 
        *
        */
        Referred(11),

        /**
        * Error. 
        *
        */
        Error(12),

        /**
        * Call end. 
        *
        */
        End(13),

        /**
        * Paused by remote. 
        *
        */
        PausedByRemote(14),

        /**
        * The call's parameters are updated for example when video is asked by remote. 
        *
        */
        UpdatedByRemote(15),

        /**
        * We are proposing early media to an incoming call. 
        *
        */
        IncomingEarlyMedia(16),

        /**
        * We have initiated a call update. 
        *
        */
        Updating(17),

        /**
        * The call object is now released. 
        *
        */
        Released(18),

        /**
        * The call is updated by remote while not yet answered (SIP UPDATE in early
        * dialog received) 
        *
        */
        EarlyUpdatedByRemote(19),

        /**
        * We are updating the call while not yet answered (SIP UPDATE in early dialog
        * sent) 
        *
        */
        EarlyUpdating(20);

        protected final int mValue;

        private State (int value) {
            mValue = value;
        }

        static public State fromInt(int value) throws RuntimeException {
            switch(value) {
            case 0: return Idle;
            case 1: return IncomingReceived;
            case 2: return OutgoingInit;
            case 3: return OutgoingProgress;
            case 4: return OutgoingRinging;
            case 5: return OutgoingEarlyMedia;
            case 6: return Connected;
            case 7: return StreamsRunning;
            case 8: return Pausing;
            case 9: return Paused;
            case 10: return Resuming;
            case 11: return Referred;
            case 12: return Error;
            case 13: return End;
            case 14: return PausedByRemote;
            case 15: return UpdatedByRemote;
            case 16: return IncomingEarlyMedia;
            case 17: return Updating;
            case 18: return Released;
            case 19: return EarlyUpdatedByRemote;
            case 20: return EarlyUpdating;
            default:
                throw new RuntimeException("Unhandled enum value " + value + " for State");
            }
        }

        public int toInt() {
            return mValue;
        }
    };

    /**
    * Returns the ZRTP authentication token to verify. 
    *
    * @return the authentication token to verify. 
    */
    public String getAuthenticationToken();

    /**
    * Returns whether ZRTP authentication token is verified. 
    *
    * If not, it must be verified by users as described in ZRTP procedure. Once done,
    * the application must inform of the results with {@link
    * Call#setAuthenticationTokenVerified}. 
    * 
    * @return  if authentication token is verifed, false otherwise. 
    */
    public boolean getAuthenticationTokenVerified();

    /**
    * Set the result of ZRTP short code verification by user. 
    *
    * If remote party also does the same, it will update the ZRTP cache so that
    * user's verification will not be required for the two users. 
    * 
    * @param verified whether the ZRTP SAS is verified. 
    */
    public void setAuthenticationTokenVerified(boolean verified);

    /**
    * Returns call quality averaged over all the duration of the call. 
    *
    * See {@link Call#getCurrentQuality} for more details about quality measurement. 
    */
    public float getAverageQuality();

    /**
    * Gets the call log associated to this call. 
    *
    * @return The {@link CallLog} associated with the specified {@link Call} 
    */
    public CallLog getCallLog();

    /**
    * Returns true if camera pictures are allowed to be sent to the remote party. 
    *
    */
    public boolean cameraEnabled();

    /**
    * Indicate whether camera input should be sent to remote end. 
    *
    */
    public void enableCamera(boolean enabled);

    /**
    * Create a new chat room for messaging from a call if not already existing, else
    * return existing one. 
    *
    * No reference is given to the caller: the chat room will be deleted when the
    * call is ended. 
    * 
    * @return {@link ChatRoom} where messaging can take place. 
    */
    public ChatRoom getChatRoom();

    /**
    * Return the associated conference object. 
    *
    * @return A pointer on {@link Conference} or null if the call is not part of any
    * conference. 
    */
    public Conference getConference();

    /**
    * Get the core that has created the specified call. 
    *
    * @return The {@link Core} object that has created the specified call. 
    */
    public Core getCore();

    /**
    * Returns current parameters associated to the call. 
    *
    */
    public CallParams getCurrentParams();

    /**
    * Obtain real-time quality rating of the call. 
    *
    * Based on local RTP statistics and RTCP feedback, a quality rating is computed
    * and updated during all the duration of the call. This function returns its
    * value at the time of the function call. It is expected that the rating is
    * updated at least every 5 seconds or so. The rating is a floating point number
    * comprised between 0 and 5.
    * 
    * 4-5 = good quality  3-4 = average quality  2-3 = poor quality  1-2 = very poor
    * quality  0-1 = can't be worse, mostly unusable 
    * 
    * @return The function returns -1 if no quality measurement is available, for
    * example if no active audio stream exist. Otherwise it returns the quality
    * rating. 
    */
    public float getCurrentQuality();

    /**
    * Returns direction of the call (incoming or outgoing). 
    *
    */
    public Call.Dir getDir();

    /**
    * Returns the diversion address associated to this call. 
    *
    */
    public Address getDiversionAddress();

    /**
    * Returns call's duration in seconds. 
    *
    */
    public int getDuration();

    /**
    * Returns true if echo cancellation is enabled. 
    *
    */
    public boolean echoCancellationEnabled();

    /**
    * Enables or disable echo cancellation for this call. 
    *
    * @param val 
    */
    public void enableEchoCancellation(boolean val);

    /**
    * Returns true if echo limiter is enabled. 
    *
    */
    public boolean echoLimiterEnabled();

    /**
    * Enables or disable echo limiter for this call. 
    *
    * @param val 
    */
    public void enableEchoLimiter(boolean val);

    /**
    * Returns full details about call errors or termination reasons. 
    *
    * @return {@link ErrorInfo} object holding the reason error. 
    */
    public ErrorInfo getErrorInfo();

    /**
    * Returns whether or not the call is currently being recorded. 
    *
    * @return true if recording is in progress, false otherwise 
    */
    public boolean isRecording();

    /**
    * Get microphone muted state. 
    *
    * @return The microphone muted state. 
    */
    public boolean getMicrophoneMuted();

    /**
    * Set microphone muted state. 
    *
    * @param muted The microphone muted state. 
    */
    public void setMicrophoneMuted(boolean muted);

    /**
    * Get microphone volume gain. 
    *
    * If the sound backend supports it, the returned gain is equal to the gain set
    * with the system mixer. 
    * 
    * @return double Percentage of the max supported volume gain. Valid values are in
    * [ 0.0 : 1.0 ]. In case of failure, a negative value is returned 
    */
    public float getMicrophoneVolumeGain();

    /**
    * Set microphone volume gain. 
    *
    * If the sound backend supports it, the new gain will synchronized with the
    * system mixer. 
    * 
    * @param volume Percentage of the max supported gain. Valid values are in [ 0.0 :
    * 1.0 ]. 
    */
    public void setMicrophoneVolumeGain(float volume);

    /**
    * Get the native window handle of the video window, casted as an unsigned long. 
    *
    */
    public Object getNativeVideoWindowId();

    /**
    * Set the native video window id where the video is to be displayed. 
    *
    * For MacOS, Linux, Windows: if not set or 0 a window will be automatically
    * created, unless the special id -1 is given. 
    */
    public void setNativeVideoWindowId(Object id);

    /**
    * Returns local parameters associated with the call. 
    *
    * This is typically the parameters passed at call initiation to {@link
    * Core#inviteAddressWithParams} or {@link Call#acceptWithParams}, or some default
    * parameters if no {@link CallParams} was explicitely passed during call
    * initiation. 
    * 
    * @return the call's local parameters. 
    */
    public CallParams getParams();

    /**
    * Set call parameters - advanced and not recommended feature - use with caution. 
    *
    * Local call parameters applicable to an outgoing or incoming shall usually be
    * passed to {@link Core#inviteAddressWithParams} or {@link
    * Call#acceptWithParams}. However, in some cases it might be desirable from a
    * software design standpoint to modify local parameters outside of the
    * application layer, typically in the purpose of implementing a custom logic
    * including special headers in INVITE or 200Ok requests, driven by a
    * call_state_changed listener method. This function accepts to assign a new
    * {@link CallParams} only in #LinphoneCallOutgoingInit and
    * #LinphoneCallIncomingReceived states. 
    */
    public void setParams(CallParams params);

    /**
    * Get the mesured playback volume level (received from remote) in dbm0. 
    *
    * @return float Volume level in percentage. 
    */
    public float getPlayVolume();

    /**
    * Get a player associated with the call to play a local file and stream it to the
    * remote peer. 
    *
    * @return A {@link Player} object 
    */
    public Player getPlayer();

    /**
    * Returns the reason for a call termination (either error or normal termination) 
    *
    */
    public Reason getReason();

    /**
    * Get the mesured record volume level (sent to remote) in dbm0. 
    *
    * @return float Volume level in percentage. 
    */
    public float getRecordVolume();

    /**
    * Gets the refer-to uri (if the call was transfered). 
    *
    * @return The refer-to uri of the call (if it was transfered) 
    */
    public String getReferTo();

    /**
    * Returns the remote address associated to this call. 
    *
    */
    public Address getRemoteAddress();

    /**
    * Returns the remote address associated to this call as a string. 
    *
    * The result string must be freed by user using ms_free(). 
    */
    public String getRemoteAddressAsString();

    /**
    * Returns the far end's sip contact as a string, if available. 
    *
    */
    public String getRemoteContact();

    /**
    * Returns call parameters proposed by remote. 
    *
    * This is useful when receiving an incoming call, to know whether the remote
    * party supports video, encryption or whatever. 
    */
    public CallParams getRemoteParams();

    /**
    * Returns the far end's user agent description string, if available. 
    *
    */
    public String getRemoteUserAgent();

    /**
    * Returns the call object this call is replacing, if any. 
    *
    * Call replacement can occur during call transfers. By default, the core
    * automatically terminates the replaced call and accept the new one. This
    * function allows the application to know whether a new incoming call is a one
    * that replaces another one. 
    */
    public Call getReplacedCall();

    /**
    * Get speaker muted state. 
    *
    * @return The speaker muted state. 
    */
    public boolean getSpeakerMuted();

    /**
    * Set speaker muted state. 
    *
    * @param muted The speaker muted state. 
    */
    public void setSpeakerMuted(boolean muted);

    /**
    * Get speaker volume gain. 
    *
    * If the sound backend supports it, the returned gain is equal to the gain set
    * with the system mixer. 
    * 
    * @return Percentage of the max supported volume gain. Valid values are in [ 0.0
    * : 1.0 ]. In case of failure, a negative value is returned 
    */
    public float getSpeakerVolumeGain();

    /**
    * Set speaker volume gain. 
    *
    * If the sound backend supports it, the new gain will synchronized with the
    * system mixer. 
    * 
    * @param volume Percentage of the max supported gain. Valid values are in [ 0.0 :
    * 1.0 ]. 
    */
    public void setSpeakerVolumeGain(float volume);

    /**
    * Retrieves the call's current state. 
    *
    */
    public Call.State getState();

    /**
    * Returns the number of stream for the given call. 
    *
    * Currently there is only two (Audio, Video), but later there will be more. 
    * 
    * @return 2 
    */
    public int getStreamCount();

    /**
    * Returns the to address with its headers associated to this call. 
    *
    */
    public Address getToAddress();

    /**
    * Returns the current transfer state, if a transfer has been initiated from this
    * call. 
    *
    * See also: {@link Core#transferCall} , {@link Core#transferCallToAnother} 
    */
    public Call.State getTransferState();

    /**
    * When this call has received a transfer request, returns the new call that was
    * automatically created as a result of the transfer. 
    *
    */
    public Call getTransferTargetCall();

    /**
    * Gets the transferer if this call was started automatically as a result of an
    * incoming transfer request. 
    *
    * The call in which the transfer request was received is returned in this case. 
    * 
    * @return The transferer call if the specified call was started automatically as
    * a result of an incoming transfer request, null otherwise 
    */
    public Call getTransfererCall();

    /**
    * Accept an incoming call. 
    *
    * Basically the application is notified of incoming calls within the
    * call_state_changed callback of the LinphoneCoreVTable structure, where it will
    * receive a LinphoneCallIncoming event with the associated {@link Call} object.
    * The application can later accept the call using this method. 
    * 
    * @return 0 on success, -1 on failure 
    */
    public int accept();

    /**
    * Accept an early media session for an incoming call. 
    *
    * This is identical as calling {@link Call#acceptEarlyMediaWithParams} with null
    * parameters. 
    * 
    * @return 0 if successful, -1 otherwise 
    * 
    * See also: {@link Call#acceptEarlyMediaWithParams} 
    */
    public int acceptEarlyMedia();

    /**
    * When receiving an incoming, accept to start a media session as early-media. 
    *
    * This means the call is not accepted but audio & video streams can be
    * established if the remote party supports early media. However, unlike after
    * call acceptance, mic and camera input are not sent during early-media, though
    * received audio & video are played normally. The call can then later be fully
    * accepted using {@link Call#accept} or {@link Call#acceptWithParams}. 
    * 
    * @param params The call parameters to use (can be null)    
    * 
    * @return 0 if successful, -1 otherwise 
    */
    public int acceptEarlyMediaWithParams(@Nullable CallParams params);

    /**
    * Accept call modifications initiated by other end. 
    *
    * This call may be performed in response to a #LinphoneCallUpdatedByRemote state
    * notification. When such notification arrives, the application can decide to
    * call {@link Call#deferUpdate} so that it can have the time to prompt the user.
    * {@link Call#getRemoteParams} can be used to get information about the call
    * parameters requested by the other party, such as whether a video stream is
    * requested.
    * 
    * When the user accepts or refuse the change, {@link Call#acceptUpdate} can be
    * done to answer to the other party. If params is null, then the same call
    * parameters established before the update request will continue to be used (no
    * change). If params is not null, then the update will be accepted according to
    * the parameters passed. Typical example is when a user accepts to start video,
    * then params should indicate that video stream should be used (see {@link
    * CallParams#enableVideo}). 
    * 
    * @param params A {@link CallParams} object describing the call parameters to
    * accept    
    * 
    * @return 0 if successful, -1 otherwise (actually when this function call is
    * performed outside ot #LinphoneCallUpdatedByRemote state) 
    */
    public int acceptUpdate(@Nullable CallParams params);

    /**
    * Accept an incoming call, with parameters. 
    *
    * Basically the application is notified of incoming calls within the
    * call_state_changed callback of the LinphoneCoreVTable structure, where it will
    * receive a LinphoneCallIncoming event with the associated {@link Call} object.
    * The application can later accept the call using this method. 
    * 
    * @param params The specific parameters for this call, for example whether video
    * is accepted or not. Use null to use default parameters    
    * 
    * @return 0 on success, -1 on failure 
    */
    public int acceptWithParams(@Nullable CallParams params);

    /**
    * Tell whether a call has been asked to autoanswer. 
    *
    * @return A boolean value telling whether the call has been asked to autoanswer 
    */
    public boolean askedToAutoanswer();

    /**
    * Stop current DTMF sequence sending. 
    *
    * Please note that some DTMF could be already sent, depending on when this
    * function call is delayed from linphone_call_send_dtmfs. This function will be
    * automatically called if call state change to anything but
    * LinphoneCallStreamsRunning. 
    */
    public void cancelDtmfs();

    /**
    * Decline a pending incoming call, with a reason. 
    *
    * @param reason The reason for rejecting the call: LinphoneReasonDeclined or
    * LinphoneReasonBusy 
    * 
    * @return 0 on success, -1 on failure 
    */
    public int decline(Reason reason);

    /**
    * Decline a pending incoming call, with a {@link ErrorInfo} object. 
    *
    * @param ei {@link ErrorInfo} containing more information on the call rejection. 
    * 
    * @return 0 on success, -1 on failure 
    */
    public int declineWithErrorInfo(ErrorInfo ei);

    /**
    * When receiving a #LinphoneCallUpdatedByRemote state notification, prevent
    * {@link Core} from performing an automatic answer. 
    *
    * When receiving a #LinphoneCallUpdatedByRemote state notification (ie an
    * incoming reINVITE), the default behaviour of {@link Core} is defined by the
    * "defer_update_default" option of the "sip" section of the config. If this
    * option is 0 (the default) then the {@link Core} automatically answers the
    * reINIVTE with call parameters unchanged. However when for example when the
    * remote party updated the call to propose a video stream, it can be useful to
    * prompt the user before answering. This can be achieved by calling {@link
    * Core#deferCallUpdate} during the call state notification, to deactivate the
    * automatic answer that would just confirm the audio but reject the video. Then,
    * when the user responds to dialog prompt, it becomes possible to call {@link
    * Call#acceptUpdate} to answer the reINVITE, with eventually video enabled in the
    * {@link CallParams} argument.
    * 
    * The #LinphoneCallUpdatedByRemote notification can also arrive when receiving an
    * INVITE without SDP. In such case, an unchanged offer is made in the 200Ok, and
    * when the ACK containing the SDP answer is received,
    * #LinphoneCallUpdatedByRemote is triggered to notify the application of possible
    * changes in the media session. However in such case defering the update has no
    * meaning since we just generating an offer.
    * 
    * @return 0 if successful, -1 if the {@link Call#deferUpdate} was done outside a
    * valid #LinphoneCallUpdatedByRemote notification 
    */
    public int deferUpdate();

    /**
    * Return a copy of the call statistics for a particular stream type. 
    *
    * @param type the stream type 
    */
    public CallStats getStats(StreamType type);

    /**
    * Returns the value of the header name. 
    *
    */
    public String getToHeader(String name);

    /**
    * Returns true if this calls has received a transfer that has not been executed
    * yet. 
    *
    * Pending transfers are executed when this call is being paused or closed,
    * locally or by remote endpoint. If the call is already paused while receiving
    * the transfer request, the transfer immediately occurs. 
    */
    public boolean hasTransferPending();

    /**
    * Indicates whether an operation is in progress at the media side. 
    *
    * It can be a bad idea to initiate signaling operations (adding video, pausing
    * the call, removing video, changing video parameters) while the media is busy in
    * establishing the connection (typically ICE connectivity checks). It can result
    * in failures generating loss of time in future operations in the call.
    * Applications are invited to check this function after each call state change to
    * decide whether certain operations are permitted or not. 
    * 
    * @return  if media is busy in establishing the connection, true otherwise. 
    */
    public boolean mediaInProgress();

    /**
    * Call generic OpenGL render for a given call. 
    *
    */
    public void oglRender();

    /**
    * Pauses the call. 
    *
    * If a music file has been setup using {@link Core#setPlayFile}, this file will
    * be played to the remote user. The only way to resume a paused call is to call
    * {@link Call#resume}. 
    * 
    * @return 0 on success, -1 on failure 
    * 
    * See also: {@link Call#resume} 
    */
    public int pause();

    /**
    * Redirect the specified call to the given redirect URI. 
    *
    * @param redirectUri The URI to redirect the call to 
    * 
    * @return 0 if successful, -1 on error. 
    */
    public int redirect(String redirectUri);

    /**
    * Request the callback passed to linphone_call_cbs_set_next_video_frame_decoded
    * to be called the next time the video decoder properly decodes a video frame. 
    *
    */
    public void requestNotifyNextVideoFrameDecoded();

    /**
    * Resumes a call. 
    *
    * The call needs to have been paused previously with {@link Call#pause}. 
    * 
    * @return 0 on success, -1 on failure 
    * 
    * See also: {@link Call#pause} 
    */
    public int resume();

    /**
    * Send the specified dtmf. 
    *
    * The dtmf is automatically played to the user. 
    * 
    * @param dtmf The dtmf name specified as a char, such as '0', '#' etc... 
    * 
    * @return 0 if successful, -1 on error. 
    */
    public int sendDtmf(char dtmf);

    /**
    * Send a list of dtmf. 
    *
    * The dtmfs are automatically sent to remote, separated by some needed
    * customizable delay. Sending is canceled if the call state changes to something
    * not LinphoneCallStreamsRunning. 
    * 
    * @param dtmfs A dtmf sequence such as '123#123123' 
    * 
    * @return -2 if there is already a DTMF sequence, -1 if call is not ready, 0
    * otherwise. 
    */
    public int sendDtmfs(String dtmfs);

    /**
    * Send a {@link InfoMessage} through an established call. 
    *
    * @param info the info message 
    */
    public int sendInfoMessage(InfoMessage info);

    /**
    * Request remote side to send us a Video Fast Update. 
    *
    */
    public void sendVfuRequest();

    /**
    * Start call recording. 
    *
    * Video record is only available if this function is called in state
    * StreamRunning. The output file where audio is recorded must be previously
    * specified with {@link CallParams#setRecordFile}. 
    */
    public void startRecording();

    /**
    * Stop call recording. 
    *
    */
    public void stopRecording();

    /**
    * Take a photo of currently captured video and write it into a jpeg file. 
    *
    * Note that the snapshot is asynchronous, an application shall not assume that
    * the file is created when the function returns. 
    * 
    * @param file a path where to write the jpeg content. 
    * 
    * @return 0 if successfull, -1 otherwise (typically if jpeg format is not
    * supported). 
    */
    public int takePreviewSnapshot(String file);

    /**
    * Take a photo of currently received video and write it into a jpeg file. 
    *
    * Note that the snapshot is asynchronous, an application shall not assume that
    * the file is created when the function returns. 
    * 
    * @param file a path where to write the jpeg content. 
    * 
    * @return 0 if successfull, -1 otherwise (typically if jpeg format is not
    * supported). 
    */
    public int takeVideoSnapshot(String file);

    /**
    * Terminates a call. 
    *
    * @return 0 on success, -1 on failure 
    */
    public int terminate();

    /**
    * Terminates a call. 
    *
    * @param ei {@link ErrorInfo} 
    * 
    * @return 0 on success, -1 on failure 
    */
    public int terminateWithErrorInfo(ErrorInfo ei);

    /**
    * Performs a simple call transfer to the specified destination. 
    *
    * The remote endpoint is expected to issue a new call to the specified
    * destination. The current call remains active and thus can be later paused or
    * terminated. It is possible to follow the progress of the transfer provided that
    * transferee sends notification about it. In this case, the
    * transfer_state_changed callback of the LinphoneCoreVTable is invoked to notify
    * of the state of the new call at the other party. The notified states are
    * #LinphoneCallOutgoingInit , #LinphoneCallOutgoingProgress,
    * #LinphoneCallOutgoingRinging and #LinphoneCallConnected. 
    * 
    * @param referTo The destination the call is to be refered to 
    * 
    * @return 0 on success, -1 on failure 
    */
    public int transfer(String referTo);

    /**
    * Transfers a call to destination of another running call. 
    *
    * This is used for "attended transfer" scenarios. The transfered call is supposed
    * to be in paused state, so that it is able to accept the transfer immediately.
    * The destination call is a call previously established to introduce the
    * transfered person. This method will send a transfer request to the transfered
    * person. The phone of the transfered is then expected to automatically call to
    * the destination of the transfer. The receiver of the transfer will then
    * automatically close the call with us (the 'dest' call). It is possible to
    * follow the progress of the transfer provided that transferee sends notification
    * about it. In this case, the transfer_state_changed callback of the
    * LinphoneCoreVTable is invoked to notify of the state of the new call at the
    * other party. The notified states are #LinphoneCallOutgoingInit ,
    * #LinphoneCallOutgoingProgress, #LinphoneCallOutgoingRinging and
    * #LinphoneCallConnected. 
    * 
    * @param dest A running call whose remote person will receive the transfer 
    * 
    * @return 0 on success, -1 on failure 
    */
    public int transferToAnother(Call dest);

    /**
    * Updates a running call according to supplied call parameters or parameters
    * changed in the LinphoneCore. 
    *
    * It triggers a SIP reINVITE in order to perform a new offer/answer of media
    * capabilities. Changing the size of the transmitted video after calling
    * linphone_core_set_preferred_video_size can be used by passing null as params
    * argument. In case no changes are requested through the {@link CallParams}
    * argument, then this argument can be omitted and set to null. WARNING: Updating
    * a call in the #LinphoneCallPaused state will still result in a paused call even
    * if the media directions set in the params are sendrecv. To resume a paused
    * call, you need to call {@link Call#resume}.
    * 
    * @param params The new call parameters to use (may be null)    
    * 
    * @return 0 if successful, -1 otherwise. 
    */
    public int update(@Nullable CallParams params);

    /**
    * Perform a zoom of the video displayed during a call. 
    *
    * The zoom ensures that all the screen is fullfilled with the video. 
    * 
    * @param cy a floating point number pointing the vertical center of the zoom to
    * be applied. This value should be between 0.0 and 1.0. 
    */
    public void zoom(float zoomFactor, float cx, float cy);

    /**
    */
    public void addListener(CallListener listener);

    /**
    */
    public void removeListener(CallListener listener);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class CallImpl implements Call {

    protected long nativePtr = 0;
    protected Object userData = null;
    protected Core core = null;

    protected CallImpl(long ptr) {
        nativePtr = ptr;
        core = getCore();
    }


    private native String getAuthenticationToken(long nativePtr);
    @Override
    synchronized public String getAuthenticationToken()  {
        synchronized(core) { 
        return getAuthenticationToken(nativePtr);
        }
    }

    private native boolean getAuthenticationTokenVerified(long nativePtr);
    @Override
    synchronized public boolean getAuthenticationTokenVerified()  {
        synchronized(core) { 
        return getAuthenticationTokenVerified(nativePtr);
        }
    }

    private native void setAuthenticationTokenVerified(long nativePtr, boolean verified);
    @Override
    synchronized public void setAuthenticationTokenVerified(boolean verified)  {
        synchronized(core) { 
        setAuthenticationTokenVerified(nativePtr, verified);
        }
    }

    private native float getAverageQuality(long nativePtr);
    @Override
    synchronized public float getAverageQuality()  {
        synchronized(core) { 
        return getAverageQuality(nativePtr);
        }
    }

    private native CallLog getCallLog(long nativePtr);
    @Override
    synchronized public CallLog getCallLog()  {
        synchronized(core) { 
        return (CallLog)getCallLog(nativePtr);
        }
    }

    private native boolean cameraEnabled(long nativePtr);
    @Override
    synchronized public boolean cameraEnabled()  {
        synchronized(core) { 
        return cameraEnabled(nativePtr);
        }
    }

    private native void enableCamera(long nativePtr, boolean enabled);
    @Override
    synchronized public void enableCamera(boolean enabled)  {
        synchronized(core) { 
        enableCamera(nativePtr, enabled);
        }
    }

    private native ChatRoom getChatRoom(long nativePtr);
    @Override
    synchronized public ChatRoom getChatRoom()  {
        synchronized(core) { 
        return (ChatRoom)getChatRoom(nativePtr);
        }
    }

    private native Conference getConference(long nativePtr);
    @Override
    synchronized public Conference getConference()  {
        synchronized(core) { 
        return (Conference)getConference(nativePtr);
        }
    }

    private native Core getCore(long nativePtr);
    @Override
    synchronized public Core getCore()  {
        
        return (Core)getCore(nativePtr);
    }

    private native CallParams getCurrentParams(long nativePtr);
    @Override
    synchronized public CallParams getCurrentParams()  {
        synchronized(core) { 
        return (CallParams)getCurrentParams(nativePtr);
        }
    }

    private native float getCurrentQuality(long nativePtr);
    @Override
    synchronized public float getCurrentQuality()  {
        synchronized(core) { 
        return getCurrentQuality(nativePtr);
        }
    }

    private native int getDir(long nativePtr);
    @Override
    synchronized public Call.Dir getDir()  {
        synchronized(core) { 
        return Call.Dir.fromInt(getDir(nativePtr));
        }
    }

    private native Address getDiversionAddress(long nativePtr);
    @Override
    synchronized public Address getDiversionAddress()  {
        synchronized(core) { 
        return (Address)getDiversionAddress(nativePtr);
        }
    }

    private native int getDuration(long nativePtr);
    @Override
    synchronized public int getDuration()  {
        synchronized(core) { 
        return getDuration(nativePtr);
        }
    }

    private native boolean echoCancellationEnabled(long nativePtr);
    @Override
    synchronized public boolean echoCancellationEnabled()  {
        synchronized(core) { 
        return echoCancellationEnabled(nativePtr);
        }
    }

    private native void enableEchoCancellation(long nativePtr, boolean val);
    @Override
    synchronized public void enableEchoCancellation(boolean val)  {
        synchronized(core) { 
        enableEchoCancellation(nativePtr, val);
        }
    }

    private native boolean echoLimiterEnabled(long nativePtr);
    @Override
    synchronized public boolean echoLimiterEnabled()  {
        synchronized(core) { 
        return echoLimiterEnabled(nativePtr);
        }
    }

    private native void enableEchoLimiter(long nativePtr, boolean val);
    @Override
    synchronized public void enableEchoLimiter(boolean val)  {
        synchronized(core) { 
        enableEchoLimiter(nativePtr, val);
        }
    }

    private native ErrorInfo getErrorInfo(long nativePtr);
    @Override
    synchronized public ErrorInfo getErrorInfo()  {
        synchronized(core) { 
        return (ErrorInfo)getErrorInfo(nativePtr);
        }
    }

    private native boolean isRecording(long nativePtr);
    @Override
    synchronized public boolean isRecording()  {
        synchronized(core) { 
        return isRecording(nativePtr);
        }
    }

    private native boolean getMicrophoneMuted(long nativePtr);
    @Override
    synchronized public boolean getMicrophoneMuted()  {
        synchronized(core) { 
        return getMicrophoneMuted(nativePtr);
        }
    }

    private native void setMicrophoneMuted(long nativePtr, boolean muted);
    @Override
    synchronized public void setMicrophoneMuted(boolean muted)  {
        synchronized(core) { 
        setMicrophoneMuted(nativePtr, muted);
        }
    }

    private native float getMicrophoneVolumeGain(long nativePtr);
    @Override
    synchronized public float getMicrophoneVolumeGain()  {
        synchronized(core) { 
        return getMicrophoneVolumeGain(nativePtr);
        }
    }

    private native void setMicrophoneVolumeGain(long nativePtr, float volume);
    @Override
    synchronized public void setMicrophoneVolumeGain(float volume)  {
        synchronized(core) { 
        setMicrophoneVolumeGain(nativePtr, volume);
        }
    }

    private native Object getNativeVideoWindowId(long nativePtr);
    @Override
    synchronized public Object getNativeVideoWindowId()  {
        synchronized(core) { 
        return getNativeVideoWindowId(nativePtr);
        }
    }

    private native void setNativeVideoWindowId(long nativePtr, Object id);
    @Override
    synchronized public void setNativeVideoWindowId(Object id)  {
        synchronized(core) { 
        setNativeVideoWindowId(nativePtr, id);
        }
    }

    private native CallParams getParams(long nativePtr);
    @Override
    synchronized public CallParams getParams()  {
        synchronized(core) { 
        return (CallParams)getParams(nativePtr);
        }
    }

    private native void setParams(long nativePtr, CallParams params);
    @Override
    synchronized public void setParams(CallParams params)  {
        synchronized(core) { 
        setParams(nativePtr, params);
        }
    }

    private native float getPlayVolume(long nativePtr);
    @Override
    synchronized public float getPlayVolume()  {
        synchronized(core) { 
        return getPlayVolume(nativePtr);
        }
    }

    private native Player getPlayer(long nativePtr);
    @Override
    synchronized public Player getPlayer()  {
        synchronized(core) { 
        return (Player)getPlayer(nativePtr);
        }
    }

    private native int getReason(long nativePtr);
    @Override
    synchronized public Reason getReason()  {
        synchronized(core) { 
        return Reason.fromInt(getReason(nativePtr));
        }
    }

    private native float getRecordVolume(long nativePtr);
    @Override
    synchronized public float getRecordVolume()  {
        synchronized(core) { 
        return getRecordVolume(nativePtr);
        }
    }

    private native String getReferTo(long nativePtr);
    @Override
    synchronized public String getReferTo()  {
        synchronized(core) { 
        return getReferTo(nativePtr);
        }
    }

    private native Address getRemoteAddress(long nativePtr);
    @Override
    synchronized public Address getRemoteAddress()  {
        synchronized(core) { 
        return (Address)getRemoteAddress(nativePtr);
        }
    }

    private native String getRemoteAddressAsString(long nativePtr);
    @Override
    synchronized public String getRemoteAddressAsString()  {
        synchronized(core) { 
        return getRemoteAddressAsString(nativePtr);
        }
    }

    private native String getRemoteContact(long nativePtr);
    @Override
    synchronized public String getRemoteContact()  {
        synchronized(core) { 
        return getRemoteContact(nativePtr);
        }
    }

    private native CallParams getRemoteParams(long nativePtr);
    @Override
    synchronized public CallParams getRemoteParams()  {
        synchronized(core) { 
        return (CallParams)getRemoteParams(nativePtr);
        }
    }

    private native String getRemoteUserAgent(long nativePtr);
    @Override
    synchronized public String getRemoteUserAgent()  {
        synchronized(core) { 
        return getRemoteUserAgent(nativePtr);
        }
    }

    private native Call getReplacedCall(long nativePtr);
    @Override
    synchronized public Call getReplacedCall()  {
        synchronized(core) { 
        return (Call)getReplacedCall(nativePtr);
        }
    }

    private native boolean getSpeakerMuted(long nativePtr);
    @Override
    synchronized public boolean getSpeakerMuted()  {
        synchronized(core) { 
        return getSpeakerMuted(nativePtr);
        }
    }

    private native void setSpeakerMuted(long nativePtr, boolean muted);
    @Override
    synchronized public void setSpeakerMuted(boolean muted)  {
        synchronized(core) { 
        setSpeakerMuted(nativePtr, muted);
        }
    }

    private native float getSpeakerVolumeGain(long nativePtr);
    @Override
    synchronized public float getSpeakerVolumeGain()  {
        synchronized(core) { 
        return getSpeakerVolumeGain(nativePtr);
        }
    }

    private native void setSpeakerVolumeGain(long nativePtr, float volume);
    @Override
    synchronized public void setSpeakerVolumeGain(float volume)  {
        synchronized(core) { 
        setSpeakerVolumeGain(nativePtr, volume);
        }
    }

    private native int getState(long nativePtr);
    @Override
    synchronized public Call.State getState()  {
        synchronized(core) { 
        return Call.State.fromInt(getState(nativePtr));
        }
    }

    private native int getStreamCount(long nativePtr);
    @Override
    synchronized public int getStreamCount()  {
        synchronized(core) { 
        return getStreamCount(nativePtr);
        }
    }

    private native Address getToAddress(long nativePtr);
    @Override
    synchronized public Address getToAddress()  {
        synchronized(core) { 
        return (Address)getToAddress(nativePtr);
        }
    }

    private native int getTransferState(long nativePtr);
    @Override
    synchronized public Call.State getTransferState()  {
        synchronized(core) { 
        return Call.State.fromInt(getTransferState(nativePtr));
        }
    }

    private native Call getTransferTargetCall(long nativePtr);
    @Override
    synchronized public Call getTransferTargetCall()  {
        synchronized(core) { 
        return (Call)getTransferTargetCall(nativePtr);
        }
    }

    private native Call getTransfererCall(long nativePtr);
    @Override
    synchronized public Call getTransfererCall()  {
        synchronized(core) { 
        return (Call)getTransfererCall(nativePtr);
        }
    }

    private native int accept(long nativePtr);
    @Override
    synchronized public int accept()  {
        synchronized(core) { 
        return accept(nativePtr);
        }
    }

    private native int acceptEarlyMedia(long nativePtr);
    @Override
    synchronized public int acceptEarlyMedia()  {
        synchronized(core) { 
        return acceptEarlyMedia(nativePtr);
        }
    }

    private native int acceptEarlyMediaWithParams(long nativePtr, CallParams params);
    @Override
    synchronized public int acceptEarlyMediaWithParams(@Nullable CallParams params)  {
        synchronized(core) { 
        return acceptEarlyMediaWithParams(nativePtr, params);
        }
    }

    private native int acceptUpdate(long nativePtr, CallParams params);
    @Override
    synchronized public int acceptUpdate(@Nullable CallParams params)  {
        synchronized(core) { 
        return acceptUpdate(nativePtr, params);
        }
    }

    private native int acceptWithParams(long nativePtr, CallParams params);
    @Override
    synchronized public int acceptWithParams(@Nullable CallParams params)  {
        synchronized(core) { 
        return acceptWithParams(nativePtr, params);
        }
    }

    private native boolean askedToAutoanswer(long nativePtr);
    @Override
    synchronized public boolean askedToAutoanswer()  {
        synchronized(core) { 
        return askedToAutoanswer(nativePtr);
        }
    }

    private native void cancelDtmfs(long nativePtr);
    @Override
    synchronized public void cancelDtmfs()  {
        synchronized(core) { 
        cancelDtmfs(nativePtr);
        }
    }

    private native int decline(long nativePtr, int reason);
    @Override
    synchronized public int decline(Reason reason)  {
        synchronized(core) { 
        return decline(nativePtr, reason.toInt());
        }
    }

    private native int declineWithErrorInfo(long nativePtr, ErrorInfo ei);
    @Override
    synchronized public int declineWithErrorInfo(ErrorInfo ei)  {
        synchronized(core) { 
        return declineWithErrorInfo(nativePtr, ei);
        }
    }

    private native int deferUpdate(long nativePtr);
    @Override
    synchronized public int deferUpdate()  {
        synchronized(core) { 
        return deferUpdate(nativePtr);
        }
    }

    private native CallStats getStats(long nativePtr, int type);
    @Override
    synchronized public CallStats getStats(StreamType type)  {
        synchronized(core) { 
        return (CallStats)getStats(nativePtr, type.toInt());
        }
    }

    private native String getToHeader(long nativePtr, String name);
    @Override
    synchronized public String getToHeader(String name)  {
        synchronized(core) { 
        return getToHeader(nativePtr, name);
        }
    }

    private native boolean hasTransferPending(long nativePtr);
    @Override
    synchronized public boolean hasTransferPending()  {
        synchronized(core) { 
        return hasTransferPending(nativePtr);
        }
    }

    private native boolean mediaInProgress(long nativePtr);
    @Override
    synchronized public boolean mediaInProgress()  {
        synchronized(core) { 
        return mediaInProgress(nativePtr);
        }
    }

    private native void oglRender(long nativePtr);
    @Override
    synchronized public void oglRender()  {
        synchronized(core) { 
        oglRender(nativePtr);
        }
    }

    private native int pause(long nativePtr);
    @Override
    synchronized public int pause()  {
        synchronized(core) { 
        return pause(nativePtr);
        }
    }

    private native int redirect(long nativePtr, String redirectUri);
    @Override
    synchronized public int redirect(String redirectUri)  {
        synchronized(core) { 
        return redirect(nativePtr, redirectUri);
        }
    }

    private native void requestNotifyNextVideoFrameDecoded(long nativePtr);
    @Override
    synchronized public void requestNotifyNextVideoFrameDecoded()  {
        synchronized(core) { 
        requestNotifyNextVideoFrameDecoded(nativePtr);
        }
    }

    private native int resume(long nativePtr);
    @Override
    synchronized public int resume()  {
        synchronized(core) { 
        return resume(nativePtr);
        }
    }

    private native int sendDtmf(long nativePtr, char dtmf);
    @Override
    synchronized public int sendDtmf(char dtmf)  {
        synchronized(core) { 
        return sendDtmf(nativePtr, dtmf);
        }
    }

    private native int sendDtmfs(long nativePtr, String dtmfs);
    @Override
    synchronized public int sendDtmfs(String dtmfs)  {
        synchronized(core) { 
        return sendDtmfs(nativePtr, dtmfs);
        }
    }

    private native int sendInfoMessage(long nativePtr, InfoMessage info);
    @Override
    synchronized public int sendInfoMessage(InfoMessage info)  {
        synchronized(core) { 
        return sendInfoMessage(nativePtr, info);
        }
    }

    private native void sendVfuRequest(long nativePtr);
    @Override
    synchronized public void sendVfuRequest()  {
        synchronized(core) { 
        sendVfuRequest(nativePtr);
        }
    }

    private native void startRecording(long nativePtr);
    @Override
    synchronized public void startRecording()  {
        synchronized(core) { 
        startRecording(nativePtr);
        }
    }

    private native void stopRecording(long nativePtr);
    @Override
    synchronized public void stopRecording()  {
        synchronized(core) { 
        stopRecording(nativePtr);
        }
    }

    private native int takePreviewSnapshot(long nativePtr, String file);
    @Override
    synchronized public int takePreviewSnapshot(String file)  {
        synchronized(core) { 
        return takePreviewSnapshot(nativePtr, file);
        }
    }

    private native int takeVideoSnapshot(long nativePtr, String file);
    @Override
    synchronized public int takeVideoSnapshot(String file)  {
        synchronized(core) { 
        return takeVideoSnapshot(nativePtr, file);
        }
    }

    private native int terminate(long nativePtr);
    @Override
    synchronized public int terminate()  {
        synchronized(core) { 
        return terminate(nativePtr);
        }
    }

    private native int terminateWithErrorInfo(long nativePtr, ErrorInfo ei);
    @Override
    synchronized public int terminateWithErrorInfo(ErrorInfo ei)  {
        synchronized(core) { 
        return terminateWithErrorInfo(nativePtr, ei);
        }
    }

    private native int transfer(long nativePtr, String referTo);
    @Override
    synchronized public int transfer(String referTo)  {
        synchronized(core) { 
        return transfer(nativePtr, referTo);
        }
    }

    private native int transferToAnother(long nativePtr, Call dest);
    @Override
    synchronized public int transferToAnother(Call dest)  {
        synchronized(core) { 
        return transferToAnother(nativePtr, dest);
        }
    }

    private native int update(long nativePtr, CallParams params);
    @Override
    synchronized public int update(@Nullable CallParams params)  {
        synchronized(core) { 
        return update(nativePtr, params);
        }
    }

    private native void zoom(long nativePtr, float zoomFactor, float cx, float cy);
    @Override
    synchronized public void zoom(float zoomFactor, float cx, float cy)  {
        synchronized(core) { 
        zoom(nativePtr, zoomFactor, cx, cy);
        }
    }

    private native void addListener(long nativePtr, CallListener listener);
    @Override
    synchronized public void addListener(CallListener listener)  {
        
        addListener(nativePtr, listener);
    }

    private native void removeListener(long nativePtr, CallListener listener);
    @Override
    synchronized public void removeListener(CallListener listener)  {
        
        removeListener(nativePtr, listener);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
