/*
CallListener.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;


/**
* That class holds all the callbacks which are called by LinphoneCall objects. 
*
* Use linphone_factory_create_call_cbs to create an instance. Then, call the
* callback setters on the events you need to monitor and pass the object to a
* LinphoneCall instance through linphone_call_add_callbacks. 
*/
public interface CallListener {
    /**
    * Callback to notify that the camera is not working and has been changed to "No
    * Webcam". 
    *
    * A camera is detected as mis-functionning as soon as it outputs no frames at all
    * during a period of 5 seconds. This check is only performed on desktop
    * platforms, in the purpose of notifying camera failures, for example if when a
    * usb cable gets disconnected.
    * 
    * @param cameraName the name of the non-working camera 
    */
    public void onCameraNotWorking(Call call, String cameraName);

    /**
    * Callback for notifying a snapshot taken. 
    *
    * @param filepath the name of the saved file 
    */
    public void onSnapshotTaken(Call call, String filepath);

    /**
    * Call state notification callback. 
    *
    * @param message An informational message about the state. 
    */
    public void onStateChanged(Call call, Call.State cstate, String message);

    /**
    * Callback for notifying progresses of transfers. 
    *
    * @param cstate The state of the call to transfer target at the far end. 
    */
    public void onTransferStateChanged(Call call, Call.State cstate);

    /**
    * Callback for notifying a received TMMBR. 
    *
    * @param tmmbr the value of the received TMMBR 
    */
    public void onTmmbrReceived(Call call, int streamIndex, int tmmbr);

    /**
    * Callback for receiving info messages. 
    *
    * @param msg LinphoneInfoMessage object. 
    */
    public void onInfoMessageReceived(Call call, InfoMessage msg);

    /**
    * Call encryption changed callback. 
    *
    * @param authenticationToken An authentication_token, currently set for ZRTP kind
    * of encryption only. 
    */
    public void onEncryptionChanged(Call call, boolean on, String authenticationToken);

    /**
    * Callback for notifying the processing SIP ACK messages. 
    *
    * @param isReceived if true this ACK is an incoming one, otherwise it is an ACK
    * about to be sent. 
    */
    public void onAckProcessing(Call call, Headers ack, boolean isReceived);

    /**
    * Callback for being notified of received DTMFs. 
    *
    * @param dtmf The ascii code of the dtmf 
    */
    public void onDtmfReceived(Call call, int dtmf);

    /**
    * Callback to notify a next video frame has been decoded. 
    *
    * @param call LinphoneCall for which the next video frame has been decoded 
    */
    public void onNextVideoFrameDecoded(Call call);

    /**
    * Callback for receiving quality statistics for calls. 
    *
    * @param stats LinphoneCallStats object 
    */
    public void onStatsUpdated(Call call, CallStats stats);

}