/*
CallLog.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* Structure representing a call log. 
*
*/
public interface CallLog {
    /**
    * Get the call ID used by the call. 
    *
    * @return The call ID used by the call as a string. 
    */
    public String getCallId();

    /**
    * Get the direction of the call. 
    *
    * @return The direction of the call. 
    */
    public Call.Dir getDir();

    /**
    * Get the duration of the call since connected. 
    *
    * @return The duration of the call in seconds. 
    */
    public int getDuration();

    /**
    * When the call was failed, return an object describing the failure. 
    *
    * @return information about the error encountered by the call associated with
    * this call log. 
    */
    public ErrorInfo getErrorInfo();

    /**
    * Get the origin address (ie from) of the call. 
    *
    * @return The origin address (ie from) of the call. 
    */
    public Address getFromAddress();

    /**
    * Get the local address (that is from or to depending on call direction) 
    *
    * @return The local address of the call 
    */
    public Address getLocalAddress();

    /**
    * Get the overall quality indication of the call. 
    *
    * @return The overall quality indication of the call. 
    */
    public float getQuality();

    /**
    * Get the persistent reference key associated to the call log. 
    *
    * The reference key can be for example an id to an external database. It is
    * stored in the config file, thus can survive to process exits/restarts.
    * 
    * @return The reference key string that has been associated to the call log, or
    * null if none has been associated. 
    */
    public String getRefKey();

    /**
    * Associate a persistent reference key to the call log. 
    *
    * The reference key can be for example an id to an external database. It is
    * stored in the config file, thus can survive to process exits/restarts.
    * 
    * @param refkey The reference key string to associate to the call log. 
    */
    public void setRefKey(String refkey);

    /**
    * Get the remote address (that is from or to depending on call direction). 
    *
    * @return The remote address of the call. 
    */
    public Address getRemoteAddress();

    /**
    * Get the start date of the call. 
    *
    * @return The date of the beginning of the call. 
    */
    public long getStartDate();

    /**
    * Get the status of the call. 
    *
    * @return The status of the call. 
    */
    public Call.Status getStatus();

    /**
    * Get the destination address (ie to) of the call. 
    *
    * @return The destination address (ie to) of the call. 
    */
    public Address getToAddress();

    /**
    * Tell whether video was enabled at the end of the call or not. 
    *
    * @return A boolean value telling whether video was enabled at the end of the
    * call. 
    */
    public boolean videoEnabled();

    /**
    * Get a human readable string describing the call. 
    *
    * Note: : the returned string must be freed by the application (use ms_free()). 
    * 
    * @return A human readable string describing the call. 
    */
    public String toStr();

    /**
    * Tells whether that call was a call to a conference server. 
    *
    * @return  if the call was a call to a conference server 
    */
    public boolean wasConference();

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class CallLogImpl implements CallLog {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected CallLogImpl(long ptr) {
        nativePtr = ptr;
    }


    private native String getCallId(long nativePtr);
    @Override
    synchronized public String getCallId()  {
        
        return getCallId(nativePtr);
    }

    private native int getDir(long nativePtr);
    @Override
    synchronized public Call.Dir getDir()  {
        
        return Call.Dir.fromInt(getDir(nativePtr));
    }

    private native int getDuration(long nativePtr);
    @Override
    synchronized public int getDuration()  {
        
        return getDuration(nativePtr);
    }

    private native ErrorInfo getErrorInfo(long nativePtr);
    @Override
    synchronized public ErrorInfo getErrorInfo()  {
        
        return (ErrorInfo)getErrorInfo(nativePtr);
    }

    private native Address getFromAddress(long nativePtr);
    @Override
    synchronized public Address getFromAddress()  {
        
        return (Address)getFromAddress(nativePtr);
    }

    private native Address getLocalAddress(long nativePtr);
    @Override
    synchronized public Address getLocalAddress()  {
        
        return (Address)getLocalAddress(nativePtr);
    }

    private native float getQuality(long nativePtr);
    @Override
    synchronized public float getQuality()  {
        
        return getQuality(nativePtr);
    }

    private native String getRefKey(long nativePtr);
    @Override
    synchronized public String getRefKey()  {
        
        return getRefKey(nativePtr);
    }

    private native void setRefKey(long nativePtr, String refkey);
    @Override
    synchronized public void setRefKey(String refkey)  {
        
        setRefKey(nativePtr, refkey);
    }

    private native Address getRemoteAddress(long nativePtr);
    @Override
    synchronized public Address getRemoteAddress()  {
        
        return (Address)getRemoteAddress(nativePtr);
    }

    private native long getStartDate(long nativePtr);
    @Override
    synchronized public long getStartDate()  {
        
        return getStartDate(nativePtr);
    }

    private native int getStatus(long nativePtr);
    @Override
    synchronized public Call.Status getStatus()  {
        
        return Call.Status.fromInt(getStatus(nativePtr));
    }

    private native Address getToAddress(long nativePtr);
    @Override
    synchronized public Address getToAddress()  {
        
        return (Address)getToAddress(nativePtr);
    }

    private native boolean videoEnabled(long nativePtr);
    @Override
    synchronized public boolean videoEnabled()  {
        
        return videoEnabled(nativePtr);
    }

    private native String toStr(long nativePtr);
    @Override
    synchronized public String toStr()  {
        
        return toStr(nativePtr);
    }

    private native boolean wasConference(long nativePtr);
    @Override
    synchronized public boolean wasConference()  {
        
        return wasConference(nativePtr);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
