/*
CallParams.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* The {@link CallParams} is an object containing various call related parameters. 
*
* It can be used to retrieve parameters from a currently running call or modify
* the call's characteristics dynamically. 
*/
public interface CallParams {
    /**
    * Refine bandwidth settings for this call by setting a bandwidth limit for audio
    * streams. 
    *
    * As a consequence, codecs whose bitrates are not compatible with this limit
    * won't be used. 
    * 
    * @param bw The audio bandwidth limit to set in kbit/s. 
    */
    public void setAudioBandwidthLimit(int bw);

    /**
    * Get the audio stream direction. 
    *
    * @return The audio stream direction associated with the call params. 
    */
    public MediaDirection getAudioDirection();

    /**
    * Set the audio stream direction. 
    *
    * @param dir The audio stream direction associated with this call params. 
    */
    public void setAudioDirection(MediaDirection dir);

    /**
    * Tell whether audio is enabled or not. 
    *
    * @return A boolean value telling whether audio is enabled or not. 
    */
    public boolean audioEnabled();

    /**
    * Enable audio stream. 
    *
    * @param enabled A boolean value telling whether to enable audio or not. 
    */
    public void enableAudio(boolean enabled);

    /**
    * Use to get multicast state of audio stream. 
    *
    * @return true if subsequent calls will propose multicast ip set by
    * linphone_core_set_audio_multicast_addr 
    */
    public boolean audioMulticastEnabled();

    /**
    * Use to enable multicast rtp for audio stream. 
    *
    * 
    * 
    */
    public void enableAudioMulticast(boolean yesno);

    /**
    * Gets a list of {@link Content} set if exists. 
    *
    * @return A list of {@link Content} objects. LinphoneContent  A list of {@link
    * Content} set if exists, null otherwise. 
    */
    public Content[] getCustomContents();

    /**
    * Indicate whether sending of early media was enabled. 
    *
    * @return A boolean value telling whether sending of early media was enabled. 
    */
    public boolean earlyMediaSendingEnabled();

    /**
    * Enable sending of real early media (during outgoing calls). 
    *
    * @param enabled A boolean value telling whether to enable early media sending or
    * not. 
    */
    public void enableEarlyMediaSending(boolean enabled);

    /**
    * Tell whether the call is part of the locally managed conference. 
    *
    * Warning: If a conference server is used to manage conferences, that function
    * does not return true even if the conference is running. If you want to test
    * whether the conference is running, you should test whether {@link
    * Core#getConference} return a non-null pointer. 
    * 
    * @return A boolean value telling whether the call is part of the locally managed
    * conference. 
    */
    public boolean getLocalConferenceMode();

    /**
    * Tell whether the call has been configured in low bandwidth mode or not. 
    *
    * This mode can be automatically discovered thanks to a stun server when
    * activate_edge_workarounds=1 in section [net] of configuration file. An
    * application that would have reliable way to know network capacity may not use
    * activate_edge_workarounds=1 but instead manually configure low bandwidth mode
    * with {@link CallParams#enableLowBandwidth}. When enabled, this param may
    * transform a call request with video in audio only mode. 
    * 
    * @return A boolean value telling whether the low bandwidth mode has been
    * configured/detected. 
    */
    public boolean lowBandwidthEnabled();

    /**
    * Indicate low bandwith mode. 
    *
    * Configuring a call to low bandwidth mode will result in the core to activate
    * several settings for the call in order to ensure that bitrate usage is lowered
    * to the minimum possible. Typically, ptime (packetization time) will be
    * increased, audio codec's output bitrate will be targetted to 20kbit/s provided
    * that it is achievable by the codec selected after SDP handshake. Video is
    * automatically disabled. 
    * 
    * @param enabled A boolean value telling whether to activate the low bandwidth
    * mode or not. 
    */
    public void enableLowBandwidth(boolean enabled);

    /**
    * Get the kind of media encryption selected for the call. 
    *
    * @return The kind of media encryption selected for the call. 
    */
    public MediaEncryption getMediaEncryption();

    /**
    * Set requested media encryption for a call. 
    *
    * @param enc The media encryption to use for the call. 
    */
    public void setMediaEncryption(MediaEncryption enc);

    /**
    * Get requested level of privacy for the call. 
    *
    * @return The privacy mode used for the call. 
    */
    public int getPrivacy();

    /**
    * Set requested level of privacy for the call. 
    *
    *  javascript  
    * 
    * @param privacy The privacy mode to used for the call. 
    */
    public void setPrivacy(int privacy);

    /**
    * Use to get real time text following rfc4103. 
    *
    * @return returns true if call rtt is activated. 
    */
    public boolean realtimeTextEnabled();

    /**
    * Use to enable real time text following rfc4103. 
    *
    * If enabled, outgoing calls put a m=text line in SDP offer . 
    * 
    * @param yesno if yes, subsequent outgoing calls will propose rtt 
    */
    public int enableRealtimeText(boolean yesno);

    /**
    * Use to get keep alive interval of real time text following rfc4103. 
    *
    * @return returns keep alive interval of real time text. 
    */
    public int getRealtimeTextKeepaliveInterval();

    /**
    * Use to set keep alive interval for real time text following rfc4103. 
    *
    * @param interval The keep alive interval for real time text, 25000 by default. 
    */
    public void setRealtimeTextKeepaliveInterval(int interval);

    /**
    * Get the framerate of the video that is received. 
    *
    * @return The actual received framerate in frames per seconds, 0 if not
    * available. 
    */
    public float getReceivedFramerate();

    /**
    * Get the definition of the received video. 
    *
    * @return The received {@link VideoDefinition} 
    */
    public VideoDefinition getReceivedVideoDefinition();

    /**
    * Get the path for the audio recording of the call. 
    *
    * @return The path to the audio recording of the call. 
    */
    public String getRecordFile();

    /**
    * Enable recording of the call. 
    *
    * This function must be used before the call parameters are assigned to the call.
    * The call recording can be started and paused after the call is established with
    * {@link Call#startRecording} and linphone_call_pause_recording(). 
    * 
    * @param path A string containing the path and filename of the file where
    * audio/video streams are to be written. The filename must have either .mkv or
    * .wav extention. The video stream will be written only if a MKV file is given. 
    */
    public void setRecordFile(String path);

    /**
    * Indicates whether RTP bundle mode (also known as Media Multiplexing) is
    * enabled. 
    *
    * See https://tools.ietf.org/html/draft-ietf-mmusic-sdp-bundle-negotiation-54 for
    * more information. 
    * 
    * @return a boolean indicating the enablement of rtp bundle mode. 
    */
    public boolean rtpBundleEnabled();

    /**
    * Enables or disables RTP bundle mode (Media Multiplexing). 
    *
    * See https://tools.ietf.org/html/draft-ietf-mmusic-sdp-bundle-negotiation-54 for
    * more information about the feature. When enabled, liblinphone will try to
    * negociate the use of a single port for all streams. It automatically enables
    * rtcp-mux. 
    * 
    * @param value a boolean to indicate whether the feature is to be enabled. 
    */
    public void enableRtpBundle(boolean value);

    /**
    * Get the RTP profile being used. 
    *
    * @return The RTP profile. 
    */
    public String getRtpProfile();

    /**
    * Get the framerate of the video that is sent. 
    *
    * @return The actual sent framerate in frames per seconds, 0 if not available. 
    */
    public float getSentFramerate();

    /**
    * Get the definition of the sent video. 
    *
    * @return The sent {@link VideoDefinition} 
    */
    public VideoDefinition getSentVideoDefinition();

    /**
    * Get the session name of the media session (ie in SDP). 
    *
    * Subject from the SIP message can be retrieved using {@link
    * CallParams#getCustomHeader} and is different. 
    * 
    * @return The session name of the media session. 
    */
    public String getSessionName();

    /**
    * Set the session name of the media session (ie in SDP). 
    *
    * Subject from the SIP message (which is different) can be set using
    * linphone_call_params_set_custom_header(). 
    * 
    * @param name The session name to be used. 
    */
    public void setSessionName(String name);

    /**
    * Get the audio payload type that has been selected by a call. 
    *
    * @return The selected payload type. null is returned if no audio payload type
    * has been seleced by the call. 
    */
    public PayloadType getUsedAudioPayloadType();

    /**
    * Get the text payload type that has been selected by a call. 
    *
    * @return The selected payload type. null is returned if no text payload type has
    * been seleced by the call. 
    */
    public PayloadType getUsedTextPayloadType();

    /**
    * Get the video payload type that has been selected by a call. 
    *
    * @return The selected payload type. null is returned if no video payload type
    * has been seleced by the call. 
    */
    public PayloadType getUsedVideoPayloadType();

    /**
    * Get the video stream direction. 
    *
    * @return The video stream direction associated with the call params. 
    */
    public MediaDirection getVideoDirection();

    /**
    * Set the video stream direction. 
    *
    * @param dir The video stream direction associated with this call params. 
    */
    public void setVideoDirection(MediaDirection dir);

    /**
    * Tell whether video is enabled or not. 
    *
    * @return A boolean value telling whether video is enabled or not. 
    */
    public boolean videoEnabled();

    /**
    * Enable video stream. 
    *
    * @param enabled A boolean value telling whether to enable video or not. 
    */
    public void enableVideo(boolean enabled);

    /**
    * Use to get multicast state of video stream. 
    *
    * @return true if subsequent calls will propose multicast ip set by
    * linphone_core_set_video_multicast_addr 
    */
    public boolean videoMulticastEnabled();

    /**
    * Use to enable multicast rtp for video stream. 
    *
    * If enabled, outgoing calls put a multicast address from
    * linphone_core_get_video_multicast_addr into video cline. In case of outgoing
    * call video stream is sent to this multicast address.  For incoming calls
    * behavior is unchanged. 
    * 
    * @param yesno if yes, subsequent outgoing calls will propose multicast ip set by
    * linphone_core_set_video_multicast_addr 
    */
    public void enableVideoMulticast(boolean yesno);

    /**
    * Adds a {@link Content} to be added to the INVITE SDP. 
    *
    * @param content The {@link Content} to be added. 
    */
    public void addCustomContent(Content content);

    /**
    * Add a custom SIP header in the INVITE for a call. 
    *
    * @param headerValue The content of the header to add. 
    */
    public void addCustomHeader(String headerName, String headerValue);

    /**
    * Add a custom attribute related to all the streams in the SDP exchanged within
    * SIP messages during a call. 
    *
    * @param attributeValue The content value of the attribute to add. 
    */
    public void addCustomSdpAttribute(String attributeName, String attributeValue);

    /**
    * Add a custom attribute related to a specific stream in the SDP exchanged within
    * SIP messages during a call. 
    *
    * @param attributeValue The content value of the attribute to add. 
    */
    public void addCustomSdpMediaAttribute(StreamType type, String attributeName, String attributeValue);

    /**
    * Clear the custom SDP attributes related to all the streams in the SDP exchanged
    * within SIP messages during a call. 
    *
    */
    public void clearCustomSdpAttributes();

    /**
    * Clear the custom SDP attributes related to a specific stream in the SDP
    * exchanged within SIP messages during a call. 
    *
    * @param type The type of the stream to clear the custom SDP attributes from. 
    */
    public void clearCustomSdpMediaAttributes(StreamType type);

    /**
    * Copy an existing {@link CallParams} object to a new {@link CallParams} object. 
    *
    * @return A copy of the {@link CallParams} object. 
    */
    public CallParams copy();

    /**
    * Get a custom SIP header. 
    *
    * @param headerName The name of the header to get. 
    * 
    * @return The content of the header or null if not found. 
    */
    public String getCustomHeader(String headerName);

    /**
    * Get a custom SDP attribute that is related to all the streams. 
    *
    * @param attributeName The name of the attribute to get. 
    * 
    * @return The content value of the attribute or null if not found. 
    */
    public String getCustomSdpAttribute(String attributeName);

    /**
    * Get a custom SDP attribute that is related to a specific stream. 
    *
    * @param attributeName The name of the attribute to get. 
    * 
    * @return The content value of the attribute or null if not found. 
    */
    public String getCustomSdpMediaAttribute(StreamType type, String attributeName);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class CallParamsImpl implements CallParams {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected CallParamsImpl(long ptr) {
        nativePtr = ptr;
    }


    private native void setAudioBandwidthLimit(long nativePtr, int bw);
    @Override
    synchronized public void setAudioBandwidthLimit(int bw)  {
        
        setAudioBandwidthLimit(nativePtr, bw);
    }

    private native int getAudioDirection(long nativePtr);
    @Override
    synchronized public MediaDirection getAudioDirection()  {
        
        return MediaDirection.fromInt(getAudioDirection(nativePtr));
    }

    private native void setAudioDirection(long nativePtr, int dir);
    @Override
    synchronized public void setAudioDirection(MediaDirection dir)  {
        
        setAudioDirection(nativePtr, dir.toInt());
    }

    private native boolean audioEnabled(long nativePtr);
    @Override
    synchronized public boolean audioEnabled()  {
        
        return audioEnabled(nativePtr);
    }

    private native void enableAudio(long nativePtr, boolean enabled);
    @Override
    synchronized public void enableAudio(boolean enabled)  {
        
        enableAudio(nativePtr, enabled);
    }

    private native boolean audioMulticastEnabled(long nativePtr);
    @Override
    synchronized public boolean audioMulticastEnabled()  {
        
        return audioMulticastEnabled(nativePtr);
    }

    private native void enableAudioMulticast(long nativePtr, boolean yesno);
    @Override
    synchronized public void enableAudioMulticast(boolean yesno)  {
        
        enableAudioMulticast(nativePtr, yesno);
    }

    private native Content[] getCustomContents(long nativePtr);
    @Override
    synchronized public Content[] getCustomContents()  {
        
        return getCustomContents(nativePtr);
    }

    private native boolean earlyMediaSendingEnabled(long nativePtr);
    @Override
    synchronized public boolean earlyMediaSendingEnabled()  {
        
        return earlyMediaSendingEnabled(nativePtr);
    }

    private native void enableEarlyMediaSending(long nativePtr, boolean enabled);
    @Override
    synchronized public void enableEarlyMediaSending(boolean enabled)  {
        
        enableEarlyMediaSending(nativePtr, enabled);
    }

    private native boolean getLocalConferenceMode(long nativePtr);
    @Override
    synchronized public boolean getLocalConferenceMode()  {
        
        return getLocalConferenceMode(nativePtr);
    }

    private native boolean lowBandwidthEnabled(long nativePtr);
    @Override
    synchronized public boolean lowBandwidthEnabled()  {
        
        return lowBandwidthEnabled(nativePtr);
    }

    private native void enableLowBandwidth(long nativePtr, boolean enabled);
    @Override
    synchronized public void enableLowBandwidth(boolean enabled)  {
        
        enableLowBandwidth(nativePtr, enabled);
    }

    private native int getMediaEncryption(long nativePtr);
    @Override
    synchronized public MediaEncryption getMediaEncryption()  {
        
        return MediaEncryption.fromInt(getMediaEncryption(nativePtr));
    }

    private native void setMediaEncryption(long nativePtr, int enc);
    @Override
    synchronized public void setMediaEncryption(MediaEncryption enc)  {
        
        setMediaEncryption(nativePtr, enc.toInt());
    }

    private native int getPrivacy(long nativePtr);
    @Override
    synchronized public int getPrivacy()  {
        
        return getPrivacy(nativePtr);
    }

    private native void setPrivacy(long nativePtr, int privacy);
    @Override
    synchronized public void setPrivacy(int privacy)  {
        
        setPrivacy(nativePtr, privacy);
    }

    private native boolean realtimeTextEnabled(long nativePtr);
    @Override
    synchronized public boolean realtimeTextEnabled()  {
        
        return realtimeTextEnabled(nativePtr);
    }

    private native int enableRealtimeText(long nativePtr, boolean yesno);
    @Override
    synchronized public int enableRealtimeText(boolean yesno)  {
        
        return enableRealtimeText(nativePtr, yesno);
    }

    private native int getRealtimeTextKeepaliveInterval(long nativePtr);
    @Override
    synchronized public int getRealtimeTextKeepaliveInterval()  {
        
        return getRealtimeTextKeepaliveInterval(nativePtr);
    }

    private native void setRealtimeTextKeepaliveInterval(long nativePtr, int interval);
    @Override
    synchronized public void setRealtimeTextKeepaliveInterval(int interval)  {
        
        setRealtimeTextKeepaliveInterval(nativePtr, interval);
    }

    private native float getReceivedFramerate(long nativePtr);
    @Override
    synchronized public float getReceivedFramerate()  {
        
        return getReceivedFramerate(nativePtr);
    }

    private native VideoDefinition getReceivedVideoDefinition(long nativePtr);
    @Override
    synchronized public VideoDefinition getReceivedVideoDefinition()  {
        
        return (VideoDefinition)getReceivedVideoDefinition(nativePtr);
    }

    private native String getRecordFile(long nativePtr);
    @Override
    synchronized public String getRecordFile()  {
        
        return getRecordFile(nativePtr);
    }

    private native void setRecordFile(long nativePtr, String path);
    @Override
    synchronized public void setRecordFile(String path)  {
        
        setRecordFile(nativePtr, path);
    }

    private native boolean rtpBundleEnabled(long nativePtr);
    @Override
    synchronized public boolean rtpBundleEnabled()  {
        
        return rtpBundleEnabled(nativePtr);
    }

    private native void enableRtpBundle(long nativePtr, boolean value);
    @Override
    synchronized public void enableRtpBundle(boolean value)  {
        
        enableRtpBundle(nativePtr, value);
    }

    private native String getRtpProfile(long nativePtr);
    @Override
    synchronized public String getRtpProfile()  {
        
        return getRtpProfile(nativePtr);
    }

    private native float getSentFramerate(long nativePtr);
    @Override
    synchronized public float getSentFramerate()  {
        
        return getSentFramerate(nativePtr);
    }

    private native VideoDefinition getSentVideoDefinition(long nativePtr);
    @Override
    synchronized public VideoDefinition getSentVideoDefinition()  {
        
        return (VideoDefinition)getSentVideoDefinition(nativePtr);
    }

    private native String getSessionName(long nativePtr);
    @Override
    synchronized public String getSessionName()  {
        
        return getSessionName(nativePtr);
    }

    private native void setSessionName(long nativePtr, String name);
    @Override
    synchronized public void setSessionName(String name)  {
        
        setSessionName(nativePtr, name);
    }

    private native PayloadType getUsedAudioPayloadType(long nativePtr);
    @Override
    synchronized public PayloadType getUsedAudioPayloadType()  {
        
        return (PayloadType)getUsedAudioPayloadType(nativePtr);
    }

    private native PayloadType getUsedTextPayloadType(long nativePtr);
    @Override
    synchronized public PayloadType getUsedTextPayloadType()  {
        
        return (PayloadType)getUsedTextPayloadType(nativePtr);
    }

    private native PayloadType getUsedVideoPayloadType(long nativePtr);
    @Override
    synchronized public PayloadType getUsedVideoPayloadType()  {
        
        return (PayloadType)getUsedVideoPayloadType(nativePtr);
    }

    private native int getVideoDirection(long nativePtr);
    @Override
    synchronized public MediaDirection getVideoDirection()  {
        
        return MediaDirection.fromInt(getVideoDirection(nativePtr));
    }

    private native void setVideoDirection(long nativePtr, int dir);
    @Override
    synchronized public void setVideoDirection(MediaDirection dir)  {
        
        setVideoDirection(nativePtr, dir.toInt());
    }

    private native boolean videoEnabled(long nativePtr);
    @Override
    synchronized public boolean videoEnabled()  {
        
        return videoEnabled(nativePtr);
    }

    private native void enableVideo(long nativePtr, boolean enabled);
    @Override
    synchronized public void enableVideo(boolean enabled)  {
        
        enableVideo(nativePtr, enabled);
    }

    private native boolean videoMulticastEnabled(long nativePtr);
    @Override
    synchronized public boolean videoMulticastEnabled()  {
        
        return videoMulticastEnabled(nativePtr);
    }

    private native void enableVideoMulticast(long nativePtr, boolean yesno);
    @Override
    synchronized public void enableVideoMulticast(boolean yesno)  {
        
        enableVideoMulticast(nativePtr, yesno);
    }

    private native void addCustomContent(long nativePtr, Content content);
    @Override
    synchronized public void addCustomContent(Content content)  {
        
        addCustomContent(nativePtr, content);
    }

    private native void addCustomHeader(long nativePtr, String headerName, String headerValue);
    @Override
    synchronized public void addCustomHeader(String headerName, String headerValue)  {
        
        addCustomHeader(nativePtr, headerName, headerValue);
    }

    private native void addCustomSdpAttribute(long nativePtr, String attributeName, String attributeValue);
    @Override
    synchronized public void addCustomSdpAttribute(String attributeName, String attributeValue)  {
        
        addCustomSdpAttribute(nativePtr, attributeName, attributeValue);
    }

    private native void addCustomSdpMediaAttribute(long nativePtr, int type, String attributeName, String attributeValue);
    @Override
    synchronized public void addCustomSdpMediaAttribute(StreamType type, String attributeName, String attributeValue)  {
        
        addCustomSdpMediaAttribute(nativePtr, type.toInt(), attributeName, attributeValue);
    }

    private native void clearCustomSdpAttributes(long nativePtr);
    @Override
    synchronized public void clearCustomSdpAttributes()  {
        
        clearCustomSdpAttributes(nativePtr);
    }

    private native void clearCustomSdpMediaAttributes(long nativePtr, int type);
    @Override
    synchronized public void clearCustomSdpMediaAttributes(StreamType type)  {
        
        clearCustomSdpMediaAttributes(nativePtr, type.toInt());
    }

    private native CallParams copy(long nativePtr);
    @Override
    synchronized public CallParams copy()  {
        
        return (CallParams)copy(nativePtr);
    }

    private native String getCustomHeader(long nativePtr, String headerName);
    @Override
    synchronized public String getCustomHeader(String headerName)  {
        
        return getCustomHeader(nativePtr, headerName);
    }

    private native String getCustomSdpAttribute(long nativePtr, String attributeName);
    @Override
    synchronized public String getCustomSdpAttribute(String attributeName)  {
        
        return getCustomSdpAttribute(nativePtr, attributeName);
    }

    private native String getCustomSdpMediaAttribute(long nativePtr, int type, String attributeName);
    @Override
    synchronized public String getCustomSdpMediaAttribute(StreamType type, String attributeName)  {
        
        return getCustomSdpMediaAttribute(nativePtr, type.toInt(), attributeName);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
