/*
ChatMessage.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* An chat message is the object that is sent and received through
* LinphoneChatRooms. 
*
*/
public interface ChatMessage {
    public enum Direction {
        /**
        * Incoming message. 
        *
        */
        Incoming(0),

        /**
        * Outgoing message. 
        *
        */
        Outgoing(1);

        protected final int mValue;

        private Direction (int value) {
            mValue = value;
        }

        static public Direction fromInt(int value) throws RuntimeException {
            switch(value) {
            case 0: return Incoming;
            case 1: return Outgoing;
            default:
                throw new RuntimeException("Unhandled enum value " + value + " for Direction");
            }
        }

        public int toInt() {
            return mValue;
        }
    };

    public enum State {
        /**
        * Initial state. 
        *
        */
        Idle(0),

        /**
        * Delivery in progress. 
        *
        */
        InProgress(1),

        /**
        * Message successfully delivered and acknowledged by the server. 
        *
        */
        Delivered(2),

        /**
        * Message was not delivered. 
        *
        */
        NotDelivered(3),

        /**
        * Message was received and acknowledged but cannot get file from server. 
        *
        */
        FileTransferError(4),

        /**
        * File transfer has been completed successfully. 
        *
        */
        FileTransferDone(5),

        /**
        * Message successfully delivered an acknowledged by the remote user. 
        *
        */
        DeliveredToUser(6),

        /**
        * Message successfully displayed to the remote user. 
        *
        */
        Displayed(7),

        /**
        *
        */
        FileTransferInProgress(8);

        protected final int mValue;

        private State (int value) {
            mValue = value;
        }

        static public State fromInt(int value) throws RuntimeException {
            switch(value) {
            case 0: return Idle;
            case 1: return InProgress;
            case 2: return Delivered;
            case 3: return NotDelivered;
            case 4: return FileTransferError;
            case 5: return FileTransferDone;
            case 6: return DeliveredToUser;
            case 7: return Displayed;
            case 8: return FileTransferInProgress;
            default:
                throw new RuntimeException("Unhandled enum value " + value + " for State");
            }
        }

        public int toInt() {
            return mValue;
        }
    };

    /**
    * Linphone message has an app-specific field that can store a text. 
    *
    * The application might want to use it for keeping data over restarts, like
    * thumbnail image path. 
    * 
    * @return the application-specific data or null if none has been stored. 
    */
    public String getAppdata();

    /**
    * Linphone message has an app-specific field that can store a text. 
    *
    * The application might want to use it for keeping data over restarts, like
    * thumbnail image path.
    * 
    * Invoking this function will attempt to update the message storage to reflect
    * the changeif it is enabled.
    * 
    * @param data the data to store into the message 
    */
    public void setAppdata(String data);

    /**
    * Returns the chatroom this message belongs to. 
    *
    */
    public ChatRoom getChatRoom();

    /**
    * Get the content type of a chat message. 
    *
    * @return The content type of the chat message 
    */
    public String getContentType();

    /**
    * Set the content type of a chat message. 
    *
    * This content type must match a content that is text representable, such as
    * text/plain, text/html or image/svg+xml. 
    * 
    * @param contentType The new content type of the chat message 
    */
    public void setContentType(String contentType);

    /**
    * Returns the list of contents in the message. 
    *
    * @return A list of {@link Content} objects. LinphoneContent  the list of {@link
    * Content}. 
    */
    public Content[] getContents();

    /**
    * Returns back pointer to {@link Core} object. 
    *
    */
    public Core getCore();

    /**
    * Returns the real time at which an ephemeral message expires and will be
    * deleted. 
    *
    * @return the time at which an ephemeral message expires. 0 means the message has
    * not been read. 
    */
    public long getEphemeralExpireTime();

    /**
    * Returns lifetime of an ephemeral message. 
    *
    * The lifetime is the duration after which the ephemeral message will disappear
    * once viewed. 
    * 
    * @return the lifetime of an ephemeral message, by default 86400s. 
    */
    public long getEphemeralLifetime();

    /**
    * Get full details about delivery error of a chat message. 
    *
    * @return a {@link ErrorInfo} describing the details. 
    */
    public ErrorInfo getErrorInfo();

    /**
    * Linphone message can carry external body as defined by rfc2017. 
    *
    * @return external body url or null if not present. 
    */
    public String getExternalBodyUrl();

    /**
    * Linphone message can carry external body as defined by rfc2017. 
    *
    * @param url ex: access-type=URL; URL="http://www.foo.com/file" 
    */
    public void setExternalBodyUrl(String url);

    /**
    * Get the path to the file to read from or write to during the file transfer. 
    *
    * @return The path to the file to use for the file transfer. 
    * 
    * @deprecated use {@link Content#getFilePath} instead. 
    */
    @Deprecated
    public String getFileTransferFilepath();

    /**
    * Set the path to the file to read from or write to during the file transfer. 
    *
    * @param filepath The path to the file to use for the file transfer. 
    * 
    * @deprecated use {@link Content#setFilePath} instead. 
    */
    @Deprecated
    public void setFileTransferFilepath(String filepath);

    /**
    * Get the file_transfer_information (used by call backs to recover informations
    * during a rcs file transfer) 
    *
    * @return a pointer to the {@link Content} structure or null if not present. 
    */
    public Content getFileTransferInformation();

    /**
    * Gets the forward info if available as a string. 
    *
    * @return the original sender of the message if it has been forwarded, null
    * otherwise 
    */
    public String getForwardInfo();

    /**
    * Get origin of the message. 
    *
    * @return {@link Address} 
    */
    public Address getFromAddress();

    /**
    * Returns true if the chat message is an ephemeral message. 
    *
    * An ephemeral message will automatically disappear from the recipient's screen
    * after the message has been viewed. 
    * 
    * @return true if it is an ephemeral message, false otherwise 
    */
    public boolean isEphemeral();

    /**
    * Return whether or not a chat message is a file transfer. 
    *
    * @return Whether or not the message is a file transfer 
    */
    public boolean isFileTransfer();

    /**
    * Gets whether or not a file is currently being downloaded or uploaded. 
    *
    * @return true if download or upload is in progress, false otherwise 
    */
    public boolean isFileTransferInProgress();

    /**
    * Returns true if the chat message is a forward message. 
    *
    * @return true if it is a forward message, false otherwise 
    */
    public boolean isForward();

    /**
    * Returns true if the message has been sent, returns true if the message has been
    * received. 
    *
    */
    public boolean isOutgoing();

    /**
    * Returns true if the message has been read, otherwise returns true. 
    *
    */
    public boolean isRead();

    /**
    * Get if the message was encrypted when transfered. 
    *
    * @return whether the message was encrypted when transfered or not 
    */
    public boolean isSecured();

    /**
    * Return whether or not a chat message is a text. 
    *
    * @return Whether or not the message is a text 
    */
    public boolean isText();

    /**
    * Returns the origin address of a message if it was a outgoing message, or the
    * destination address if it was an incoming message. 
    *
    * @return {@link Address} 
    */
    public Address getLocalAddress();

    /**
    * Get the message identifier. 
    *
    * It is used to identify a message so that it can be notified as delivered and/or
    * displayed. 
    * 
    * @return The message identifier. 
    */
    public String getMessageId();

    /**
    * Get the state of the message. 
    *
    * @return LinphoneChatMessageState 
    */
    public ChatMessage.State getState();

    /**
    * Get text part of this message. 
    *
    * @return text or null if no text. 
    * 
    * @deprecated use getTextContent() instead 
    */
    @Deprecated
    public String getText();

    /**
    * Gets the text content if available as a string. 
    *
    * @return the {@link Content} buffer if available, null otherwise 
    */
    public String getTextContent();

    /**
    * Get the time the message was sent. 
    *
    */
    public long getTime();

    /**
    * Get destination of the message. 
    *
    * @return {@link Address} 
    */
    public Address getToAddress();

    /**
    * Get if a chat message is to be stored. 
    *
    * @return Whether or not the message is to be stored 
    */
    public boolean getToBeStored();

    /**
    * Set if a chat message is to be stored. 
    *
    * This content type must match a content that is text representable, such as
    * text/plain, text/html or image/svg+xml. 
    * 
    * @param toBeStored Whether or not the chat message is to be stored 
    */
    public void setToBeStored(boolean toBeStored);

    /**
    * Add custom headers to the message. 
    *
    * @param headerValue header value 
    */
    public void addCustomHeader(String headerName, String headerValue);

    /**
    * Adds a file content to the ChatMessage. 
    *
    * @param cContent {@link Content} object to add. 
    */
    public void addFileContent(Content cContent);

    /**
    * Adds a text content to the ChatMessage. 
    *
    * @param text The text to add to the message. 
    */
    public void addTextContent(String text);

    /**
    * Cancel an ongoing file transfer attached to this message. 
    *
    * (upload or download) 
    */
    public void cancelFileTransfer();

    /**
    * Start the download of the {@link Content} referenced in the {@link ChatMessage}
    * from remote server. 
    *
    * @param cContent {@link Content} object to download. 
    */
    public boolean downloadContent(Content cContent);

    /**
    * Start the download of the file referenced in a {@link ChatMessage} from remote
    * server. 
    *
    * @deprecated Use {@link ChatMessage#downloadContent} instead 
    */
    @Deprecated
    public boolean downloadFile();

    /**
    * Retrieve a custom header value given its name. 
    *
    * @param headerName header name searched 
    */
    public String getCustomHeader(String headerName);

    /**
    * Gets the list of participants for which the imdn state has reached the
    * specified state and the time at which they did. 
    *
    * @param state The LinphoneChatMessageState the imdn have reached (only use
    * LinphoneChatMessageStateDelivered, LinphoneChatMessageStateDeliveredToUser,
    * LinphoneChatMessageStateDisplayed and LinphoneChatMessageStateNotDelivered) 
    * 
    * @return A list of {@link ParticipantImdnState} objects.
    * LinphoneParticipantImdnState  The objects inside the list are freshly allocated
    * with a reference counter equal to one, so they need to be freed on list
    * destruction with bctbx_list_free_with_data() for instance.   
    */
    public ParticipantImdnState[] getParticipantsByImdnState(ChatMessage.State state);

    /**
    * Returns true if the chat message has a text content. 
    *
    * @return true if it has one, false otherwise 
    */
    public boolean hasTextContent();

    /**
    * Fulfill a chat message char by char. 
    *
    * Message linked to a Real Time Text Call send char in realtime following RFC
    * 4103/T.140 To commit a message, use linphone_chat_room_send_message 
    * 
    * @param character T.140 char 
    * 
    * @return 0 if succeed. 
    */
    public int putChar(int character);

    /**
    * Removes a content from the ChatMessage. 
    *
    * @param content the {@link Content} object to remove. 
    */
    public void removeContent(Content content);

    /**
    * Removes a custom header from the message. 
    *
    * @param headerName name of the header to remove 
    */
    public void removeCustomHeader(String headerName);

    /**
    * Resend a chat message if it is in the 'not delivered' state for whatever
    * reason. 
    *
    * Note: Unlike linphone_chat_message_resend, that function only takes a reference
    * on the {@link ChatMessage} instead of totaly takes ownership on it. Thus, the
    * {@link ChatMessage} object must be released by the API user after calling that
    * function.
    */
    public void resend();

    /**
    * Send a chat message. 
    *
    */
    public void send();

    /**
    */
    public void addListener(ChatMessageListener listener);

    /**
    */
    public void removeListener(ChatMessageListener listener);

    /**
    */
    public void setListener(ChatMessageListener listener);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class ChatMessageImpl implements ChatMessage {

    protected long nativePtr = 0;
    protected Object userData = null;
    protected Core core = null;

    protected ChatMessageImpl(long ptr) {
        nativePtr = ptr;
        core = getCore();
    }


    private native String getAppdata(long nativePtr);
    @Override
    synchronized public String getAppdata()  {
        synchronized(core) { 
        return getAppdata(nativePtr);
        }
    }

    private native void setAppdata(long nativePtr, String data);
    @Override
    synchronized public void setAppdata(String data)  {
        synchronized(core) { 
        setAppdata(nativePtr, data);
        }
    }

    private native ChatRoom getChatRoom(long nativePtr);
    @Override
    synchronized public ChatRoom getChatRoom()  {
        synchronized(core) { 
        return (ChatRoom)getChatRoom(nativePtr);
        }
    }

    private native String getContentType(long nativePtr);
    @Override
    synchronized public String getContentType()  {
        synchronized(core) { 
        return getContentType(nativePtr);
        }
    }

    private native void setContentType(long nativePtr, String contentType);
    @Override
    synchronized public void setContentType(String contentType)  {
        synchronized(core) { 
        setContentType(nativePtr, contentType);
        }
    }

    private native Content[] getContents(long nativePtr);
    @Override
    synchronized public Content[] getContents()  {
        synchronized(core) { 
        return getContents(nativePtr);
        }
    }

    private native Core getCore(long nativePtr);
    @Override
    synchronized public Core getCore()  {
        
        return (Core)getCore(nativePtr);
    }

    private native long getEphemeralExpireTime(long nativePtr);
    @Override
    synchronized public long getEphemeralExpireTime()  {
        synchronized(core) { 
        return getEphemeralExpireTime(nativePtr);
        }
    }

    private native long getEphemeralLifetime(long nativePtr);
    @Override
    synchronized public long getEphemeralLifetime()  {
        synchronized(core) { 
        return getEphemeralLifetime(nativePtr);
        }
    }

    private native ErrorInfo getErrorInfo(long nativePtr);
    @Override
    synchronized public ErrorInfo getErrorInfo()  {
        synchronized(core) { 
        return (ErrorInfo)getErrorInfo(nativePtr);
        }
    }

    private native String getExternalBodyUrl(long nativePtr);
    @Override
    synchronized public String getExternalBodyUrl()  {
        synchronized(core) { 
        return getExternalBodyUrl(nativePtr);
        }
    }

    private native void setExternalBodyUrl(long nativePtr, String url);
    @Override
    synchronized public void setExternalBodyUrl(String url)  {
        synchronized(core) { 
        setExternalBodyUrl(nativePtr, url);
        }
    }

    private native String getFileTransferFilepath(long nativePtr);
    @Override
    synchronized public String getFileTransferFilepath()  {
        synchronized(core) { 
        return getFileTransferFilepath(nativePtr);
        }
    }

    private native void setFileTransferFilepath(long nativePtr, String filepath);
    @Override
    synchronized public void setFileTransferFilepath(String filepath)  {
        synchronized(core) { 
        setFileTransferFilepath(nativePtr, filepath);
        }
    }

    private native Content getFileTransferInformation(long nativePtr);
    @Override
    synchronized public Content getFileTransferInformation()  {
        synchronized(core) { 
        return (Content)getFileTransferInformation(nativePtr);
        }
    }

    private native String getForwardInfo(long nativePtr);
    @Override
    synchronized public String getForwardInfo()  {
        synchronized(core) { 
        return getForwardInfo(nativePtr);
        }
    }

    private native Address getFromAddress(long nativePtr);
    @Override
    synchronized public Address getFromAddress()  {
        synchronized(core) { 
        return (Address)getFromAddress(nativePtr);
        }
    }

    private native boolean isEphemeral(long nativePtr);
    @Override
    synchronized public boolean isEphemeral()  {
        synchronized(core) { 
        return isEphemeral(nativePtr);
        }
    }

    private native boolean isFileTransfer(long nativePtr);
    @Override
    synchronized public boolean isFileTransfer()  {
        synchronized(core) { 
        return isFileTransfer(nativePtr);
        }
    }

    private native boolean isFileTransferInProgress(long nativePtr);
    @Override
    synchronized public boolean isFileTransferInProgress()  {
        synchronized(core) { 
        return isFileTransferInProgress(nativePtr);
        }
    }

    private native boolean isForward(long nativePtr);
    @Override
    synchronized public boolean isForward()  {
        synchronized(core) { 
        return isForward(nativePtr);
        }
    }

    private native boolean isOutgoing(long nativePtr);
    @Override
    synchronized public boolean isOutgoing()  {
        synchronized(core) { 
        return isOutgoing(nativePtr);
        }
    }

    private native boolean isRead(long nativePtr);
    @Override
    synchronized public boolean isRead()  {
        synchronized(core) { 
        return isRead(nativePtr);
        }
    }

    private native boolean isSecured(long nativePtr);
    @Override
    synchronized public boolean isSecured()  {
        synchronized(core) { 
        return isSecured(nativePtr);
        }
    }

    private native boolean isText(long nativePtr);
    @Override
    synchronized public boolean isText()  {
        synchronized(core) { 
        return isText(nativePtr);
        }
    }

    private native Address getLocalAddress(long nativePtr);
    @Override
    synchronized public Address getLocalAddress()  {
        synchronized(core) { 
        return (Address)getLocalAddress(nativePtr);
        }
    }

    private native String getMessageId(long nativePtr);
    @Override
    synchronized public String getMessageId()  {
        synchronized(core) { 
        return getMessageId(nativePtr);
        }
    }

    private native int getState(long nativePtr);
    @Override
    synchronized public ChatMessage.State getState()  {
        synchronized(core) { 
        return ChatMessage.State.fromInt(getState(nativePtr));
        }
    }

    private native String getText(long nativePtr);
    @Override
    synchronized public String getText()  {
        synchronized(core) { 
        return getText(nativePtr);
        }
    }

    private native String getTextContent(long nativePtr);
    @Override
    synchronized public String getTextContent()  {
        synchronized(core) { 
        return getTextContent(nativePtr);
        }
    }

    private native long getTime(long nativePtr);
    @Override
    synchronized public long getTime()  {
        synchronized(core) { 
        return getTime(nativePtr);
        }
    }

    private native Address getToAddress(long nativePtr);
    @Override
    synchronized public Address getToAddress()  {
        synchronized(core) { 
        return (Address)getToAddress(nativePtr);
        }
    }

    private native boolean getToBeStored(long nativePtr);
    @Override
    synchronized public boolean getToBeStored()  {
        synchronized(core) { 
        return getToBeStored(nativePtr);
        }
    }

    private native void setToBeStored(long nativePtr, boolean toBeStored);
    @Override
    synchronized public void setToBeStored(boolean toBeStored)  {
        synchronized(core) { 
        setToBeStored(nativePtr, toBeStored);
        }
    }

    private native void addCustomHeader(long nativePtr, String headerName, String headerValue);
    @Override
    synchronized public void addCustomHeader(String headerName, String headerValue)  {
        synchronized(core) { 
        addCustomHeader(nativePtr, headerName, headerValue);
        }
    }

    private native void addFileContent(long nativePtr, Content cContent);
    @Override
    synchronized public void addFileContent(Content cContent)  {
        synchronized(core) { 
        addFileContent(nativePtr, cContent);
        }
    }

    private native void addTextContent(long nativePtr, String text);
    @Override
    synchronized public void addTextContent(String text)  {
        synchronized(core) { 
        addTextContent(nativePtr, text);
        }
    }

    private native void cancelFileTransfer(long nativePtr);
    @Override
    synchronized public void cancelFileTransfer()  {
        synchronized(core) { 
        cancelFileTransfer(nativePtr);
        }
    }

    private native boolean downloadContent(long nativePtr, Content cContent);
    @Override
    synchronized public boolean downloadContent(Content cContent)  {
        synchronized(core) { 
        return downloadContent(nativePtr, cContent);
        }
    }

    private native boolean downloadFile(long nativePtr);
    @Override
    synchronized public boolean downloadFile()  {
        synchronized(core) { 
        return downloadFile(nativePtr);
        }
    }

    private native String getCustomHeader(long nativePtr, String headerName);
    @Override
    synchronized public String getCustomHeader(String headerName)  {
        synchronized(core) { 
        return getCustomHeader(nativePtr, headerName);
        }
    }

    private native ParticipantImdnState[] getParticipantsByImdnState(long nativePtr, int state);
    @Override
    synchronized public ParticipantImdnState[] getParticipantsByImdnState(ChatMessage.State state)  {
        synchronized(core) { 
        return getParticipantsByImdnState(nativePtr, state.toInt());
        }
    }

    private native boolean hasTextContent(long nativePtr);
    @Override
    synchronized public boolean hasTextContent()  {
        synchronized(core) { 
        return hasTextContent(nativePtr);
        }
    }

    private native int putChar(long nativePtr, int character);
    @Override
    synchronized public int putChar(int character)  {
        synchronized(core) { 
        return putChar(nativePtr, character);
        }
    }

    private native void removeContent(long nativePtr, Content content);
    @Override
    synchronized public void removeContent(Content content)  {
        synchronized(core) { 
        removeContent(nativePtr, content);
        }
    }

    private native void removeCustomHeader(long nativePtr, String headerName);
    @Override
    synchronized public void removeCustomHeader(String headerName)  {
        synchronized(core) { 
        removeCustomHeader(nativePtr, headerName);
        }
    }

    private native void resend2(long nativePtr);
    @Override
    synchronized public void resend()  {
        synchronized(core) { 
        resend2(nativePtr);
        }
    }

    private native void send(long nativePtr);
    @Override
    synchronized public void send()  {
        synchronized(core) { 
        send(nativePtr);
        }
    }

    private native void addListener(long nativePtr, ChatMessageListener listener);
    @Override
    synchronized public void addListener(ChatMessageListener listener)  {
        
        addListener(nativePtr, listener);
    }

    private native void removeListener(long nativePtr, ChatMessageListener listener);
    @Override
    synchronized public void removeListener(ChatMessageListener listener)  {
        
        removeListener(nativePtr, listener);
    }

    private native void setListener(long nativePtr, ChatMessageListener listener);
    @Override
    synchronized public void setListener(ChatMessageListener listener)  {
        
        setListener(nativePtr, listener);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
