/*
ChatMessageListener.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;


/**
* An object to handle the callbacks for the handling a LinphoneChatMessage
* objects. 
*
*/
public interface ChatMessageListener {
    /**
    * Call back used to notify participant IMDN state. 
    *
    * @param state LinphoneParticipantImdnState 
    */
    public void onParticipantImdnStateChanged(ChatMessage msg, ParticipantImdnState state);

    /**
    * File transfer receive callback prototype. 
    *
    * This function is called by the core upon an incoming File transfer is started.
    * This function may be call several time for the same file in case of large file. 
    * 
    * @param buffer LinphoneBuffer holding the received data. Empty buffer means end
    * of file. 
    */
    public void onFileTransferRecv(ChatMessage msg, Content content, Buffer buffer);

    /**
    * File transfer send callback prototype. 
    *
    * This function is called by the core when an outgoing file transfer is started.
    * This function is called until size is set to 0. 
    * 
    * @param size the number of bytes expected by the framework 
    * 
    * @return A LinphoneBuffer object holding the data written by the application. An
    * empty buffer means end of file. 
    */
    public Buffer onFileTransferSend(ChatMessage msg, Content content, int offset, int size);

    /**
    * Callback used to notify an ephemeral message that its lifespan before
    * disappearing has started to decrease. 
    *
    * This callback is called when the ephemeral message is read by the receiver. 
    * 
    * @param msg LinphoneChatMessage object 
    */
    public void onEphemeralMessageTimerStarted(ChatMessage msg);

    /**
    * File transfer progress indication callback prototype. 
    *
    * @param total The total number of bytes to be sent/received. 
    */
    public void onFileTransferProgressIndication(ChatMessage msg, Content content, int offset, int total);

    /**
    * Call back used to notify message delivery status. 
    *
    * @param msg LinphoneChatMessage object 
    */
    public void onMsgStateChanged(ChatMessage msg, ChatMessage.State state);

    /**
    * Call back used to notify ephemeral message is deleted. 
    *
    * @param msg LinphoneChatMessage object 
    */
    public void onEphemeralMessageDeleted(ChatMessage msg);

}