/*
ChatRoom.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* A chat room is the place where text messages are exchanged. 
*
* Can be created by {@link Core#createChatRoom}. 
*/
public interface ChatRoom {
    public enum State {
        /**
        * Initial state. 
        *
        */
        None(0),

        /**
        * Chat room is now instantiated on local. 
        *
        */
        Instantiated(1),

        /**
        * One creation request was sent to the server. 
        *
        */
        CreationPending(2),

        /**
        * Chat room was created on the server. 
        *
        */
        Created(3),

        /**
        * Chat room creation failed. 
        *
        */
        CreationFailed(4),

        /**
        * Wait for chat room termination. 
        *
        */
        TerminationPending(5),

        /**
        * Chat room exists on server but not in local. 
        *
        */
        Terminated(6),

        /**
        * The chat room termination failed. 
        *
        */
        TerminationFailed(7),

        /**
        * Chat room was deleted on the server. 
        *
        */
        Deleted(8);

        protected final int mValue;

        private State (int value) {
            mValue = value;
        }

        static public State fromInt(int value) throws RuntimeException {
            switch(value) {
            case 0: return None;
            case 1: return Instantiated;
            case 2: return CreationPending;
            case 3: return Created;
            case 4: return CreationFailed;
            case 5: return TerminationPending;
            case 6: return Terminated;
            case 7: return TerminationFailed;
            case 8: return Deleted;
            default:
                throw new RuntimeException("Unhandled enum value " + value + " for State");
            }
        }

        public int toInt() {
            return mValue;
        }
    };

    /**
    * get Curent Call associated to this chatroom if any To commit a message, use
    * linphone_chat_room_send_message 
    *
    * @return {@link Call} or null. 
    */
    public Call getCall();

    /**
    * Get the capabilities of a chat room. 
    *
    * @return The capabilities of the chat room 
    */
    public int getCapabilities();

    /**
    * When realtime text is enabled linphone_call_params_realtime_text_enabled,
    * LinphoneCoreIsComposingReceivedCb is call everytime a char is received from
    * peer. 
    *
    * At the end of remote typing a regular {@link ChatMessage} is received with
    * committed data from LinphoneCoreMessageReceivedCb. 
    * 
    * @return RFC 4103/T.140 char 
    */
    public int getChar();

    /**
    * Gets the list of participants that are currently composing. 
    *
    * @return A list of {@link Address} objects. LinphoneAddress  list of addresses
    * that are in the is_composing state 
    */
    public Address[] getComposingAddresses();

    /**
    * Get the conference address of the chat room. 
    *
    * @return The conference address of the chat room or null if this type of chat
    * room is not conference based 
    */
    public Address getConferenceAddress();

    /**
    * Set the conference address of a group chat room. 
    *
    * This function needs to be called from the
    * LinphoneChatRoomCbsConferenceAddressGenerationCb callback and only there. This
    * function is meaningful only for server implementation of chatroom, and shall
    * not by used by client applications. 
    * 
    * @param confAddr The conference address to be used by the group chat room 
    */
    public void setConferenceAddress(Address confAddr);

    /**
    * Returns back pointer to {@link Core} object. 
    *
    */
    public Core getCore();

    /**
    * Returns current parameters associated with the chat room. 
    *
    * This is typically the parameters passed at chat room creation to {@link
    * Core#createChatRoom} or some default parameters if no {@link ChatRoomParams}
    * was explicitely passed during chat room creation. 
    * 
    * @return the chat room current parameters. 
    */
    public ChatRoomParams getCurrentParams();

    /**
    * Returns whether or not the ephemeral message feature is enabled in the chat
    * room. 
    *
    * @return true if ephemeral is enabled, false otherwise. 
    */
    public boolean ephemeralEnabled();

    /**
    * Enable or disable the ephemeral message feature in the chat room. 
    *
    * Works only for conference based chat room. An ephemeral message will
    * automatically disappear from the recipient's screen after the message has been
    * viewed. 
    * 
    * @param ephem  if the ephemeral message feature is enabled, true otherwise. 
    */
    public void enableEphemeral(boolean ephem);

    /**
    * Get lifetime (in seconds) for all new ephemeral messages in the chat room. 
    *
    * After the message is read, it will be deleted after "time" seconds. 
    * 
    * @return the ephemeral lifetime (in secoonds) 
    */
    public long getEphemeralLifetime();

    /**
    * Set lifetime (in seconds) for all new ephemral messages in the chat room. 
    *
    * After the message is read, it will be deleted after "time" seconds. 
    * 
    * @param time The ephemral lifetime, default 24h, 86400s 
    */
    public void setEphemeralLifetime(long time);

    /**
    * Gets the number of events in a chat room. 
    *
    * @return the number of events. 
    */
    public int getHistoryEventsSize();

    /**
    * Gets the number of messages in a chat room. 
    *
    * @return the number of messages. 
    */
    public int getHistorySize();

    /**
    * Returns whether or not a {@link ChatRoom} has at least one {@link ChatMessage}
    * or not. 
    *
    * @return true if there are no {@link ChatMessage}, false otherwise. 
    */
    public boolean isEmpty();

    /**
    * Tells whether the remote is currently composing a message. 
    *
    * @return  if the remote is currently composing a message, true otherwise. 
    */
    public boolean isRemoteComposing();

    /**
    * Gets the last chat message sent or received in this chat room. 
    *
    * @return the latest {@link ChatMessage} 
    */
    public ChatMessage getLastMessageInHistory();

    /**
    * Return the last updated time for the chat room. 
    *
    * @return the last updated time 
    */
    public long getLastUpdateTime();

    /**
    * get local address associated to  this {@link ChatRoom} 
    *
    * @return {@link Address} local address 
    */
    public Address getLocalAddress();

    /**
    * Get the participant representing myself in the chat room. 
    *
    * @return The participant representing myself in the conference. 
    */
    public Participant getMe();

    /**
    * Get the number of participants in the chat room (that is without ourselves). 
    *
    * @return The number of participants in the chat room 
    */
    public int getNbParticipants();

    /**
    * Get the list of participants of a chat room. 
    *
    * @return A list of LinphoneParticipant objects. LinphoneParticipant  
    */
    public Participant[] getParticipants();

    /**
    * get peer address associated to  this {@link ChatRoom} 
    *
    * @return {@link Address} peer address 
    */
    public Address getPeerAddress();

    /**
    * Get the security level of a chat room. 
    *
    * @return The security level of the chat room 
    */
    public ChatRoomSecurityLevel getSecurityLevel();

    /**
    * Get the state of the chat room. 
    *
    * @return The state of the chat room 
    */
    public ChatRoom.State getState();

    /**
    * Get the subject of a chat room. 
    *
    * @return The subject of the chat room 
    */
    public String getSubject();

    /**
    * Set the subject of a chat room. 
    *
    * @param subject The new subject to set for the chat room 
    */
    public void setSubject(String subject);

    /**
    * Gets the number of unread messages in the chatroom. 
    *
    * @return the number of unread messages. 
    */
    public int getUnreadMessagesCount();

    /**
    * Add a participant to a chat room. 
    *
    * This may fail if this type of chat room does not handle participants. Use
    * {@link ChatRoom#canHandleParticipants} to know if this chat room handles
    * participants. 
    * 
    * @param addr The address of the participant to add to the chat room 
    */
    public void addParticipant(Address addr);

    /**
    * Add several participants to a chat room at once. 
    *
    * This may fail if this type of chat room does not handle participants. Use
    * {@link ChatRoom#canHandleParticipants} to know if this chat room handles
    * participants. 
    * 
    * @param addresses A list of {@link Address} objects. LinphoneAddress  
    * 
    * @return True if everything is OK, False otherwise 
    */
    public boolean addParticipants(Address[] addresses);

    /**
    * Allow cpim on a basic chat room   . 
    *
    */
    public void allowCpim();

    /**
    * Allow multipart on a basic chat room   . 
    *
    */
    public void allowMultipart();

    /**
    * Tells whether a chat room is able to handle participants. 
    *
    * @return A boolean value telling whether the chat room can handle participants
    * or not 
    */
    public boolean canHandleParticipants();

    /**
    * Notifies the destination of the chat message being composed that the user is
    * typing a new message. 
    *
    */
    public void compose();

    /**
    * Creates an empty message attached to a dedicated chat room. 
    *
    * @return a new {@link ChatMessage} 
    */
    public ChatMessage createEmptyMessage();

    /**
    * Creates a message attached to a dedicated chat room with a particular content. 
    *
    * Use linphone_chat_room_send_message to initiate the transfer 
    * 
    * @param initialContent {@link Content} initial content.
    * LinphoneCoreVTable.file_transfer_send is invoked later to notify file transfer
    * progress and collect next chunk of the message if LinphoneContent.data is null. 
    * 
    * @return a new {@link ChatMessage} 
    */
    public ChatMessage createFileTransferMessage(Content initialContent);

    /**
    * Creates a forward message attached to a dedicated chat room with a particular
    * message. 
    *
    * @param msg {@link ChatMessage} message to be forwarded. 
    * 
    * @return a new {@link ChatMessage} 
    */
    public ChatMessage createForwardMessage(ChatMessage msg);

    /**
    * Creates a message attached to a dedicated chat room. 
    *
    * @param message text message, null if absent. 
    * 
    * @return a new {@link ChatMessage} 
    */
    public ChatMessage createMessage(String message);

    /**
    * Delete all messages from the history. 
    *
    */
    public void deleteHistory();

    /**
    * Delete a message from the chat room history. 
    *
    * @param msg The {@link ChatMessage} object to remove. 
    */
    public void deleteMessage(ChatMessage msg);

    /**
    * Uses linphone spec to check if all participants support ephemeral messages. 
    *
    * It doesn't prevent to send ephemeral messages in the room but those who don't
    * support it won't delete messages after lifetime has expired. 
    * 
    * @return true if all participants in the chat room support ephemeral messages,
    * false otherwise 
    */
    public boolean ephemeralSupportedByAllParticipants();

    /**
    * Gets the chat message sent or received in this chat room that matches the
    * message_id. 
    *
    * @param messageId The id of the message to find 
    * 
    * @return the {@link ChatMessage} 
    */
    public ChatMessage findMessage(String messageId);

    /**
    * Find a participant of a chat room from its address. 
    *
    * @param addr The address to search in the list of participants of the chat room 
    * 
    * @return The participant if found, null otherwise. 
    */
    public Participant findParticipant(Address addr);

    /**
    * Gets nb_message most recent messages from cr chat room, sorted from oldest to
    * most recent. 
    *
    * @param nbMessage Number of message to retrieve. 0 means everything. 
    * 
    * @return A list of {@link ChatMessage} objects. LinphoneChatMessage  The objects
    * inside the list are freshly allocated with a reference counter equal to one, so
    * they need to be freed on list destruction with bctbx_list_free_with_data() for
    * instance.   
    */
    public ChatMessage[] getHistory(int nbMessage);

    /**
    * Gets nb_events most recent events from cr chat room, sorted from oldest to most
    * recent. 
    *
    * @param nbEvents Number of events to retrieve. 0 means everything. 
    * 
    * @return A list of {@link EventLog} objects. LinphoneEventLog  The objects
    * inside the list are freshly allocated with a reference counter equal to one, so
    * they need to be freed on list destruction with bctbx_list_free_with_data() for
    * instance.   
    */
    public EventLog[] getHistoryEvents(int nbEvents);

    /**
    * Gets nb_events most recent chat message events from cr chat room, sorted from
    * oldest to most recent. 
    *
    * @param nbEvents Number of events to retrieve. 0 means everything. 
    * 
    * @return A list of {@link EventLog} objects. LinphoneEventLog  The objects
    * inside the list are freshly allocated with a reference counter equal to one, so
    * they need to be freed on list destruction with bctbx_list_free_with_data() for
    * instance.   
    */
    public EventLog[] getHistoryMessageEvents(int nbEvents);

    /**
    * Gets the partial list of messages in the given range, sorted from oldest to
    * most recent. 
    *
    * @param end The last message of the range to be retrieved. History oldest
    * message has index of history size - 1 (use linphone_chat_room_get_history_size
    * to retrieve history size) 
    * 
    * @return A list of {@link ChatMessage} objects. LinphoneChatMessage  The objects
    * inside the list are freshly allocated with a reference counter equal to one, so
    * they need to be freed on list destruction with bctbx_list_free_with_data() for
    * instance.   
    */
    public ChatMessage[] getHistoryRange(int begin, int end);

    /**
    * Gets the partial list of events in the given range, sorted from oldest to most
    * recent. 
    *
    * @param end The last event of the range to be retrieved. History oldest event
    * has index of history size - 1 
    * 
    * @return A list of {@link EventLog} objects. LinphoneEventLog  The objects
    * inside the list are freshly allocated with a reference counter equal to one, so
    * they need to be freed on list destruction with bctbx_list_free_with_data() for
    * instance.   
    */
    public EventLog[] getHistoryRangeEvents(int begin, int end);

    /**
    * Gets the partial list of chat message events in the given range, sorted from
    * oldest to most recent. 
    *
    * @param end The last event of the range to be retrieved. History oldest event
    * has index of history size - 1 
    * 
    * @return A list of {@link EventLog} objects. LinphoneEventLog  The objects
    * inside the list are freshly allocated with a reference counter equal to one, so
    * they need to be freed on list destruction with bctbx_list_free_with_data() for
    * instance.   
    */
    public EventLog[] getHistoryRangeMessageEvents(int begin, int end);

    /**
    * Return whether or not the chat room has been left. 
    *
    * @return whether or not the chat room has been left 
    */
    public boolean hasBeenLeft();

    /**
    * Check if a chat room has given capabilities. 
    *
    * @param mask A Capabilities mask 
    * 
    * @return True if the mask matches, false otherwise 
    */
    public boolean hasCapability(int mask);

    /**
    * Leave a chat room. 
    *
    */
    public void leave();

    /**
    * Returns true if lime is available for given peer. 
    *
    * @return true if zrtp secrets have already been shared and ready to use 
    */
    public boolean limeAvailable();

    /**
    * Mark all messages of the conversation as read. 
    *
    */
    public void markAsRead();

    /**
    * Notify the chatroom that a participant device has just registered. 
    *
    * This function is meaningful only for server implementation of chatroom, and
    * shall not by used by client applications. 
    */
    public void notifyParticipantDeviceRegistration(Address participantDevice);

    /**
    * Used to receive a chat message when using async mechanism with IM encryption
    * engine. 
    *
    * @param msg {@link ChatMessage} object 
    */
    public void receiveChatMessage(ChatMessage msg);

    /**
    * Remove a participant of a chat room. 
    *
    * @param participant The participant to remove from the chat room 
    */
    public void removeParticipant(Participant participant);

    /**
    * Remove several participants of a chat room at once. 
    *
    * @param participants A list of LinphoneParticipant objects. LinphoneParticipant  
    */
    public void removeParticipants(Participant[] participants);

    /**
    * Send a message to peer member of this chat room. 
    *
    * The state of the sending message will be notified via the callbacks defined in
    * the LinphoneChatMessageCbs object that can be obtained by calling
    * linphone_chat_message_get_callbacks. Note: Unlike
    * linphone_chat_room_send_chat_message, that function only takes a reference on
    * the {@link ChatMessage} instead of totaly takes ownership on it. Thus, the
    * {@link ChatMessage} object must be released by the API user after calling that
    * function.
    * 
    * @param msg The message to send. 
    */
    public void sendChatMessage(ChatMessage msg);

    /**
    * Change the admin status of a participant of a chat room (you need to be an
    * admin yourself to do this). 
    *
    * @param isAdmin A boolean value telling whether the participant should now be an
    * admin or not 
    */
    public void setParticipantAdminStatus(Participant participant, boolean isAdmin);

    /**
    * Set the list of participant devices in the form of SIP URIs with GRUUs for a
    * given participant. 
    *
    * This function is meaningful only for server implementation of chatroom, and
    * shall not by used by client applications. 
    * 
    * @param deviceIdentities A list of LinphoneParticipantDeviceIdentity objects.
    * LinphoneParticipantDeviceIdentity  list of the participant devices to be used
    * by the group chat room 
    */
    public void setParticipantDevices(Address partAddr, ParticipantDeviceIdentity[] deviceIdentities);

    /**
    */
    public void addListener(ChatRoomListener listener);

    /**
    */
    public void removeListener(ChatRoomListener listener);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class ChatRoomImpl implements ChatRoom {

    protected long nativePtr = 0;
    protected Object userData = null;
    protected Core core = null;

    protected ChatRoomImpl(long ptr) {
        nativePtr = ptr;
        core = getCore();
    }


    private native Call getCall(long nativePtr);
    @Override
    synchronized public Call getCall()  {
        synchronized(core) { 
        return (Call)getCall(nativePtr);
        }
    }

    private native int getCapabilities(long nativePtr);
    @Override
    synchronized public int getCapabilities()  {
        synchronized(core) { 
        return getCapabilities(nativePtr);
        }
    }

    private native int getChar(long nativePtr);
    @Override
    synchronized public int getChar()  {
        synchronized(core) { 
        return getChar(nativePtr);
        }
    }

    private native Address[] getComposingAddresses(long nativePtr);
    @Override
    synchronized public Address[] getComposingAddresses()  {
        synchronized(core) { 
        return getComposingAddresses(nativePtr);
        }
    }

    private native Address getConferenceAddress(long nativePtr);
    @Override
    synchronized public Address getConferenceAddress()  {
        synchronized(core) { 
        return (Address)getConferenceAddress(nativePtr);
        }
    }

    private native void setConferenceAddress(long nativePtr, Address confAddr);
    @Override
    synchronized public void setConferenceAddress(Address confAddr)  {
        synchronized(core) { 
        setConferenceAddress(nativePtr, confAddr);
        }
    }

    private native Core getCore(long nativePtr);
    @Override
    synchronized public Core getCore()  {
        
        return (Core)getCore(nativePtr);
    }

    private native ChatRoomParams getCurrentParams(long nativePtr);
    @Override
    synchronized public ChatRoomParams getCurrentParams()  {
        synchronized(core) { 
        return (ChatRoomParams)getCurrentParams(nativePtr);
        }
    }

    private native boolean ephemeralEnabled(long nativePtr);
    @Override
    synchronized public boolean ephemeralEnabled()  {
        synchronized(core) { 
        return ephemeralEnabled(nativePtr);
        }
    }

    private native void enableEphemeral(long nativePtr, boolean ephem);
    @Override
    synchronized public void enableEphemeral(boolean ephem)  {
        synchronized(core) { 
        enableEphemeral(nativePtr, ephem);
        }
    }

    private native long getEphemeralLifetime(long nativePtr);
    @Override
    synchronized public long getEphemeralLifetime()  {
        synchronized(core) { 
        return getEphemeralLifetime(nativePtr);
        }
    }

    private native void setEphemeralLifetime(long nativePtr, long time);
    @Override
    synchronized public void setEphemeralLifetime(long time)  {
        synchronized(core) { 
        setEphemeralLifetime(nativePtr, time);
        }
    }

    private native int getHistoryEventsSize(long nativePtr);
    @Override
    synchronized public int getHistoryEventsSize()  {
        synchronized(core) { 
        return getHistoryEventsSize(nativePtr);
        }
    }

    private native int getHistorySize(long nativePtr);
    @Override
    synchronized public int getHistorySize()  {
        synchronized(core) { 
        return getHistorySize(nativePtr);
        }
    }

    private native boolean isEmpty(long nativePtr);
    @Override
    synchronized public boolean isEmpty()  {
        synchronized(core) { 
        return isEmpty(nativePtr);
        }
    }

    private native boolean isRemoteComposing(long nativePtr);
    @Override
    synchronized public boolean isRemoteComposing()  {
        synchronized(core) { 
        return isRemoteComposing(nativePtr);
        }
    }

    private native ChatMessage getLastMessageInHistory(long nativePtr);
    @Override
    synchronized public ChatMessage getLastMessageInHistory()  {
        synchronized(core) { 
        return (ChatMessage)getLastMessageInHistory(nativePtr);
        }
    }

    private native long getLastUpdateTime(long nativePtr);
    @Override
    synchronized public long getLastUpdateTime()  {
        synchronized(core) { 
        return getLastUpdateTime(nativePtr);
        }
    }

    private native Address getLocalAddress(long nativePtr);
    @Override
    synchronized public Address getLocalAddress()  {
        synchronized(core) { 
        return (Address)getLocalAddress(nativePtr);
        }
    }

    private native Participant getMe(long nativePtr);
    @Override
    synchronized public Participant getMe()  {
        synchronized(core) { 
        return (Participant)getMe(nativePtr);
        }
    }

    private native int getNbParticipants(long nativePtr);
    @Override
    synchronized public int getNbParticipants()  {
        synchronized(core) { 
        return getNbParticipants(nativePtr);
        }
    }

    private native Participant[] getParticipants(long nativePtr);
    @Override
    synchronized public Participant[] getParticipants()  {
        synchronized(core) { 
        return getParticipants(nativePtr);
        }
    }

    private native Address getPeerAddress(long nativePtr);
    @Override
    synchronized public Address getPeerAddress()  {
        synchronized(core) { 
        return (Address)getPeerAddress(nativePtr);
        }
    }

    private native int getSecurityLevel(long nativePtr);
    @Override
    synchronized public ChatRoomSecurityLevel getSecurityLevel()  {
        synchronized(core) { 
        return ChatRoomSecurityLevel.fromInt(getSecurityLevel(nativePtr));
        }
    }

    private native int getState(long nativePtr);
    @Override
    synchronized public ChatRoom.State getState()  {
        synchronized(core) { 
        return ChatRoom.State.fromInt(getState(nativePtr));
        }
    }

    private native String getSubject(long nativePtr);
    @Override
    synchronized public String getSubject()  {
        synchronized(core) { 
        return getSubject(nativePtr);
        }
    }

    private native void setSubject(long nativePtr, String subject);
    @Override
    synchronized public void setSubject(String subject)  {
        synchronized(core) { 
        setSubject(nativePtr, subject);
        }
    }

    private native int getUnreadMessagesCount(long nativePtr);
    @Override
    synchronized public int getUnreadMessagesCount()  {
        synchronized(core) { 
        return getUnreadMessagesCount(nativePtr);
        }
    }

    private native void addParticipant(long nativePtr, Address addr);
    @Override
    synchronized public void addParticipant(Address addr)  {
        synchronized(core) { 
        addParticipant(nativePtr, addr);
        }
    }

    private native boolean addParticipants(long nativePtr, Address[] addresses);
    @Override
    synchronized public boolean addParticipants(Address[] addresses)  {
        synchronized(core) { 
        return addParticipants(nativePtr, addresses);
        }
    }

    private native void allowCpim(long nativePtr);
    @Override
    synchronized public void allowCpim()  {
        synchronized(core) { 
        allowCpim(nativePtr);
        }
    }

    private native void allowMultipart(long nativePtr);
    @Override
    synchronized public void allowMultipart()  {
        synchronized(core) { 
        allowMultipart(nativePtr);
        }
    }

    private native boolean canHandleParticipants(long nativePtr);
    @Override
    synchronized public boolean canHandleParticipants()  {
        synchronized(core) { 
        return canHandleParticipants(nativePtr);
        }
    }

    private native void compose(long nativePtr);
    @Override
    synchronized public void compose()  {
        synchronized(core) { 
        compose(nativePtr);
        }
    }

    private native ChatMessage createEmptyMessage(long nativePtr);
    @Override
    synchronized public ChatMessage createEmptyMessage()  {
        synchronized(core) { 
        return (ChatMessage)createEmptyMessage(nativePtr);
        }
    }

    private native ChatMessage createFileTransferMessage(long nativePtr, Content initialContent);
    @Override
    synchronized public ChatMessage createFileTransferMessage(Content initialContent)  {
        synchronized(core) { 
        return (ChatMessage)createFileTransferMessage(nativePtr, initialContent);
        }
    }

    private native ChatMessage createForwardMessage(long nativePtr, ChatMessage msg);
    @Override
    synchronized public ChatMessage createForwardMessage(ChatMessage msg)  {
        synchronized(core) { 
        return (ChatMessage)createForwardMessage(nativePtr, msg);
        }
    }

    private native ChatMessage createMessage(long nativePtr, String message);
    @Override
    synchronized public ChatMessage createMessage(String message)  {
        synchronized(core) { 
        return (ChatMessage)createMessage(nativePtr, message);
        }
    }

    private native void deleteHistory(long nativePtr);
    @Override
    synchronized public void deleteHistory()  {
        synchronized(core) { 
        deleteHistory(nativePtr);
        }
    }

    private native void deleteMessage(long nativePtr, ChatMessage msg);
    @Override
    synchronized public void deleteMessage(ChatMessage msg)  {
        synchronized(core) { 
        deleteMessage(nativePtr, msg);
        }
    }

    private native boolean ephemeralSupportedByAllParticipants(long nativePtr);
    @Override
    synchronized public boolean ephemeralSupportedByAllParticipants()  {
        synchronized(core) { 
        return ephemeralSupportedByAllParticipants(nativePtr);
        }
    }

    private native ChatMessage findMessage(long nativePtr, String messageId);
    @Override
    synchronized public ChatMessage findMessage(String messageId)  {
        synchronized(core) { 
        return (ChatMessage)findMessage(nativePtr, messageId);
        }
    }

    private native Participant findParticipant(long nativePtr, Address addr);
    @Override
    synchronized public Participant findParticipant(Address addr)  {
        synchronized(core) { 
        return (Participant)findParticipant(nativePtr, addr);
        }
    }

    private native ChatMessage[] getHistory(long nativePtr, int nbMessage);
    @Override
    synchronized public ChatMessage[] getHistory(int nbMessage)  {
        synchronized(core) { 
        return getHistory(nativePtr, nbMessage);
        }
    }

    private native EventLog[] getHistoryEvents(long nativePtr, int nbEvents);
    @Override
    synchronized public EventLog[] getHistoryEvents(int nbEvents)  {
        synchronized(core) { 
        return getHistoryEvents(nativePtr, nbEvents);
        }
    }

    private native EventLog[] getHistoryMessageEvents(long nativePtr, int nbEvents);
    @Override
    synchronized public EventLog[] getHistoryMessageEvents(int nbEvents)  {
        synchronized(core) { 
        return getHistoryMessageEvents(nativePtr, nbEvents);
        }
    }

    private native ChatMessage[] getHistoryRange(long nativePtr, int begin, int end);
    @Override
    synchronized public ChatMessage[] getHistoryRange(int begin, int end)  {
        synchronized(core) { 
        return getHistoryRange(nativePtr, begin, end);
        }
    }

    private native EventLog[] getHistoryRangeEvents(long nativePtr, int begin, int end);
    @Override
    synchronized public EventLog[] getHistoryRangeEvents(int begin, int end)  {
        synchronized(core) { 
        return getHistoryRangeEvents(nativePtr, begin, end);
        }
    }

    private native EventLog[] getHistoryRangeMessageEvents(long nativePtr, int begin, int end);
    @Override
    synchronized public EventLog[] getHistoryRangeMessageEvents(int begin, int end)  {
        synchronized(core) { 
        return getHistoryRangeMessageEvents(nativePtr, begin, end);
        }
    }

    private native boolean hasBeenLeft(long nativePtr);
    @Override
    synchronized public boolean hasBeenLeft()  {
        synchronized(core) { 
        return hasBeenLeft(nativePtr);
        }
    }

    private native boolean hasCapability(long nativePtr, int mask);
    @Override
    synchronized public boolean hasCapability(int mask)  {
        synchronized(core) { 
        return hasCapability(nativePtr, mask);
        }
    }

    private native void leave(long nativePtr);
    @Override
    synchronized public void leave()  {
        synchronized(core) { 
        leave(nativePtr);
        }
    }

    private native boolean limeAvailable(long nativePtr);
    @Override
    synchronized public boolean limeAvailable()  {
        synchronized(core) { 
        return limeAvailable(nativePtr);
        }
    }

    private native void markAsRead(long nativePtr);
    @Override
    synchronized public void markAsRead()  {
        synchronized(core) { 
        markAsRead(nativePtr);
        }
    }

    private native void notifyParticipantDeviceRegistration(long nativePtr, Address participantDevice);
    @Override
    synchronized public void notifyParticipantDeviceRegistration(Address participantDevice)  {
        synchronized(core) { 
        notifyParticipantDeviceRegistration(nativePtr, participantDevice);
        }
    }

    private native void receiveChatMessage(long nativePtr, ChatMessage msg);
    @Override
    synchronized public void receiveChatMessage(ChatMessage msg)  {
        synchronized(core) { 
        receiveChatMessage(nativePtr, msg);
        }
    }

    private native void removeParticipant(long nativePtr, Participant participant);
    @Override
    synchronized public void removeParticipant(Participant participant)  {
        synchronized(core) { 
        removeParticipant(nativePtr, participant);
        }
    }

    private native void removeParticipants(long nativePtr, Participant[] participants);
    @Override
    synchronized public void removeParticipants(Participant[] participants)  {
        synchronized(core) { 
        removeParticipants(nativePtr, participants);
        }
    }

    private native void sendChatMessage2(long nativePtr, ChatMessage msg);
    @Override
    synchronized public void sendChatMessage(ChatMessage msg)  {
        synchronized(core) { 
        sendChatMessage2(nativePtr, msg);
        }
    }

    private native void setParticipantAdminStatus(long nativePtr, Participant participant, boolean isAdmin);
    @Override
    synchronized public void setParticipantAdminStatus(Participant participant, boolean isAdmin)  {
        synchronized(core) { 
        setParticipantAdminStatus(nativePtr, participant, isAdmin);
        }
    }

    private native void setParticipantDevices(long nativePtr, Address partAddr, ParticipantDeviceIdentity[] deviceIdentities);
    @Override
    synchronized public void setParticipantDevices(Address partAddr, ParticipantDeviceIdentity[] deviceIdentities)  {
        synchronized(core) { 
        setParticipantDevices(nativePtr, partAddr, deviceIdentities);
        }
    }

    private native void addListener(long nativePtr, ChatRoomListener listener);
    @Override
    synchronized public void addListener(ChatRoomListener listener)  {
        
        addListener(nativePtr, listener);
    }

    private native void removeListener(long nativePtr, ChatRoomListener listener);
    @Override
    synchronized public void removeListener(ChatRoomListener listener)  {
        
        removeListener(nativePtr, listener);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
