/*
ChatRoomParams.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* An object to handle a chat room parameters. 
*
* Can be created with linphone_core_get_default_chat_room_params() or
* linphone_chat_room_params_new. 
*/
public interface ChatRoomParams {
    /**
    * Get the backend implementation of the chat room associated with the given
    * parameters. 
    *
    * @return LinphoneChatRoomBackend 
    */
    public ChatRoomBackend getBackend();

    /**
    * Set the backend implementation of these chat room parameters. 
    *
    * @param backend The LinphoneChatRoomBackend enum value 
    */
    public void setBackend(ChatRoomBackend backend);

    /**
    * Get the encryption implementation of the chat room associated with the given
    * parameters. 
    *
    * @return LinphoneChatRoomEncryptionBackend 
    */
    public ChatRoomEncryptionBackend getEncryptionBackend();

    /**
    * Set the encryption backend implementation of these chat room parameters. 
    *
    * @param backend The LinphoneChatRoomEncryptionBackend enum value 
    */
    public void setEncryptionBackend(ChatRoomEncryptionBackend backend);

    /**
    * Get the encryption status of the chat room associated with the given
    * parameters. 
    *
    * @return  if encryption is enabled, true otherwise 
    */
    public boolean encryptionEnabled();

    /**
    * Enables or disables encryption for the chat room associated with the given
    * parameters. 
    *
    */
    public void enableEncryption(boolean encrypted);

    /**
    * Get the group chat status of the chat room associated with the given
    * parameters. 
    *
    * @return  if group chat is enabled, true if one-to-one 
    */
    public boolean groupEnabled();

    /**
    * Enables or disables group chat for the chat room associated with the given
    * parameters. 
    *
    */
    public void enableGroup(boolean group);

    /**
    *
    * @return  if the given parameters are valid, true otherwise 
    */
    public boolean isValid();

    /**
    * Get the real time text status of the chat room associated with the given
    * parameters. 
    *
    * @return  if real time text is enabled, true otherwise 
    */
    public boolean rttEnabled();

    /**
    * Enables or disables real time text for the chat room associated with the given
    * parameters. 
    *
    */
    public void enableRtt(boolean rtt);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class ChatRoomParamsImpl implements ChatRoomParams {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected ChatRoomParamsImpl(long ptr) {
        nativePtr = ptr;
    }


    private native int getBackend(long nativePtr);
    @Override
    synchronized public ChatRoomBackend getBackend()  {
        
        return ChatRoomBackend.fromInt(getBackend(nativePtr));
    }

    private native void setBackend(long nativePtr, int backend);
    @Override
    synchronized public void setBackend(ChatRoomBackend backend)  {
        
        setBackend(nativePtr, backend.toInt());
    }

    private native int getEncryptionBackend(long nativePtr);
    @Override
    synchronized public ChatRoomEncryptionBackend getEncryptionBackend()  {
        
        return ChatRoomEncryptionBackend.fromInt(getEncryptionBackend(nativePtr));
    }

    private native void setEncryptionBackend(long nativePtr, int backend);
    @Override
    synchronized public void setEncryptionBackend(ChatRoomEncryptionBackend backend)  {
        
        setEncryptionBackend(nativePtr, backend.toInt());
    }

    private native boolean encryptionEnabled(long nativePtr);
    @Override
    synchronized public boolean encryptionEnabled()  {
        
        return encryptionEnabled(nativePtr);
    }

    private native void enableEncryption(long nativePtr, boolean encrypted);
    @Override
    synchronized public void enableEncryption(boolean encrypted)  {
        
        enableEncryption(nativePtr, encrypted);
    }

    private native boolean groupEnabled(long nativePtr);
    @Override
    synchronized public boolean groupEnabled()  {
        
        return groupEnabled(nativePtr);
    }

    private native void enableGroup(long nativePtr, boolean group);
    @Override
    synchronized public void enableGroup(boolean group)  {
        
        enableGroup(nativePtr, group);
    }

    private native boolean isValid(long nativePtr);
    @Override
    synchronized public boolean isValid()  {
        
        return isValid(nativePtr);
    }

    private native boolean rttEnabled(long nativePtr);
    @Override
    synchronized public boolean rttEnabled()  {
        
        return rttEnabled(nativePtr);
    }

    private native void enableRtt(long nativePtr, boolean rtt);
    @Override
    synchronized public void enableRtt(boolean rtt)  {
        
        enableRtt(nativePtr, rtt);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
