/*
Conference.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* {@link Conference} class The _LinphoneConference struct does not exists, it's
* the Conference C++ class that is used behind 
*
*/
public interface Conference {
    /**
    * Get the conference id as string. 
    *
    */
    public String getId();

    /**
    * Set the conference id as string. 
    *
    */
    public void setId(String conferenceID);

    /**
    * Get URIs of all participants of one conference The returned bctbx_list_t
    * contains URIs of all participant. 
    *
    * That list must be freed after use and each URI must be unref with
    * linphone_address_unref 
    * 
    * @return A list of {@link Address} objects. LinphoneAddress  
    */
    public Address[] getParticipants();

    /**
    * Invite participants to the conference, by supplying a list of {@link Address}. 
    *
    * @param params {@link CallParams} to use for inviting the participants. 
    */
    public int inviteParticipants(Address[] addresses, CallParams params);

    /**
    * Remove a participant from a conference. 
    *
    * @param uri SIP URI of the participant to remove 
    * 
    * Warning: The passed SIP URI must be one of the URIs returned by {@link
    * Conference#getParticipants} 
    * 
    * @return 0 if succeeded, -1 if failed 
    */
    public int removeParticipant(Address uri);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class ConferenceImpl implements Conference {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected ConferenceImpl(long ptr) {
        nativePtr = ptr;
    }


    private native String getId(long nativePtr);
    @Override
    synchronized public String getId()  {
        
        return getId(nativePtr);
    }

    private native void setId(long nativePtr, String conferenceID);
    @Override
    synchronized public void setId(String conferenceID)  {
        
        setId(nativePtr, conferenceID);
    }

    private native Address[] getParticipants(long nativePtr);
    @Override
    synchronized public Address[] getParticipants()  {
        
        return getParticipants(nativePtr);
    }

    private native int inviteParticipants(long nativePtr, Address[] addresses, CallParams params);
    @Override
    synchronized public int inviteParticipants(Address[] addresses, CallParams params)  {
        
        return inviteParticipants(nativePtr, addresses, params);
    }

    private native int removeParticipant(long nativePtr, Address uri);
    @Override
    synchronized public int removeParticipant(Address uri)  {
        
        return removeParticipant(nativePtr, uri);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
