/*
Content.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* The LinphoneContent object holds data that can be embedded in a signaling
* message. 
*
*/
public interface Content {
    /**
    * Get the content data buffer, usually a string. 
    *
    * @return The content data buffer. 
    */
    public byte[] getBuffer();

    /**
    * Get the encoding of the data buffer, for example "gzip". 
    *
    * @return The encoding of the data buffer. 
    */
    public String getEncoding();

    /**
    * Set the encoding of the data buffer, for example "gzip". 
    *
    * @param encoding The encoding of the data buffer. 
    */
    public void setEncoding(String encoding);

    /**
    * Get the file transfer filepath set for this content (replace
    * linphone_chat_message_get_file_transfer_filepath). 
    *
    * @return The file path set for this content if it has been set, null otherwise. 
    */
    public String getFilePath();

    /**
    * Set the file transfer filepath for this content (replace
    * linphone_chat_message_set_file_transfer_filepath). 
    *
    * @param filePath the file transfer filepath. 
    */
    public void setFilePath(String filePath);

    /**
    * Get the file size if content is either a FileContent or a FileTransferContent. 
    *
    * @return The represented file size. 
    */
    public int getFileSize();

    /**
    * Tells whether or not this content contains a file. 
    *
    * @return True if this content contains a file, false otherwise. 
    */
    public boolean isFile();

    /**
    * Tells whether or not this content is a file transfer. 
    *
    * @return True if this content is a file transfer, false otherwise. 
    */
    public boolean isFileTransfer();

    /**
    * Tell whether a content is a multipart content. 
    *
    * @return A boolean value telling whether the content is multipart or not. 
    */
    public boolean isMultipart();

    /**
    * Tells whether or not this content contains text. 
    *
    * @return True if this content contains plain text, false otherwise. 
    */
    public boolean isText();

    /**
    * Get the key associated with a RCS file transfer message if encrypted. 
    *
    * @return The key to encrypt/decrypt the file associated to this content. 
    */
    public String getKey();

    /**
    * Get the size of key associated with a RCS file transfer message if encrypted. 
    *
    * @return The key size in bytes 
    */
    public int getKeySize();

    /**
    * Get the name associated with a RCS file transfer message. 
    *
    * It is used to store the original filename of the file to be downloaded from
    * server. 
    * 
    * @return The name of the content. 
    */
    public String getName();

    /**
    * Set the name associated with a RCS file transfer message. 
    *
    * It is used to store the original filename of the file to be downloaded from
    * server. 
    * 
    * @param name The name of the content. 
    */
    public void setName(String name);

    /**
    * Get all the parts from a multipart content. 
    *
    * @return A A list of {@link Content} objects. LinphoneContent  The objects
    * inside the list are freshly allocated with a reference counter equal to one, so
    * they need to be freed on list destruction with bctbx_list_free_with_data() for
    * instance.   object holding the part if found, null otherwise. 
    */
    public Content[] getParts();

    /**
    * Get the content data buffer size, excluding null character despite null
    * character is always set for convenience. 
    *
    * @return The content data buffer size. 
    */
    public int getSize();

    /**
    * Set the content data size, excluding null character despite null character is
    * always set for convenience. 
    *
    * @param size The content data buffer size. 
    */
    public void setSize(int size);

    /**
    * Get the string content data buffer. 
    *
    * @return The string content data buffer. 
    */
    public String getStringBuffer();

    /**
    * Set the string content data buffer. 
    *
    * @param buffer The string content data buffer. 
    */
    public void setStringBuffer(String buffer);

    /**
    * Get the mime subtype of the content data. 
    *
    * @return The mime subtype of the content data, for example "html". 
    */
    public String getSubtype();

    /**
    * Set the mime subtype of the content data. 
    *
    * @param subtype The mime subtype of the content data, for example "html". 
    */
    public void setSubtype(String subtype);

    /**
    * Get the mime type of the content data. 
    *
    * @return The mime type of the content data, for example "application". 
    */
    public String getType();

    /**
    * Set the mime type of the content data. 
    *
    * @param type The mime type of the content data, for example "application". 
    */
    public void setType(String type);

    /**
    * Adds a parameter to the ContentType header. 
    *
    * @param value the value of the parameter to add. 
    */
    public void addContentTypeParameter(String name, String value);

    /**
    * Find a part from a multipart content looking for a part header with a specified
    * value. 
    *
    * @param headerValue The value of the header to look for. 
    * 
    * @return A {@link Content} object object the part if found, null otherwise. 
    */
    public Content findPartByHeader(String headerName, String headerValue);

    /**
    * Get a custom header value of a content. 
    *
    * @param headerName The name of the header to get the value from. 
    * 
    * @return The value of the header if found, null otherwise. 
    */
    public String getCustomHeader(String headerName);

    /**
    * Get a part from a multipart content according to its index. 
    *
    * @param idx The index of the part to get. 
    * 
    * @return A {@link Content} object holding the part if found, null otherwise. 
    */
    public Content getPart(int idx);

    /**
    * Set the content data buffer, usually a string. 
    *
    * @param size The size of the content data buffer. 
    */
    public void setBuffer(byte[] buffer, int size);

    /**
    * Set the key associated with a RCS file transfer message if encrypted. 
    *
    * @param keyLength The lengh of the key. 
    */
    public void setKey(String key, int keyLength);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class ContentImpl implements Content {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected ContentImpl(long ptr) {
        nativePtr = ptr;
    }


    private native byte[] getBuffer(long nativePtr);
    @Override
    synchronized public byte[] getBuffer()  {
        
        return getBuffer(nativePtr);
    }

    private native String getEncoding(long nativePtr);
    @Override
    synchronized public String getEncoding()  {
        
        return getEncoding(nativePtr);
    }

    private native void setEncoding(long nativePtr, String encoding);
    @Override
    synchronized public void setEncoding(String encoding)  {
        
        setEncoding(nativePtr, encoding);
    }

    private native String getFilePath(long nativePtr);
    @Override
    synchronized public String getFilePath()  {
        
        return getFilePath(nativePtr);
    }

    private native void setFilePath(long nativePtr, String filePath);
    @Override
    synchronized public void setFilePath(String filePath)  {
        
        setFilePath(nativePtr, filePath);
    }

    private native int getFileSize(long nativePtr);
    @Override
    synchronized public int getFileSize()  {
        
        return getFileSize(nativePtr);
    }

    private native boolean isFile(long nativePtr);
    @Override
    synchronized public boolean isFile()  {
        
        return isFile(nativePtr);
    }

    private native boolean isFileTransfer(long nativePtr);
    @Override
    synchronized public boolean isFileTransfer()  {
        
        return isFileTransfer(nativePtr);
    }

    private native boolean isMultipart(long nativePtr);
    @Override
    synchronized public boolean isMultipart()  {
        
        return isMultipart(nativePtr);
    }

    private native boolean isText(long nativePtr);
    @Override
    synchronized public boolean isText()  {
        
        return isText(nativePtr);
    }

    private native String getKey(long nativePtr);
    @Override
    synchronized public String getKey()  {
        
        return getKey(nativePtr);
    }

    private native int getKeySize(long nativePtr);
    @Override
    synchronized public int getKeySize()  {
        
        return getKeySize(nativePtr);
    }

    private native String getName(long nativePtr);
    @Override
    synchronized public String getName()  {
        
        return getName(nativePtr);
    }

    private native void setName(long nativePtr, String name);
    @Override
    synchronized public void setName(String name)  {
        
        setName(nativePtr, name);
    }

    private native Content[] getParts(long nativePtr);
    @Override
    synchronized public Content[] getParts()  {
        
        return getParts(nativePtr);
    }

    private native int getSize(long nativePtr);
    @Override
    synchronized public int getSize()  {
        
        return getSize(nativePtr);
    }

    private native void setSize(long nativePtr, int size);
    @Override
    synchronized public void setSize(int size)  {
        
        setSize(nativePtr, size);
    }

    private native String getStringBuffer(long nativePtr);
    @Override
    synchronized public String getStringBuffer()  {
        
        return getStringBuffer(nativePtr);
    }

    private native void setStringBuffer(long nativePtr, String buffer);
    @Override
    synchronized public void setStringBuffer(String buffer)  {
        
        setStringBuffer(nativePtr, buffer);
    }

    private native String getSubtype(long nativePtr);
    @Override
    synchronized public String getSubtype()  {
        
        return getSubtype(nativePtr);
    }

    private native void setSubtype(long nativePtr, String subtype);
    @Override
    synchronized public void setSubtype(String subtype)  {
        
        setSubtype(nativePtr, subtype);
    }

    private native String getType(long nativePtr);
    @Override
    synchronized public String getType()  {
        
        return getType(nativePtr);
    }

    private native void setType(long nativePtr, String type);
    @Override
    synchronized public void setType(String type)  {
        
        setType(nativePtr, type);
    }

    private native void addContentTypeParameter(long nativePtr, String name, String value);
    @Override
    synchronized public void addContentTypeParameter(String name, String value)  {
        
        addContentTypeParameter(nativePtr, name, value);
    }

    private native Content findPartByHeader(long nativePtr, String headerName, String headerValue);
    @Override
    synchronized public Content findPartByHeader(String headerName, String headerValue)  {
        
        return (Content)findPartByHeader(nativePtr, headerName, headerValue);
    }

    private native String getCustomHeader(long nativePtr, String headerName);
    @Override
    synchronized public String getCustomHeader(String headerName)  {
        
        return getCustomHeader(nativePtr, headerName);
    }

    private native Content getPart(long nativePtr, int idx);
    @Override
    synchronized public Content getPart(int idx)  {
        
        return (Content)getPart(nativePtr, idx);
    }

    private native void setBuffer(long nativePtr, byte[] buffer, int size);
    @Override
    synchronized public void setBuffer(byte[] buffer, int size)  {
        
        setBuffer(nativePtr, buffer, size);
    }

    private native void setKey(long nativePtr, String key, int keyLength);
    @Override
    synchronized public void setKey(String key, int keyLength)  {
        
        setKey(nativePtr, key, keyLength);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
