/*
CoreListener.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;


/**
* That class holds all the callbacks which are called by LinphoneCore. 
*
* Use linphone_factory_create_core_cbs to create an instance. Then, call the
* callback setters on the events you need to monitor and pass the object to a
* LinphoneCore instance through linphone_core_add_callbacks.
* 
* That class is inherited from belle_sip_object_t. 
*/
public interface CoreListener {
    /**
    * Callback for notifying progresses of transfers. 
    *
    * @param newCallState the state of the call to transfer target at the far end. 
    */
    public void onTransferStateChanged(Core lc, Call transfered, Call.State newCallState);

    /**
    * Callback prototype for reporting when a friend list has been added to the core
    * friends list. 
    *
    * @param list LinphoneFriendList object 
    */
    public void onFriendListCreated(Core lc, FriendList list);

    /**
    * Callback prototype for notifying the application about changes of subscription
    * states, including arrival of new subscriptions. 
    *
    */
    public void onSubscriptionStateChanged(Core lc, Event lev, SubscriptionState state);

    /**
    * Callback to notify a new call-log entry has been added. 
    *
    * This is done typically when a call terminates. 
    * 
    * @param newcl the new call log entry added. 
    */
    public void onCallLogUpdated(Core lc, CallLog newcl);

    /**
    * Call state notification callback. 
    *
    * @param message a non null informational message about the state. 
    */
    public void onCallStateChanged(Core lc, Call call, Call.State cstate, String message);

    /**
    * Callback for requesting authentication information to application or user. 
    *
    * @param method the type of authentication requested Application shall reply to
    * this callback using linphone_core_add_auth_info. 
    */
    public void onAuthenticationRequested(Core lc, AuthInfo authInfo, AuthMethod method);

    /**
    * Reports presence model change for a specific URI or phone number of a friend. 
    *
    * @param presenceModel The new presence model 
    */
    public void onNotifyPresenceReceivedForUriOrTel(Core lc, Friend lf, String uriOrTel, PresenceModel presenceModel);

    /**
    * Callback prototype telling that a LinphoneChatRoom state has changed. 
    *
    * @param cr The LinphoneChatRoom object for which the state has changed 
    */
    public void onChatRoomStateChanged(Core lc, ChatRoom cr, ChatRoom.State state);

    /**
    * Callback prototype. 
    *
    */
    public void onBuddyInfoUpdated(Core lc, Friend lf);

    /**
    * Callback prototype for reporting network change either automatically detected
    * or notified by linphone_core_set_network_reachable. 
    *
    * @param reachable true if network is reachable. 
    */
    public void onNetworkReachable(Core lc, boolean reachable);

    /**
    * Callback prototype for notifying the application about notification received
    * from the network. 
    *
    */
    public void onNotifyReceived(Core lc, Event lev, String notifiedEvent, Content body);

    /**
    * Reports that a new subscription request has been received and wait for a
    * decision. 
    *
    * Status on this subscription request is notified by changing policy  for this
    * friend 
    * 
    * @param url of the subscriber 
    */
    public void onNewSubscriptionRequested(Core lc, Friend lf, String url);

    /**
    * Callback for receiving quality statistics for calls. 
    *
    * @param stats the call statistics. 
    */
    public void onCallStatsUpdated(Core lc, Call call, CallStats stats);

    /**
    * Report status change for a friend previously added  to LinphoneCore. 
    *
    * @param lf Updated LinphoneFriend . 
    */
    public void onNotifyPresenceReceived(Core lc, Friend lf);

    /**
    * Function prototype used by #linphone_core_cbs_set_ec_calibrator_audio_init(). 
    *
    * @param lc The core. 
    */
    public void onEcCalibrationAudioInit(Core lc);

    /**
    * Chat message callback prototype. 
    *
    * @param room LinphoneChatRoom involved in this conversation. Can be be created
    * by the framework in case the from  is not present in any chat room. 
    */
    public void onMessageReceived(Core lc, ChatRoom room, ChatMessage message);

    /**
    * Function prototype used by #linphone_core_cbs_set_ec_calibrator_result(). 
    *
    * @param delayMs The measured delay if available. 
    */
    public void onEcCalibrationResult(Core lc, EcCalibratorStatus status, int delayMs);

    /**
    * Callback prototype for notifying the application about subscription received
    * from the network. 
    *
    */
    public void onSubscribeReceived(Core lc, Event lev, String subscribeEvent, Content body);

    /**
    * Callback prototype for receiving info messages. 
    *
    * @param msg the info message. 
    */
    public void onInfoReceived(Core lc, Call call, InfoMessage msg);

    /**
    * Chat room marked as read callback. 
    *
    * @param room LinphoneChatRoom that has been marked as read. 
    */
    public void onChatRoomRead(Core lc, ChatRoom room);

    /**
    * Registration state notification callback prototype. 
    *
    */
    public void onRegistrationStateChanged(Core lc, ProxyConfig cfg, RegistrationState cstate, String message);

    /**
    * Callback prototype for reporting when a friend list has been removed from the
    * core friends list. 
    *
    * @param list LinphoneFriendList object 
    */
    public void onFriendListRemoved(Core lc, FriendList list);

    /**
    * Callback prototype. 
    *
    */
    public void onReferReceived(Core lc, String referTo);

    /**
    * Callback prototype telling the result of decoded qrcode. 
    *
    * @param result The result of the decoded qrcode 
    */
    public void onQrcodeFound(Core lc, String result);

    /**
    * Callback prototype for configuring status changes notification. 
    *
    * @param message informational message. 
    */
    public void onConfiguringStatus(Core lc, ConfiguringState status, String message);

    /**
    * Callback notifying that a new LinphoneCall (either incoming or outgoing) has
    * been created. 
    *
    * @param call The newly created LinphoneCall object 
    */
    public void onCallCreated(Core lc, Call call);

    /**
    * Callback prototype for notifying the application about changes of publish
    * states. 
    *
    */
    public void onPublishStateChanged(Core lc, Event lev, PublishState state);

    /**
    * Call encryption changed callback. 
    *
    * @param authenticationToken an authentication_token, currently set for ZRTP kind
    * of encryption only. 
    */
    public void onCallEncryptionChanged(Core lc, Call call, boolean on, String authenticationToken);

    /**
    * Is composing notification callback prototype. 
    *
    * @param room LinphoneChatRoom involved in the conversation. 
    */
    public void onIsComposingReceived(Core lc, ChatRoom room);

    /**
    * Chat message not decrypted callback prototype. 
    *
    * @param room LinphoneChatRoom involved in this conversation. Can be be created
    * by the framework in case the from  is not present in any chat room. 
    */
    public void onMessageReceivedUnableDecrypt(Core lc, ChatRoom room, ChatMessage message);

    /**
    * Callback prototype for reporting log collection upload progress indication. 
    *
    * @param lc LinphoneCore object 
    */
    public void onLogCollectionUploadProgressIndication(Core lc, int offset, int total);

    /**
    * Callback prototype telling that a LinphoneChatRoom subject has changed. 
    *
    * @param cr The LinphoneChatRoom object for which the subject has changed 
    */
    public void onChatRoomSubjectChanged(Core lc, ChatRoom cr);

    /**
    * Callback prototype for reporting the result of a version update check. 
    *
    * @param url The url where to download the new version if the result is
    * #LinphoneVersionUpdateCheckNewVersionAvailable 
    */
    public void onVersionUpdateCheckResultReceived(Core lc, VersionUpdateCheckResult result, String version, String url);

    /**
    * Function prototype used by #linphone_core_cbs_set_ec_calibrator_audio_uninit(). 
    *
    * @param lc The core. 
    */
    public void onEcCalibrationAudioUninit(Core lc);

    /**
    * Global state notification callback. 
    *
    * @param message informational message. 
    */
    public void onGlobalStateChanged(Core lc, GlobalState gstate, String message);

    /**
    * Callback prototype for reporting log collection upload state change. 
    *
    * @param info Additional information: error message in case of error state, URL
    * of uploaded file in case of success. 
    */
    public void onLogCollectionUploadStateChanged(Core lc, Core.LogCollectionUploadState state, String info);

    /**
    * Callback for being notified of DTMFs received. 
    *
    * @param dtmf the ascii code of the dtmf 
    */
    public void onDtmfReceived(Core lc, Call call, int dtmf);

    /**
    * Callback prototype telling that a LinphoneChatRoom ephemeral message has
    * expired. 
    *
    * @param cr The LinphoneChatRoom object for which a message has expired. 
    */
    public void onChatRoomEphemeralMessageDeleted(Core lc, ChatRoom cr);

    /**
    * Called after the #send method of the LinphoneChatMessage was called. 
    *
    * The message will be in state InProgress. In case of resend this callback won't
    * be called. 
    * 
    * @param room LinphoneChatRoom involved in this conversation. Can be be created
    * by the framework in case the from  is not present in any chat room. 
    */
    public void onMessageSent(Core lc, ChatRoom room, ChatMessage message);

}