/*
DialPlan.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* Represents a dial plan. 
*
*/
public interface DialPlan {
    /**
    * Returns the country name of the dialplan. 
    *
    * @return the country name 
    */
    public String getCountry();

    /**
    * Returns the country calling code of the dialplan. 
    *
    * @return the country calling code 
    */
    public String getCountryCallingCode();

    /**
    * Returns the international call prefix of the dialplan. 
    *
    * @return the international call prefix 
    */
    public String getInternationalCallPrefix();

    /**
    * Return if given plan is generic. 
    *
    */
    public boolean isGeneric();

    /**
    * Returns the iso country code of the dialplan. 
    *
    * @return the iso country code 
    */
    public String getIsoCountryCode();

    /**
    * Returns the national number length of the dialplan. 
    *
    * @return the national number length 
    */
    public int getNationalNumberLength();

    /**
    * Find best match for given CCC. 
    *
    * @return Return matching dial plan, or a generic one if none found 
    */
    public DialPlan byCcc(String ccc);

    /**
    * Find best match for given CCC. 
    *
    * @return Return matching dial plan, or a generic one if none found 
    */
    public DialPlan byCccAsInt(int ccc);

    /**
    *
    * @return A list of {@link DialPlan} objects. LinphoneDialPlan  of all known dial
    * plans 
    */
    public DialPlan[] getAllList();

    /**
    * Function to get call country code from an e164 number, ex: +33952650121 will
    * return 33. 
    *
    * @param e164 phone number 
    * 
    * @return call country code or -1 if not found 
    */
    public int lookupCccFromE164(String e164);

    /**
    * Function to get call country code from ISO 3166-1 alpha-2 code, ex: FR returns
    * 33. 
    *
    * @param iso country code alpha2 
    * 
    * @return call country code or -1 if not found 
    */
    public int lookupCccFromIso(String iso);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class DialPlanImpl implements DialPlan {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected DialPlanImpl(long ptr) {
        nativePtr = ptr;
    }


    private native String getCountry(long nativePtr);
    @Override
    synchronized public String getCountry()  {
        
        return getCountry(nativePtr);
    }

    private native String getCountryCallingCode(long nativePtr);
    @Override
    synchronized public String getCountryCallingCode()  {
        
        return getCountryCallingCode(nativePtr);
    }

    private native String getInternationalCallPrefix(long nativePtr);
    @Override
    synchronized public String getInternationalCallPrefix()  {
        
        return getInternationalCallPrefix(nativePtr);
    }

    private native boolean isGeneric(long nativePtr);
    @Override
    synchronized public boolean isGeneric()  {
        
        return isGeneric(nativePtr);
    }

    private native String getIsoCountryCode(long nativePtr);
    @Override
    synchronized public String getIsoCountryCode()  {
        
        return getIsoCountryCode(nativePtr);
    }

    private native int getNationalNumberLength(long nativePtr);
    @Override
    synchronized public int getNationalNumberLength()  {
        
        return getNationalNumberLength(nativePtr);
    }

    private native DialPlan byCcc(long nativePtr, String ccc);
    @Override
    synchronized public DialPlan byCcc(String ccc)  {
        
        return (DialPlan)byCcc(nativePtr, ccc);
    }

    private native DialPlan byCccAsInt(long nativePtr, int ccc);
    @Override
    synchronized public DialPlan byCccAsInt(int ccc)  {
        
        return (DialPlan)byCccAsInt(nativePtr, ccc);
    }

    private native DialPlan[] getAllList(long nativePtr);
    @Override
    synchronized public DialPlan[] getAllList()  {
        
        return getAllList(nativePtr);
    }

    private native int lookupCccFromE1644(long nativePtr, String e164);
    @Override
    synchronized public int lookupCccFromE164(String e164)  {
        
        return lookupCccFromE1644(nativePtr, e164);
    }

    private native int lookupCccFromIso(long nativePtr, String iso);
    @Override
    synchronized public int lookupCccFromIso(String iso)  {
        
        return lookupCccFromIso(nativePtr, iso);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
