/*
Event.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* Object representing an event state, which is subcribed or published. 
*
* See also: {@link Core#publish} 
* 
* See also: {@link Core#subscribe} 
*/
public interface Event {
    /**
    * Returns back pointer to the {@link Core} that created this {@link Event}. 
    *
    */
    public Core getCore();

    /**
    * Get full details about an error occured. 
    *
    */
    public ErrorInfo getErrorInfo();

    /**
    * Get the "from" address of the subscription. 
    *
    */
    public Address getFrom();

    /**
    * Get the name of the event as specified in the event package RFC. 
    *
    */
    public String getName();

    /**
    * Get publish state. 
    *
    * If the event object was not created by a publish mechanism, LinphonePublishNone
    * is returned. 
    */
    public PublishState getPublishState();

    /**
    * Return reason code (in case of error state reached). 
    *
    */
    public Reason getReason();

    /**
    * Get the "contact" address of the subscription. 
    *
    * @return The "contact" address of the subscription 
    */
    public Address getRemoteContact();

    /**
    * Get the resource address of the subscription or publish. 
    *
    */
    public Address getResource();

    /**
    * Get subscription direction. 
    *
    * If the object wasn't created by a subscription mechanism,
    * LinphoneSubscriptionInvalidDir is returned. 
    */
    public SubscriptionDir getSubscriptionDir();

    /**
    * Get subscription state. 
    *
    * If the event object was not created by a subscription mechanism,
    * LinphoneSubscriptionNone is returned. 
    */
    public SubscriptionState getSubscriptionState();

    /**
    * Accept an incoming subcription. 
    *
    */
    public int acceptSubscription();

    /**
    * Add a custom header to an outgoing susbscription or publish. 
    *
    * @param value the header's value. 
    */
    public void addCustomHeader(String name, String value);

    /**
    * Deny an incoming subscription with given reason. 
    *
    */
    public int denySubscription(Reason reason);

    /**
    * Obtain the value of a given header for an incoming subscription. 
    *
    * @param name header's name 
    * 
    * @return the header's value or null if such header doesn't exist. 
    */
    public String getCustomHeader(String name);

    /**
    * Send a notification. 
    *
    * @param body an optional body containing the actual notification data. 
    * 
    * @return 0 if successful, -1 otherwise. 
    */
    public int notify(Content body);

    /**
    * Prevent an event from refreshing its publish. 
    *
    * This is useful to let registrations to expire naturally (or) when the
    * application wants to keep control on when refreshes are sent. The refreshing
    * operations can be resumed with {@link ProxyConfig#refreshRegister}. 
    */
    public void pausePublish();

    /**
    * Refresh an outgoing publish keeping the same body. 
    *
    * @return 0 if successful, -1 otherwise. 
    */
    public int refreshPublish();

    /**
    * Refresh an outgoing subscription keeping the same body. 
    *
    * @return 0 if successful, -1 otherwise. 
    */
    public int refreshSubscribe();

    /**
    * Send a publish created by {@link Core#createPublish}. 
    *
    * @param body the new data to be published 
    */
    public int sendPublish(Content body);

    /**
    * Send a subscription previously created by {@link Core#createSubscribe}. 
    *
    * @param body optional content to attach with the subscription. 
    * 
    * @return 0 if successful, -1 otherwise. 
    */
    public int sendSubscribe(Content body);

    /**
    * Terminate an incoming or outgoing subscription that was previously acccepted,
    * or a previous publication. 
    *
    * The {@link Event} shall not be used anymore after this operation, unless the
    * application explicitely took a reference on the object with linphone_event_ref. 
    */
    public void terminate();

    /**
    * Update (refresh) a publish. 
    *
    * @param body the new data to be published 
    */
    public int updatePublish(Content body);

    /**
    * Update (refresh) an outgoing subscription, changing the body. 
    *
    * @param body an optional body to include in the subscription update, may be
    * null.    
    */
    public int updateSubscribe(@Nullable Content body);

    /**
    */
    public void addListener(EventListener listener);

    /**
    */
    public void removeListener(EventListener listener);

    /**
    */
    public void setListener(EventListener listener);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class EventImpl implements Event {

    protected long nativePtr = 0;
    protected Object userData = null;
    protected Core core = null;

    protected EventImpl(long ptr) {
        nativePtr = ptr;
        core = getCore();
    }


    private native Core getCore(long nativePtr);
    @Override
    synchronized public Core getCore()  {
        
        return (Core)getCore(nativePtr);
    }

    private native ErrorInfo getErrorInfo(long nativePtr);
    @Override
    synchronized public ErrorInfo getErrorInfo()  {
        synchronized(core) { 
        return (ErrorInfo)getErrorInfo(nativePtr);
        }
    }

    private native Address getFrom(long nativePtr);
    @Override
    synchronized public Address getFrom()  {
        synchronized(core) { 
        return (Address)getFrom(nativePtr);
        }
    }

    private native String getName(long nativePtr);
    @Override
    synchronized public String getName()  {
        synchronized(core) { 
        return getName(nativePtr);
        }
    }

    private native int getPublishState(long nativePtr);
    @Override
    synchronized public PublishState getPublishState()  {
        synchronized(core) { 
        return PublishState.fromInt(getPublishState(nativePtr));
        }
    }

    private native int getReason(long nativePtr);
    @Override
    synchronized public Reason getReason()  {
        synchronized(core) { 
        return Reason.fromInt(getReason(nativePtr));
        }
    }

    private native Address getRemoteContact(long nativePtr);
    @Override
    synchronized public Address getRemoteContact()  {
        synchronized(core) { 
        return (Address)getRemoteContact(nativePtr);
        }
    }

    private native Address getResource(long nativePtr);
    @Override
    synchronized public Address getResource()  {
        synchronized(core) { 
        return (Address)getResource(nativePtr);
        }
    }

    private native int getSubscriptionDir(long nativePtr);
    @Override
    synchronized public SubscriptionDir getSubscriptionDir()  {
        synchronized(core) { 
        return SubscriptionDir.fromInt(getSubscriptionDir(nativePtr));
        }
    }

    private native int getSubscriptionState(long nativePtr);
    @Override
    synchronized public SubscriptionState getSubscriptionState()  {
        synchronized(core) { 
        return SubscriptionState.fromInt(getSubscriptionState(nativePtr));
        }
    }

    private native int acceptSubscription(long nativePtr);
    @Override
    synchronized public int acceptSubscription()  {
        synchronized(core) { 
        return acceptSubscription(nativePtr);
        }
    }

    private native void addCustomHeader(long nativePtr, String name, String value);
    @Override
    synchronized public void addCustomHeader(String name, String value)  {
        synchronized(core) { 
        addCustomHeader(nativePtr, name, value);
        }
    }

    private native int denySubscription(long nativePtr, int reason);
    @Override
    synchronized public int denySubscription(Reason reason)  {
        synchronized(core) { 
        return denySubscription(nativePtr, reason.toInt());
        }
    }

    private native String getCustomHeader(long nativePtr, String name);
    @Override
    synchronized public String getCustomHeader(String name)  {
        synchronized(core) { 
        return getCustomHeader(nativePtr, name);
        }
    }

    private native int notify(long nativePtr, Content body);
    @Override
    synchronized public int notify(Content body)  {
        synchronized(core) { 
        return notify(nativePtr, body);
        }
    }

    private native void pausePublish(long nativePtr);
    @Override
    synchronized public void pausePublish()  {
        synchronized(core) { 
        pausePublish(nativePtr);
        }
    }

    private native int refreshPublish(long nativePtr);
    @Override
    synchronized public int refreshPublish()  {
        synchronized(core) { 
        return refreshPublish(nativePtr);
        }
    }

    private native int refreshSubscribe(long nativePtr);
    @Override
    synchronized public int refreshSubscribe()  {
        synchronized(core) { 
        return refreshSubscribe(nativePtr);
        }
    }

    private native int sendPublish(long nativePtr, Content body);
    @Override
    synchronized public int sendPublish(Content body)  {
        synchronized(core) { 
        return sendPublish(nativePtr, body);
        }
    }

    private native int sendSubscribe(long nativePtr, Content body);
    @Override
    synchronized public int sendSubscribe(Content body)  {
        synchronized(core) { 
        return sendSubscribe(nativePtr, body);
        }
    }

    private native void terminate(long nativePtr);
    @Override
    synchronized public void terminate()  {
        synchronized(core) { 
        terminate(nativePtr);
        }
    }

    private native int updatePublish(long nativePtr, Content body);
    @Override
    synchronized public int updatePublish(Content body)  {
        synchronized(core) { 
        return updatePublish(nativePtr, body);
        }
    }

    private native int updateSubscribe(long nativePtr, Content body);
    @Override
    synchronized public int updateSubscribe(@Nullable Content body)  {
        synchronized(core) { 
        return updateSubscribe(nativePtr, body);
        }
    }

    private native void addListener(long nativePtr, EventListener listener);
    @Override
    synchronized public void addListener(EventListener listener)  {
        
        addListener(nativePtr, listener);
    }

    private native void removeListener(long nativePtr, EventListener listener);
    @Override
    synchronized public void removeListener(EventListener listener)  {
        
        removeListener(nativePtr, listener);
    }

    private native void setListener(long nativePtr, EventListener listener);
    @Override
    synchronized public void setListener(EventListener listener)  {
        
        setListener(nativePtr, listener);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
