/*
EventLog.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* Base object of events. 
*
*/
public interface EventLog {
    public enum SecurityEventType {
        /**
        * Event is not a security event. 
        *
        */
        None(0),

        /**
        * Chatroom security level downgraded event. 
        *
        */
        SecurityLevelDowngraded(1),

        /**
        * Participant has exceeded the maximum number of device event. 
        *
        */
        ParticipantMaxDeviceCountExceeded(2),

        /**
        * Peer device instant messaging encryption identity key has changed event. 
        *
        */
        EncryptionIdentityKeyChanged(3),

        /**
        * Man in the middle detected event. 
        *
        */
        ManInTheMiddleDetected(4);

        protected final int mValue;

        private SecurityEventType (int value) {
            mValue = value;
        }

        static public SecurityEventType fromInt(int value) throws RuntimeException {
            switch(value) {
            case 0: return None;
            case 1: return SecurityLevelDowngraded;
            case 2: return ParticipantMaxDeviceCountExceeded;
            case 3: return EncryptionIdentityKeyChanged;
            case 4: return ManInTheMiddleDetected;
            default:
                throw new RuntimeException("Unhandled enum value " + value + " for SecurityEventType");
            }
        }

        public int toInt() {
            return mValue;
        }
    };

    public enum Type {
        /**
        * No defined event. 
        *
        */
        None(0),

        /**
        * Conference (created) event. 
        *
        */
        ConferenceCreated(1),

        /**
        * Conference (terminated) event. 
        *
        */
        ConferenceTerminated(2),

        /**
        * Conference call (start) event. 
        *
        */
        ConferenceCallStart(3),

        /**
        * Conference call (end) event. 
        *
        */
        ConferenceCallEnd(4),

        /**
        * Conference chat message event. 
        *
        */
        ConferenceChatMessage(5),

        /**
        * Conference participant (added) event. 
        *
        */
        ConferenceParticipantAdded(6),

        /**
        * Conference participant (removed) event. 
        *
        */
        ConferenceParticipantRemoved(7),

        /**
        * Conference participant (set admin) event. 
        *
        */
        ConferenceParticipantSetAdmin(8),

        /**
        * Conference participant (unset admin) event. 
        *
        */
        ConferenceParticipantUnsetAdmin(9),

        /**
        * Conference participant device (added) event. 
        *
        */
        ConferenceParticipantDeviceAdded(10),

        /**
        * Conference participant device (removed) event. 
        *
        */
        ConferenceParticipantDeviceRemoved(11),

        /**
        * Conference subject event. 
        *
        */
        ConferenceSubjectChanged(12),

        /**
        * Conference encryption security event. 
        *
        */
        ConferenceSecurityEvent(13),

        /**
        * Conference ephemeral message (ephemeral message lifetime changed) event. 
        *
        */
        ConferenceEphemeralMessageLifetimeChanged(14),

        /**
        * Conference ephemeral message (ephemeral message enabled) event. 
        *
        */
        ConferenceEphemeralMessageEnabled(15),

        /**
        * Conference ephemeral message (ephemeral message disabled) event. 
        *
        */
        ConferenceEphemeralMessageDisabled(16);

        protected final int mValue;

        private Type (int value) {
            mValue = value;
        }

        static public Type fromInt(int value) throws RuntimeException {
            switch(value) {
            case 0: return None;
            case 1: return ConferenceCreated;
            case 2: return ConferenceTerminated;
            case 3: return ConferenceCallStart;
            case 4: return ConferenceCallEnd;
            case 5: return ConferenceChatMessage;
            case 6: return ConferenceParticipantAdded;
            case 7: return ConferenceParticipantRemoved;
            case 8: return ConferenceParticipantSetAdmin;
            case 9: return ConferenceParticipantUnsetAdmin;
            case 10: return ConferenceParticipantDeviceAdded;
            case 11: return ConferenceParticipantDeviceRemoved;
            case 12: return ConferenceSubjectChanged;
            case 13: return ConferenceSecurityEvent;
            case 14: return ConferenceEphemeralMessageLifetimeChanged;
            case 15: return ConferenceEphemeralMessageEnabled;
            case 16: return ConferenceEphemeralMessageDisabled;
            default:
                throw new RuntimeException("Unhandled enum value " + value + " for Type");
            }
        }

        public int toInt() {
            return mValue;
        }
    };

    /**
    * Returns the call of a conference call event. 
    *
    * @return The conference call. 
    */
    public Call getCall();

    /**
    * Returns the chat message of a conference chat message event. 
    *
    * @return The conference chat message. 
    */
    public ChatMessage getChatMessage();

    /**
    * Returns the creation time of a event log. 
    *
    * @return The event creation time 
    */
    public long getCreationTime();

    /**
    * Returns the device address of a conference participant device event. 
    *
    * @return The conference device address. 
    */
    public Address getDeviceAddress();

    /**
    * Returns the ephemeral message lifetime of a conference ephemeral message event. 
    *
    * Ephemeral lifetime means the time before an ephemeral message which has been
    * viewed gets deleted. 
    * 
    * @return The ephemeral message lifetime. 
    */
    public long getEphemeralMessageLifetime();

    /**
    * Returns the local address of a conference event. 
    *
    * @return The local address. 
    */
    public Address getLocalAddress();

    /**
    * Returns the notify id of a conference notified event. 
    *
    * @return The conference notify id. 
    */
    public int getNotifyId();

    /**
    * Returns the participant address of a conference participant event. 
    *
    * @return The conference participant address. 
    */
    public Address getParticipantAddress();

    /**
    * Returns the peer address of a conference event. 
    *
    * @return The peer address. 
    */
    public Address getPeerAddress();

    /**
    * Returns the faulty device address of a conference security event. 
    *
    * @return The address of the faulty device. 
    */
    public Address getSecurityEventFaultyDeviceAddress();

    /**
    * Returns the type of security event. 
    *
    * @return The security event type. 
    */
    public EventLog.SecurityEventType getSecurityEventType();

    /**
    * Returns the subject of a conference subject event. 
    *
    * @return The conference subject. 
    */
    public String getSubject();

    /**
    * Returns the type of a event log. 
    *
    * @return The event type 
    */
    public EventLog.Type getType();

    /**
    * Delete event log from database. 
    *
    */
    public void deleteFromDatabase();

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class EventLogImpl implements EventLog {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected EventLogImpl(long ptr) {
        nativePtr = ptr;
    }


    private native Call getCall(long nativePtr);
    @Override
    synchronized public Call getCall()  {
        
        return (Call)getCall(nativePtr);
    }

    private native ChatMessage getChatMessage(long nativePtr);
    @Override
    synchronized public ChatMessage getChatMessage()  {
        
        return (ChatMessage)getChatMessage(nativePtr);
    }

    private native long getCreationTime(long nativePtr);
    @Override
    synchronized public long getCreationTime()  {
        
        return getCreationTime(nativePtr);
    }

    private native Address getDeviceAddress(long nativePtr);
    @Override
    synchronized public Address getDeviceAddress()  {
        
        return (Address)getDeviceAddress(nativePtr);
    }

    private native long getEphemeralMessageLifetime(long nativePtr);
    @Override
    synchronized public long getEphemeralMessageLifetime()  {
        
        return getEphemeralMessageLifetime(nativePtr);
    }

    private native Address getLocalAddress(long nativePtr);
    @Override
    synchronized public Address getLocalAddress()  {
        
        return (Address)getLocalAddress(nativePtr);
    }

    private native int getNotifyId(long nativePtr);
    @Override
    synchronized public int getNotifyId()  {
        
        return getNotifyId(nativePtr);
    }

    private native Address getParticipantAddress(long nativePtr);
    @Override
    synchronized public Address getParticipantAddress()  {
        
        return (Address)getParticipantAddress(nativePtr);
    }

    private native Address getPeerAddress(long nativePtr);
    @Override
    synchronized public Address getPeerAddress()  {
        
        return (Address)getPeerAddress(nativePtr);
    }

    private native Address getSecurityEventFaultyDeviceAddress(long nativePtr);
    @Override
    synchronized public Address getSecurityEventFaultyDeviceAddress()  {
        
        return (Address)getSecurityEventFaultyDeviceAddress(nativePtr);
    }

    private native int getSecurityEventType(long nativePtr);
    @Override
    synchronized public EventLog.SecurityEventType getSecurityEventType()  {
        
        return EventLog.SecurityEventType.fromInt(getSecurityEventType(nativePtr));
    }

    private native String getSubject(long nativePtr);
    @Override
    synchronized public String getSubject()  {
        
        return getSubject(nativePtr);
    }

    private native int getType(long nativePtr);
    @Override
    synchronized public EventLog.Type getType()  {
        
        return EventLog.Type.fromInt(getType(nativePtr));
    }

    private native void deleteFromDatabase(long nativePtr);
    @Override
    synchronized public void deleteFromDatabase()  {
        
        deleteFromDatabase(nativePtr);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
