/*
Factory.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;
import android.content.Context;
import android.os.Build;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.linphone.mediastream.Log;
import org.linphone.mediastream.Version;
import org.linphone.core.tools.OpenH264DownloadHelper;

/**
* {@link Factory} is a singleton object devoted to the creation of all the object
* of Liblinphone that cannot be created by {@link Core} itself. 
*
*/
public abstract class Factory {
    static Factory _Factory;

    public static final synchronized Factory instance() {
		try {
			if (_Factory == null) {
				_Factory = new FactoryImpl(0); // This value is not relevant, correct factory pointer will be used in JNI layer
			}
		} catch (Exception e) {
			System.err.println("Cannot instanciate factory");
		}
		return _Factory;
	}

    abstract public OpenH264DownloadHelper createOpenH264DownloadHelper(Context context);

    /**
      * Gets the LoggingService singleton
      */
    abstract public LoggingService getLoggingService();

    abstract public void setDebugMode(boolean enable, String tag);

    abstract public Core getCore(long ptr);

    /**
    * Get the directory where the data resources are located. 
    *
    * @return The path to the directory where the data resources are located 
    */
    abstract public String getDataResourcesDir();

    /**
    * Set the directory where the data resources are located. 
    *
    * @param path The path where the data resources are located 
    */
    abstract public void setDataResourcesDir(String path);

    /**
    * Returns a bctbx_list_t of all DialPlans. 
    *
    * @return A list of {@link DialPlan} objects. LinphoneDialPlan  a list of
    * DialPlan 
    */
    abstract public DialPlan[] getDialPlans();

    /**
    * Get the directory where the image resources are located. 
    *
    * @return The path to the directory where the image resources are located 
    */
    abstract public String getImageResourcesDir();

    /**
    * Set the directory where the image resources are located. 
    *
    * @param path The path where the image resources are located 
    */
    abstract public void setImageResourcesDir(String path);

    /**
    * Indicates if the storage in database is available. 
    *
    * @return  if the database storage is available, true otherwise 
    */
    abstract public boolean isDatabaseStorageAvailable();

    /**
    * Indicates if IMDN are available. 
    *
    * @return  if IDMN are available 
    */
    abstract public boolean isImdnAvailable();

    /**
    * Sets the log collection path. 
    *
    * @param path the path of the logs 
    */
    abstract public void setLogCollectionPath(String path);

    /**
    * Get the directory where the mediastreamer2 plugins are located. 
    *
    * @return The path to the directory where the mediastreamer2 plugins are located,
    * or null if it has not been set 
    */
    abstract public String getMspluginsDir();

    /**
    * Set the directory where the mediastreamer2 plugins are located. 
    *
    * @param path The path to the directory where the mediastreamer2 plugins are
    * located 
    */
    abstract public void setMspluginsDir(String path);

    /**
    * Get the directory where the ring resources are located. 
    *
    * @return The path to the directory where the ring resources are located 
    */
    abstract public String getRingResourcesDir();

    /**
    * Set the directory where the ring resources are located. 
    *
    * @param path The path where the ring resources are located 
    */
    abstract public void setRingResourcesDir(String path);

    /**
    * Get the directory where the sound resources are located. 
    *
    * @return The path to the directory where the sound resources are located 
    */
    abstract public String getSoundResourcesDir();

    /**
    * Set the directory where the sound resources are located. 
    *
    * @param path The path where the sound resources are located 
    */
    abstract public void setSoundResourcesDir(String path);

    /**
    * Get the list of standard video definitions supported by Linphone. 
    *
    * @return A list of {@link VideoDefinition} objects. LinphoneVideoDefinition  
    */
    abstract public VideoDefinition[] getSupportedVideoDefinitions();

    /**
    * Get the top directory where the resources are located. 
    *
    * @return The path to the top directory where the resources are located 
    */
    abstract public String getTopResourcesDir();

    /**
    * Set the top directory where the resources are located. 
    *
    * If you only define this top directory, the other resources directory will
    * automatically be derived form this one. 
    * 
    * @param path The path to the top directory where the resources are located 
    */
    abstract public void setTopResourcesDir(String path);

    /**
    * Parse a string holding a SIP URI and create the according {@link Address}
    * object. 
    *
    * @param addr A string holding the SIP URI to parse. 
    * 
    * @return A new {@link Address}. 
    */
    abstract public Address createAddress(String addr);

    /**
    * Creates a {@link AuthInfo} object. 
    *
    * The object can be created empty, that is with all arguments set to null.
    * Username, userid, password, realm and domain can be set later using specific
    * methods. At the end, username and passwd (or ha1) are required. 
    * 
    * @param algorithm The algorithm for encrypting password. 
    * 
    * @return A {@link AuthInfo} object. linphone_auth_info_destroy() must be used to
    * destroy it when no longer needed. The {@link Core} makes a copy of {@link
    * AuthInfo} passed through {@link Core#addAuthInfo}. 
    */
    abstract public AuthInfo createAuthInfo(String username, String userid, String passwd, String ha1, String realm, String domain, String algorithm);

    /**
    * Creates a {@link AuthInfo} object. 
    *
    * The object can be created empty, that is with all arguments set to null.
    * Username, userid, password, realm and domain can be set later using specific
    * methods. At the end, username and passwd (or ha1) are required. 
    * 
    * @param domain The SIP domain for which this authentication information is
    * valid, if it has to be restricted for a single SIP domain. 
    * 
    * @return A {@link AuthInfo} object. linphone_auth_info_destroy() must be used to
    * destroy it when no longer needed. The {@link Core} makes a copy of {@link
    * AuthInfo} passed through {@link Core#addAuthInfo}. 
    */
    abstract public AuthInfo createAuthInfo(String username, String userid, String passwd, String ha1, String realm, String domain);

    /**
    * Creates an object {@link Buffer}. 
    *
    * @return a {@link Buffer} 
    */
    abstract public Buffer createBuffer();

    /**
    * Creates an object {@link Buffer}. 
    *
    * @param size the size of the data 
    * 
    * @return a {@link Buffer} 
    */
    abstract public Buffer createBufferFromData(byte[] data, int size);

    /**
    * Creates an object {@link Buffer}. 
    *
    * @param data the data to set in the buffer 
    * 
    * @return a {@link Buffer} 
    */
    abstract public Buffer createBufferFromString(String data);

    /**
    * Creates an object {@link Config}. 
    *
    * @param path the path of the config 
    * 
    * @return a {@link Config} 
    */
    abstract public Config createConfig(String path);

    /**
    * Creates an object {@link Config}. 
    *
    * @param data the config data 
    * 
    * @return a {@link Config} 
    */
    abstract public Config createConfigFromString(String data);

    /**
    * Creates an object {@link Config}. 
    *
    * @param path the path of the factory 
    * 
    * @return a {@link Config} 
    */
    abstract public Config createConfigWithFactory(String path, String factoryPath);

    /**
    * Creates an object {@link Content}. 
    *
    * @return a {@link Content} 
    */
    abstract public Content createContent();

    /**
    * Instantiate a {@link Core} object. 
    *
    * The {@link Core} object is the primary handle for doing all phone actions. It
    * should be unique within your application. The {@link Core} object is not
    * started automatically, you need to call {@link Core#start} to that effect. The
    * returned {@link Core} will be in LinphoneGlobalState Ready. Core ressources can
    * be released using {@link Core#stop} which is strongly encouraged on garbage
    * collected languages. 
    * 
    * @param systemContext A pointer to a system object required by the core to
    * operate. Currently it is required to pass an android Context on android, pass
    * null on other platforms. 
    * 
    * See also: linphone_core_new_with_config_3 
    */
    abstract public Core createCore(String configPath, String factoryConfigPath, Object systemContext);

    /**
    * Instantiate a {@link Core} object with a given LinphoneConfig. 
    *
    * The {@link Core} object is the primary handle for doing all phone actions. It
    * should be unique within your application. The {@link Core} object is not
    * started automatically, you need to call {@link Core#start} to that effect. The
    * returned {@link Core} will be in LinphoneGlobalState Ready. Core ressources can
    * be released using {@link Core#stop} which is strongly encouraged on garbage
    * collected languages. 
    * 
    * @param systemContext A pointer to a system object required by the core to
    * operate. Currently it is required to pass an android Context on android, pass
    * null on other platforms. 
    * 
    * See also: linphone_factory_create_core_3 
    */
    abstract public Core createCoreWithConfig(Config config, Object systemContext);

    /**
    * Creates an object LinphoneErrorInfo. 
    *
    * @return {@link ErrorInfo} object. 
    */
    abstract public ErrorInfo createErrorInfo();

    /**
    * Create a #LinphoneParticipantDeviceIdentity object. 
    *
    * @param name the name given to the device. 
    * 
    * @return A new #LinphoneParticipantDeviceIdentity. 
    */
    abstract public ParticipantDeviceIdentity createParticipantDeviceIdentity(Address address, String name);

    /**
    * Creates an object LinphoneRange. 
    *
    * @return {@link Range} object. 
    */
    abstract public Range createRange();

    /**
    * Creates an object LinphoneTransports. 
    *
    * @return {@link Transports} object. 
    */
    abstract public Transports createTransports();

    /**
    * Creates an object {@link TunnelConfig}. 
    *
    * @return a {@link TunnelConfig} 
    */
    abstract public TunnelConfig createTunnelConfig();

    /**
    * Create an empty {@link Vcard}. 
    *
    * @return a new {@link Vcard}. 
    */
    abstract public Vcard createVcard();

    /**
    * Creates an object LinphoneVideoActivationPolicy. 
    *
    * @return {@link VideoActivationPolicy} object. 
    */
    abstract public VideoActivationPolicy createVideoActivationPolicy();

    /**
    * Create a {@link VideoDefinition} from a given width and height. 
    *
    * @param height The height of the created video definition 
    * 
    * @return A new {@link VideoDefinition} object 
    */
    abstract public VideoDefinition createVideoDefinition(int width, int height);

    /**
    * Create a {@link VideoDefinition} from a given standard definition name. 
    *
    * @param name The standard definition name of the video definition to create 
    * 
    * @return A new {@link VideoDefinition} object 
    */
    abstract public VideoDefinition createVideoDefinitionFromName(String name);

    /**
    * Enables or disables log collection. 
    *
    * @param state the policy for log collection 
    */
    abstract public void enableLogCollection(LogCollectionState state);

    /**
    * Indicates if the given LinphoneChatRoomBackend is available. 
    *
    * @param chatroomBackend the LinphoneChatRoomBackend 
    * 
    * @return  if the chatroom backend is available, true otherwise 
    */
    abstract public boolean isChatroomBackendAvailable(ChatRoomBackend chatroomBackend);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    abstract public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    abstract public Object getUserData();
}

class FactoryImpl extends Factory {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected FactoryImpl(long ptr) {
        nativePtr = ptr;
    }

	private static boolean loadOptionalLibrary(String s) {
		try {
			System.loadLibrary(s);
			return true;
		} catch (Throwable e) {
			android.util.Log.w("FactoryImpl", "Unable to load optional library " + s + ": " + e.getMessage());
		}
		return false;
	}

	static {
		System.loadLibrary("c++_shared");
		loadOptionalLibrary("ffmpeg-linphone");
		System.loadLibrary("bctoolbox");
		System.loadLibrary("ortp");
		System.loadLibrary("mediastreamer");
		System.loadLibrary("linphone");
		Version.dumpCapabilities();
	}

    public OpenH264DownloadHelper createOpenH264DownloadHelper(Context context) {
		if (context == null) {
			new CoreException("Cannot create OpenH264DownloadHelper");
			return null;
		}
		return new OpenH264DownloadHelper(context);
	}

	private native Core getCore(long nativePtr, long ptr);
	@Override
	public Core getCore(long ptr) {
	    return getCore(nativePtr, ptr);
	}

    @Override
	public LoggingService getLoggingService() {
		LoggingService l = new LoggingServiceImpl(0);
		return l.get();
	}

    @Override
    public native void setDebugMode(boolean enable, String tag);

    private native String getDataResourcesDir(long nativePtr);
    @Override
    synchronized public String getDataResourcesDir()  {
        
        return getDataResourcesDir(nativePtr);
    }

    private native void setDataResourcesDir(long nativePtr, String path);
    @Override
    synchronized public void setDataResourcesDir(String path)  {
        
        setDataResourcesDir(nativePtr, path);
    }

    private native DialPlan[] getDialPlans(long nativePtr);
    @Override
    synchronized public DialPlan[] getDialPlans()  {
        
        return getDialPlans(nativePtr);
    }

    private native String getImageResourcesDir(long nativePtr);
    @Override
    synchronized public String getImageResourcesDir()  {
        
        return getImageResourcesDir(nativePtr);
    }

    private native void setImageResourcesDir(long nativePtr, String path);
    @Override
    synchronized public void setImageResourcesDir(String path)  {
        
        setImageResourcesDir(nativePtr, path);
    }

    private native boolean isDatabaseStorageAvailable(long nativePtr);
    @Override
    synchronized public boolean isDatabaseStorageAvailable()  {
        
        return isDatabaseStorageAvailable(nativePtr);
    }

    private native boolean isImdnAvailable(long nativePtr);
    @Override
    synchronized public boolean isImdnAvailable()  {
        
        return isImdnAvailable(nativePtr);
    }

    private native void setLogCollectionPath(long nativePtr, String path);
    @Override
    synchronized public void setLogCollectionPath(String path)  {
        
        setLogCollectionPath(nativePtr, path);
    }

    private native String getMspluginsDir(long nativePtr);
    @Override
    synchronized public String getMspluginsDir()  {
        
        return getMspluginsDir(nativePtr);
    }

    private native void setMspluginsDir(long nativePtr, String path);
    @Override
    synchronized public void setMspluginsDir(String path)  {
        
        setMspluginsDir(nativePtr, path);
    }

    private native String getRingResourcesDir(long nativePtr);
    @Override
    synchronized public String getRingResourcesDir()  {
        
        return getRingResourcesDir(nativePtr);
    }

    private native void setRingResourcesDir(long nativePtr, String path);
    @Override
    synchronized public void setRingResourcesDir(String path)  {
        
        setRingResourcesDir(nativePtr, path);
    }

    private native String getSoundResourcesDir(long nativePtr);
    @Override
    synchronized public String getSoundResourcesDir()  {
        
        return getSoundResourcesDir(nativePtr);
    }

    private native void setSoundResourcesDir(long nativePtr, String path);
    @Override
    synchronized public void setSoundResourcesDir(String path)  {
        
        setSoundResourcesDir(nativePtr, path);
    }

    private native VideoDefinition[] getSupportedVideoDefinitions(long nativePtr);
    @Override
    synchronized public VideoDefinition[] getSupportedVideoDefinitions()  {
        
        return getSupportedVideoDefinitions(nativePtr);
    }

    private native String getTopResourcesDir(long nativePtr);
    @Override
    synchronized public String getTopResourcesDir()  {
        
        return getTopResourcesDir(nativePtr);
    }

    private native void setTopResourcesDir(long nativePtr, String path);
    @Override
    synchronized public void setTopResourcesDir(String path)  {
        
        setTopResourcesDir(nativePtr, path);
    }

    private native Address createAddress(long nativePtr, String addr);
    @Override
    synchronized public Address createAddress(String addr)  {
        
        return (Address)createAddress(nativePtr, addr);
    }

    private native AuthInfo createAuthInfo2(long nativePtr, String username, String userid, String passwd, String ha1, String realm, String domain, String algorithm);
    @Override
    synchronized public AuthInfo createAuthInfo(String username, String userid, String passwd, String ha1, String realm, String domain, String algorithm)  {
        
        return (AuthInfo)createAuthInfo2(nativePtr, username, userid, passwd, ha1, realm, domain, algorithm);
    }

    private native AuthInfo createAuthInfo(long nativePtr, String username, String userid, String passwd, String ha1, String realm, String domain);
    @Override
    synchronized public AuthInfo createAuthInfo(String username, String userid, String passwd, String ha1, String realm, String domain)  {
        
        return (AuthInfo)createAuthInfo(nativePtr, username, userid, passwd, ha1, realm, domain);
    }

    private native Buffer createBuffer(long nativePtr);
    @Override
    synchronized public Buffer createBuffer()  {
        
        return (Buffer)createBuffer(nativePtr);
    }

    private native Buffer createBufferFromData(long nativePtr, byte[] data, int size);
    @Override
    synchronized public Buffer createBufferFromData(byte[] data, int size)  {
        
        return (Buffer)createBufferFromData(nativePtr, data, size);
    }

    private native Buffer createBufferFromString(long nativePtr, String data);
    @Override
    synchronized public Buffer createBufferFromString(String data)  {
        
        return (Buffer)createBufferFromString(nativePtr, data);
    }

    private native Config createConfig(long nativePtr, String path);
    @Override
    synchronized public Config createConfig(String path)  {
        
        return (Config)createConfig(nativePtr, path);
    }

    private native Config createConfigFromString(long nativePtr, String data);
    @Override
    synchronized public Config createConfigFromString(String data)  {
        
        return (Config)createConfigFromString(nativePtr, data);
    }

    private native Config createConfigWithFactory(long nativePtr, String path, String factoryPath);
    @Override
    synchronized public Config createConfigWithFactory(String path, String factoryPath)  {
        
        return (Config)createConfigWithFactory(nativePtr, path, factoryPath);
    }

    private native Content createContent(long nativePtr);
    @Override
    synchronized public Content createContent()  {
        
        return (Content)createContent(nativePtr);
    }

    private native Core createCore3(long nativePtr, String configPath, String factoryConfigPath, Object systemContext);
    @Override
    synchronized public Core createCore(String configPath, String factoryConfigPath, Object systemContext)  {
        
        return (Core)createCore3(nativePtr, configPath, factoryConfigPath, systemContext);
    }

    private native Core createCoreWithConfig3(long nativePtr, Config config, Object systemContext);
    @Override
    synchronized public Core createCoreWithConfig(Config config, Object systemContext)  {
        
        return (Core)createCoreWithConfig3(nativePtr, config, systemContext);
    }

    private native ErrorInfo createErrorInfo(long nativePtr);
    @Override
    synchronized public ErrorInfo createErrorInfo()  {
        
        return (ErrorInfo)createErrorInfo(nativePtr);
    }

    private native ParticipantDeviceIdentity createParticipantDeviceIdentity(long nativePtr, Address address, String name);
    @Override
    synchronized public ParticipantDeviceIdentity createParticipantDeviceIdentity(Address address, String name)  {
        
        return (ParticipantDeviceIdentity)createParticipantDeviceIdentity(nativePtr, address, name);
    }

    private native Range createRange(long nativePtr);
    @Override
    synchronized public Range createRange()  {
        
        return (Range)createRange(nativePtr);
    }

    private native Transports createTransports(long nativePtr);
    @Override
    synchronized public Transports createTransports()  {
        
        return (Transports)createTransports(nativePtr);
    }

    private native TunnelConfig createTunnelConfig(long nativePtr);
    @Override
    synchronized public TunnelConfig createTunnelConfig()  {
        
        return (TunnelConfig)createTunnelConfig(nativePtr);
    }

    private native Vcard createVcard(long nativePtr);
    @Override
    synchronized public Vcard createVcard()  {
        
        return (Vcard)createVcard(nativePtr);
    }

    private native VideoActivationPolicy createVideoActivationPolicy(long nativePtr);
    @Override
    synchronized public VideoActivationPolicy createVideoActivationPolicy()  {
        
        return (VideoActivationPolicy)createVideoActivationPolicy(nativePtr);
    }

    private native VideoDefinition createVideoDefinition(long nativePtr, int width, int height);
    @Override
    synchronized public VideoDefinition createVideoDefinition(int width, int height)  {
        
        return (VideoDefinition)createVideoDefinition(nativePtr, width, height);
    }

    private native VideoDefinition createVideoDefinitionFromName(long nativePtr, String name);
    @Override
    synchronized public VideoDefinition createVideoDefinitionFromName(String name)  {
        
        return (VideoDefinition)createVideoDefinitionFromName(nativePtr, name);
    }

    private native void enableLogCollection(long nativePtr, int state);
    @Override
    synchronized public void enableLogCollection(LogCollectionState state)  {
        
        enableLogCollection(nativePtr, state.toInt());
    }

    private native boolean isChatroomBackendAvailable(long nativePtr, int chatroomBackend);
    @Override
    synchronized public boolean isChatroomBackendAvailable(ChatRoomBackend chatroomBackend)  {
        
        return isChatroomBackendAvailable(nativePtr, chatroomBackend.toInt());
    }


    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
