/*
Friend.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* Represents a buddy, all presence actions like subscription and status change
* notification are performed on this object. 
*
*/
public interface Friend {
    /**
    * Get address of this friend. 
    *
    * Note: the {@link Address} object returned is hold by the LinphoneFriend,
    * however calling several time this function may return different objects. 
    * 
    * @return {@link Address} 
    */
    public Address getAddress();

    /**
    * Set {@link Address} for this friend. 
    *
    * @param address {@link Address} 
    */
    public int setAddress(Address address);

    /**
    * Returns a list of {@link Address} for this friend. 
    *
    * @return A list of {@link Address} objects. LinphoneAddress  
    */
    public Address[] getAddresses();

    /**
    * Returns the capabilities associated to this friend. 
    *
    * @return an int representing the capabilities of the friend 
    */
    public int getCapabilities();

    /**
    * Get the consolidated presence of a friend. 
    *
    * @return The consolidated presence of the friend 
    */
    public ConsolidatedPresence getConsolidatedPresence();

    /**
    * Returns the {@link Core} object managing this friend, if any. 
    *
    */
    public Core getCore();

    /**
    * get current subscription policy for this {@link Friend} 
    *
    * @return LinphoneSubscribePolicy 
    */
    public SubscribePolicy getIncSubscribePolicy();

    /**
    * Configure incoming subscription policy for this friend. 
    *
    * @param pol LinphoneSubscribePolicy policy to apply. 
    */
    public int setIncSubscribePolicy(SubscribePolicy pol);

    /**
    * Tells whether we already received presence information for a friend. 
    *
    * @return  if presence information has been received for the friend, true
    * otherwise. 
    */
    public boolean isPresenceReceived();

    /**
    * Get the display name for this friend. 
    *
    * @return The display name of this friend 
    */
    public String getName();

    /**
    * Set the display name for this friend. 
    *
    * @param name 
    */
    public int setName(String name);

    /**
    * Returns a list of phone numbers for this friend. 
    *
    * @return A list of const char * objects. const char *  
    */
    public String[] getPhoneNumbers();

    /**
    * Get the presence model of a friend. 
    *
    * @return A {@link PresenceModel} object, or null if the friend do not have
    * presence information (in which case he is considered offline) 
    */
    public PresenceModel getPresenceModel();

    /**
    * Set the presence model of a friend. 
    *
    * @param presence The {@link PresenceModel} object to set for the friend 
    */
    public void setPresenceModel(PresenceModel presence);

    /**
    * Get the reference key of a friend. 
    *
    * @return The reference key of the friend. 
    */
    public String getRefKey();

    /**
    * Set the reference key of a friend. 
    *
    * @param key The reference key to use for the friend. 
    */
    public void setRefKey(String key);

    /**
    * get subscription flag value 
    *
    * @return returns true is subscription is activated for this friend 
    */
    public boolean subscribesEnabled();

    /**
    * Configure {@link Friend} to subscribe to presence information. 
    *
    * @param val if true this friend will receive subscription message 
    */
    public int enableSubscribes(boolean val);

    /**
    * Get subscription state of a friend. 
    *
    * @return LinphoneSubscriptionState 
    */
    public SubscriptionState getSubscriptionState();

    /**
    * Returns the vCard object associated to this friend, if any. 
    *
    */
    public Vcard getVcard();

    /**
    * Binds a vCard object to a friend. 
    *
    * @param vcard The vCard object to bind 
    */
    public void setVcard(Vcard vcard);

    /**
    * Adds an address in this friend. 
    *
    * @param addr {@link Address} object 
    */
    public void addAddress(Address addr);

    /**
    * Adds a phone number in this friend. 
    *
    * @param phone number to add 
    */
    public void addPhoneNumber(String phone);

    /**
    * Creates a vCard object associated to this friend if there isn't one yet and if
    * the full name is available, either by the parameter or the one in the friend's
    * SIP URI. 
    *
    * @param name The full name of the friend or null to use the one from the
    * friend's SIP URI 
    * 
    * @return true if the vCard has been created, false if it wasn't possible (for
    * exemple if name and the friend's SIP URI are null or if the friend's SIP URI
    * doesn't have a display name), or if there is already one vcard 
    */
    public boolean createVcard(String name);

    /**
    * Commits modification made to the friend configuration. 
    *
    */
    public void done();

    /**
    * Starts editing a friend configuration. 
    *
    * Because friend configuration must be consistent, applications MUST call {@link
    * Friend#edit} before doing any attempts to modify friend configuration (such as
    * address  or subscription policy and so on). Once the modifications are done,
    * then the application must call {@link Friend#done} to commit the changes. 
    */
    public void edit();

    /**
    * Returns the version of a friend's capbility. 
    *
    * @param capability LinphoneFriendCapability object 
    * 
    * @return the version of a friend's capbility. 
    */
    public float getCapabilityVersion(FriendCapability capability);

    /**
    * Get the presence model for a specific SIP URI or phone number of a friend. 
    *
    * @param uriOrTel The SIP URI or phone number for which to get the presence model 
    * 
    * @return A {@link PresenceModel} object, or null if the friend do not have
    * presence information for this SIP URI or phone number 
    */
    public PresenceModel getPresenceModelForUriOrTel(String uriOrTel);

    /**
    * Returns whether or not a friend has a capbility. 
    *
    * @param capability LinphoneFriendCapability object 
    * 
    * @return whether or not a friend has a capbility 
    */
    public boolean hasCapability(FriendCapability capability);

    /**
    * Returns whether or not a friend has a capbility with a given version. 
    *
    * @param version the version to test 
    * 
    * @return whether or not a friend has a capbility with a given version or -1.0 if
    * friend has not capability. 
    */
    public boolean hasCapabilityWithVersion(FriendCapability capability, float version);

    /**
    * Returns whether or not a friend has a capbility with a given version or more. 
    *
    * @param version the version to test 
    * 
    * @return whether or not a friend has a capbility with a given version or more. 
    */
    public boolean hasCapabilityWithVersionOrMore(FriendCapability capability, float version);

    /**
    * Check that the given friend is in a friend list. 
    *
    * @return  if the friend is in a friend list, true otherwise. 
    */
    public boolean inList();

    /**
    * Removes a friend from it's friend list and from the rc if exists. 
    *
    */
    public void remove();

    /**
    * Removes an address in this friend. 
    *
    * @param addr {@link Address} object 
    */
    public void removeAddress(Address addr);

    /**
    * Removes a phone number in this friend. 
    *
    * @param phone number to remove 
    */
    public void removePhoneNumber(String phone);

    /**
    * Saves a friend either in database if configured, otherwise in linphonerc. 
    *
    * @param lc the linphone core 
    */
    public void save(Core lc);

    /**
    * Set the presence model for a specific SIP URI or phone number of a friend. 
    *
    * @param presence The {@link PresenceModel} object to set 
    */
    public void setPresenceModelForUriOrTel(String uriOrTel, PresenceModel presence);

    /**
    * Contructor same as linphone_friend_new + {@link Friend#setAddress} 
    *
    * @param vcard a vCard object 
    * 
    * @return a new {@link Friend} with vCard initialized  
    */
    public Friend newFromVcard(Vcard vcard);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class FriendImpl implements Friend {

    protected long nativePtr = 0;
    protected Object userData = null;
    protected Core core = null;

    protected FriendImpl(long ptr) {
        nativePtr = ptr;
        core = getCore();
    }


    private native Address getAddress(long nativePtr);
    @Override
    synchronized public Address getAddress()  {
        synchronized(core) { 
        return (Address)getAddress(nativePtr);
        }
    }

    private native int setAddress(long nativePtr, Address address);
    @Override
    synchronized public int setAddress(Address address)  {
        synchronized(core) { 
        return setAddress(nativePtr, address);
        }
    }

    private native Address[] getAddresses(long nativePtr);
    @Override
    synchronized public Address[] getAddresses()  {
        synchronized(core) { 
        return getAddresses(nativePtr);
        }
    }

    private native int getCapabilities(long nativePtr);
    @Override
    synchronized public int getCapabilities()  {
        synchronized(core) { 
        return getCapabilities(nativePtr);
        }
    }

    private native int getConsolidatedPresence(long nativePtr);
    @Override
    synchronized public ConsolidatedPresence getConsolidatedPresence()  {
        synchronized(core) { 
        return ConsolidatedPresence.fromInt(getConsolidatedPresence(nativePtr));
        }
    }

    private native Core getCore(long nativePtr);
    @Override
    synchronized public Core getCore()  {
        
        return (Core)getCore(nativePtr);
    }

    private native int getIncSubscribePolicy(long nativePtr);
    @Override
    synchronized public SubscribePolicy getIncSubscribePolicy()  {
        synchronized(core) { 
        return SubscribePolicy.fromInt(getIncSubscribePolicy(nativePtr));
        }
    }

    private native int setIncSubscribePolicy(long nativePtr, int pol);
    @Override
    synchronized public int setIncSubscribePolicy(SubscribePolicy pol)  {
        synchronized(core) { 
        return setIncSubscribePolicy(nativePtr, pol.toInt());
        }
    }

    private native boolean isPresenceReceived(long nativePtr);
    @Override
    synchronized public boolean isPresenceReceived()  {
        synchronized(core) { 
        return isPresenceReceived(nativePtr);
        }
    }

    private native String getName(long nativePtr);
    @Override
    synchronized public String getName()  {
        synchronized(core) { 
        return getName(nativePtr);
        }
    }

    private native int setName(long nativePtr, String name);
    @Override
    synchronized public int setName(String name)  {
        synchronized(core) { 
        return setName(nativePtr, name);
        }
    }

    private native String[] getPhoneNumbers(long nativePtr);
    @Override
    synchronized public String[] getPhoneNumbers()  {
        synchronized(core) { 
        return getPhoneNumbers(nativePtr);
        }
    }

    private native PresenceModel getPresenceModel(long nativePtr);
    @Override
    synchronized public PresenceModel getPresenceModel()  {
        synchronized(core) { 
        return (PresenceModel)getPresenceModel(nativePtr);
        }
    }

    private native void setPresenceModel(long nativePtr, PresenceModel presence);
    @Override
    synchronized public void setPresenceModel(PresenceModel presence)  {
        synchronized(core) { 
        setPresenceModel(nativePtr, presence);
        }
    }

    private native String getRefKey(long nativePtr);
    @Override
    synchronized public String getRefKey()  {
        synchronized(core) { 
        return getRefKey(nativePtr);
        }
    }

    private native void setRefKey(long nativePtr, String key);
    @Override
    synchronized public void setRefKey(String key)  {
        synchronized(core) { 
        setRefKey(nativePtr, key);
        }
    }

    private native boolean subscribesEnabled(long nativePtr);
    @Override
    synchronized public boolean subscribesEnabled()  {
        synchronized(core) { 
        return subscribesEnabled(nativePtr);
        }
    }

    private native int enableSubscribes(long nativePtr, boolean val);
    @Override
    synchronized public int enableSubscribes(boolean val)  {
        synchronized(core) { 
        return enableSubscribes(nativePtr, val);
        }
    }

    private native int getSubscriptionState(long nativePtr);
    @Override
    synchronized public SubscriptionState getSubscriptionState()  {
        synchronized(core) { 
        return SubscriptionState.fromInt(getSubscriptionState(nativePtr));
        }
    }

    private native Vcard getVcard(long nativePtr);
    @Override
    synchronized public Vcard getVcard()  {
        synchronized(core) { 
        return (Vcard)getVcard(nativePtr);
        }
    }

    private native void setVcard(long nativePtr, Vcard vcard);
    @Override
    synchronized public void setVcard(Vcard vcard)  {
        synchronized(core) { 
        setVcard(nativePtr, vcard);
        }
    }

    private native void addAddress(long nativePtr, Address addr);
    @Override
    synchronized public void addAddress(Address addr)  {
        synchronized(core) { 
        addAddress(nativePtr, addr);
        }
    }

    private native void addPhoneNumber(long nativePtr, String phone);
    @Override
    synchronized public void addPhoneNumber(String phone)  {
        synchronized(core) { 
        addPhoneNumber(nativePtr, phone);
        }
    }

    private native boolean createVcard(long nativePtr, String name);
    @Override
    synchronized public boolean createVcard(String name)  {
        synchronized(core) { 
        return createVcard(nativePtr, name);
        }
    }

    private native void done(long nativePtr);
    @Override
    synchronized public void done()  {
        synchronized(core) { 
        done(nativePtr);
        }
    }

    private native void edit(long nativePtr);
    @Override
    synchronized public void edit()  {
        synchronized(core) { 
        edit(nativePtr);
        }
    }

    private native float getCapabilityVersion(long nativePtr, int capability);
    @Override
    synchronized public float getCapabilityVersion(FriendCapability capability)  {
        synchronized(core) { 
        return getCapabilityVersion(nativePtr, capability.toInt());
        }
    }

    private native PresenceModel getPresenceModelForUriOrTel(long nativePtr, String uriOrTel);
    @Override
    synchronized public PresenceModel getPresenceModelForUriOrTel(String uriOrTel)  {
        synchronized(core) { 
        return (PresenceModel)getPresenceModelForUriOrTel(nativePtr, uriOrTel);
        }
    }

    private native boolean hasCapability(long nativePtr, int capability);
    @Override
    synchronized public boolean hasCapability(FriendCapability capability)  {
        synchronized(core) { 
        return hasCapability(nativePtr, capability.toInt());
        }
    }

    private native boolean hasCapabilityWithVersion(long nativePtr, int capability, float version);
    @Override
    synchronized public boolean hasCapabilityWithVersion(FriendCapability capability, float version)  {
        synchronized(core) { 
        return hasCapabilityWithVersion(nativePtr, capability.toInt(), version);
        }
    }

    private native boolean hasCapabilityWithVersionOrMore(long nativePtr, int capability, float version);
    @Override
    synchronized public boolean hasCapabilityWithVersionOrMore(FriendCapability capability, float version)  {
        synchronized(core) { 
        return hasCapabilityWithVersionOrMore(nativePtr, capability.toInt(), version);
        }
    }

    private native boolean inList(long nativePtr);
    @Override
    synchronized public boolean inList()  {
        synchronized(core) { 
        return inList(nativePtr);
        }
    }

    private native void remove(long nativePtr);
    @Override
    synchronized public void remove()  {
        synchronized(core) { 
        remove(nativePtr);
        }
    }

    private native void removeAddress(long nativePtr, Address addr);
    @Override
    synchronized public void removeAddress(Address addr)  {
        synchronized(core) { 
        removeAddress(nativePtr, addr);
        }
    }

    private native void removePhoneNumber(long nativePtr, String phone);
    @Override
    synchronized public void removePhoneNumber(String phone)  {
        synchronized(core) { 
        removePhoneNumber(nativePtr, phone);
        }
    }

    private native void save(long nativePtr, Core lc);
    @Override
    synchronized public void save(Core lc)  {
        synchronized(core) { 
        save(nativePtr, lc);
        }
    }

    private native void setPresenceModelForUriOrTel(long nativePtr, String uriOrTel, PresenceModel presence);
    @Override
    synchronized public void setPresenceModelForUriOrTel(String uriOrTel, PresenceModel presence)  {
        synchronized(core) { 
        setPresenceModelForUriOrTel(nativePtr, uriOrTel, presence);
        }
    }

    private native Friend newFromVcard(long nativePtr, Vcard vcard);
    @Override
    synchronized public Friend newFromVcard(Vcard vcard)  {
        synchronized(core) { 
        return (Friend)newFromVcard(nativePtr, vcard);
        }
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
