/*
FriendList.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* The {@link FriendList} object representing a list of friends. 
*
*/
public interface FriendList {
    public enum SyncStatus {
        /**
        *
        */
        Started(0),

        /**
        *
        */
        Successful(1),

        /**
        *
        */
        Failure(2);

        protected final int mValue;

        private SyncStatus (int value) {
            mValue = value;
        }

        static public SyncStatus fromInt(int value) throws RuntimeException {
            switch(value) {
            case 0: return Started;
            case 1: return Successful;
            case 2: return Failure;
            default:
                throw new RuntimeException("Unhandled enum value " + value + " for SyncStatus");
            }
        }

        public int toInt() {
            return mValue;
        }
    };

    public enum Status {
        /**
        *
        */
        OK(0),

        /**
        *
        */
        NonExistentFriend(1),

        /**
        *
        */
        InvalidFriend(2);

        protected final int mValue;

        private Status (int value) {
            mValue = value;
        }

        static public Status fromInt(int value) throws RuntimeException {
            switch(value) {
            case 0: return OK;
            case 1: return NonExistentFriend;
            case 2: return InvalidFriend;
            default:
                throw new RuntimeException("Unhandled enum value " + value + " for Status");
            }
        }

        public int toInt() {
            return mValue;
        }
    };

    /**
    * Returns the {@link Core} object attached to this LinphoneFriendList. 
    *
    * @return a {@link Core} object 
    */
    public Core getCore();

    /**
    * Get the display name of the friend list. 
    *
    * @return The display name of the friend list. 
    */
    public String getDisplayName();

    /**
    * Set the display name of the friend list. 
    *
    * @param displayName The new display name of the friend list. 
    */
    public void setDisplayName(String displayName);

    /**
    * Retrieves the list of {@link Friend} from this LinphoneFriendList. 
    *
    * @return A list of {@link Friend} objects. LinphoneFriend  a list of {@link
    * Friend} 
    */
    public Friend[] getFriends();

    /**
    * Get wheter the subscription of the friend list is bodyless or not. 
    *
    * @return Wheter the subscription of the friend list is bodyless or not. 
    */
    public boolean isSubscriptionBodyless();

    /**
    * Get the RLS (Resource List Server) URI associated with the friend list to
    * subscribe to these friends presence. 
    *
    * @return The RLS URI associated with the friend list. 
    */
    public Address getRlsAddress();

    /**
    * Set the RLS (Resource List Server) URI associated with the friend list to
    * subscribe to these friends presence. 
    *
    * @param rlsAddr The RLS URI to associate with the friend list. 
    */
    public void setRlsAddress(Address rlsAddr);

    /**
    * Get the RLS (Resource List Server) URI associated with the friend list to
    * subscribe to these friends presence. 
    *
    * @return The RLS URI associated with the friend list. 
    */
    public String getRlsUri();

    /**
    * Set the RLS (Resource List Server) URI associated with the friend list to
    * subscribe to these friends presence. 
    *
    * @param rlsUri The RLS URI to associate with the friend list. 
    */
    public void setRlsUri(String rlsUri);

    /**
    * Set wheter the subscription of the friend list is bodyless or not. 
    *
    */
    public void setSubscriptionBodyless(boolean bodyless);

    /**
    * Gets whether subscription to NOTIFYes of all friends list are enabled or not. 
    *
    * @return Whether subscriptions are enabled or not 
    */
    public boolean subscriptionsEnabled();

    /**
    * Enable subscription to NOTIFYes of all friends list. 
    *
    * @param enabled should subscription be enabled or not 
    */
    public void enableSubscriptions(boolean enabled);

    /**
    * Get the URI associated with the friend list. 
    *
    * @return The URI associated with the friend list. 
    */
    public String getUri();

    /**
    * Set the URI associated with the friend list. 
    *
    * @param uri The URI to associate with the friend list. 
    */
    public void setUri(String uri);

    /**
    * Add a friend to a friend list. 
    *
    * If or when a remote CardDAV server will be attached to the list, the friend
    * will be sent to the server. 
    * 
    * @param lf {@link Friend} object to add to the friend list. 
    * 
    * @return #LinphoneFriendListOK if successfully added,
    * #LinphoneFriendListInvalidFriend if the friend is not valid. 
    */
    public FriendList.Status addFriend(Friend lf);

    /**
    * Add a friend to a friend list. 
    *
    * The friend will never be sent to a remote CardDAV server. Warning!
    * #LinphoneFriends added this way will be removed on the next synchronization,
    * and the callback contact_deleted will be called. 
    * 
    * @param lf {@link Friend} object to add to the friend list. 
    * 
    * @return #LinphoneFriendListOK if successfully added,
    * #LinphoneFriendListInvalidFriend if the friend is not valid. 
    */
    public FriendList.Status addLocalFriend(Friend lf);

    /**
    * Creates and export {@link Friend} objects from {@link FriendList} to a file
    * using vCard 4 format. 
    *
    * @param vcardFile the path to a file that will contain the vCards 
    */
    public void exportFriendsAsVcard4File(String vcardFile);

    /**
    * Find a friend in the friend list using a LinphoneAddress. 
    *
    * @param address {@link Address} object of the friend we want to search for. 
    * 
    * @return A {@link Friend} if found, null otherwise. 
    */
    public Friend findFriendByAddress(Address address);

    /**
    * Find a friend in the friend list using a ref key. 
    *
    * @param refKey The ref key string of the friend we want to search for. 
    * 
    * @return A {@link Friend} if found, null otherwise. 
    */
    public Friend findFriendByRefKey(String refKey);

    /**
    * Find a friend in the friend list using an URI string. 
    *
    * @param uri A string containing the URI of the friend we want to search for. 
    * 
    * @return A {@link Friend} if found, null otherwise. 
    */
    public Friend findFriendByUri(String uri);

    /**
    * Find all friends in the friend list using a LinphoneAddress. 
    *
    * @param address {@link Address} object of the friends we want to search for. 
    * 
    * @return A list of {@link Friend} objects. LinphoneFriend  as a list of {@link
    * Friend} if found, null otherwise. 
    */
    public Friend[] findFriendsByAddress(Address address);

    /**
    * Find all friends in the friend list using an URI string. 
    *
    * @param uri A string containing the URI of the friends we want to search for. 
    * 
    * @return A list of {@link Friend} objects. LinphoneFriend  as a list of {@link
    * Friend} if found, null otherwise. 
    */
    public Friend[] findFriendsByUri(String uri);

    /**
    * Creates and adds {@link Friend} objects to {@link FriendList} from a buffer
    * that contains the vCard(s) to parse. 
    *
    * @param vcardBuffer the buffer that contains the vCard(s) to parse 
    * 
    * @return the amount of linphone friends created 
    */
    public int importFriendsFromVcard4Buffer(String vcardBuffer);

    /**
    * Creates and adds {@link Friend} objects to {@link FriendList} from a file that
    * contains the vCard(s) to parse. 
    *
    * @param vcardFile the path to a file that contains the vCard(s) to parse 
    * 
    * @return the amount of linphone friends created 
    */
    public int importFriendsFromVcard4File(String vcardFile);

    /**
    * Notify our presence to all the friends in the friend list that have subscribed
    * to our presence directly (not using a RLS). 
    *
    * @param presence {@link PresenceModel} object. 
    */
    public void notifyPresence(PresenceModel presence);

    /**
    * Remove a friend from a friend list. 
    *
    * @param lf {@link Friend} object to remove from the friend list. 
    * 
    * @return #LinphoneFriendListOK if removed successfully,
    * #LinphoneFriendListNonExistentFriend if the friend is not in the list. 
    */
    public FriendList.Status removeFriend(Friend lf);

    /**
    * Starts a CardDAV synchronization using value set using
    * linphone_friend_list_set_uri. 
    *
    */
    public void synchronizeFriendsFromServer();

    /**
    * Goes through all the {@link Friend} that are dirty and does a CardDAV PUT to
    * update the server. 
    *
    */
    public void updateDirtyFriends();

    /**
    * Sets the revision from the last synchronization. 
    *
    * @param rev The revision 
    */
    public void updateRevision(int rev);

    /**
    * Update presence subscriptions for the entire list. 
    *
    * Calling this function is necessary when list subscriptions are enabled, ie when
    * a RLS presence server is used. 
    */
    public void updateSubscriptions();

    /**
    */
    public void addListener(FriendListListener listener);

    /**
    */
    public void removeListener(FriendListListener listener);

    /**
    */
    public void setListener(FriendListListener listener);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class FriendListImpl implements FriendList {

    protected long nativePtr = 0;
    protected Object userData = null;
    protected Core core = null;

    protected FriendListImpl(long ptr) {
        nativePtr = ptr;
        core = getCore();
    }


    private native Core getCore(long nativePtr);
    @Override
    synchronized public Core getCore()  {
        
        return (Core)getCore(nativePtr);
    }

    private native String getDisplayName(long nativePtr);
    @Override
    synchronized public String getDisplayName()  {
        synchronized(core) { 
        return getDisplayName(nativePtr);
        }
    }

    private native void setDisplayName(long nativePtr, String displayName);
    @Override
    synchronized public void setDisplayName(String displayName)  {
        synchronized(core) { 
        setDisplayName(nativePtr, displayName);
        }
    }

    private native Friend[] getFriends(long nativePtr);
    @Override
    synchronized public Friend[] getFriends()  {
        synchronized(core) { 
        return getFriends(nativePtr);
        }
    }

    private native boolean isSubscriptionBodyless(long nativePtr);
    @Override
    synchronized public boolean isSubscriptionBodyless()  {
        synchronized(core) { 
        return isSubscriptionBodyless(nativePtr);
        }
    }

    private native Address getRlsAddress(long nativePtr);
    @Override
    synchronized public Address getRlsAddress()  {
        synchronized(core) { 
        return (Address)getRlsAddress(nativePtr);
        }
    }

    private native void setRlsAddress(long nativePtr, Address rlsAddr);
    @Override
    synchronized public void setRlsAddress(Address rlsAddr)  {
        synchronized(core) { 
        setRlsAddress(nativePtr, rlsAddr);
        }
    }

    private native String getRlsUri(long nativePtr);
    @Override
    synchronized public String getRlsUri()  {
        synchronized(core) { 
        return getRlsUri(nativePtr);
        }
    }

    private native void setRlsUri(long nativePtr, String rlsUri);
    @Override
    synchronized public void setRlsUri(String rlsUri)  {
        synchronized(core) { 
        setRlsUri(nativePtr, rlsUri);
        }
    }

    private native void setSubscriptionBodyless(long nativePtr, boolean bodyless);
    @Override
    synchronized public void setSubscriptionBodyless(boolean bodyless)  {
        synchronized(core) { 
        setSubscriptionBodyless(nativePtr, bodyless);
        }
    }

    private native boolean subscriptionsEnabled(long nativePtr);
    @Override
    synchronized public boolean subscriptionsEnabled()  {
        synchronized(core) { 
        return subscriptionsEnabled(nativePtr);
        }
    }

    private native void enableSubscriptions(long nativePtr, boolean enabled);
    @Override
    synchronized public void enableSubscriptions(boolean enabled)  {
        synchronized(core) { 
        enableSubscriptions(nativePtr, enabled);
        }
    }

    private native String getUri(long nativePtr);
    @Override
    synchronized public String getUri()  {
        synchronized(core) { 
        return getUri(nativePtr);
        }
    }

    private native void setUri(long nativePtr, String uri);
    @Override
    synchronized public void setUri(String uri)  {
        synchronized(core) { 
        setUri(nativePtr, uri);
        }
    }

    private native int addFriend(long nativePtr, Friend lf);
    @Override
    synchronized public FriendList.Status addFriend(Friend lf)  {
        synchronized(core) { 
        return FriendList.Status.fromInt(addFriend(nativePtr, lf));
        }
    }

    private native int addLocalFriend(long nativePtr, Friend lf);
    @Override
    synchronized public FriendList.Status addLocalFriend(Friend lf)  {
        synchronized(core) { 
        return FriendList.Status.fromInt(addLocalFriend(nativePtr, lf));
        }
    }

    private native void exportFriendsAsVcard4File(long nativePtr, String vcardFile);
    @Override
    synchronized public void exportFriendsAsVcard4File(String vcardFile)  {
        synchronized(core) { 
        exportFriendsAsVcard4File(nativePtr, vcardFile);
        }
    }

    private native Friend findFriendByAddress(long nativePtr, Address address);
    @Override
    synchronized public Friend findFriendByAddress(Address address)  {
        synchronized(core) { 
        return (Friend)findFriendByAddress(nativePtr, address);
        }
    }

    private native Friend findFriendByRefKey(long nativePtr, String refKey);
    @Override
    synchronized public Friend findFriendByRefKey(String refKey)  {
        synchronized(core) { 
        return (Friend)findFriendByRefKey(nativePtr, refKey);
        }
    }

    private native Friend findFriendByUri(long nativePtr, String uri);
    @Override
    synchronized public Friend findFriendByUri(String uri)  {
        synchronized(core) { 
        return (Friend)findFriendByUri(nativePtr, uri);
        }
    }

    private native Friend[] findFriendsByAddress(long nativePtr, Address address);
    @Override
    synchronized public Friend[] findFriendsByAddress(Address address)  {
        synchronized(core) { 
        return findFriendsByAddress(nativePtr, address);
        }
    }

    private native Friend[] findFriendsByUri(long nativePtr, String uri);
    @Override
    synchronized public Friend[] findFriendsByUri(String uri)  {
        synchronized(core) { 
        return findFriendsByUri(nativePtr, uri);
        }
    }

    private native int importFriendsFromVcard4Buffer(long nativePtr, String vcardBuffer);
    @Override
    synchronized public int importFriendsFromVcard4Buffer(String vcardBuffer)  {
        synchronized(core) { 
        return importFriendsFromVcard4Buffer(nativePtr, vcardBuffer);
        }
    }

    private native int importFriendsFromVcard4File(long nativePtr, String vcardFile);
    @Override
    synchronized public int importFriendsFromVcard4File(String vcardFile)  {
        synchronized(core) { 
        return importFriendsFromVcard4File(nativePtr, vcardFile);
        }
    }

    private native void notifyPresence(long nativePtr, PresenceModel presence);
    @Override
    synchronized public void notifyPresence(PresenceModel presence)  {
        synchronized(core) { 
        notifyPresence(nativePtr, presence);
        }
    }

    private native int removeFriend(long nativePtr, Friend lf);
    @Override
    synchronized public FriendList.Status removeFriend(Friend lf)  {
        synchronized(core) { 
        return FriendList.Status.fromInt(removeFriend(nativePtr, lf));
        }
    }

    private native void synchronizeFriendsFromServer(long nativePtr);
    @Override
    synchronized public void synchronizeFriendsFromServer()  {
        synchronized(core) { 
        synchronizeFriendsFromServer(nativePtr);
        }
    }

    private native void updateDirtyFriends(long nativePtr);
    @Override
    synchronized public void updateDirtyFriends()  {
        synchronized(core) { 
        updateDirtyFriends(nativePtr);
        }
    }

    private native void updateRevision(long nativePtr, int rev);
    @Override
    synchronized public void updateRevision(int rev)  {
        synchronized(core) { 
        updateRevision(nativePtr, rev);
        }
    }

    private native void updateSubscriptions(long nativePtr);
    @Override
    synchronized public void updateSubscriptions()  {
        synchronized(core) { 
        updateSubscriptions(nativePtr);
        }
    }

    private native void addListener(long nativePtr, FriendListListener listener);
    @Override
    synchronized public void addListener(FriendListListener listener)  {
        
        addListener(nativePtr, listener);
    }

    private native void removeListener(long nativePtr, FriendListListener listener);
    @Override
    synchronized public void removeListener(FriendListListener listener)  {
        
        removeListener(nativePtr, listener);
    }

    private native void setListener(long nativePtr, FriendListListener listener);
    @Override
    synchronized public void setListener(FriendListListener listener)  {
        
        setListener(nativePtr, listener);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
