/*
LoggingService.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* Singleton class giving access to logging features. 
*
*/
public interface LoggingService {
    /**
    * Set the verbosity of the log. 
    *
    * For instance, a level of LinphoneLogLevelMessage will let pass fatal, error,
    * warning and message-typed messages whereas trace and debug messages will be
    * dumped out. 
    */
    public void setLogLevel(LogLevel level);

    /**
    * Gets the log level mask. 
    *
    */
    public int getLogLevelMask();

    /**
    * Sets the types of messages that will be authorized to be written in the log. 
    *
    * @param mask Example: LinphoneLogLevelMessage|LinphoneLogLevelError will ONLY
    * let pass message-typed and error messages. 
    * 
    * Note: Calling that function reset the log level that has been specified by
    * {@link LoggingService#setLogLevel}. 
    */
    public void setLogLevelMask(int mask);

    /**
    * Enables logging in a file. 
    *
    * That function enables an internal log handler that writes log messages in
    * log-rotated files.
    * 
    * @param maxSize The maximal size of each part of the log. The log rotating is
    * triggered each time the currently opened log part reach that limit. 
    */
    public void setLogFile(String dir, String filename, int maxSize);

    /**
    * Gets the singleton logging service object. 
    *
    * The singleton is automatically instantiated if it hasn't been done yet.
    * 
    * @return A pointer on the singleton. 
    */
    public LoggingService get();

    /**
    */
    public void addListener(LoggingServiceListener listener);

    /**
    */
    public void removeListener(LoggingServiceListener listener);

    /**
    */
    public void setListener(LoggingServiceListener listener);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class LoggingServiceImpl implements LoggingService {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected LoggingServiceImpl(long ptr) {
        nativePtr = ptr;
    }


    private native void setLogLevel(long nativePtr, int level);
    @Override
    synchronized public void setLogLevel(LogLevel level)  {
        
        setLogLevel(nativePtr, level.toInt());
    }

    private native int getLogLevelMask(long nativePtr);
    @Override
    synchronized public int getLogLevelMask()  {
        
        return getLogLevelMask(nativePtr);
    }

    private native void setLogLevelMask(long nativePtr, int mask);
    @Override
    synchronized public void setLogLevelMask(int mask)  {
        
        setLogLevelMask(nativePtr, mask);
    }

    private native void setLogFile(long nativePtr, String dir, String filename, int maxSize);
    @Override
    synchronized public void setLogFile(String dir, String filename, int maxSize)  {
        
        setLogFile(nativePtr, dir, filename, maxSize);
    }

    private native LoggingService get(long nativePtr);
    @Override
    synchronized public LoggingService get()  {
        
        return (LoggingService)get(nativePtr);
    }

    private native void addListener(long nativePtr, LoggingServiceListener listener);
    @Override
    synchronized public void addListener(LoggingServiceListener listener)  {
        
        addListener(nativePtr, listener);
    }

    private native void removeListener(long nativePtr, LoggingServiceListener listener);
    @Override
    synchronized public void removeListener(LoggingServiceListener listener)  {
        
        removeListener(nativePtr, listener);
    }

    private native void setListener(long nativePtr, LoggingServiceListener listener);
    @Override
    synchronized public void setListener(LoggingServiceListener listener)  {
        
        setListener(nativePtr, listener);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
