/*
NatPolicy.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* Policy to use to pass through NATs/firewalls. 
*
*/
public interface NatPolicy {
    /**
    * Returns the {@link Core} object managing this nat policy, if any. 
    *
    */
    public Core getCore();

    /**
    * Tell whether ICE is enabled. 
    *
    * @return Boolean value telling whether ICE is enabled. 
    */
    public boolean iceEnabled();

    /**
    * Enable ICE. 
    *
    * ICE can be enabled without STUN/TURN, in which case only the local candidates
    * will be used. 
    * 
    * @param enable Boolean value telling whether to enable ICE. 
    */
    public void enableIce(boolean enable);

    /**
    * Tell whether STUN is enabled. 
    *
    * @return Boolean value telling whether STUN is enabled. 
    */
    public boolean stunEnabled();

    /**
    * Enable STUN. 
    *
    * If TURN is also enabled, TURN will be used instead of STUN. 
    * 
    * @param enable Boolean value telling whether to enable STUN. 
    */
    public void enableStun(boolean enable);

    /**
    * Get the STUN/TURN server to use with this NAT policy. 
    *
    * Used when STUN or TURN are enabled. 
    * 
    * @return The STUN server used by this NAT policy. 
    */
    public String getStunServer();

    /**
    * Set the STUN/TURN server to use with this NAT policy. 
    *
    * Used when STUN or TURN are enabled. 
    * 
    * @param stunServer The STUN server to use with this NAT policy. 
    */
    public void setStunServer(String stunServer);

    /**
    * Get the username used to authenticate with the STUN/TURN server. 
    *
    * The authentication will search for a {@link AuthInfo} with this username. If it
    * is not set the username of the currently used {@link ProxyConfig} is used to
    * search for a LinphoneAuthInfo. 
    * 
    * @return The username used to authenticate with the STUN/TURN server. 
    */
    public String getStunServerUsername();

    /**
    * Set the username used to authenticate with the STUN/TURN server. 
    *
    * The authentication will search for a {@link AuthInfo} with this username. If it
    * is not set the username of the currently used {@link ProxyConfig} is used to
    * search for a LinphoneAuthInfo. 
    * 
    * @param username The username used to authenticate with the STUN/TURN server. 
    */
    public void setStunServerUsername(String username);

    /**
    * Tell whether TURN is enabled. 
    *
    * @return Boolean value telling whether TURN is enabled. 
    */
    public boolean turnEnabled();

    /**
    * Enable TURN. 
    *
    * If STUN is also enabled, it is ignored and TURN is used. 
    * 
    * @param enable Boolean value telling whether to enable TURN. 
    */
    public void enableTurn(boolean enable);

    /**
    * Tell whether uPnP is enabled. 
    *
    * @return Boolean value telling whether uPnP is enabled. 
    */
    public boolean upnpEnabled();

    /**
    * Enable uPnP. 
    *
    * This has the effect to disable every other policies (ICE, STUN and TURN). 
    * 
    * @param enable Boolean value telling whether to enable uPnP. 
    */
    public void enableUpnp(boolean enable);

    /**
    * Clear a NAT policy (deactivate all protocols and unset the STUN server). 
    *
    */
    public void clear();

    /**
    * Start a STUN server DNS resolution. 
    *
    */
    public void resolveStunServer();

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class NatPolicyImpl implements NatPolicy {

    protected long nativePtr = 0;
    protected Object userData = null;
    protected Core core = null;

    protected NatPolicyImpl(long ptr) {
        nativePtr = ptr;
        core = getCore();
    }


    private native Core getCore(long nativePtr);
    @Override
    synchronized public Core getCore()  {
        
        return (Core)getCore(nativePtr);
    }

    private native boolean iceEnabled(long nativePtr);
    @Override
    synchronized public boolean iceEnabled()  {
        synchronized(core) { 
        return iceEnabled(nativePtr);
        }
    }

    private native void enableIce(long nativePtr, boolean enable);
    @Override
    synchronized public void enableIce(boolean enable)  {
        synchronized(core) { 
        enableIce(nativePtr, enable);
        }
    }

    private native boolean stunEnabled(long nativePtr);
    @Override
    synchronized public boolean stunEnabled()  {
        synchronized(core) { 
        return stunEnabled(nativePtr);
        }
    }

    private native void enableStun(long nativePtr, boolean enable);
    @Override
    synchronized public void enableStun(boolean enable)  {
        synchronized(core) { 
        enableStun(nativePtr, enable);
        }
    }

    private native String getStunServer(long nativePtr);
    @Override
    synchronized public String getStunServer()  {
        synchronized(core) { 
        return getStunServer(nativePtr);
        }
    }

    private native void setStunServer(long nativePtr, String stunServer);
    @Override
    synchronized public void setStunServer(String stunServer)  {
        synchronized(core) { 
        setStunServer(nativePtr, stunServer);
        }
    }

    private native String getStunServerUsername(long nativePtr);
    @Override
    synchronized public String getStunServerUsername()  {
        synchronized(core) { 
        return getStunServerUsername(nativePtr);
        }
    }

    private native void setStunServerUsername(long nativePtr, String username);
    @Override
    synchronized public void setStunServerUsername(String username)  {
        synchronized(core) { 
        setStunServerUsername(nativePtr, username);
        }
    }

    private native boolean turnEnabled(long nativePtr);
    @Override
    synchronized public boolean turnEnabled()  {
        synchronized(core) { 
        return turnEnabled(nativePtr);
        }
    }

    private native void enableTurn(long nativePtr, boolean enable);
    @Override
    synchronized public void enableTurn(boolean enable)  {
        synchronized(core) { 
        enableTurn(nativePtr, enable);
        }
    }

    private native boolean upnpEnabled(long nativePtr);
    @Override
    synchronized public boolean upnpEnabled()  {
        synchronized(core) { 
        return upnpEnabled(nativePtr);
        }
    }

    private native void enableUpnp(long nativePtr, boolean enable);
    @Override
    synchronized public void enableUpnp(boolean enable)  {
        synchronized(core) { 
        enableUpnp(nativePtr, enable);
        }
    }

    private native void clear(long nativePtr);
    @Override
    synchronized public void clear()  {
        synchronized(core) { 
        clear(nativePtr);
        }
    }

    private native void resolveStunServer(long nativePtr);
    @Override
    synchronized public void resolveStunServer()  {
        synchronized(core) { 
        resolveStunServer(nativePtr);
        }
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
