/*
PayloadType.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* Object representing an RTP payload type. 
*
*/
public interface PayloadType {
    /**
    * Get the number of channels. 
    *
    * @return The number of channels. 
    */
    public int getChannels();

    /**
    * Get the clock rate of a payload type. 
    *
    * @return [in] The clock rate in Hz. 
    */
    public int getClockRate();

    /**
    * Return a string describing a payload type. 
    *
    * The format of the string is <mime_type>/<clock_rate>/<channels>. 
    * 
    * @return The description of the payload type. Must be release after use. 
    */
    public String getDescription();

    /**
    * Get a description of the encoder used to provide a payload type. 
    *
    * @return The description of the encoder. Can be null if the payload type is not
    * supported by Mediastreamer2. 
    */
    public String getEncoderDescription();

    /**
    * Check whether the payload is usable according the bandwidth targets set in the
    * core. 
    *
    * @return  if the payload type is usable. 
    */
    public boolean isUsable();

    /**
    * Tells whether the specified payload type represents a variable bitrate codec. 
    *
    * @return  if the payload type represents a VBR codec, true instead. 
    */
    public boolean isVbr();

    /**
    * Get the mime type. 
    *
    * @return The mime type. 
    */
    public String getMimeType();

    /**
    * Get the normal bitrate in bits/s. 
    *
    * @return The normal bitrate in bits/s or -1 if an error has occured. 
    */
    public int getNormalBitrate();

    /**
    * Change the normal bitrate of a payload type. 
    *
    * @param bitrate The new bitrate in bits/s. 
    */
    public void setNormalBitrate(int bitrate);

    /**
    * Returns the payload type number assigned for this codec. 
    *
    * @return The number of the payload type. 
    */
    public int getNumber();

    /**
    * Force a number for a payload type. 
    *
    * The {@link Core} does payload type number assignment automatically. This
    * function is mainly to be used for tests, in order to override the automatic
    * assignment mechanism. 
    * 
    * @param number The number to assign to the payload type. 
    */
    public void setNumber(int number);

    /**
    * Get the format parameters for incoming streams. 
    *
    * @return The format parameters as string. 
    */
    public String getRecvFmtp();

    /**
    * Set the format parameters for incoming streams. 
    *
    * @param recvFmtp The new format parameters as string. The string will be copied. 
    */
    public void setRecvFmtp(String recvFmtp);

    /**
    * Get the format parameters for outgoing streams. 
    *
    * @return The format parameters as string. 
    */
    public String getSendFmtp();

    /**
    * Set the format parameters for outgoing streams. 
    *
    * @param sendFmtp The new format parameters as string. The string will be copied. 
    */
    public void setSendFmtp(String sendFmtp);

    /**
    * Get the type of a payload type. 
    *
    * @return The type of the payload e.g. PAYLOAD_AUDIO_CONTINUOUS or PAYLOAD_VIDEO. 
    */
    public int getType();

    /**
    * Enable/disable a payload type. 
    *
    * @param enabled Set true for enabling and true for disabling. 
    * 
    * @return 0 for success, -1 for failure. 
    */
    public int enable(boolean enabled);

    /**
    * Check whether a palyoad type is enabled. 
    *
    * @return  if enabled, true if disabled. 
    */
    public boolean enabled();

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class PayloadTypeImpl implements PayloadType {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected PayloadTypeImpl(long ptr) {
        nativePtr = ptr;
    }


    private native int getChannels(long nativePtr);
    @Override
    synchronized public int getChannels()  {
        
        return getChannels(nativePtr);
    }

    private native int getClockRate(long nativePtr);
    @Override
    synchronized public int getClockRate()  {
        
        return getClockRate(nativePtr);
    }

    private native String getDescription(long nativePtr);
    @Override
    synchronized public String getDescription()  {
        
        return getDescription(nativePtr);
    }

    private native String getEncoderDescription(long nativePtr);
    @Override
    synchronized public String getEncoderDescription()  {
        
        return getEncoderDescription(nativePtr);
    }

    private native boolean isUsable(long nativePtr);
    @Override
    synchronized public boolean isUsable()  {
        
        return isUsable(nativePtr);
    }

    private native boolean isVbr(long nativePtr);
    @Override
    synchronized public boolean isVbr()  {
        
        return isVbr(nativePtr);
    }

    private native String getMimeType(long nativePtr);
    @Override
    synchronized public String getMimeType()  {
        
        return getMimeType(nativePtr);
    }

    private native int getNormalBitrate(long nativePtr);
    @Override
    synchronized public int getNormalBitrate()  {
        
        return getNormalBitrate(nativePtr);
    }

    private native void setNormalBitrate(long nativePtr, int bitrate);
    @Override
    synchronized public void setNormalBitrate(int bitrate)  {
        
        setNormalBitrate(nativePtr, bitrate);
    }

    private native int getNumber(long nativePtr);
    @Override
    synchronized public int getNumber()  {
        
        return getNumber(nativePtr);
    }

    private native void setNumber(long nativePtr, int number);
    @Override
    synchronized public void setNumber(int number)  {
        
        setNumber(nativePtr, number);
    }

    private native String getRecvFmtp(long nativePtr);
    @Override
    synchronized public String getRecvFmtp()  {
        
        return getRecvFmtp(nativePtr);
    }

    private native void setRecvFmtp(long nativePtr, String recvFmtp);
    @Override
    synchronized public void setRecvFmtp(String recvFmtp)  {
        
        setRecvFmtp(nativePtr, recvFmtp);
    }

    private native String getSendFmtp(long nativePtr);
    @Override
    synchronized public String getSendFmtp()  {
        
        return getSendFmtp(nativePtr);
    }

    private native void setSendFmtp(long nativePtr, String sendFmtp);
    @Override
    synchronized public void setSendFmtp(String sendFmtp)  {
        
        setSendFmtp(nativePtr, sendFmtp);
    }

    private native int getType(long nativePtr);
    @Override
    synchronized public int getType()  {
        
        return getType(nativePtr);
    }

    private native int enable(long nativePtr, boolean enabled);
    @Override
    synchronized public int enable(boolean enabled)  {
        
        return enable(nativePtr, enabled);
    }

    private native boolean enabled(long nativePtr);
    @Override
    synchronized public boolean enabled()  {
        
        return enabled(nativePtr);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
