/*
PresenceModel.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* Presence model type holding information about the presence of a person. 
*
*/
public interface PresenceModel {
    /**
    * Gets the first activity of a presence model (there is usually only one). 
    *
    * @return A {@link PresenceActivity} object if successful, null otherwise. 
    */
    public PresenceActivity getActivity();

    /**
    * Gets the basic status of a presence model. 
    *
    * @return The LinphonePresenceBasicStatus of the {@link PresenceModel} object
    * given as parameter. 
    */
    public PresenceBasicStatus getBasicStatus();

    /**
    * Sets the basic status of a presence model. 
    *
    * @param basicStatus The LinphonePresenceBasicStatus to set for the {@link
    * PresenceModel} object. 
    * 
    * @return 0 if successful, a value < 0 in case of error. 
    */
    public int setBasicStatus(PresenceBasicStatus basicStatus);

    /**
    * Gets the capabilities of a {@link PresenceModel} object. 
    *
    * @return the capabilities. 
    */
    public int getCapabilities();

    /**
    * Get the consolidated presence from a presence model. 
    *
    * @return The LinphoneConsolidatedPresence corresponding to the presence model 
    */
    public ConsolidatedPresence getConsolidatedPresence();

    /**
    * Gets the contact of a presence model. 
    *
    * @return A pointer to a dynamically allocated string containing the contact, or
    * null if no contact is found.
    * 
    * The returned string is to be freed by calling ms_free(). 
    */
    public String getContact();

    /**
    * Sets the contact of a presence model. 
    *
    * @param contact The contact string to set. 
    * 
    * @return 0 if successful, a value < 0 in case of error. 
    */
    public int setContact(String contact);

    /**
    * Tells whether a presence model is considered online. 
    *
    * It is any of theses cases:
    * 
    */
    public boolean isOnline();

    /**
    * Gets the number of activities included in the presence model. 
    *
    * @return The number of activities included in the {@link PresenceModel} object. 
    */
    public int getNbActivities();

    /**
    * Gets the number of persons included in the presence model. 
    *
    * @return The number of persons included in the {@link PresenceModel} object. 
    */
    public int getNbPersons();

    /**
    * Gets the number of services included in the presence model. 
    *
    * @return The number of services included in the {@link PresenceModel} object. 
    */
    public int getNbServices();

    /**
    * Gets the presentity of a presence model. 
    *
    * @return A pointer to a const LinphoneAddress, or null if no contact is found. 
    */
    public Address getPresentity();

    /**
    * Sets the presentity of a presence model. 
    *
    * @param presentity The presentity address to set (presentity is copied). 
    * 
    * @return 0 if successful, a value < 0 in case of error. 
    */
    public int setPresentity(Address presentity);

    /**
    * Gets the timestamp of a presence model. 
    *
    * @return The timestamp of the {@link PresenceModel} object or -1 on error. 
    */
    public long getTimestamp();

    /**
    * Adds an activity to a presence model. 
    *
    * @param activity The {@link PresenceActivity} object to add to the model. 
    * 
    * @return 0 if successful, a value < 0 in case of error. 
    */
    public int addActivity(PresenceActivity activity);

    /**
    * Adds a note to a presence model. 
    *
    * @param lang The language of the note to be added. Can be null if no language is
    * to be specified for the note. 
    * 
    * @return 0 if successful, a value < 0 in case of error.
    * 
    * Only one note for each language can be set, so e.g. setting a note for the 'fr'
    * language if there is only one will replace the existing one. 
    */
    public int addNote(String noteContent, String lang);

    /**
    * Adds a person to a presence model. 
    *
    * @param person The {@link PresencePerson} object to add to the model. 
    * 
    * @return 0 if successful, a value < 0 in case of error. 
    */
    public int addPerson(PresencePerson person);

    /**
    * Adds a service to a presence model. 
    *
    * @param service The {@link PresenceService} object to add to the model. 
    * 
    * @return 0 if successful, a value < 0 in case of error. 
    */
    public int addService(PresenceService service);

    /**
    * Clears the activities of a presence model. 
    *
    * @return 0 if successful, a value < 0 in case of error. 
    */
    public int clearActivities();

    /**
    * Clears all the notes of a presence model. 
    *
    * @return 0 if successful, a value < 0 in case of error. 
    */
    public int clearNotes();

    /**
    * Clears the persons of a presence model. 
    *
    * @return 0 if successful, a value < 0 in case of error. 
    */
    public int clearPersons();

    /**
    * Clears the services of a presence model. 
    *
    * @return 0 if successful, a value < 0 in case of error. 
    */
    public int clearServices();

    /**
    * Returns the version of the capability of a {@link PresenceModel}. 
    *
    * @param capability The capability to test. 
    * 
    * @return the version of the capability of a {@link PresenceModel} or -1.0 if the
    * model has not the capability. 
    */
    public float getCapabilityVersion(FriendCapability capability);

    /**
    * Gets the first note of a presence model (there is usually only one). 
    *
    * @param lang The language of the note to get. Can be null to get a note that has
    * no language specified or to get the first note whatever language it is written
    * into. 
    * 
    * @return A pointer to a {@link PresenceNote} object if successful, null
    * otherwise. 
    */
    public PresenceNote getNote(String lang);

    /**
    * Gets the nth activity of a presence model. 
    *
    * @param idx The index of the activity to get (the first activity having the
    * index 0). 
    * 
    * @return A pointer to a {@link PresenceActivity} object if successful, null
    * otherwise. 
    */
    public PresenceActivity getNthActivity(int idx);

    /**
    * Gets the nth person of a presence model. 
    *
    * @param idx The index of the person to get (the first person having the index
    * 0). 
    * 
    * @return A pointer to a {@link PresencePerson} object if successful, null
    * otherwise. 
    */
    public PresencePerson getNthPerson(int idx);

    /**
    * Gets the nth service of a presence model. 
    *
    * @param idx The index of the service to get (the first service having the index
    * 0). 
    * 
    * @return A pointer to a {@link PresenceService} object if successful, null
    * otherwise. 
    */
    public PresenceService getNthService(int idx);

    /**
    * Returns whether or not the {@link PresenceModel} object has a given capability. 
    *
    * @param capability The capability to test. 
    * 
    * @return whether or not the {@link PresenceModel} object has a given capability. 
    */
    public boolean hasCapability(FriendCapability capability);

    /**
    * Returns whether or not the {@link PresenceModel} object has a given capability
    * with a certain version. 
    *
    * @param version The wanted version to test. 
    * 
    * @return whether or not the {@link PresenceModel} object has a given capability
    * with a certain version. 
    */
    public boolean hasCapabilityWithVersion(FriendCapability capability, float version);

    /**
    * Returns whether or not the {@link PresenceModel} object has a given capability
    * with a certain version or more. 
    *
    * @param version The wanted version to test. 
    * 
    * @return whether or not the {@link PresenceModel} object has a given capability
    * with a certain version or more. 
    */
    public boolean hasCapabilityWithVersionOrMore(FriendCapability capability, float version);

    /**
    * Sets the activity of a presence model (limits to only one activity). 
    *
    * @param description An additional description of the activity to set for the
    * model. Can be null if no additional description is to be added. 
    * 
    * @return 0 if successful, a value < 0 in case of error.
    * 
    * WARNING: This function will modify the basic status of the model according to
    * the activity being set. If you don't want the basic status to be modified
    * automatically, you can use the combination of {@link
    * PresenceModel#setBasicStatus}, {@link PresenceModel#clearActivities} and {@link
    * PresenceModel#addActivity}. 
    */
    public int setActivity(PresenceActivity.Type activity, String description);

    /**
    * Creates a presence model specifying an activity. 
    *
    * @param description An additional description of the activity (mainly useful for
    * the 'other' activity). Set it to null to not add a description. 
    * 
    * @return The created presence model, or null if an error occured. 
    * 
    * See also: linphone_presence_model_new 
    * 
    * See also: linphone_presence_model_new_with_activity_and_note
    * 
    * The created presence model has the activity specified in the parameters. 
    */
    public PresenceModel newWithActivity(PresenceActivity.Type activity, String description);

    /**
    * Creates a presence model specifying an activity and adding a note. 
    *
    * @param lang The language the note is written in. It can be set to null in order
    * to not specify the language of the note. 
    * 
    * @return The created presence model, or null if an error occured. 
    * 
    * See also: linphone_presence_model_new_with_activity 
    * 
    * See also: linphone_presence_model_new_with_activity_and_note
    * 
    * The created presence model has the activity and the note specified in the
    * parameters. 
    */
    public PresenceModel newWithActivityAndNote(PresenceActivity.Type activity, String description, String note, String lang);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class PresenceModelImpl implements PresenceModel {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected PresenceModelImpl(long ptr) {
        nativePtr = ptr;
    }


    private native PresenceActivity getActivity(long nativePtr);
    @Override
    synchronized public PresenceActivity getActivity()  {
        
        return (PresenceActivity)getActivity(nativePtr);
    }

    private native int getBasicStatus(long nativePtr);
    @Override
    synchronized public PresenceBasicStatus getBasicStatus()  {
        
        return PresenceBasicStatus.fromInt(getBasicStatus(nativePtr));
    }

    private native int setBasicStatus(long nativePtr, int basicStatus);
    @Override
    synchronized public int setBasicStatus(PresenceBasicStatus basicStatus)  {
        
        return setBasicStatus(nativePtr, basicStatus.toInt());
    }

    private native int getCapabilities(long nativePtr);
    @Override
    synchronized public int getCapabilities()  {
        
        return getCapabilities(nativePtr);
    }

    private native int getConsolidatedPresence(long nativePtr);
    @Override
    synchronized public ConsolidatedPresence getConsolidatedPresence()  {
        
        return ConsolidatedPresence.fromInt(getConsolidatedPresence(nativePtr));
    }

    private native String getContact(long nativePtr);
    @Override
    synchronized public String getContact()  {
        
        return getContact(nativePtr);
    }

    private native int setContact(long nativePtr, String contact);
    @Override
    synchronized public int setContact(String contact)  {
        
        return setContact(nativePtr, contact);
    }

    private native boolean isOnline(long nativePtr);
    @Override
    synchronized public boolean isOnline()  {
        
        return isOnline(nativePtr);
    }

    private native int getNbActivities(long nativePtr);
    @Override
    synchronized public int getNbActivities()  {
        
        return getNbActivities(nativePtr);
    }

    private native int getNbPersons(long nativePtr);
    @Override
    synchronized public int getNbPersons()  {
        
        return getNbPersons(nativePtr);
    }

    private native int getNbServices(long nativePtr);
    @Override
    synchronized public int getNbServices()  {
        
        return getNbServices(nativePtr);
    }

    private native Address getPresentity(long nativePtr);
    @Override
    synchronized public Address getPresentity()  {
        
        return (Address)getPresentity(nativePtr);
    }

    private native int setPresentity(long nativePtr, Address presentity);
    @Override
    synchronized public int setPresentity(Address presentity)  {
        
        return setPresentity(nativePtr, presentity);
    }

    private native long getTimestamp(long nativePtr);
    @Override
    synchronized public long getTimestamp()  {
        
        return getTimestamp(nativePtr);
    }

    private native int addActivity(long nativePtr, PresenceActivity activity);
    @Override
    synchronized public int addActivity(PresenceActivity activity)  {
        
        return addActivity(nativePtr, activity);
    }

    private native int addNote(long nativePtr, String noteContent, String lang);
    @Override
    synchronized public int addNote(String noteContent, String lang)  {
        
        return addNote(nativePtr, noteContent, lang);
    }

    private native int addPerson(long nativePtr, PresencePerson person);
    @Override
    synchronized public int addPerson(PresencePerson person)  {
        
        return addPerson(nativePtr, person);
    }

    private native int addService(long nativePtr, PresenceService service);
    @Override
    synchronized public int addService(PresenceService service)  {
        
        return addService(nativePtr, service);
    }

    private native int clearActivities(long nativePtr);
    @Override
    synchronized public int clearActivities()  {
        
        return clearActivities(nativePtr);
    }

    private native int clearNotes(long nativePtr);
    @Override
    synchronized public int clearNotes()  {
        
        return clearNotes(nativePtr);
    }

    private native int clearPersons(long nativePtr);
    @Override
    synchronized public int clearPersons()  {
        
        return clearPersons(nativePtr);
    }

    private native int clearServices(long nativePtr);
    @Override
    synchronized public int clearServices()  {
        
        return clearServices(nativePtr);
    }

    private native float getCapabilityVersion(long nativePtr, int capability);
    @Override
    synchronized public float getCapabilityVersion(FriendCapability capability)  {
        
        return getCapabilityVersion(nativePtr, capability.toInt());
    }

    private native PresenceNote getNote(long nativePtr, String lang);
    @Override
    synchronized public PresenceNote getNote(String lang)  {
        
        return (PresenceNote)getNote(nativePtr, lang);
    }

    private native PresenceActivity getNthActivity(long nativePtr, int idx);
    @Override
    synchronized public PresenceActivity getNthActivity(int idx)  {
        
        return (PresenceActivity)getNthActivity(nativePtr, idx);
    }

    private native PresencePerson getNthPerson(long nativePtr, int idx);
    @Override
    synchronized public PresencePerson getNthPerson(int idx)  {
        
        return (PresencePerson)getNthPerson(nativePtr, idx);
    }

    private native PresenceService getNthService(long nativePtr, int idx);
    @Override
    synchronized public PresenceService getNthService(int idx)  {
        
        return (PresenceService)getNthService(nativePtr, idx);
    }

    private native boolean hasCapability(long nativePtr, int capability);
    @Override
    synchronized public boolean hasCapability(FriendCapability capability)  {
        
        return hasCapability(nativePtr, capability.toInt());
    }

    private native boolean hasCapabilityWithVersion(long nativePtr, int capability, float version);
    @Override
    synchronized public boolean hasCapabilityWithVersion(FriendCapability capability, float version)  {
        
        return hasCapabilityWithVersion(nativePtr, capability.toInt(), version);
    }

    private native boolean hasCapabilityWithVersionOrMore(long nativePtr, int capability, float version);
    @Override
    synchronized public boolean hasCapabilityWithVersionOrMore(FriendCapability capability, float version)  {
        
        return hasCapabilityWithVersionOrMore(nativePtr, capability.toInt(), version);
    }

    private native int setActivity(long nativePtr, int activity, String description);
    @Override
    synchronized public int setActivity(PresenceActivity.Type activity, String description)  {
        
        return setActivity(nativePtr, activity.toInt(), description);
    }

    private native PresenceModel newWithActivity(long nativePtr, int activity, String description);
    @Override
    synchronized public PresenceModel newWithActivity(PresenceActivity.Type activity, String description)  {
        
        return (PresenceModel)newWithActivity(nativePtr, activity.toInt(), description);
    }

    private native PresenceModel newWithActivityAndNote(long nativePtr, int activity, String description, String note, String lang);
    @Override
    synchronized public PresenceModel newWithActivityAndNote(PresenceActivity.Type activity, String description, String note, String lang)  {
        
        return (PresenceModel)newWithActivityAndNote(nativePtr, activity.toInt(), description, note, lang);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
