/*
ProxyConfig.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* The {@link ProxyConfig} object represents a proxy configuration to be used by
* the {@link Core} object. 
*
* Its fields must not be used directly in favour of the accessors methods. Once
* created and filled properly the {@link ProxyConfig} can be given to {@link
* Core} with {@link Core#addProxyConfig}. This will automatically triggers the
* registration, if enabled.
* 
* The proxy configuration are persistent to restarts because they are saved in
* the configuration file. As a consequence, after linphone_core_new there might
* already be a list of configured proxy that can be examined with {@link
* Core#getProxyConfigList}.
* 
* The default proxy (see linphone_core_set_default_proxy ) is the one of the list
* that is used by default for calls. 
*/
public interface ProxyConfig {
    /**
    * Indicates whether AVPF/SAVPF is being used for calls using this proxy config. 
    *
    * @return True if AVPF/SAVPF is enabled, false otherwise. 
    */
    public boolean avpfEnabled();

    /**
    * Get enablement status of RTCP feedback (also known as AVPF profile). 
    *
    * @return the enablement mode, which can be LinphoneAVPFDefault (use
    * LinphoneCore's mode), LinphoneAVPFEnabled (avpf is enabled), or
    * LinphoneAVPFDisabled (disabled). 
    */
    public AVPFMode getAvpfMode();

    /**
    * Enable the use of RTCP feedback (also known as AVPF profile). 
    *
    * @param mode the enablement mode, which can be LinphoneAVPFDefault (use
    * LinphoneCore's mode), LinphoneAVPFEnabled (avpf is enabled), or
    * LinphoneAVPFDisabled (disabled). 
    */
    public void setAvpfMode(AVPFMode mode);

    /**
    * Get the interval between regular RTCP reports when using AVPF/SAVPF. 
    *
    * @return The interval in seconds. 
    */
    public int getAvpfRrInterval();

    /**
    * Set the interval between regular RTCP reports when using AVPF/SAVPF. 
    *
    * @param interval The interval in seconds (between 0 and 5 seconds). 
    */
    public void setAvpfRrInterval(int interval);

    /**
    * Get the conference factory uri. 
    *
    * @return The uri of the conference factory 
    */
    public String getConferenceFactoryUri();

    /**
    * Set the conference factory uri. 
    *
    * @param uri The uri of the conference factory 
    */
    public void setConferenceFactoryUri(String uri);

    /**
    * Return the contact address of the proxy config. 
    *
    * @return a {@link Address} correspong to the contact address of the proxy
    * config. 
    */
    public Address getContact();

    /**
    *
    * @return previously set contact parameters. 
    */
    public String getContactParameters();

    /**
    * Set optional contact parameters that will be added to the contact information
    * sent in the registration. 
    *
    * @param contactParams a string contaning the additional parameters in text form,
    * like "myparam=something;myparam2=something_else"
    * 
    * The main use case for this function is provide the proxy additional information
    * regarding the user agent, like for example unique identifier or apple push id.
    * As an example, the contact address in the SIP register sent will look like
    * <sip:joe@15.128.128.93:50421>;apple-push-id=43143-DFE23F-2323-FA2232. 
    */
    public void setContactParameters(String contactParams);

    /**
    *
    * @return previously set contact URI parameters. 
    */
    public String getContactUriParameters();

    /**
    * Set optional contact parameters that will be added to the contact information
    * sent in the registration, inside the URI. 
    *
    * @param contactUriParams a string containing the additional parameters in text
    * form, like "myparam=something;myparam2=something_else"
    * 
    * The main use case for this function is provide the proxy additional information
    * regarding the user agent, like for example unique identifier or apple push id.
    * As an example, the contact address in the SIP register sent will look like
    * <sip:joe@15.128.128.93:50421;apple-push-id=43143-DFE23F-2323-FA2232>. 
    */
    public void setContactUriParameters(String contactUriParams);

    /**
    * Get the {@link Core} object to which is associated the {@link ProxyConfig}. 
    *
    * @return The {@link Core} object to which is associated the {@link ProxyConfig}. 
    */
    public Core getCore();

    /**
    * Get the dependency of a {@link ProxyConfig}. 
    *
    * @return The proxy config this one is dependent upon, or null if not marked
    * dependent 
    */
    public ProxyConfig getDependency();

    /**
    * Mark this proxy configuration as being dependent on the given one. 
    *
    * The dependency must refer to a proxy config previously added to the core and
    * which idkey property is defined.
    * 
    * See also: {@link ProxyConfig#setIdkey}
    * 
    * The proxy configuration marked as dependent will wait for successful
    * registration on its dependency before triggering its own.
    * 
    * Once registered, both proxy configurations will share the same contact address
    * (the 'dependency' one).
    * 
    * This mecanism must be enabled before the proxy configuration is added to the
    * core
    */
    public void setDependency(ProxyConfig dependency);

    /**
    *
    * @return whether liblinphone should replace "+" by "00" in dialed numbers
    * (passed to linphone_core_invite ). 
    */
    public boolean getDialEscapePlus();

    /**
    * Sets whether liblinphone should replace "+" by international calling prefix in
    * dialed numbers (passed to linphone_core_invite ). 
    *
    */
    public void setDialEscapePlus(boolean val);

    /**
    *
    * @return dialing prefix. 
    */
    public String getDialPrefix();

    /**
    * Sets a dialing prefix to be automatically prepended when inviting a number with
    * {@link Core#invite}; This dialing prefix shall usually be the country code of
    * the country where the user is living, without "+". 
    *
    */
    public void setDialPrefix(String prefix);

    /**
    * Get the domain name of the given proxy config. 
    *
    * @return The domain name of the proxy config. 
    */
    public String getDomain();

    /**
    * Get the reason why registration failed when the proxy config state is
    * LinphoneRegistrationFailed. 
    *
    * @return The reason why registration failed for this proxy config. 
    */
    public Reason getError();

    /**
    * Get detailed information why registration failed when the proxy config state is
    * LinphoneRegistrationFailed. 
    *
    * @return The details why registration failed for this proxy config. 
    */
    public ErrorInfo getErrorInfo();

    /**
    *
    * @return the duration of registration. 
    */
    public int getExpires();

    /**
    * Sets the registration expiration time in seconds. 
    *
    */
    public void setExpires(int expires);

    /**
    *
    * @return the SIP identity that belongs to this proxy configuration. 
    */
    public Address getIdentityAddress();

    /**
    * Sets the user identity as a SIP address. 
    *
    * This identity is normally formed with display name, username and domain, such
    * as: Alice <sip:alice@example.net> The REGISTER messages will have from and to
    * set to this identity. 
    */
    public int setIdentityAddress(Address identity);

    /**
    * Get the idkey property of a {@link ProxyConfig}. 
    *
    * @return The idkey string, or null 
    */
    public String getIdkey();

    /**
    * Set the idkey property on the given proxy configuration. 
    *
    * This property can the be referenced by another proxy config 'depends_on' to
    * create a dependency relation between them. See also:
    * linphone_proxy_config_set_depends_on()
    * 
    * @param idkey The idkey string to associate to the given {@link ProxyConfig} 
    */
    public void setIdkey(String idkey);

    /**
    * Indicates whether to add to the contact parameters the push notification
    * information. 
    *
    * @return True if push notification informations should be added, false
    * otherwise. 
    */
    public boolean isPushNotificationAllowed();

    /**
    * Get The policy that is used to pass through NATs/firewalls when using this
    * proxy config. 
    *
    * If it is set to null, the default NAT policy from the core will be used
    * instead. 
    * 
    * @return {@link NatPolicy} object in use. 
    * 
    * See also: {@link Core#getNatPolicy} 
    */
    public NatPolicy getNatPolicy();

    /**
    * Set the policy to use to pass through NATs/firewalls when using this proxy
    * config. 
    *
    * If it is set to null, the default NAT policy from the core will be used
    * instead. 
    * 
    * @param policy {@link NatPolicy} object 
    * 
    * See also: {@link Core#setNatPolicy} 
    */
    public void setNatPolicy(NatPolicy policy);

    /**
    * Get default privacy policy for all calls routed through this proxy. 
    *
    * @return Privacy mode 
    */
    public int getPrivacy();

    /**
    * Set default privacy policy for all calls routed through this proxy. 
    *
    * @param privacy LinphonePrivacy to configure privacy 
    */
    public void setPrivacy(int privacy);

    /**
    *
    * @return  if PUBLISH request is enabled for this proxy. 
    */
    public boolean publishEnabled();

    /**
    * Indicates either or not, PUBLISH must be issued for this {@link ProxyConfig}. 
    *
    * In case this {@link ProxyConfig} has been added to {@link Core}, follows the
    * {@link ProxyConfig#edit} rule. 
    * 
    * @param val if true, publish will be engaged 
    */
    public void enablePublish(boolean val);

    /**
    * get the publish expiration time in second. 
    *
    * Default value is the registration expiration value. 
    * 
    * @return expires in second 
    */
    public int getPublishExpires();

    /**
    * Set the publish expiration time in second. 
    *
    * @param expires in second 
    */
    public void setPublishExpires(int expires);

    /**
    * Indicates whether to add to the contact parameters the push notification
    * information. 
    *
    * @param allow True to allow push notification information, false otherwise. 
    */
    public void setPushNotificationAllowed(boolean allow);

    /**
    * Get the route of the collector end-point when using quality reporting. 
    *
    * This SIP address should be used on server-side to process packets directly
    * before discarding packets. Collector address should be a non existing account
    * and will not receive any messages. If null, reports will be send to the proxy
    * domain. 
    * 
    * @return The SIP address of the collector end-point. 
    */
    public String getQualityReportingCollector();

    /**
    * Set the route of the collector end-point when using quality reporting. 
    *
    * This SIP address should be used on server-side to process packets directly
    * before discarding packets. Collector address should be a non existing account
    * and will not receive any messages. If null, reports will be send to the proxy
    * domain. 
    * 
    * @param collector route of the collector end-point, if null PUBLISH will be sent
    * to the proxy domain. 
    */
    public void setQualityReportingCollector(String collector);

    /**
    * Indicates whether quality statistics during call should be stored and sent to a
    * collector according to RFC 6035. 
    *
    * @return True if quality repotring is enabled, false otherwise. 
    */
    public boolean qualityReportingEnabled();

    /**
    * Indicates whether quality statistics during call should be stored and sent to a
    * collector according to RFC 6035. 
    *
    * @param enable True to store quality statistics and send them to the collector,
    * false to disable it. 
    */
    public void enableQualityReporting(boolean enable);

    /**
    * Get the interval between interval reports when using quality reporting. 
    *
    * @return The interval in seconds, 0 means interval reports are disabled. 
    */
    public int getQualityReportingInterval();

    /**
    * Set the interval between 2 interval reports sending when using quality
    * reporting. 
    *
    * If call exceed interval size, an interval report will be sent to the collector.
    * On call termination, a session report will be sent for the remaining period.
    * Value must be 0 (disabled) or positive. 
    * 
    * @param interval The interval in seconds, 0 means interval reports are disabled. 
    */
    public void setQualityReportingInterval(int interval);

    /**
    * Get the realm of the given proxy config. 
    *
    * @return The realm of the proxy config. 
    */
    public String getRealm();

    /**
    * Set the realm of the given proxy config. 
    *
    * @param realm New realm value. 
    */
    public void setRealm(String realm);

    /**
    * Get the persistent reference key associated to the proxy config. 
    *
    * The reference key can be for example an id to an external database. It is
    * stored in the config file, thus can survive to process exits/restarts.
    * 
    * @return The reference key string that has been associated to the proxy config,
    * or null if none has been associated. 
    */
    public String getRefKey();

    /**
    * Associate a persistent reference key to the proxy config. 
    *
    * The reference key can be for example an id to an external database. It is
    * stored in the config file, thus can survive to process exits/restarts.
    * 
    * @param refkey The reference key string to associate to the proxy config. 
    */
    public void setRefKey(String refkey);

    /**
    *
    * @return  if registration to the proxy is enabled. 
    */
    public boolean registerEnabled();

    /**
    * Indicates either or not, REGISTRATION must be issued for this {@link
    * ProxyConfig}. 
    *
    * In case this {@link ProxyConfig} has been added to {@link Core}, follows the
    * {@link ProxyConfig#edit} rule. 
    * 
    * @param val if true, registration will be engaged 
    */
    public void enableRegister(boolean val);

    /**
    *
    * @return the route set for this proxy configuration. 
    * 
    * @deprecated Use {@link ProxyConfig#getRoutes} instead. 
    */
    @Deprecated
    public String getRoute();

    /**
    * Sets a SIP route. 
    *
    * When a route is set, all outgoing calls will go to the route's destination if
    * this proxy is the default one (see linphone_core_set_default_proxy ). @return
    * -1 if route is invalid, 0 otherwise. 
    */
    public int setRoute(String route);

    /**
    * Gets the list of the routes set for this proxy config. 
    *
    * @return A list of const char * objects. const char *  the list of routes. 
    */
    public String[] getRoutes();

    /**
    * Sets a list of SIP route. 
    *
    * When a route is set, all outgoing calls will go to the route's destination if
    * this proxy is the default one (see linphone_core_set_default_proxy ). 
    * 
    * @param routes A A list of const char * objects. const char *  of routes 
    * 
    * @return -1 if routes are invalid, 0 otherwise. 
    */
    public int setRoutes(String[] routes);

    /**
    *
    * @return the proxy's SIP address. 
    */
    public String getServerAddr();

    /**
    * Sets the proxy address. 
    *
    * Examples of valid sip proxy address are:
    * 
    */
    public int setServerAddr(String serverAddr);

    /**
    * Get the registration state of the given proxy config. 
    *
    * @return The registration state of the proxy config. 
    */
    public RegistrationState getState();

    /**
    * Get the transport from either service route, route or addr. 
    *
    * @return The transport as a string (I.E udp, tcp, tls, dtls) 
    */
    public String getTransport();

    /**
    * Return the unread chat message count for a given proxy config. 
    *
    * @return The unread chat message count. 
    */
    public int getUnreadChatMessageCount();

    /**
    * Commits modification made to the proxy configuration. 
    *
    */
    public int done();

    /**
    * Starts editing a proxy configuration. 
    *
    * Because proxy configuration must be consistent, applications MUST call {@link
    * ProxyConfig#edit} before doing any attempts to modify proxy configuration (such
    * as identity, proxy address and so on). Once the modifications are done, then
    * the application must call {@link ProxyConfig#done} to commit the changes. 
    */
    public void edit();

    /**
    * Find authentication info matching proxy config, if any, similarly to
    * linphone_core_find_auth_info. 
    *
    * @return a {@link AuthInfo} matching proxy config criteria if possible, null if
    * nothing can be found. 
    */
    public AuthInfo findAuthInfo();

    /**
    * Obtain the value of a header sent by the server in last answer to REGISTER. 
    *
    * @param headerName the header name for which to fetch corresponding value 
    * 
    * @return the value of the queried header. 
    */
    public String getCustomHeader(String headerName);

    /**
    * Detect if the given input is a phone number or not. 
    *
    * @param username string to parse. 
    * 
    * @return  if input is a phone number, true otherwise. 
    */
    public boolean isPhoneNumber(String username);

    /**
    * Normalize a human readable phone number into a basic string. 
    *
    * 888-444-222 becomes 888444222 or +33888444222 depending on the {@link
    * ProxyConfig} object. This function will always generate a normalized username
    * if input is a phone number. 
    * 
    * @param username the string to parse 
    * 
    * @return  if input is an invalid phone number, normalized phone number from
    * username input otherwise. 
    */
    public String normalizePhoneNumber(String username);

    /**
    * Normalize a human readable sip uri into a fully qualified LinphoneAddress. 
    *
    * A sip address should look like DisplayName <sip:username@domain:port> .
    * Basically this function performs the following tasks
    * 
    * 
    * The result is a syntactically correct SIP address. 
    * 
    * @param username the string to parse 
    * 
    * @return  if invalid input, normalized sip address otherwise. 
    */
    public Address normalizeSipUri(String username);

    /**
    * Prevent a proxy config from refreshing its registration. 
    *
    * This is useful to let registrations to expire naturally (or) when the
    * application wants to keep control on when refreshes are sent. However,
    * linphone_core_set_network_reachable(lc,true) will always request the proxy
    * configs to refresh their registrations. The refreshing operations can be
    * resumed with {@link ProxyConfig#refreshRegister}. 
    */
    public void pauseRegister();

    /**
    * Refresh a proxy registration. 
    *
    * This is useful if for example you resuming from suspend, thus IP address may
    * have changed. 
    */
    public void refreshRegister();

    /**
    * Set the value of a custom header sent to the server in REGISTERs request. 
    *
    * @param headerValue the header's value 
    */
    public void setCustomHeader(String headerName, String headerValue);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class ProxyConfigImpl implements ProxyConfig {

    protected long nativePtr = 0;
    protected Object userData = null;
    protected Core core = null;

    protected ProxyConfigImpl(long ptr) {
        nativePtr = ptr;
        core = getCore();
    }


    private native boolean avpfEnabled(long nativePtr);
    @Override
    synchronized public boolean avpfEnabled()  {
        synchronized(core) { 
        return avpfEnabled(nativePtr);
        }
    }

    private native int getAvpfMode(long nativePtr);
    @Override
    synchronized public AVPFMode getAvpfMode()  {
        synchronized(core) { 
        return AVPFMode.fromInt(getAvpfMode(nativePtr));
        }
    }

    private native void setAvpfMode(long nativePtr, int mode);
    @Override
    synchronized public void setAvpfMode(AVPFMode mode)  {
        synchronized(core) { 
        setAvpfMode(nativePtr, mode.toInt());
        }
    }

    private native int getAvpfRrInterval(long nativePtr);
    @Override
    synchronized public int getAvpfRrInterval()  {
        synchronized(core) { 
        return getAvpfRrInterval(nativePtr);
        }
    }

    private native void setAvpfRrInterval(long nativePtr, int interval);
    @Override
    synchronized public void setAvpfRrInterval(int interval)  {
        synchronized(core) { 
        setAvpfRrInterval(nativePtr, interval);
        }
    }

    private native String getConferenceFactoryUri(long nativePtr);
    @Override
    synchronized public String getConferenceFactoryUri()  {
        synchronized(core) { 
        return getConferenceFactoryUri(nativePtr);
        }
    }

    private native void setConferenceFactoryUri(long nativePtr, String uri);
    @Override
    synchronized public void setConferenceFactoryUri(String uri)  {
        synchronized(core) { 
        setConferenceFactoryUri(nativePtr, uri);
        }
    }

    private native Address getContact(long nativePtr);
    @Override
    synchronized public Address getContact()  {
        synchronized(core) { 
        return (Address)getContact(nativePtr);
        }
    }

    private native String getContactParameters(long nativePtr);
    @Override
    synchronized public String getContactParameters()  {
        synchronized(core) { 
        return getContactParameters(nativePtr);
        }
    }

    private native void setContactParameters(long nativePtr, String contactParams);
    @Override
    synchronized public void setContactParameters(String contactParams)  {
        synchronized(core) { 
        setContactParameters(nativePtr, contactParams);
        }
    }

    private native String getContactUriParameters(long nativePtr);
    @Override
    synchronized public String getContactUriParameters()  {
        synchronized(core) { 
        return getContactUriParameters(nativePtr);
        }
    }

    private native void setContactUriParameters(long nativePtr, String contactUriParams);
    @Override
    synchronized public void setContactUriParameters(String contactUriParams)  {
        synchronized(core) { 
        setContactUriParameters(nativePtr, contactUriParams);
        }
    }

    private native Core getCore(long nativePtr);
    @Override
    synchronized public Core getCore()  {
        
        return (Core)getCore(nativePtr);
    }

    private native ProxyConfig getDependency(long nativePtr);
    @Override
    synchronized public ProxyConfig getDependency()  {
        synchronized(core) { 
        return (ProxyConfig)getDependency(nativePtr);
        }
    }

    private native void setDependency(long nativePtr, ProxyConfig dependency);
    @Override
    synchronized public void setDependency(ProxyConfig dependency)  {
        synchronized(core) { 
        setDependency(nativePtr, dependency);
        }
    }

    private native boolean getDialEscapePlus(long nativePtr);
    @Override
    synchronized public boolean getDialEscapePlus()  {
        synchronized(core) { 
        return getDialEscapePlus(nativePtr);
        }
    }

    private native void setDialEscapePlus(long nativePtr, boolean val);
    @Override
    synchronized public void setDialEscapePlus(boolean val)  {
        synchronized(core) { 
        setDialEscapePlus(nativePtr, val);
        }
    }

    private native String getDialPrefix(long nativePtr);
    @Override
    synchronized public String getDialPrefix()  {
        synchronized(core) { 
        return getDialPrefix(nativePtr);
        }
    }

    private native void setDialPrefix(long nativePtr, String prefix);
    @Override
    synchronized public void setDialPrefix(String prefix)  {
        synchronized(core) { 
        setDialPrefix(nativePtr, prefix);
        }
    }

    private native String getDomain(long nativePtr);
    @Override
    synchronized public String getDomain()  {
        synchronized(core) { 
        return getDomain(nativePtr);
        }
    }

    private native int getError(long nativePtr);
    @Override
    synchronized public Reason getError()  {
        synchronized(core) { 
        return Reason.fromInt(getError(nativePtr));
        }
    }

    private native ErrorInfo getErrorInfo(long nativePtr);
    @Override
    synchronized public ErrorInfo getErrorInfo()  {
        synchronized(core) { 
        return (ErrorInfo)getErrorInfo(nativePtr);
        }
    }

    private native int getExpires(long nativePtr);
    @Override
    synchronized public int getExpires()  {
        synchronized(core) { 
        return getExpires(nativePtr);
        }
    }

    private native void setExpires(long nativePtr, int expires);
    @Override
    synchronized public void setExpires(int expires)  {
        synchronized(core) { 
        setExpires(nativePtr, expires);
        }
    }

    private native Address getIdentityAddress(long nativePtr);
    @Override
    synchronized public Address getIdentityAddress()  {
        synchronized(core) { 
        return (Address)getIdentityAddress(nativePtr);
        }
    }

    private native int setIdentityAddress(long nativePtr, Address identity);
    @Override
    synchronized public int setIdentityAddress(Address identity)  {
        synchronized(core) { 
        return setIdentityAddress(nativePtr, identity);
        }
    }

    private native String getIdkey(long nativePtr);
    @Override
    synchronized public String getIdkey()  {
        synchronized(core) { 
        return getIdkey(nativePtr);
        }
    }

    private native void setIdkey(long nativePtr, String idkey);
    @Override
    synchronized public void setIdkey(String idkey)  {
        synchronized(core) { 
        setIdkey(nativePtr, idkey);
        }
    }

    private native boolean isPushNotificationAllowed(long nativePtr);
    @Override
    synchronized public boolean isPushNotificationAllowed()  {
        synchronized(core) { 
        return isPushNotificationAllowed(nativePtr);
        }
    }

    private native NatPolicy getNatPolicy(long nativePtr);
    @Override
    synchronized public NatPolicy getNatPolicy()  {
        synchronized(core) { 
        return (NatPolicy)getNatPolicy(nativePtr);
        }
    }

    private native void setNatPolicy(long nativePtr, NatPolicy policy);
    @Override
    synchronized public void setNatPolicy(NatPolicy policy)  {
        synchronized(core) { 
        setNatPolicy(nativePtr, policy);
        }
    }

    private native int getPrivacy(long nativePtr);
    @Override
    synchronized public int getPrivacy()  {
        synchronized(core) { 
        return getPrivacy(nativePtr);
        }
    }

    private native void setPrivacy(long nativePtr, int privacy);
    @Override
    synchronized public void setPrivacy(int privacy)  {
        synchronized(core) { 
        setPrivacy(nativePtr, privacy);
        }
    }

    private native boolean publishEnabled(long nativePtr);
    @Override
    synchronized public boolean publishEnabled()  {
        synchronized(core) { 
        return publishEnabled(nativePtr);
        }
    }

    private native void enablePublish(long nativePtr, boolean val);
    @Override
    synchronized public void enablePublish(boolean val)  {
        synchronized(core) { 
        enablePublish(nativePtr, val);
        }
    }

    private native int getPublishExpires(long nativePtr);
    @Override
    synchronized public int getPublishExpires()  {
        synchronized(core) { 
        return getPublishExpires(nativePtr);
        }
    }

    private native void setPublishExpires(long nativePtr, int expires);
    @Override
    synchronized public void setPublishExpires(int expires)  {
        synchronized(core) { 
        setPublishExpires(nativePtr, expires);
        }
    }

    private native void setPushNotificationAllowed(long nativePtr, boolean allow);
    @Override
    synchronized public void setPushNotificationAllowed(boolean allow)  {
        synchronized(core) { 
        setPushNotificationAllowed(nativePtr, allow);
        }
    }

    private native String getQualityReportingCollector(long nativePtr);
    @Override
    synchronized public String getQualityReportingCollector()  {
        synchronized(core) { 
        return getQualityReportingCollector(nativePtr);
        }
    }

    private native void setQualityReportingCollector(long nativePtr, String collector);
    @Override
    synchronized public void setQualityReportingCollector(String collector)  {
        synchronized(core) { 
        setQualityReportingCollector(nativePtr, collector);
        }
    }

    private native boolean qualityReportingEnabled(long nativePtr);
    @Override
    synchronized public boolean qualityReportingEnabled()  {
        synchronized(core) { 
        return qualityReportingEnabled(nativePtr);
        }
    }

    private native void enableQualityReporting(long nativePtr, boolean enable);
    @Override
    synchronized public void enableQualityReporting(boolean enable)  {
        synchronized(core) { 
        enableQualityReporting(nativePtr, enable);
        }
    }

    private native int getQualityReportingInterval(long nativePtr);
    @Override
    synchronized public int getQualityReportingInterval()  {
        synchronized(core) { 
        return getQualityReportingInterval(nativePtr);
        }
    }

    private native void setQualityReportingInterval(long nativePtr, int interval);
    @Override
    synchronized public void setQualityReportingInterval(int interval)  {
        synchronized(core) { 
        setQualityReportingInterval(nativePtr, interval);
        }
    }

    private native String getRealm(long nativePtr);
    @Override
    synchronized public String getRealm()  {
        synchronized(core) { 
        return getRealm(nativePtr);
        }
    }

    private native void setRealm(long nativePtr, String realm);
    @Override
    synchronized public void setRealm(String realm)  {
        synchronized(core) { 
        setRealm(nativePtr, realm);
        }
    }

    private native String getRefKey(long nativePtr);
    @Override
    synchronized public String getRefKey()  {
        synchronized(core) { 
        return getRefKey(nativePtr);
        }
    }

    private native void setRefKey(long nativePtr, String refkey);
    @Override
    synchronized public void setRefKey(String refkey)  {
        synchronized(core) { 
        setRefKey(nativePtr, refkey);
        }
    }

    private native boolean registerEnabled(long nativePtr);
    @Override
    synchronized public boolean registerEnabled()  {
        synchronized(core) { 
        return registerEnabled(nativePtr);
        }
    }

    private native void enableRegister(long nativePtr, boolean val);
    @Override
    synchronized public void enableRegister(boolean val)  {
        synchronized(core) { 
        enableRegister(nativePtr, val);
        }
    }

    private native String getRoute(long nativePtr);
    @Override
    synchronized public String getRoute()  {
        synchronized(core) { 
        return getRoute(nativePtr);
        }
    }

    private native int setRoute(long nativePtr, String route);
    @Override
    synchronized public int setRoute(String route)  {
        synchronized(core) { 
        return setRoute(nativePtr, route);
        }
    }

    private native String[] getRoutes(long nativePtr);
    @Override
    synchronized public String[] getRoutes()  {
        synchronized(core) { 
        return getRoutes(nativePtr);
        }
    }

    private native int setRoutes(long nativePtr, String[] routes);
    @Override
    synchronized public int setRoutes(String[] routes)  {
        synchronized(core) { 
        return setRoutes(nativePtr, routes);
        }
    }

    private native String getServerAddr(long nativePtr);
    @Override
    synchronized public String getServerAddr()  {
        synchronized(core) { 
        return getServerAddr(nativePtr);
        }
    }

    private native int setServerAddr(long nativePtr, String serverAddr);
    @Override
    synchronized public int setServerAddr(String serverAddr)  {
        synchronized(core) { 
        return setServerAddr(nativePtr, serverAddr);
        }
    }

    private native int getState(long nativePtr);
    @Override
    synchronized public RegistrationState getState()  {
        synchronized(core) { 
        return RegistrationState.fromInt(getState(nativePtr));
        }
    }

    private native String getTransport(long nativePtr);
    @Override
    synchronized public String getTransport()  {
        synchronized(core) { 
        return getTransport(nativePtr);
        }
    }

    private native int getUnreadChatMessageCount(long nativePtr);
    @Override
    synchronized public int getUnreadChatMessageCount()  {
        synchronized(core) { 
        return getUnreadChatMessageCount(nativePtr);
        }
    }

    private native int done(long nativePtr);
    @Override
    synchronized public int done()  {
        synchronized(core) { 
        return done(nativePtr);
        }
    }

    private native void edit(long nativePtr);
    @Override
    synchronized public void edit()  {
        synchronized(core) { 
        edit(nativePtr);
        }
    }

    private native AuthInfo findAuthInfo(long nativePtr);
    @Override
    synchronized public AuthInfo findAuthInfo()  {
        synchronized(core) { 
        return (AuthInfo)findAuthInfo(nativePtr);
        }
    }

    private native String getCustomHeader(long nativePtr, String headerName);
    @Override
    synchronized public String getCustomHeader(String headerName)  {
        synchronized(core) { 
        return getCustomHeader(nativePtr, headerName);
        }
    }

    private native boolean isPhoneNumber(long nativePtr, String username);
    @Override
    synchronized public boolean isPhoneNumber(String username)  {
        synchronized(core) { 
        return isPhoneNumber(nativePtr, username);
        }
    }

    private native String normalizePhoneNumber(long nativePtr, String username);
    @Override
    synchronized public String normalizePhoneNumber(String username)  {
        synchronized(core) { 
        return normalizePhoneNumber(nativePtr, username);
        }
    }

    private native Address normalizeSipUri(long nativePtr, String username);
    @Override
    synchronized public Address normalizeSipUri(String username)  {
        synchronized(core) { 
        return (Address)normalizeSipUri(nativePtr, username);
        }
    }

    private native void pauseRegister(long nativePtr);
    @Override
    synchronized public void pauseRegister()  {
        synchronized(core) { 
        pauseRegister(nativePtr);
        }
    }

    private native void refreshRegister(long nativePtr);
    @Override
    synchronized public void refreshRegister()  {
        synchronized(core) { 
        refreshRegister(nativePtr);
        }
    }

    private native void setCustomHeader(long nativePtr, String headerName, String headerValue);
    @Override
    synchronized public void setCustomHeader(String headerName, String headerValue)  {
        synchronized(core) { 
        setCustomHeader(nativePtr, headerName, headerValue);
        }
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
