/*
Tunnel.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* Linphone tunnel object. 
*
*/
public interface Tunnel {
    public enum Mode {
        /**
        * The tunnel is disabled. 
        *
        */
        Disable(0),

        /**
        * The tunnel is enabled. 
        *
        */
        Enable(1),

        /**
        * The tunnel is enabled automatically if it is required. 
        *
        */
        Auto(2);

        protected final int mValue;

        private Mode (int value) {
            mValue = value;
        }

        static public Mode fromInt(int value) throws RuntimeException {
            switch(value) {
            case 0: return Disable;
            case 1: return Enable;
            case 2: return Auto;
            default:
                throw new RuntimeException("Unhandled enum value " + value + " for Mode");
            }
        }

        public int toInt() {
            return mValue;
        }
    };

    /**
    * Returns whether the tunnel is activated. 
    *
    * If mode is set to auto, this gives indication whether the automatic detection
    * determined that tunnel was necessary or not. 
    * 
    * @return  if tunnel is in use, true otherwise. 
    */
    public boolean getActivated();

    /**
    * Get the dual tunnel client mode. 
    *
    * @return  if dual tunnel client mode is enabled, true otherwise 
    */
    public boolean dualModeEnabled();

    /**
    * Sets whether or not to use the dual tunnel client mode. 
    *
    * By default this feature is disabled. After enabling it, add a server with 2
    * hosts and 2 ports for the feature to work. 
    * 
    * @param dualModeEnabled  to enable it, true to disable it 
    */
    public void enableDualMode(boolean dualModeEnabled);

    /**
    * Get the tunnel mode. 
    *
    * @return The current LinphoneTunnelMode 
    */
    public Tunnel.Mode getMode();

    /**
    * Set the tunnel mode. 
    *
    * The tunnel mode can be 'enable', 'disable' or 'auto' If the mode is set to
    * 'auto', the tunnel manager will try to established an RTP session with the
    * tunnel server on the UdpMirrorPort. If the connection fail, the tunnel is
    * automatically activated whereas the tunnel is automatically disabled if the
    * connection succeed. 
    * 
    * @param mode The desired LinphoneTunnelMode 
    */
    public void setMode(Tunnel.Mode mode);

    /**
    * Get added servers. 
    *
    * @return A list of {@link TunnelConfig} objects. LinphoneTunnelConfig  
    */
    public TunnelConfig[] getServers();

    /**
    * Check whether tunnel is set to transport SIP packets. 
    *
    * @return A boolean value telling whether SIP packets shall pass through the
    * tunnel 
    */
    public boolean sipEnabled();

    /**
    * Set whether SIP packets must be directly sent to a UA or pass through the
    * tunnel. 
    *
    * @param enable If true, SIP packets shall pass through the tunnel 
    */
    public void enableSip(boolean enable);

    /**
    * Add a tunnel server configuration. 
    *
    * @param tunnelConfig {@link TunnelConfig} object 
    */
    public void addServer(TunnelConfig tunnelConfig);

    /**
    * Remove all tunnel server addresses previously entered with {@link
    * Tunnel#addServer} 
    *
    */
    public void cleanServers();

    /**
    * Check whether the tunnel is connected. 
    *
    * @return A boolean value telling if the tunnel is connected 
    */
    public boolean connected();

    /**
    * Force reconnection to the tunnel server. 
    *
    * This method is useful when the device switches from wifi to Edge/3G or vice
    * versa. In most cases the tunnel client socket won't be notified promptly that
    * its connection is now zombie, so it is recommended to call this method that
    * will cause the lost connection to be closed and new connection to be issued. 
    */
    public void reconnect();

    /**
    * Remove a tunnel server configuration. 
    *
    * @param tunnelConfig {@link TunnelConfig} object 
    */
    public void removeServer(TunnelConfig tunnelConfig);

    /**
    * Set an optional http proxy to go through when connecting to tunnel server. 
    *
    * @param passwd Optional http proxy password. Use null if not needed. 
    */
    public void setHttpProxy(String host, int port, String username, String passwd);

    /**
    * Set authentication info for the http proxy. 
    *
    * @param passwd Password 
    */
    public void setHttpProxyAuthInfo(String username, String passwd);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class TunnelImpl implements Tunnel {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected TunnelImpl(long ptr) {
        nativePtr = ptr;
    }


    private native boolean getActivated(long nativePtr);
    @Override
    synchronized public boolean getActivated()  {
        
        return getActivated(nativePtr);
    }

    private native boolean dualModeEnabled(long nativePtr);
    @Override
    synchronized public boolean dualModeEnabled()  {
        
        return dualModeEnabled(nativePtr);
    }

    private native void enableDualMode(long nativePtr, boolean dualModeEnabled);
    @Override
    synchronized public void enableDualMode(boolean dualModeEnabled)  {
        
        enableDualMode(nativePtr, dualModeEnabled);
    }

    private native int getMode(long nativePtr);
    @Override
    synchronized public Tunnel.Mode getMode()  {
        
        return Tunnel.Mode.fromInt(getMode(nativePtr));
    }

    private native void setMode(long nativePtr, int mode);
    @Override
    synchronized public void setMode(Tunnel.Mode mode)  {
        
        setMode(nativePtr, mode.toInt());
    }

    private native TunnelConfig[] getServers(long nativePtr);
    @Override
    synchronized public TunnelConfig[] getServers()  {
        
        return getServers(nativePtr);
    }

    private native boolean sipEnabled(long nativePtr);
    @Override
    synchronized public boolean sipEnabled()  {
        
        return sipEnabled(nativePtr);
    }

    private native void enableSip(long nativePtr, boolean enable);
    @Override
    synchronized public void enableSip(boolean enable)  {
        
        enableSip(nativePtr, enable);
    }

    private native void addServer(long nativePtr, TunnelConfig tunnelConfig);
    @Override
    synchronized public void addServer(TunnelConfig tunnelConfig)  {
        
        addServer(nativePtr, tunnelConfig);
    }

    private native void cleanServers(long nativePtr);
    @Override
    synchronized public void cleanServers()  {
        
        cleanServers(nativePtr);
    }

    private native boolean connected(long nativePtr);
    @Override
    synchronized public boolean connected()  {
        
        return connected(nativePtr);
    }

    private native void reconnect(long nativePtr);
    @Override
    synchronized public void reconnect()  {
        
        reconnect(nativePtr);
    }

    private native void removeServer(long nativePtr, TunnelConfig tunnelConfig);
    @Override
    synchronized public void removeServer(TunnelConfig tunnelConfig)  {
        
        removeServer(nativePtr, tunnelConfig);
    }

    private native void setHttpProxy(long nativePtr, String host, int port, String username, String passwd);
    @Override
    synchronized public void setHttpProxy(String host, int port, String username, String passwd)  {
        
        setHttpProxy(nativePtr, host, port, username, passwd);
    }

    private native void setHttpProxyAuthInfo(long nativePtr, String username, String passwd);
    @Override
    synchronized public void setHttpProxyAuthInfo(String username, String passwd)  {
        
        setHttpProxyAuthInfo(nativePtr, username, passwd);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
