/*
Vcard.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* The {@link Vcard} object. 
*
*/
public interface Vcard {
    /**
    * Gets the eTag of the vCard. 
    *
    * @return the eTag of the vCard in the CardDAV server, otherwise null 
    */
    public String getEtag();

    /**
    * Sets the eTAG of the vCard. 
    *
    * @param etag the eTAG 
    */
    public void setEtag(String etag);

    /**
    * Returns the family name in the N attribute of the vCard, or null if it isn't
    * set yet. 
    *
    * @return the family name of the vCard, or null 
    */
    public String getFamilyName();

    /**
    * Sets the family name in the N attribute of the vCard. 
    *
    * @param name the family name to set for the vCard 
    */
    public void setFamilyName(String name);

    /**
    * Returns the FN attribute of the vCard, or null if it isn't set yet. 
    *
    * @return the display name of the vCard, or null 
    */
    public String getFullName();

    /**
    * Sets the FN attribute of the vCard (which is mandatory). 
    *
    * @param name the display name to set for the vCard 
    */
    public void setFullName(String name);

    /**
    * Returns the given name in the N attribute of the vCard, or null if it isn't set
    * yet. 
    *
    * @return the given name of the vCard, or null 
    */
    public String getGivenName();

    /**
    * Sets the given name in the N attribute of the vCard. 
    *
    * @param name the given name to set for the vCard 
    */
    public void setGivenName(String name);

    /**
    * Gets the Organization of the vCard. 
    *
    * @return the Organization of the vCard or null 
    */
    public String getOrganization();

    /**
    * Fills the Organization field of the vCard. 
    *
    * @param organization the Organization 
    */
    public void setOrganization(String organization);

    /**
    * Returns the list of phone numbers (as string) in the vCard (all the TEL
    * attributes) or null. 
    *
    * @return A list of const char * objects. const char *  
    */
    public String[] getPhoneNumbers();

    /**
    * Returns the list of SIP addresses (as LinphoneAddress) in the vCard (all the
    * IMPP attributes that has an URI value starting by "sip:") or null. 
    *
    * @return A list of {@link Address} objects. LinphoneAddress  
    */
    public Address[] getSipAddresses();

    /**
    * Returns the skipFieldValidation property of the vcard. 
    *
    * @return the skipFieldValidation property of the vcard 
    */
    public boolean getSkipValidation();

    /**
    * Sets the skipFieldValidation property of the vcard. 
    *
    * @param skip skipFieldValidation property of the vcard 
    */
    public void setSkipValidation(boolean skip);

    /**
    * Gets the UID of the vCard. 
    *
    * @return the UID of the vCard, otherwise null 
    */
    public String getUid();

    /**
    * Sets the unique ID of the vCard. 
    *
    * @param uid the unique id 
    */
    public void setUid(String uid);

    /**
    * Gets the URL of the vCard. 
    *
    * @return the URL of the vCard in the CardDAV server, otherwise null 
    */
    public String getUrl();

    /**
    * Sets the URL of the vCard. 
    *
    * @param url the URL 
    */
    public void setUrl(String url);

    /**
    * Adds a phone number in the vCard, using the TEL property. 
    *
    * @param phone the phone number to add 
    */
    public void addPhoneNumber(String phone);

    /**
    * Adds a SIP address in the vCard, using the IMPP property. 
    *
    * @param sipAddress the SIP address to add 
    */
    public void addSipAddress(String sipAddress);

    /**
    * Returns the vCard4 representation of the LinphoneVcard. 
    *
    * @return a const char * that represents the vCard 
    */
    public String asVcard4String();

    /**
    * Clone a {@link Vcard}. 
    *
    * @return a new {@link Vcard} object 
    */
    public Vcard clone();

    /**
    * Edits the preferred SIP address in the vCard (or the first one), using the IMPP
    * property. 
    *
    * @param sipAddress the new SIP address 
    */
    public void editMainSipAddress(String sipAddress);

    /**
    * Generates a random unique id for the vCard. 
    *
    * If is required to be able to synchronize the vCard with a CardDAV server 
    * 
    * @return  if operation is successful, otherwise true (for example if it already
    * has an unique ID) 
    */
    public boolean generateUniqueId();

    /**
    * Removes a phone number in the vCard (if it exists), using the TEL property. 
    *
    * @param phone the phone number to remove 
    */
    public void removePhoneNumber(String phone);

    /**
    * Removes a SIP address in the vCard (if it exists), using the IMPP property. 
    *
    * @param sipAddress the SIP address to remove 
    */
    public void removeSipAddress(String sipAddress);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class VcardImpl implements Vcard {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected VcardImpl(long ptr) {
        nativePtr = ptr;
    }


    private native String getEtag(long nativePtr);
    @Override
    synchronized public String getEtag()  {
        
        return getEtag(nativePtr);
    }

    private native void setEtag(long nativePtr, String etag);
    @Override
    synchronized public void setEtag(String etag)  {
        
        setEtag(nativePtr, etag);
    }

    private native String getFamilyName(long nativePtr);
    @Override
    synchronized public String getFamilyName()  {
        
        return getFamilyName(nativePtr);
    }

    private native void setFamilyName(long nativePtr, String name);
    @Override
    synchronized public void setFamilyName(String name)  {
        
        setFamilyName(nativePtr, name);
    }

    private native String getFullName(long nativePtr);
    @Override
    synchronized public String getFullName()  {
        
        return getFullName(nativePtr);
    }

    private native void setFullName(long nativePtr, String name);
    @Override
    synchronized public void setFullName(String name)  {
        
        setFullName(nativePtr, name);
    }

    private native String getGivenName(long nativePtr);
    @Override
    synchronized public String getGivenName()  {
        
        return getGivenName(nativePtr);
    }

    private native void setGivenName(long nativePtr, String name);
    @Override
    synchronized public void setGivenName(String name)  {
        
        setGivenName(nativePtr, name);
    }

    private native String getOrganization(long nativePtr);
    @Override
    synchronized public String getOrganization()  {
        
        return getOrganization(nativePtr);
    }

    private native void setOrganization(long nativePtr, String organization);
    @Override
    synchronized public void setOrganization(String organization)  {
        
        setOrganization(nativePtr, organization);
    }

    private native String[] getPhoneNumbers(long nativePtr);
    @Override
    synchronized public String[] getPhoneNumbers()  {
        
        return getPhoneNumbers(nativePtr);
    }

    private native Address[] getSipAddresses(long nativePtr);
    @Override
    synchronized public Address[] getSipAddresses()  {
        
        return getSipAddresses(nativePtr);
    }

    private native boolean getSkipValidation(long nativePtr);
    @Override
    synchronized public boolean getSkipValidation()  {
        
        return getSkipValidation(nativePtr);
    }

    private native void setSkipValidation(long nativePtr, boolean skip);
    @Override
    synchronized public void setSkipValidation(boolean skip)  {
        
        setSkipValidation(nativePtr, skip);
    }

    private native String getUid(long nativePtr);
    @Override
    synchronized public String getUid()  {
        
        return getUid(nativePtr);
    }

    private native void setUid(long nativePtr, String uid);
    @Override
    synchronized public void setUid(String uid)  {
        
        setUid(nativePtr, uid);
    }

    private native String getUrl(long nativePtr);
    @Override
    synchronized public String getUrl()  {
        
        return getUrl(nativePtr);
    }

    private native void setUrl(long nativePtr, String url);
    @Override
    synchronized public void setUrl(String url)  {
        
        setUrl(nativePtr, url);
    }

    private native void addPhoneNumber(long nativePtr, String phone);
    @Override
    synchronized public void addPhoneNumber(String phone)  {
        
        addPhoneNumber(nativePtr, phone);
    }

    private native void addSipAddress(long nativePtr, String sipAddress);
    @Override
    synchronized public void addSipAddress(String sipAddress)  {
        
        addSipAddress(nativePtr, sipAddress);
    }

    private native String asVcard4String(long nativePtr);
    @Override
    synchronized public String asVcard4String()  {
        
        return asVcard4String(nativePtr);
    }

    private native Vcard clone(long nativePtr);
    @Override
    synchronized public Vcard clone()  {
        
        return (Vcard)clone(nativePtr);
    }

    private native void editMainSipAddress(long nativePtr, String sipAddress);
    @Override
    synchronized public void editMainSipAddress(String sipAddress)  {
        
        editMainSipAddress(nativePtr, sipAddress);
    }

    private native boolean generateUniqueId(long nativePtr);
    @Override
    synchronized public boolean generateUniqueId()  {
        
        return generateUniqueId(nativePtr);
    }

    private native void removePhoneNumber(long nativePtr, String phone);
    @Override
    synchronized public void removePhoneNumber(String phone)  {
        
        removePhoneNumber(nativePtr, phone);
    }

    private native void removeSipAddress(long nativePtr, String sipAddress);
    @Override
    synchronized public void removeSipAddress(String sipAddress)  {
        
        removeSipAddress(nativePtr, sipAddress);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
