/*
Core.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import org.jetbrains.annotations.Nullable;

/**
* Linphone core main object created by function linphone_core_new . 
*
*/
public interface Core {
    public enum LogCollectionUploadState {
        /**
        * Delivery in progress. 
        *
        */
        InProgress(0),

        /**
        * Log collection upload successfully delivered and acknowledged by remote end
        * point. 
        *
        */
        Delivered(1),

        /**
        * Log collection upload was not delivered. 
        *
        */
        NotDelivered(2);

        protected final int mValue;

        private LogCollectionUploadState (int value) {
            mValue = value;
        }

        static public LogCollectionUploadState fromInt(int value) throws RuntimeException {
            switch(value) {
            case 0: return InProgress;
            case 1: return Delivered;
            case 2: return NotDelivered;
            default:
                throw new RuntimeException("Unhandled enum value " + value + " for LogCollectionUploadState");
            }
        }

        public int toInt() {
            return mValue;
        }
    };

    /**
      * Gets the mediastreamer's factory
      */
    public org.linphone.mediastream.Factory getMediastreamerFactory();
    /**
    * Returns which adaptive rate algorithm is currently configured for future calls. 
    *
    * See also: {@link Core#setAdaptiveRateAlgorithm} 
    */
    public String getAdaptiveRateAlgorithm();

    /**
    * Sets adaptive rate algorithm. 
    *
    * It will be used for each new calls starting from now. Calls already started
    * will not be updated. 
    * 
    * @param algorithm the adaptive rate control algorithm. Currently two values are
    * supported: 'advanced', which is the default value, or 'basic'. 
    */
    public void setAdaptiveRateAlgorithm(String algorithm);

    /**
    * Returns whether adaptive rate control is enabled. 
    *
    * See also: {@link Core#enableAdaptiveRateControl} 
    */
    public boolean adaptiveRateControlEnabled();

    /**
    * Enable adaptive rate control. 
    *
    * Adaptive rate control consists in using RTCP feedback provided information to
    * dynamically control the output bitrate of the audio and video encoders, so that
    * we can adapt to the network conditions and available bandwidth. Control of the
    * audio encoder is done in case of audio-only call, and control of the video
    * encoder is done for audio & video calls. Adaptive rate control feature is
    * enabled by default. 
    */
    public void enableAdaptiveRateControl(boolean enabled);

    /**
    * Tells whether the audio adaptive jitter compensation is enabled. 
    *
    * @return  if the audio adaptive jitter compensation is enabled, true otherwise. 
    */
    public boolean audioAdaptiveJittcompEnabled();

    /**
    * Enable or disable the audio adaptive jitter compensation. 
    *
    * @param enable  to enable the audio adaptive jitter compensation, true to
    * disable it. 
    */
    public void enableAudioAdaptiveJittcomp(boolean enable);

    /**
    * Get the DSCP field for outgoing audio streams. 
    *
    * The DSCP defines the quality of service in IP packets. 
    * 
    * @return The current DSCP value 
    */
    public int getAudioDscp();

    /**
    * Set the DSCP field for outgoing audio streams. 
    *
    * The DSCP defines the quality of service in IP packets. 
    * 
    * @param dscp The DSCP value to set 
    */
    public void setAudioDscp(int dscp);

    /**
    * Returns the nominal audio jitter buffer size in milliseconds. 
    *
    * @return The nominal audio jitter buffer size in milliseconds 
    */
    public int getAudioJittcomp();

    /**
    * Sets the nominal audio jitter buffer size in milliseconds. 
    *
    * The value takes effect immediately for all running and pending calls, if any. A
    * value of 0 disables the jitter buffer. 
    */
    public void setAudioJittcomp(int milliseconds);

    /**
    * Use to get multicast address to be used for audio stream. 
    *
    * @return an ipv4/6 multicast address or default value 
    */
    public String getAudioMulticastAddr();

    /**
    * Use to set multicast address to be used for audio stream. 
    *
    * @param ip an ipv4/6 multicast address 
    * 
    * @return 0 in case of success 
    */
    public int setAudioMulticastAddr(String ip);

    /**
    * Use to get multicast state of audio stream. 
    *
    * @return true if subsequent calls will propose multicast ip set by
    * linphone_core_set_audio_multicast_addr 
    */
    public boolean audioMulticastEnabled();

    /**
    * Use to enable multicast rtp for audio stream. 
    *
    * If enabled, outgoing calls put a multicast address from
    * linphone_core_get_video_multicast_addr into audio cline. In case of outgoing
    * call audio stream is sent to this multicast address. For incoming calls
    * behavior is unchanged. 
    * 
    * @param yesno if yes, subsequent calls will propose multicast ip set by
    * linphone_core_set_audio_multicast_addr 
    */
    public void enableAudioMulticast(boolean yesno);

    /**
    * Use to get multicast ttl to be used for audio stream. 
    *
    * @return a time to leave value 
    */
    public int getAudioMulticastTtl();

    /**
    * Use to set multicast ttl to be used for audio stream. 
    *
    * @param ttl value or -1 if not used. [0..255] default value is 1 
    * 
    * @return 0 in case of success 
    */
    public int setAudioMulticastTtl(int ttl);

    /**
    * Return the list of the available audio payload types. 
    *
    * @return A list of {@link PayloadType} objects. LinphonePayloadType  A freshly
    * allocated list of the available payload types. The list must be destroyed with
    * bctbx_list_free() after usage. The elements of the list haven't to be unref. 
    */
    public PayloadType[] getAudioPayloadTypes();

    /**
    * Redefine the list of the available payload types. 
    *
    * @param payloadTypes A list of {@link PayloadType} objects. LinphonePayloadType 
    * The new list of payload types. The core does not take ownership on it. 
    */
    public void setAudioPayloadTypes(PayloadType[] payloadTypes);

    /**
    * Gets the UDP port used for audio streaming. 
    *
    * @return The UDP port used for audio streaming 
    */
    public int getAudioPort();

    /**
    * Sets the UDP port used for audio streaming. 
    *
    * A value of -1 will request the system to allocate the local port randomly. This
    * is recommended in order to avoid firewall warnings. 
    * 
    * @param port The UDP port to use for audio streaming 
    */
    public void setAudioPort(int port);

    /**
    * Get the audio port range from which is randomly chosen the UDP port used for
    * audio streaming. 
    *
    * @return a {@link Range} object 
    */
    public Range getAudioPortsRange();

    /**
    * Returns an unmodifiable list of currently entered {@link AuthInfo}. 
    *
    * @return A list of {@link AuthInfo} objects. LinphoneAuthInfo  
    */
    public AuthInfo[] getAuthInfoList();

    /**
    * Return AVPF enablement. 
    *
    * See {@link Core#setAvpfMode} . 
    * 
    * @return The current AVPF mode 
    */
    public AVPFMode getAvpfMode();

    /**
    * Enable RTCP feedback (also known as RTP/AVPF profile). 
    *
    * Setting LinphoneAVPFDefault is equivalent to LinphoneAVPFDisabled. This setting
    * can be overriden per {@link ProxyConfig} with {@link ProxyConfig#setAvpfMode}.
    * The value set here is used for calls placed or received out of any proxy
    * configured, or if the proxy config is configured with LinphoneAVPFDefault. 
    * 
    * @param mode The AVPF mode to use. 
    */
    public void setAvpfMode(AVPFMode mode);

    /**
    * Return the avpf report interval in seconds. 
    *
    * @return The current AVPF report interval in seconds 
    */
    public int getAvpfRrInterval();

    /**
    * Set the avpf report interval in seconds. 
    *
    * This value can be overriden by the proxy config using {@link
    * ProxyConfig#setAvpfRrInterval}. 
    * 
    * @param interval The report interval in seconds 
    */
    public void setAvpfRrInterval(int interval);

    /**
    * Get the list of call logs (past calls). 
    *
    * @return A list of {@link CallLog} objects. LinphoneCallLog  
    */
    public CallLog[] getCallLogs();

    /**
    * Gets the database filename where call logs will be stored. 
    *
    * @return filesystem path 
    */
    public String getCallLogsDatabasePath();

    /**
    * Sets the database filename where call logs will be stored. 
    *
    * If the file does not exist, it will be created.
    * 
    * @param path filesystem path 
    */
    public void setCallLogsDatabasePath(String path);

    /**
    * Gets the current list of calls. 
    *
    * Note that this list is read-only and might be changed by the core after a
    * function call to {@link Core#iterate}. Similarly the {@link Call} objects
    * inside it might be destroyed without prior notice. To hold references to {@link
    * Call} object into your program, you must use linphone_call_ref. 
    * 
    * @return A list of {@link Call} objects. LinphoneCall  
    */
    public Call[] getCalls();

    /**
    * Get the number of Call. 
    *
    * @return The current number of calls 
    */
    public int getCallsNb();

    /**
    * Get the camera sensor rotation. 
    *
    * This is needed on some mobile platforms to get the number of degrees the camera
    * sensor is rotated relative to the screen. 
    * 
    * @return The camera sensor rotation in degrees (0 to 360) or -1 if it could not
    * be retrieved 
    */
    public int getCameraSensorRotation();

    /**
    * Gets the name of the currently assigned sound device for capture. 
    *
    * @return The name of the currently assigned sound device for capture 
    */
    public String getCaptureDevice();

    /**
    * Sets the sound device used for capture. 
    *
    * @param devid The device name as returned by linphone_core_get_sound_devices 
    * 
    * @return 0 
    */
    public int setCaptureDevice(String devid);

    /**
    * Get path to the database file used for storing chat messages. 
    *
    * @return file path or null if not exist 
    * 
    * @deprecated 2018-01-10 
    */
    @Deprecated
    public String getChatDatabasePath();

    /**
    * Set the chat database path. 
    *
    * @param path the database path 
    * 
    * @deprecated 2018-01-10: Use only for migration purposes 
    */
    @Deprecated
    public void setChatDatabasePath(String path);

    /**
    * Returns whether chat is enabled. 
    *
    */
    public boolean chatEnabled();

    /**
    * Returns an list of chat rooms. 
    *
    * @return A list of {@link ChatRoom} objects. LinphoneChatRoom  
    */
    public ChatRoom[] getChatRooms();

    /**
    * Get a pointer on the internal conference object. 
    *
    * @return A pointer on {@link Conference} or null if no conference are going on 
    */
    public Conference getConference();

    /**
    * Get the set input volume of the local participant. 
    *
    * @return A value inside [0.0 ; 1.0] 
    */
    public float getConferenceLocalInputVolume();

    /**
    * Tells whether the conference server feature is enabled. 
    *
    * @return A boolean value telling whether the conference server feature is
    * enabled or not 
    */
    public boolean conferenceServerEnabled();

    /**
    * Enable the conference server feature. 
    *
    * This has the effect to listen of the conference factory uri to create new
    * conferences when receiving INVITE messages there. 
    * 
    * @param enable A boolean value telling whether to enable or disable the
    * conference server feature 
    */
    public void enableConferenceServer(boolean enable);

    /**
    * Get the number of participant in the running conference. 
    *
    * The local participant is included in the count only if it is in the conference. 
    * 
    * @return The number of participant 
    */
    public int getConferenceSize();

    /**
    * Returns the LpConfig object used to manage the storage (config) file. 
    *
    */
    public Config getConfig();

    /**
    * Get my consolidated presence. 
    *
    * @return My consolidated presence 
    */
    public ConsolidatedPresence getConsolidatedPresence();

    /**
    * Set my consolidated presence. 
    *
    * @param presence LinphoneConsolidatedPresence value 
    */
    public void setConsolidatedPresence(ConsolidatedPresence presence);

    /**
    * Gets the current call. 
    *
    * @return The current call or null if no call is running 
    */
    public Call getCurrentCall();

    /**
    * Get the remote address of the current call. 
    *
    * @return The remote address of the current call or null if there is no current
    * call. 
    */
    public Address getCurrentCallRemoteAddress();

    /**
    * Get the effective video definition provided by the camera for the captured
    * video. 
    *
    * When preview is disabled or not yet started this function returns a 0x0 video
    * definition. 
    * 
    * @return The captured {@link VideoDefinition}
    * 
    * See also: {@link Core#setPreviewVideoDefinition} 
    */
    public VideoDefinition getCurrentPreviewVideoDefinition();

    /**
    * Retrieves the first list of {@link Friend} from the core. 
    *
    * @return the first {@link FriendList} object or null 
    */
    public FriendList getDefaultFriendList();

    /**
    *
    * @return the default proxy configuration, that is the one used to determine the
    * current identity. 
    * 
    * @return The default proxy configuration. 
    */
    public ProxyConfig getDefaultProxyConfig();

    /**
    * Sets the default proxy. 
    *
    * This default proxy must be part of the list of already entered
    * LinphoneProxyConfig. Toggling it as default will make {@link Core} use the
    * identity associated with the proxy configuration in all incoming and outgoing
    * calls. 
    * 
    * @param config The proxy configuration to use as the default one. 
    */
    public void setDefaultProxyConfig(ProxyConfig config);

    /**
    * Get the name of the default mediastreamer2 filter used for rendering video on
    * the current platform. 
    *
    * This is for advanced users of the library, mainly to expose mediastreamer video
    * filter name and status. 
    * 
    * @return The default video display filter 
    */
    public String getDefaultVideoDisplayFilter();

    /**
    * Gets the delayed timeout See {@link Core#setDelayedTimeout} for details. 
    *
    * @return The current delayed timeout in seconds 
    */
    public int getDelayedTimeout();

    /**
    * Set the in delayed timeout in seconds. 
    *
    * After this timeout period, a delayed call (internal call initialisation or
    * resolution) is resumed. 
    * 
    * @param seconds The new delayed timeout 
    */
    public void setDelayedTimeout(int seconds);

    /**
    * Gets the current device orientation. 
    *
    * @return The current device orientation
    * 
    * See also: {@link Core#setDeviceRotation} 
    */
    public int getDeviceRotation();

    /**
    * Tells the core the device current orientation. 
    *
    * This can be used by capture filters on mobile devices to select between
    * portrait/landscape mode and to produce properly oriented images. The exact
    * meaning of the value in rotation if left to each device specific
    * implementations. IOS supported values are 0 for UIInterfaceOrientationPortrait
    * and 270 for UIInterfaceOrientationLandscapeRight. 
    * 
    * @param rotation The orientation to use 
    */
    public void setDeviceRotation(int rotation);

    /**
    * Tells whether DNS search (use of local domain if the fully qualified name did
    * return results) is enabled. 
    *
    * @return  if DNS search is enabled, true if disabled. 
    */
    public boolean dnsSearchEnabled();

    /**
    * Enable or disable DNS search (use of local domain if the fully qualified name
    * did return results). 
    *
    * @param enable  to enable DNS search, true to disable it. 
    */
    public void enableDnsSearch(boolean enable);

    /**
    * Forces liblinphone to use the supplied list of dns servers, instead of system's
    * ones. 
    *
    * @param servers A list of const char * objects. const char *  A list of strings
    * containing the IP addresses of DNS servers to be used. Setting to null restores
    * default behaviour, which is to use the DNS server list provided by the system.
    * The list is copied internally. 
    */
    public void setDnsServers(String[] servers);

    /**
    * Forces liblinphone to use the supplied list of dns servers, instead of system's
    * ones and set dns_set_by_app at true or false according to value of servers
    * list. 
    *
    * @param servers A list of const char * objects. const char *  A list of strings
    * containing the IP addresses of DNS servers to be used. Setting to null restores
    * default behaviour, which is to use the DNS server list provided by the system.
    * The list is copied internally. 
    */
    public void setDnsServersApp(String[] servers);

    /**
    * Tells if the DNS was set by an application. 
    *
    * @return  if DNS was set by app, true otherwise. 
    */
    public boolean getDnsSetByApp();

    /**
    * Tells whether DNS SRV resolution is enabled. 
    *
    * @return  if DNS SRV resolution is enabled, true if disabled. 
    */
    public boolean dnsSrvEnabled();

    /**
    * Enable or disable DNS SRV resolution. 
    *
    * @param enable  to enable DNS SRV resolution, true to disable it. 
    */
    public void enableDnsSrv(boolean enable);

    /**
    * Retrieve the maximum available download bandwidth. 
    *
    * This value was set by {@link Core#setDownloadBandwidth}. 
    */
    public int getDownloadBandwidth();

    /**
    * Sets maximum available download bandwidth This is IP bandwidth, in kbit/s. 
    *
    * This information is used signaled to other parties during calls (within SDP
    * messages) so that the remote end can have sufficient knowledge to properly
    * configure its audio & video codec output bitrate to not overflow available
    * bandwidth.
    * 
    * @param bw the bandwidth in kbits/s, 0 for infinite 
    */
    public void setDownloadBandwidth(int bw);

    /**
    * Get audio packetization time linphone expects to receive from peer. 
    *
    * A value of zero means that ptime is not specified. 
    */
    public int getDownloadPtime();

    /**
    * Set audio packetization time linphone expects to receive from peer. 
    *
    * A value of zero means that ptime is not specified. 
    */
    public void setDownloadPtime(int ptime);

    /**
    * Returns true if echo cancellation is enabled. 
    *
    * @return A boolean value telling whether echo cancellation is enabled or
    * disabled 
    */
    public boolean echoCancellationEnabled();

    /**
    * Enables or disable echo cancellation. 
    *
    * Value is saved and used for subsequent calls. This actually controls software
    * echo cancellation. If hardware echo cancellation is available, it will be
    * always used and activated for calls, regardless of the value passed to this
    * function. When hardware echo cancellation is available, the software one is of
    * course not activated. 
    * 
    * @param val A boolean value telling whether echo cancellation is to be enabled
    * or disabled. 
    */
    public void enableEchoCancellation(boolean val);

    /**
    * Get the name of the mediastreamer2 filter used for echo cancelling. 
    *
    * @return The name of the mediastreamer2 filter used for echo cancelling 
    */
    public String getEchoCancellerFilterName();

    /**
    * Set the name of the mediastreamer2 filter to be used for echo cancelling. 
    *
    * This is for advanced users of the library. 
    * 
    * @param filtername The name of the mediastreamer2 filter to be used for echo
    * cancelling 
    */
    public void setEchoCancellerFilterName(String filtername);

    /**
    * Tells whether echo limiter is enabled. 
    *
    * @return  if the echo limiter is enabled, true otherwise. 
    */
    public boolean echoLimiterEnabled();

    /**
    * Enables or disable echo limiter. 
    *
    * @param val  to enable echo limiter, true to disable it. 
    */
    public void enableEchoLimiter(boolean val);

    /**
    * Enable or disable the UPDATE method support. 
    *
    * @param value Enable or disable it 
    */
    public void setEnableSipUpdate(int value);

    /**
    * Sets expected available upload bandwidth This is IP bandwidth, in kbit/s. 
    *
    * This information is used by liblinphone together with remote side available
    * bandwidth signaled in SDP messages to properly configure audio & video codec's
    * output bitrate.
    * 
    * @param bw the bandwidth in kbits/s, 0 for infinite 
    */
    public void setExpectedBandwidth(int bw);

    /**
    * Get the globaly set http file transfer server to be used for content type
    * application/vnd.gsma.rcs-ft-http+xml. 
    *
    * @return URL of the file server like https://file.linphone.org/upload.php 
    */
    public String getFileTransferServer();

    /**
    * Globaly set an http file transfer server to be used for content type
    * application/vnd.gsma.rcs-ft-http+xml. 
    *
    * This value can also be set for a dedicated account using
    * linphone_proxy_config_set_file_transfer_server 
    * 
    * @param serverUrl URL of the file server like
    * https://file.linphone.org/upload.php 
    */
    public void setFileTransferServer(String serverUrl);

    /**
    * Sets whether or not to start friend lists subscription when in foreground. 
    *
    * @param enable whether or not to enable the feature 
    */
    public void enableFriendListSubscription(boolean enable);

    /**
    * Gets the database filename where friends will be stored. 
    *
    * @return filesystem path 
    */
    public String getFriendsDatabasePath();

    /**
    * Sets the database filename where friends will be stored. 
    *
    * If the file does not exist, it will be created.
    * 
    * @param path filesystem path 
    */
    public void setFriendsDatabasePath(String path);

    /**
    * Retrieves the list of {@link FriendList} from the core. 
    *
    * @return A list of {@link FriendList} objects. LinphoneFriendList  a list of
    * {@link FriendList} 
    */
    public FriendList[] getFriendsLists();

    /**
    * Returns true if hostname part of primary contact is guessed automatically. 
    *
    */
    public boolean getGuessHostname();

    /**
    * Tells {@link Core} to guess local hostname automatically in primary contact. 
    *
    */
    public void setGuessHostname(boolean val);

    /**
    * Get http proxy address to be used for signaling. 
    *
    * @return hostname of IP adress of the http proxy (can be null to disable). 
    */
    public String getHttpProxyHost();

    /**
    * Set http proxy address to be used for signaling during next channel connection. 
    *
    * Use linphone_core_set_network_reachable FASLE/true to force channel restart. 
    * 
    * @param host Hostname of IP adress of the http proxy (can be null to disable). 
    */
    public void setHttpProxyHost(String host);

    /**
    * Get http proxy port to be used for signaling. 
    *
    * @return port of the http proxy. 
    */
    public int getHttpProxyPort();

    /**
    * Set http proxy port to be used for signaling. 
    *
    * @param port of the http proxy. 
    */
    public void setHttpProxyPort(int port);

    /**
    * Gets the default identity SIP address. 
    *
    * This is an helper function. If no default proxy is set, this will return the
    * primary contact ( see {@link Core#getPrimaryContact} ). If a default proxy is
    * set it returns the registered identity on the proxy. 
    * 
    * @return The default identity SIP address 
    */
    public String getIdentity();

    /**
    * Get the {@link ImNotifPolicy} object controlling the instant messaging
    * notifications. 
    *
    * @return A {@link ImNotifPolicy} object. 
    */
    public ImNotifPolicy getImNotifPolicy();

    /**
    * Gets the in call timeout See {@link Core#setInCallTimeout} for details. 
    *
    * @return The current in call timeout in seconds 
    */
    public int getInCallTimeout();

    /**
    * Set the in call timeout in seconds. 
    *
    * After this timeout period, the call is automatically hangup. 
    * 
    * @param seconds The new timeout in seconds 
    */
    public void setInCallTimeout(int seconds);

    /**
    * Returns the incoming call timeout See {@link Core#setIncTimeout} for details. 
    *
    * @return The current incoming call timeout in seconds 
    */
    public int getIncTimeout();

    /**
    * Set the incoming call timeout in seconds. 
    *
    * If an incoming call isn't answered for this timeout period, it is automatically
    * declined. 
    * 
    * @param seconds The new timeout in seconds 
    */
    public void setIncTimeout(int seconds);

    /**
    * Tells whether IPv6 is enabled or not. 
    *
    * @return A boolean value telling whether IPv6 is enabled or not
    * 
    * See also: {@link Core#enableIpv6} for more details on how IPv6 is supported in
    * liblinphone. 
    */
    public boolean ipv6Enabled();

    /**
    * Turns IPv6 support on or off. 
    *
    * @param val A boolean value telling whether to enable IPv6 support 
    */
    public void enableIpv6(boolean val);

    /**
    * Check whether the device is echo canceller calibration is required. 
    *
    * @return  if it is required, true otherwise 
    */
    public boolean isEchoCancellerCalibrationRequired();

    /**
    * Returns whether or not friend lists subscription are enabled. 
    *
    * @return whether or not the feature is enabled 
    */
    public boolean isFriendListSubscriptionEnabled();

    /**
    * Indicates whether the local participant is part of a conference. 
    *
    * Warning: That function automatically fails in the case of conferences using a
    * conferencet server (focus). If you use such a conference, you should use {@link
    * Conference#removeParticipant} instead. 
    * 
    * @return  if the local participant is in a conference, true otherwise. 
    */
    public boolean isInConference();

    /**
    * Tells whether there is an incoming invite pending. 
    *
    * @return A boolean telling whether an incoming invite is pending or not. 
    */
    public boolean isIncomingInvitePending();

    /**
    * Check if the configured media encryption is mandatory or not. 
    *
    * @return  if media encryption is mandatory; true otherwise. 
    */
    public boolean isMediaEncryptionMandatory();

    /**
    * return network state either as positioned by the application or by linphone
    * itself. 
    *
    */
    public boolean isNetworkReachable();

    /**
    * Returns whether or not sender name is hidden in forward message. 
    *
    * @return whether or not the feature 
    */
    public boolean isSenderNameHiddenInForwardMessage();

    /**
    * Is signaling keep alive enabled. 
    *
    * @return A boolean value telling whether signaling keep alive is enabled 
    */
    public boolean keepAliveEnabled();

    /**
    * Enables signaling keep alive, small udp packet sent periodically to keep udp
    * NAT association. 
    *
    * @param enable A boolean value telling whether signaling keep alive is to be
    * enabled 
    */
    public void enableKeepAlive(boolean enable);

    /**
    * Get the latest outgoing call log. 
    *
    * @return {LinphoneCallLog} 
    */
    public CallLog getLastOutgoingCallLog();

    /**
    * Tells wether LIME X3DH is enabled or not. 
    *
    * @return The current lime state 
    */
    public boolean limeX3DhEnabled();

    /**
    * Tells to LinphoneCore to use LIME X3DH. 
    *
    * @param enable A boolean value telling whether to enable or disable LIME X3DH 
    */
    public void enableLimeX3Dh(boolean enable);

    /**
    * Get the x3dh server url. 
    *
    */
    public String getLimeX3DhServerUrl();

    /**
    * Set the x3dh server url. 
    *
    * If empty, this function will disable LIME X3DH from core. Otherwise, or if
    * different from the existing value, this will (re-)initialize the LIME X3DH
    * engine. 
    * 
    * @param url The x3dh server url 
    */
    public void setLimeX3DhServerUrl(String url);

    /**
    * Get the linphone specs value telling what functionalities the linphone client
    * supports. 
    *
    * @return The linphone specs telling what functionalities the linphone client
    * supports
    * 
    * @deprecated Use linphone_core_get_linphone_specs_list instead. Deprecated since
    * 2019-02-07 
    */
    @Deprecated
    public String getLinphoneSpecs();

    /**
    * Set the linphone specs value telling what functionalities the linphone client
    * supports. 
    *
    * @param specs The linphone specs to set
    * 
    * @deprecated Use linphone_core_set_linphone_specs_list or
    * linphone_core_add_linphone_spec instead. Deprecated since 2019-02-07 
    */
    @Deprecated
    public void setLinphoneSpecs(String specs);

    /**
    * Get the list of linphone specs string values representing what functionalities
    * the linphone client supports. 
    *
    * @return A list of char * objects. char *  a list of supported specs. The list
    * must be freed with bctbx_list_free() after usage 
    */
    public String[] getLinphoneSpecsList();

    /**
    * Set the linphone specs list value telling what functionalities the linphone
    * client supports. 
    *
    * @param specs A list of char * objects. char *  The list of string specs to set 
    */
    public void setLinphoneSpecsList(String[] specs);

    /**
    * Gets the url of the server where to upload the collected log files. 
    *
    * @return The url of the server where to upload the collected log files. 
    */
    public String getLogCollectionUploadServerUrl();

    /**
    * Set the url of the server where to upload the collected log files. 
    *
    * @param serverUrl The url of the server where to upload the collected log files. 
    */
    public void setLogCollectionUploadServerUrl(String serverUrl);

    /**
    * Get the maximum number of simultaneous calls Linphone core can manage at a
    * time. 
    *
    * All new call above this limit are declined with a busy answer 
    * 
    * @return max number of simultaneous calls 
    */
    public int getMaxCalls();

    /**
    * Set the maximum number of simultaneous calls Linphone core can manage at a
    * time. 
    *
    * All new call above this limit are declined with a busy answer 
    * 
    * @param max number of simultaneous calls 
    */
    public void setMaxCalls(int max);

    /**
    * Gets the size under which incoming files in chat messages will be downloaded
    * automatically. 
    *
    * @return The size in bytes, -1 if autodownload feature is disabled, 0 to
    * download them all no matter the size 
    */
    public int getMaxSizeForAutoDownloadIncomingFiles();

    /**
    * Sets the size under which incoming files in chat messages will be downloaded
    * automatically. 
    *
    * @param size The size in bytes, -1 to disable the autodownload feature, 0 to
    * download them all no matter the size 
    */
    public void setMaxSizeForAutoDownloadIncomingFiles(int size);

    /**
    * Gets the name of the currently assigned sound device for media. 
    *
    * @return The name of the currently assigned sound device for capture 
    */
    public String getMediaDevice();

    /**
    * Sets the sound device used for media. 
    *
    * @param devid The device name as returned by linphone_core_get_sound_devices 
    * 
    * @return 0 
    */
    public int setMediaDevice(String devid);

    /**
    * Get the media encryption policy being used for RTP packets. 
    *
    * @return The media encryption policy being used. 
    */
    public MediaEncryption getMediaEncryption();

    /**
    * Choose the media encryption policy to be used for RTP packets. 
    *
    * @param menc The media encryption policy to be used. 
    * 
    * @return 0 if successful, any other value otherwise. 
    */
    public int setMediaEncryption(MediaEncryption menc);

    /**
    * Define whether the configured media encryption is mandatory, if it is and the
    * negotation cannot result in the desired media encryption then the call will
    * fail. 
    *
    * If not an INVITE will be resent with encryption disabled. 
    * 
    * @param m  to set it mandatory; true otherwise. 
    */
    public void setMediaEncryptionMandatory(boolean m);

    /**
    * This method is called by the application to notify the linphone core library
    * when the media (RTP) network is reachable. 
    *
    * This is for advanced usage, when SIP and RTP layers are required to use
    * different interfaces. Most applications just need {@link
    * Core#setNetworkReachable}. 
    */
    public void setMediaNetworkReachable(boolean value);

    /**
    * Tells whether the microphone is enabled. 
    *
    * @return  if the microphone is enabled, true if disabled. 
    */
    public boolean micEnabled();

    /**
    * Enable or disable the microphone. 
    *
    * @param enable  to enable the microphone, true to disable it. 
    */
    public void enableMic(boolean enable);

    /**
    * Get microphone gain in db. 
    *
    * @return The current microphone gain 
    */
    public float getMicGainDb();

    /**
    * Allow to control microphone level: gain in db. 
    *
    * @param level The new microphone level 
    */
    public void setMicGainDb(float level);

    /**
    * Get the number of missed calls. 
    *
    * Once checked, this counter can be reset with {@link
    * Core#resetMissedCallsCount}. 
    * 
    * @return The number of missed calls. 
    */
    public int getMissedCallsCount();

    /**
    * Returns the maximum transmission unit size in bytes. 
    *
    */
    public int getMtu();

    /**
    * Sets the maximum transmission unit size in bytes. 
    *
    * This information is useful for sending RTP packets. Default value is 1500. 
    * 
    * @param mtu The MTU in bytes 
    */
    public void setMtu(int mtu);

    /**
    * Get the public IP address of NAT being used. 
    *
    * @return The public IP address of NAT being used. 
    */
    public String getNatAddress();

    /**
    * Set the public IP address of NAT when using the firewall policy is set to use
    * NAT. 
    *
    * @param addr The public IP address of NAT to use. 
    */
    public void setNatAddress(String addr);

    /**
    * Get The policy that is used to pass through NATs/firewalls. 
    *
    * It may be overridden by a NAT policy for a specific proxy config. 
    * 
    * @return {@link NatPolicy} object in use.
    * 
    * See also: {@link ProxyConfig#getNatPolicy} 
    */
    public NatPolicy getNatPolicy();

    /**
    * Set the policy to use to pass through NATs/firewalls. 
    *
    * It may be overridden by a NAT policy for a specific proxy config. 
    * 
    * @param policy {@link NatPolicy} object
    * 
    * See also: {@link ProxyConfig#setNatPolicy} 
    */
    public void setNatPolicy(NatPolicy policy);

    /**
    * Get the native window handle of the video preview window. 
    *
    * @return The native window handle of the video preview window 
    */
    public Object getNativePreviewWindowId();

    /**
    * Set the native window id where the preview video (local camera) is to be
    * displayed. 
    *
    * This has to be used in conjonction with {@link Core#usePreviewWindow}. MacOS,
    * Linux, Windows: if not set or zero the core will create its own window, unless
    * the special id -1 is given. 
    * 
    * @param id The native window id where the preview video is to be displayed 
    */
    public void setNativePreviewWindowId(Object id);

    /**
    * Get the native window handle of the video window. 
    *
    * @return The native window handle of the video window 
    */
    public Object getNativeVideoWindowId();

    /**
    * Set the native video window id where the video is to be displayed. 
    *
    * For MacOS, Linux, Windows: if not set or LINPHONE_VIDEO_DISPLAY_AUTO the core
    * will create its own window, unless the special id LINPHONE_VIDEO_DISPLAY_NONE
    * is given. 
    */
    public void setNativeVideoWindowId(Object id);

    /**
    * This method is called by the application to notify the linphone core library
    * when network is reachable. 
    *
    * Calling this method with true trigger linphone to initiate a registration
    * process for all proxies. Calling this method disables the automatic network
    * detection mode. It means you must call this method after each network state
    * changes. 
    */
    public void setNetworkReachable(boolean value);

    /**
    * Gets the value of the no-rtp timeout. 
    *
    * When no RTP or RTCP packets have been received for a while {@link Core} will
    * consider the call is broken (remote end crashed or disconnected from the
    * network), and thus will terminate the call. The no-rtp timeout is the duration
    * above which the call is considered broken. 
    * 
    * @return The value of the no-rtp timeout in seconds 
    */
    public int getNortpTimeout();

    /**
    * Sets the no-rtp timeout value in seconds. 
    *
    * @param seconds The no-rtp timeout value to use in seconds
    * 
    * See also: {@link Core#getNortpTimeout} for details. 
    */
    public void setNortpTimeout(int seconds);

    /**
    * Get the wav file that is played when putting somebody on hold, or when files
    * are used instead of soundcards (see {@link Core#setUseFiles}). 
    *
    * The file is a 16 bit linear wav file. 
    * 
    * @return The path to the file that is played when putting somebody on hold. 
    */
    public String getPlayFile();

    /**
    * Sets a wav file to be played when putting somebody on hold, or when files are
    * used instead of soundcards (see {@link Core#setUseFiles}). 
    *
    * The file must be a 16 bit linear wav file. 
    * 
    * @param file The path to the file to be played when putting somebody on hold. 
    */
    public void setPlayFile(String file);

    /**
    * Gets the name of the currently assigned sound device for playback. 
    *
    * @return The name of the currently assigned sound device for playback 
    */
    public String getPlaybackDevice();

    /**
    * Sets the sound device used for playback. 
    *
    * @param devid The device name as returned by linphone_core_get_sound_devices 
    * 
    * @return 0 
    */
    public int setPlaybackDevice(String devid);

    /**
    * Get playback gain in db before entering sound card. 
    *
    * @return The current playback gain 
    */
    public float getPlaybackGainDb();

    /**
    * Allow to control play level before entering sound card: gain in db. 
    *
    * @param level The new play level 
    */
    public void setPlaybackGainDb(float level);

    /**
    * Returns the preferred video framerate, previously set by {@link
    * Core#setPreferredFramerate}. 
    *
    * @return frame rate in number of frames per seconds. 
    */
    public float getPreferredFramerate();

    /**
    * Set the preferred frame rate for video. 
    *
    * Based on the available bandwidth constraints and network conditions, the video
    * encoder remains free to lower the framerate. There is no warranty that the
    * preferred frame rate be the actual framerate. used during a call. Default value
    * is 0, which means "use encoder's default fps value". 
    * 
    * @param fps the target frame rate in number of frames per seconds. 
    */
    public void setPreferredFramerate(float fps);

    /**
    * Get the preferred video definition for the stream that is captured and sent to
    * the remote party. 
    *
    * @return The preferred {@link VideoDefinition} 
    */
    public VideoDefinition getPreferredVideoDefinition();

    /**
    * Set the preferred video definition for the stream that is captured and sent to
    * the remote party. 
    *
    * All standard video definitions are accepted on the receive path. 
    * 
    * @param vdef {@link VideoDefinition} object 
    */
    public void setPreferredVideoDefinition(VideoDefinition vdef);

    /**
    * Sets the preferred video size by its name. 
    *
    * This is identical to linphone_core_set_preferred_video_size except that it
    * takes the name of the video resolution as input. Video resolution names are:
    * qcif, svga, cif, vga, 4cif, svga ...
    * 
    * @deprecated Use {@link Factory#createVideoDefinitionFromName} and {@link
    * Core#setPreferredVideoDefinition} instead 
    */
    @Deprecated
    public void setPreferredVideoSizeByName(String name);

    /**
    * Get my presence model. 
    *
    * @return A {@link PresenceModel} object, or null if no presence model has been
    * set. 
    */
    public PresenceModel getPresenceModel();

    /**
    * Set my presence model. 
    *
    * @param presence {@link PresenceModel} 
    */
    public void setPresenceModel(PresenceModel presence);

    /**
    * Get the definition of the captured video. 
    *
    * @return The captured {@link VideoDefinition} if it was previously set by {@link
    * Core#setPreviewVideoDefinition}, otherwise a 0x0 LinphoneVideoDefinition. 
    * 
    * See also: {@link Core#setPreviewVideoDefinition} 
    */
    public VideoDefinition getPreviewVideoDefinition();

    /**
    * Set the video definition for the captured (preview) video. 
    *
    * This method is for advanced usage where a video capture must be set
    * independently of the definition of the stream actually sent through the call.
    * This allows for example to have the preview window in High Definition even if
    * due to bandwidth constraint the sent video definition is small. Using this
    * feature increases the CPU consumption, since a rescaling will be done
    * internally. 
    * 
    * @param vdef {@link VideoDefinition} object 
    */
    public void setPreviewVideoDefinition(VideoDefinition vdef);

    /**
    * Sets the preview video size by its name. 
    *
    * See linphone_core_set_preview_video_size for more information about this
    * feature.
    * 
    * Video resolution names are: qcif, svga, cif, vga, 4cif, svga ...
    * 
    * @deprecated Use {@link Factory#createVideoDefinitionFromName} and {@link
    * Core#setPreviewVideoDefinition} instead 
    */
    @Deprecated
    public void setPreviewVideoSizeByName(String name);

    /**
    * Returns the default identity when no proxy configuration is used. 
    *
    */
    public String getPrimaryContact();

    /**
    * Sets the local "from" identity. 
    *
    * This data is used in absence of any proxy configuration or when no default
    * proxy configuration is set. See {@link ProxyConfig} 
    */
    public int setPrimaryContact(String contact);

    /**
    * Same as {@link Core#getPrimaryContact} but the result is a {@link Address}
    * object instead of const char *. 
    *
    * @deprecated Use {@link Core#createPrimaryContactParsed} instead. Deprecated
    * since 2018-10-22. 
    */
    @Deprecated
    public Address getPrimaryContactParsed();

    /**
    * Get provisioning URI. 
    *
    * @return the provisioning URI. 
    */
    public String getProvisioningUri();

    /**
    * Set URI where to download xml configuration file at startup. 
    *
    * This can also be set from configuration file or factory config file, from
    * [misc] section, item "config-uri". Calling this function does not load the
    * configuration. It will write the value into configuration so that configuration
    * from remote URI will take place at next {@link Core} start. 
    * 
    * @param uri the http or https uri to use in order to download the configuration.
    * Passing null will disable remote provisioning. 
    * 
    * @return -1 if uri could not be parsed, 0 otherwise. Note that this does not
    * check validity of URI endpoint nor scheme and download may still fail. 
    */
    public int setProvisioningUri(String uri);

    /**
    * Returns an unmodifiable list of entered proxy configurations. 
    *
    * @return A list of {@link ProxyConfig} objects. LinphoneProxyConfig  
    */
    public ProxyConfig[] getProxyConfigList();

    /**
    * Tells whether QRCode is enabled in the preview. 
    *
    * @return A boolean value telling whether QRCode is enabled in the preview. 
    */
    public boolean qrcodeVideoPreviewEnabled();

    /**
    * Controls QRCode enablement. 
    *
    * @param val A boolean value telling whether the QRCode is enabled in the
    * preview. 
    */
    public void enableQrcodeVideoPreview(boolean val);

    /**
    * Gets if realtime text is enabled or not. 
    *
    * @return true if realtime text is enabled, false otherwise 
    */
    public boolean realtimeTextEnabled();

    /**
    * Get the wav file where incoming stream is recorded, when files are used instead
    * of soundcards (see {@link Core#setUseFiles}). 
    *
    * This feature is different from call recording ({@link
    * CallParams#setRecordFile}) The file is a 16 bit linear wav file. 
    * 
    * @return The path to the file where incoming stream is recorded. 
    */
    public String getRecordFile();

    /**
    * Sets a wav file where incoming stream is to be recorded, when files are used
    * instead of soundcards (see {@link Core#setUseFiles}). 
    *
    * This feature is different from call recording ({@link
    * CallParams#setRecordFile}) The file will be a 16 bit linear wav file. 
    * 
    * @param file The path to the file where incoming stream is to be recorded. 
    */
    public void setRecordFile(String file);

    /**
    * Get the ring back tone played to far end during incoming calls. 
    *
    */
    public String getRemoteRingbackTone();

    /**
    * Specify a ring back tone to be played to far end during incoming calls. 
    *
    * @param ring The path to the ring back tone to be played. 
    */
    public void setRemoteRingbackTone(String ring);

    /**
    * Tells whether NACK context is enabled or not. 
    *
    * @return A boolean value telling whether NACK context is enabled or not 
    */
    public boolean retransmissionOnNackEnabled();

    /**
    * Turns NACK context on or off. 
    *
    * @param val A boolean value telling whether to enable NACK context 
    */
    public void enableRetransmissionOnNack(boolean val);

    /**
    * Returns the path to the wav file used for ringing. 
    *
    * @return The path to the wav file used for ringing 
    */
    public String getRing();

    /**
    * Sets the path to a wav file used for ringing. 
    *
    * The file must be a wav 16bit linear. Local ring is disabled if null. 
    * 
    * @param path The path to a wav file to be used for ringing 
    */
    public void setRing(String path);

    /**
    * Tells whether the ring play is enabled during an incoming early media call. 
    *
    */
    public boolean getRingDuringIncomingEarlyMedia();

    /**
    * Enable or disable the ring play during an incoming early media call. 
    *
    * @param enable A boolean value telling whether to enable ringing during an
    * incoming early media call. 
    */
    public void setRingDuringIncomingEarlyMedia(boolean enable);

    /**
    * Returns the path to the wav file used for ringing back. 
    *
    * @return The path to the wav file used for ringing back 
    */
    public String getRingback();

    /**
    * Sets the path to a wav file used for ringing back. 
    *
    * Ringback means the ring that is heard when it's ringing at the remote party.
    * The file must be a wav 16bit linear. 
    * 
    * @param path The path to a wav file to be used for ringing back 
    */
    public void setRingback(String path);

    /**
    * Gets the name of the currently assigned sound device for ringing. 
    *
    * @return The name of the currently assigned sound device for ringing 
    */
    public String getRingerDevice();

    /**
    * Sets the sound device used for ringing. 
    *
    * @param devid The device name as returned by linphone_core_get_sound_devices 
    * 
    * @return 0 
    */
    public int setRingerDevice(String devid);

    /**
    * Gets the path to a file or folder containing the trusted root CAs (PEM format) 
    *
    * @return The path to a file or folder containing the trusted root CAs 
    */
    public String getRootCa();

    /**
    * Sets the path to a file or folder containing trusted root CAs (PEM format) 
    *
    * @param path The path to a file or folder containing trusted root CAs 
    */
    public void setRootCa(String path);

    /**
    * Sets the trusted root CAs (PEM format) 
    *
    * @param data The trusted root CAs as a string 
    */
    public void setRootCaData(String data);

    /**
    * Returns whether RTP bundle mode (also known as Media Multiplexing) is enabled. 
    *
    * See https://tools.ietf.org/html/draft-ietf-mmusic-sdp-bundle-negotiation-54 for
    * more information. 
    * 
    * @return a boolean indicating the enablement of rtp bundle mode. 
    */
    public boolean rtpBundleEnabled();

    /**
    * Enables or disables RTP bundle mode (Media Multiplexing). 
    *
    * See https://tools.ietf.org/html/draft-ietf-mmusic-sdp-bundle-negotiation-54 for
    * more information about the feature. When enabled, liblinphone will try to
    * negociate the use of a single port for all streams when doing an outgoing call.
    * It automatically enables rtcp-mux. This feature can also be enabled per-call
    * using {@link CallParams}. 
    * 
    * @param value a boolean to indicate whether the feature is to be enabled. 
    */
    public void enableRtpBundle(boolean value);

    /**
    * Media offer control param for SIP INVITE. 
    *
    * @return true if INVITE has to be sent whitout SDP. 
    */
    public boolean sdp200AckEnabled();

    /**
    * Control when media offer is sent in SIP INVITE. 
    *
    * @param enable true if INVITE has to be sent whitout SDP. 
    */
    public void enableSdp200Ack(boolean enable);

    /**
    * Tells whether video self view during call is enabled or not. 
    *
    * @return A boolean value telling whether self view is enabled 
    * 
    * See also: {@link Core#enableSelfView} for details. 
    */
    public boolean selfViewEnabled();

    /**
    * Enables or disable self view during calls. 
    *
    * @param val A boolean value telling whether to enable self view Self-view refers
    * to having local webcam image inserted in corner of the video window during
    * calls. This function works at any time, including during calls. 
    */
    public void enableSelfView(boolean val);

    /**
    * Enable whether or not to hide sender name in forward message. 
    *
    * @param enable whether or not to enable the feature 
    */
    public void enableSenderNameHiddenInForwardMessage(boolean enable);

    /**
    * Check if the Session Timers feature is enabled. 
    *
    */
    public boolean getSessionExpiresEnabled();

    /**
    * Enable the Session Timers support. 
    *
    * @param enabled Enable or disable it 
    */
    public void setSessionExpiresEnabled(boolean enabled);

    /**
    * Returns the session expires min value, 90 by default. 
    *
    */
    public int getSessionExpiresMinValue();

    /**
    * Sets the session expires minSE value, forced to a minimum of 90 by default. 
    *
    */
    public void setSessionExpiresMinValue(int min);

    /**
    * Returns the session expires refresher value. 
    *
    */
    public SessionExpiresRefresher getSessionExpiresRefresherValue();

    /**
    * Sets the session expires refresher value. 
    *
    * @param refresher The refresher configuration value 
    */
    public void setSessionExpiresRefresherValue(SessionExpiresRefresher refresher);

    /**
    * Returns the session expires value. 
    *
    */
    public int getSessionExpiresValue();

    /**
    * Sets the session expires value, 0 by default. 
    *
    */
    public void setSessionExpiresValue(int expires);

    /**
    * Get the DSCP field for SIP signaling channel. 
    *
    * The DSCP defines the quality of service in IP packets. 
    * 
    * @return The current DSCP value 
    */
    public int getSipDscp();

    /**
    * Set the DSCP field for SIP signaling channel. 
    *
    * The DSCP defines the quality of service in IP packets. 
    * 
    * @param dscp The DSCP value to set 
    */
    public void setSipDscp(int dscp);

    /**
    * This method is called by the application to notify the linphone core library
    * when the SIP network is reachable. 
    *
    * This is for advanced usage, when SIP and RTP layers are required to use
    * different interfaces. Most applications just need {@link
    * Core#setNetworkReachable}. 
    */
    public void setSipNetworkReachable(boolean value);

    /**
    * Get the SIP transport timeout. 
    *
    * @return The SIP transport timeout in milliseconds. 
    */
    public int getSipTransportTimeout();

    /**
    * Set the SIP transport timeout. 
    *
    * @param timeoutMs The SIP transport timeout in milliseconds. 
    */
    public void setSipTransportTimeout(int timeoutMs);

    /**
    * Gets the list of the available sound devices. 
    *
    * @return A list of char * objects. char *  An unmodifiable array of strings
    * contanining the names of the available sound devices that is null terminated 
    */
    public String[] getSoundDevicesList();

    /**
    * Get the path to the image file streamed when "Static picture" is set as the
    * video device. 
    *
    * @return The path to the image file streamed when "Static picture" is set as the
    * video device. 
    */
    public String getStaticPicture();

    /**
    * Set the path to the image file to stream when "Static picture" is set as the
    * video device. 
    *
    * @param path The path to the image file to use. 
    */
    public int setStaticPicture(String path);

    /**
    * Get the frame rate for static picture. 
    *
    * @return The frame rate used for static picture. 
    */
    public float getStaticPictureFps();

    /**
    * Set the frame rate for static picture. 
    *
    * @param fps The new frame rate to use for static picture. 
    */
    public int setStaticPictureFps(float fps);

    /**
    * Get the STUN server address being used. 
    *
    * @return The STUN server address being used. 
    */
    public String getStunServer();

    /**
    * Set the STUN server address to use when the firewall policy is set to STUN. 
    *
    * @param server The STUN server address to use. 
    */
    public void setStunServer(String server);

    /**
    * Returns a null terminated table of strings containing the file format extension
    * supported for call recording. 
    *
    * @return A list of char * objects. char *  the supported formats, typically
    * 'wav' and 'mkv' 
    */
    public String[] getSupportedFileFormatsList();

    /**
    * Return the list of the available text payload types. 
    *
    * @return A list of {@link PayloadType} objects. LinphonePayloadType  A freshly
    * allocated list of the available payload types. The list must be destroyed with
    * bctbx_list_free() after usage. The elements of the list haven't to be unref. 
    */
    public PayloadType[] getTextPayloadTypes();

    /**
    * Redefine the list of the available payload types. 
    *
    * @param payloadTypes A list of {@link PayloadType} objects. LinphonePayloadType 
    * The new list of payload types. The core does not take ownership on it. 
    */
    public void setTextPayloadTypes(PayloadType[] payloadTypes);

    /**
    * Gets the UDP port used for text streaming. 
    *
    * @return The UDP port used for text streaming 
    */
    public int getTextPort();

    /**
    * Sets the UDP port used for text streaming. 
    *
    * A value if -1 will request the system to allocate the local port randomly. This
    * is recommended in order to avoid firewall warnings. 
    * 
    * @param port The UDP port to use for text streaming 
    */
    public void setTextPort(int port);

    /**
    * Get the text port range from which is randomly chosen the UDP port used for
    * text streaming. 
    *
    * @return a {@link Range} object 
    */
    public Range getTextPortsRange();

    /**
    * Gets the TLS certificate. 
    *
    * @return the TLS certificate or null if not set yet 
    */
    public String getTlsCert();

    /**
    * Sets a TLS certificate used for TLS authentication The certificate won't be
    * stored, you have to set it after each {@link Core} startup. 
    *
    * @param tlsCert the TLS certificate 
    */
    public void setTlsCert(String tlsCert);

    /**
    * Gets the path to the TLS certificate file. 
    *
    * @return the TLS certificate path or null if not set yet 
    */
    public String getTlsCertPath();

    /**
    * Sets a TLS certificate path used for TLS authentication The path will be stored
    * in the rc file and automatically restored on startup. 
    *
    * @param tlsCertPath path to the TLS certificate 
    */
    public void setTlsCertPath(String tlsCertPath);

    /**
    * Gets the TLS key. 
    *
    * @return the TLS key or null if not set yet 
    */
    public String getTlsKey();

    /**
    * Sets a TLS key used for TLS authentication The key won't be stored, you have to
    * set it after each {@link Core} startup. 
    *
    * @param tlsKey the TLS key 
    */
    public void setTlsKey(String tlsKey);

    /**
    * Gets the path to the TLS key file. 
    *
    * @return the TLS key path or null if not set yet 
    */
    public String getTlsKeyPath();

    /**
    * Sets a TLS key path used for TLS authentication The path will be stored in the
    * rc file and automatically restored on startup. 
    *
    * @param tlsKeyPath path to the TLS key 
    */
    public void setTlsKeyPath(String tlsKeyPath);

    /**
    * Retrieves the port configuration used for each transport (udp, tcp, tls). 
    *
    * A zero value port for a given transport means the transport is not used. A
    * value of LC_SIP_TRANSPORT_RANDOM (-1) means the port is to be chosen randomly
    * by the system. 
    * 
    * @return A {@link Transports} structure with the configured ports 
    */
    public Transports getTransports();

    /**
    * Sets the ports to be used for each of transport (UDP or TCP) A zero value port
    * for a given transport means the transport is not used. 
    *
    * A value of LC_SIP_TRANSPORT_RANDOM (-1) means the port is to be choosen
    * randomly by the system. 
    * 
    * @param transports A #LinphoneSipTransports structure giving the ports to use 
    * 
    * @return 0 
    */
    public int setTransports(Transports transports);

    /**
    * Retrieves the real port number assigned for each sip transport (udp, tcp, tls). 
    *
    * A zero value means that the transport is not activated. If
    * LC_SIP_TRANSPORT_RANDOM was passed to linphone_core_set_sip_transports, the
    * random port choosed by the system is returned. 
    * 
    * @return A {@link Transports} structure with the ports being used 
    */
    public Transports getTransportsUsed();

    /**
    * get tunnel instance if available 
    *
    * @return {@link Tunnel} or null if not available 
    */
    public Tunnel getTunnel();

    /**
    * Return the global unread chat message count. 
    *
    * @return The global unread chat message count. 
    */
    public int getUnreadChatMessageCount();

    /**
    * Return the unread chat message count for all active local address. 
    *
    * (Primary contact + proxy configs.) 
    * 
    * @return The unread chat message count. 
    */
    public int getUnreadChatMessageCountFromActiveLocals();

    /**
    * Retrieve the maximum available upload bandwidth. 
    *
    * This value was set by {@link Core#setUploadBandwidth}. 
    */
    public int getUploadBandwidth();

    /**
    * Sets maximum available upload bandwidth This is IP bandwidth, in kbit/s. 
    *
    * This information is used by liblinphone together with remote side available
    * bandwidth signaled in SDP messages to properly configure audio & video codec's
    * output bitrate.
    * 
    * @param bw the bandwidth in kbits/s, 0 for infinite 
    */
    public void setUploadBandwidth(int bw);

    /**
    * Set audio packetization time linphone will send (in absence of requirement from
    * peer) A value of 0 stands for the current codec default packetization time. 
    *
    */
    public int getUploadPtime();

    /**
    * Set audio packetization time linphone will send (in absence of requirement from
    * peer) A value of 0 stands for the current codec default packetization time. 
    *
    */
    public void setUploadPtime(int ptime);

    /**
    * Return the external ip address of router. 
    *
    * In some cases the uPnP can have an external ip address but not a usable uPnP
    * (state different of Ok).
    * 
    * @return a null terminated string containing the external ip address. If the the
    * external ip address is not available return null. 
    */
    public String getUpnpExternalIpaddress();

    /**
    * Return the internal state of uPnP. 
    *
    * @return an LinphoneUpnpState. 
    */
    public UpnpState getUpnpState();

    /**
    * Gets whether linphone is currently streaming audio from and to files, rather
    * than using the soundcard. 
    *
    * @return A boolean value representing whether linphone is streaming audio from
    * and to files or not. 
    */
    public boolean getUseFiles();

    /**
    * Ask the core to stream audio from and to files, instead of using the soundcard. 
    *
    * @param yesno A boolean value asking to stream audio from and to files or not. 
    */
    public void setUseFiles(boolean yesno);

    /**
    * Indicates whether SIP INFO is used to send digits. 
    *
    * @return A boolean value telling whether SIP INFO is used to send digits 
    */
    public boolean getUseInfoForDtmf();

    /**
    * Sets whether SIP INFO is to be used to send digits. 
    *
    * @param useInfo A boolean value telling whether to use SIP INFO to send digits 
    */
    public void setUseInfoForDtmf(boolean useInfo);

    /**
    * Indicates whether RFC2833 is used to send digits. 
    *
    * @return A boolean value telling whether RFC2833 is used to send digits 
    */
    public boolean getUseRfc2833ForDtmf();

    /**
    * Sets whether RFC2833 is to be used to send digits. 
    *
    * @param useRfc2833 A boolean value telling whether to use RFC2833 to send digits 
    */
    public void setUseRfc2833ForDtmf(boolean useRfc2833);

    /**
    *
    * @return liblinphone's user agent as a string. 
    */
    public String getUserAgent();

    /**
    * Get the path to the directory storing the user's certificates. 
    *
    * @return The path to the directory storing the user's certificates. 
    */
    public String getUserCertificatesPath();

    /**
    * Set the path to the directory storing the user's x509 certificates (used by
    * dtls) 
    *
    * @param path The path to the directory to use to store the user's certificates. 
    */
    public void setUserCertificatesPath(String path);

    /**
    * Get the default policy for video. 
    *
    * See {@link Core#setVideoActivationPolicy} for more details. 
    * 
    * @return The video policy being used 
    */
    public VideoActivationPolicy getVideoActivationPolicy();

    /**
    * Sets the default policy for video. 
    *
    * This policy defines whether:
    * 
    */
    public void setVideoActivationPolicy(VideoActivationPolicy policy);

    /**
    * Tells whether the video adaptive jitter compensation is enabled. 
    *
    * @return  if the video adaptive jitter compensation is enabled, true otherwise. 
    */
    public boolean videoAdaptiveJittcompEnabled();

    /**
    * Enable or disable the video adaptive jitter compensation. 
    *
    * @param enable  to enable the video adaptive jitter compensation, true to
    * disable it. 
    */
    public void enableVideoAdaptiveJittcomp(boolean enable);

    /**
    * Tells whether video capture is enabled. 
    *
    * @return  if video capture is enabled, true if disabled. 
    */
    public boolean videoCaptureEnabled();

    /**
    * Enable or disable video capture. 
    *
    * This function does not have any effect during calls. It just indicates the
    * {@link Core} to initiate future calls with video capture or not. 
    * 
    * @param enable  to enable video capture, true to disable it. 
    */
    public void enableVideoCapture(boolean enable);

    /**
    * Returns the name of the currently active video device. 
    *
    * @return The name of the currently active video device 
    */
    public String getVideoDevice();

    /**
    * Sets the active video device. 
    *
    * @param id The name of the video device to use as returned by
    * linphone_core_get_video_devices 
    */
    public int setVideoDevice(String id);

    /**
    * Gets the list of the available video capture devices. 
    *
    * @return A list of char * objects. char *  An unmodifiable array of strings
    * contanining the names of the available video capture devices that is null
    * terminated 
    */
    public String[] getVideoDevicesList();

    /**
    * Tells whether video display is enabled. 
    *
    * @return  if video display is enabled, true if disabled. 
    */
    public boolean videoDisplayEnabled();

    /**
    * Enable or disable video display. 
    *
    * This function does not have any effect during calls. It just indicates the
    * {@link Core} to initiate future calls with video display or not. 
    * 
    * @param enable  to enable video display, true to disable it. 
    */
    public void enableVideoDisplay(boolean enable);

    /**
    * Get the name of the mediastreamer2 filter used for rendering video. 
    *
    * @return The currently selected video display filter 
    */
    public String getVideoDisplayFilter();

    /**
    * Set the name of the mediastreamer2 filter to be used for rendering video. 
    *
    * This is for advanced users of the library, mainly to workaround hardware/driver
    * bugs. 
    */
    public void setVideoDisplayFilter(String filtername);

    /**
    * Get the DSCP field for outgoing video streams. 
    *
    * The DSCP defines the quality of service in IP packets. 
    * 
    * @return The current DSCP value 
    */
    public int getVideoDscp();

    /**
    * Set the DSCP field for outgoing video streams. 
    *
    * The DSCP defines the quality of service in IP packets. 
    * 
    * @param dscp The DSCP value to set 
    */
    public void setVideoDscp(int dscp);

    /**
    * Returns true if either capture or display is enabled, true otherwise. 
    *
    * same as ( linphone_core_video_capture_enabled |
    * linphone_core_video_display_enabled ) 
    */
    public boolean videoEnabled();

    /**
    * Returns the nominal video jitter buffer size in milliseconds. 
    *
    * @return The nominal video jitter buffer size in milliseconds 
    */
    public int getVideoJittcomp();

    /**
    * Sets the nominal video jitter buffer size in milliseconds. 
    *
    * The value takes effect immediately for all running and pending calls, if any. A
    * value of 0 disables the jitter buffer. 
    */
    public void setVideoJittcomp(int milliseconds);

    /**
    * Use to get multicast address to be used for video stream. 
    *
    * @return an ipv4/6 multicast address, or default value 
    */
    public String getVideoMulticastAddr();

    /**
    * Use to set multicast address to be used for video stream. 
    *
    * @param ip an ipv4/6 multicast address 
    * 
    * @return 0 in case of success 
    */
    public int setVideoMulticastAddr(String ip);

    /**
    * Use to get multicast state of video stream. 
    *
    * @return true if subsequent calls will propose multicast ip set by
    * linphone_core_set_video_multicast_addr 
    */
    public boolean videoMulticastEnabled();

    /**
    * Use to enable multicast rtp for video stream. 
    *
    * If enabled, outgoing calls put a multicast address from
    * linphone_core_get_video_multicast_addr into video cline. In case of outgoing
    * call video stream is sent to this multicast address. For incoming calls
    * behavior is unchanged. 
    * 
    * @param yesno if yes, subsequent outgoing calls will propose multicast ip set by
    * linphone_core_set_video_multicast_addr 
    */
    public void enableVideoMulticast(boolean yesno);

    /**
    * Use to get multicast ttl to be used for video stream. 
    *
    * @return a time to leave value 
    */
    public int getVideoMulticastTtl();

    /**
    * Use to set multicast ttl to be used for video stream. 
    *
    * @param ttl value or -1 if not used. [0..255] default value is 1 
    * 
    * @return 0 in case of success 
    */
    public int setVideoMulticastTtl(int ttl);

    /**
    * Return the list of the available video payload types. 
    *
    * @return A list of {@link PayloadType} objects. LinphonePayloadType  A freshly
    * allocated list of the available payload types. The list must be destroyed with
    * bctbx_list_free() after usage. The elements of the list haven't to be unref. 
    */
    public PayloadType[] getVideoPayloadTypes();

    /**
    * Redefine the list of the available video payload types. 
    *
    * @param payloadTypes A list of {@link PayloadType} objects. LinphonePayloadType 
    * The new list of codecs. The core does not take ownership on it. 
    */
    public void setVideoPayloadTypes(PayloadType[] payloadTypes);

    /**
    * Gets the UDP port used for video streaming. 
    *
    * @return The UDP port used for video streaming 
    */
    public int getVideoPort();

    /**
    * Sets the UDP port used for video streaming. 
    *
    * A value of -1 will request the system to allocate the local port randomly. This
    * is recommended in order to avoid firewall warnings. 
    * 
    * @param port The UDP port to use for video streaming 
    */
    public void setVideoPort(int port);

    /**
    * Get the video port range from which is randomly chosen the UDP port used for
    * video streaming. 
    *
    * @return a {@link Range} object 
    */
    public Range getVideoPortsRange();

    /**
    * Get the video preset used for video calls. 
    *
    * @return The name of the video preset used for video calls (can be null if the
    * default video preset is used). 
    */
    public String getVideoPreset();

    /**
    * Set the video preset to be used for video calls. 
    *
    * @param preset The name of the video preset to be used (can be null to use the
    * default video preset). 
    */
    public void setVideoPreset(String preset);

    /**
    * Tells whether video preview is enabled. 
    *
    * @return A boolean value telling whether video preview is enabled 
    */
    public boolean videoPreviewEnabled();

    /**
    * Controls video preview enablement. 
    *
    * @param val A boolean value telling whether the video preview is to be shown
    * Video preview refers to the action of displaying the local webcam image to the
    * user while not in call. 
    */
    public void enableVideoPreview(boolean val);

    /**
    * Enable or disable video source reuse when switching from preview to actual
    * video call. 
    *
    * This source reuse is useful when you always display the preview, even before
    * calls are initiated. By keeping the video source for the transition to a real
    * video call, you will smooth out the source close/reopen cycle.
    * 
    * This function does not have any effect durfing calls. It just indicates the
    * {@link Core} to initiate future calls with video source reuse or not. Also, at
    * the end of a video call, the source will be closed whatsoever for now. 
    * 
    * @param enable  to enable video source reuse. true to disable it for subsequent
    * calls. 
    */
    public void enableVideoSourceReuse(boolean enable);

    /**
    * Tells whether Wifi only mode is enabled or not. 
    *
    * @return A boolean value telling whether Wifi only mode is enabled or not 
    */
    public boolean wifiOnlyEnabled();

    /**
    * Turns Wifi only mode on or off. 
    *
    * @param val A boolean value telling whether to enable IPv6 support 
    */
    public void enableWifiOnly(boolean val);

    /**
    * Get the path to the file storing the zrtp secrets cache. 
    *
    * @return The path to the file storing the zrtp secrets cache. 
    */
    public String getZrtpSecretsFile();

    /**
    * Set the path to the file storing the zrtp secrets cache. 
    *
    * @param file The path to the file to use to store the zrtp secrets cache. 
    */
    public void setZrtpSecretsFile(String file);

    /**
    * Accept an incoming call. 
    *
    * Basically the application is notified of incoming calls within the
    * call_state_changed callback of the LinphoneCoreVTable structure, where it will
    * receive a LinphoneCallIncoming event with the associated {@link Call} object.
    * The application can later accept the call using this method. 
    * 
    * @param call The {@link Call} object representing the call to be answered 
    * 
    * @return 0 on success, -1 on failure
    * 
    * @deprecated Use {@link Call#accept} instead. Deprecated since 2017-02-13. 
    */
    @Deprecated
    public int acceptCall(Call call);

    /**
    * Accept call modifications initiated by other end. 
    *
    * This call may be performed in response to a #LinphoneCallUpdatedByRemote state
    * notification. When such notification arrives, the application can decide to
    * call linphone_core_defer_update_call() so that it can have the time to prompt
    * the user. {@link Call#getRemoteParams} can be used to get information about the
    * call parameters requested by the other party, such as whether a video stream is
    * requested.
    * 
    * When the user accepts or refuse the change, {@link Core#acceptCallUpdate} can
    * be done to answer to the other party. If params is null, then the same call
    * parameters established before the update request will continue to be used (no
    * change). If params is not null, then the update will be accepted according to
    * the parameters passed. Typical example is when a user accepts to start video,
    * then params should indicate that video stream should be used (see {@link
    * CallParams#enableVideo}). 
    * 
    * @param params A {@link CallParams} object describing the call parameters to
    * accept 
    * 
    * @return 0 if successful, -1 otherwise (actually when this function call is
    * performed outside ot #LinphoneCallUpdatedByRemote state)
    * 
    * @deprecated Use {@link Call#acceptUpdate} instead. Deprecated since 2017-02-13. 
    */
    @Deprecated
    public int acceptCallUpdate(Call call, CallParams params);

    /**
    * Accept an incoming call, with parameters. 
    *
    * Basically the application is notified of incoming calls within the
    * call_state_changed callback of the LinphoneCoreVTable structure, where it will
    * receive a LinphoneCallIncoming event with the associated {@link Call} object.
    * The application can later accept the call using this method. 
    * 
    * @param params The specific parameters for this call, for example whether video
    * is accepted or not. Use null to use default parameters 
    * 
    * @return 0 on success, -1 on failure
    * 
    * @deprecated Use {@link Call#acceptWithParams} instead. Deprecated since
    * 2017-02-13. 
    */
    @Deprecated
    public int acceptCallWithParams(Call call, CallParams params);

    /**
    * Accept an early media session for an incoming call. 
    *
    * This is identical as calling {@link Core#acceptEarlyMediaWithParams} with null
    * call parameters. 
    * 
    * @param call The incoming call to accept 
    * 
    * @return 0 if successful, -1 otherwise
    * 
    * See also: {@link Core#acceptEarlyMediaWithParams} 
    * 
    * @deprecated Use {@link Call#acceptEarlyMedia} instead. Deprecated since
    * 2017-02-13. 
    */
    @Deprecated
    public int acceptEarlyMedia(Call call);

    /**
    * When receiving an incoming, accept to start a media session as early-media. 
    *
    * This means the call is not accepted but audio & video streams can be
    * established if the remote party supports early media. However, unlike after
    * call acceptance, mic and camera input are not sent during early-media, though
    * received audio & video are played normally. The call can then later be fully
    * accepted using {@link Core#acceptCall} or {@link Core#acceptCallWithParams}. 
    * 
    * @param params The call parameters to use (can be null) 
    * 
    * @return 0 if successful, -1 otherwise
    * 
    * @deprecated Use {@link Call#acceptEarlyMediaWithParams} instead. 
    */
    @Deprecated
    public int acceptEarlyMediaWithParams(Call call, CallParams params);

    /**
    * Add all current calls into the conference. 
    *
    * If no conference is running a new internal conference context is created and
    * all current calls are added to it. 
    * 
    * @return 0 if succeeded. Negative number if failed 
    */
    public int addAllToConference();

    /**
    * Adds authentication information to the {@link Core}. 
    *
    * That piece of information will be used during all SIP transactions that require
    * authentication. 
    * 
    * @param info The {@link AuthInfo} to add. 
    */
    public void addAuthInfo(AuthInfo info);

    /**
    * Add a friend to the current buddy list, if subscription attribute  is set, a
    * SIP SUBSCRIBE message is sent. 
    *
    * @param fr {@link Friend} to add 
    * 
    * @deprecated use {@link FriendList#addFriend} instead. 
    */
    @Deprecated
    public void addFriend(Friend fr);

    /**
    * Add a friend list. 
    *
    * @param list {@link FriendList} object 
    */
    public void addFriendList(FriendList list);

    /**
    * Add the given linphone specs to the list of functionalities the linphone client
    * supports. 
    *
    * @param spec The spec to add 
    */
    public void addLinphoneSpec(String spec);

    /**
    * Add a proxy configuration. 
    *
    * This will start registration on the proxy, if registration is enabled. 
    */
    public int addProxyConfig(ProxyConfig config);

    /**
    * This function controls signaling features supported by the core. 
    *
    * They are typically included in a SIP Supported header. 
    * 
    * @param tag The feature tag name 
    */
    public void addSupportedTag(String tag);

    /**
    * Add a participant to the conference. 
    *
    * If no conference is going on a new internal conference context is created and
    * the participant is added to it. 
    * 
    * @param call The current call with the participant to add 
    * 
    * @return 0 if succeeded. Negative number if failed 
    */
    public int addToConference(Call call);

    /**
    * Special function to indicate if the audio session is activated. 
    *
    * Must be called when ProviderDelegate of the callkit notifies that the audio
    * session is activated or deactivated. 
    */
    public void audioSessionActivated(boolean actived);

    /**
    * Special function to indicate if the callkit is enabled, False by default. 
    *
    */
    public void callkitEnabled(boolean enabled);

    /**
    * Checks if a new version of the application is available. 
    *
    * @param currentVersion The current version of the application 
    */
    public void checkForUpdate(String currentVersion);

    /**
    * Clear all authentication information. 
    *
    */
    public void clearAllAuthInfo();

    /**
    * Erase the call log. 
    *
    */
    public void clearCallLogs();

    /**
    * Erase all proxies from config. 
    *
    */
    public void clearProxyConfig();

    /**
    * Create a {@link AccountCreator} and set Linphone Request callbacks. 
    *
    * @param xmlrpcUrl The URL to the XML-RPC server. Must be NON null. 
    * 
    * @return The new {@link AccountCreator} object. 
    */
    public AccountCreator createAccountCreator(String xmlrpcUrl);

    /**
    * Create a {@link Address} object by parsing the user supplied address, given as
    * a string. 
    *
    * @param address String containing the user supplied address 
    * 
    * @return The create {@link Address} object 
    */
    public Address createAddress(String address);

    /**
    * Create an authentication information with default values from Linphone core. 
    *
    * @param domain String containing the SIP domain for which this authentication
    * information is valid, if it has to be restricted for a single SIP domain. 
    * 
    * @return {@link AuthInfo} with default values set
    * 
    * @deprecated use {@link Factory#createAuthInfo} instead. 
    */
    @Deprecated
    public AuthInfo createAuthInfo(String username, String userid, String passwd, String ha1, String realm, String domain);

    /**
    * Creates a fake LinphoneCallLog. 
    *
    * @param quality call quality 
    * 
    * @return LinphoneCallLog object 
    */
    public CallLog createCallLog(Address from, Address to, Call.Dir dir, int duration, long startTime, long connectedTime, Call.Status status, boolean videoEnabled, float quality);

    /**
    * Create a {@link CallParams} suitable for {@link Core#inviteWithParams}, {@link
    * Core#acceptCallWithParams}, {@link Core#acceptEarlyMediaWithParams}, {@link
    * Core#acceptCallUpdate}. 
    *
    * The parameters are initialized according to the current {@link Core}
    * configuration and the current state of the LinphoneCall. 
    * 
    * @param call {@link Call} for which the parameters are to be build, or null in
    * the case where the parameters are to be used for a new outgoing call.    
    * 
    * @return A new {@link CallParams} object 
    */
    public CallParams createCallParams(@Nullable Call call);

    /**
    * Create a chat room. 
    *
    * @param participants A list of {@link Address} objects. LinphoneAddress  The
    * initial list of participants of the chat room 
    * 
    * @return The newly created chat room. 
    */
    public ChatRoom createChatRoom(ChatRoomParams params, String subject, Address[] participants);

    /**
    *
    * @param participants A list of {@link Address} objects. LinphoneAddress  The
    * initial list of participants of the chat room 
    * 
    * @return The newly created chat room. 
    */
    public ChatRoom createChatRoom(String subject, Address[] participants);

    /**
    *
    * @param participant {@link Address} representing the initial participant to add
    * to the chat room 
    * 
    * @return The newly created chat room. 
    */
    public ChatRoom createChatRoom(ChatRoomParams params, Address localAddr, Address participant);

    /**
    *
    * @param participant {@link Address} representing the initial participant to add
    * to the chat room 
    * 
    * @return The newly created chat room. 
    */
    public ChatRoom createChatRoom(Address participant);

    /**
    * Create a chat room. 
    *
    * @param participants A list of {@link Address} objects. LinphoneAddress  The
    * initial list of participants of the chat room 
    * 
    * @return The newly created chat room. 
    */
    public ChatRoom createChatRoom(ChatRoomParams params, Address localAddr, String subject, Address[] participants);

    /**
    * Create a client-side group chat room. 
    *
    * When calling this function the chat room is only created at the client-side and
    * is empty. You need to call {@link ChatRoom#addParticipants} to create at the
    * server side and add participants to it. Also, the created chat room will not be
    * a one-to-one chat room even if {@link ChatRoom#addParticipants} is called with
    * only one participant.
    * 
    * @param fallback Boolean value telling whether we should plan on being able to
    * fallback to a basic chat room if the client-side group chat room creation fails 
    * 
    * @return The newly created client-side group chat room. 
    */
    public ChatRoom createClientGroupChatRoom(String subject, boolean fallback);

    /**
    * Create a client-side group chat room. 
    *
    * When calling this function the chat room is only created at the client-side and
    * is empty. You need to call {@link ChatRoom#addParticipants} to create at the
    * server side and add participants to it. Also, the created chat room will not be
    * a one-to-one chat room even if {@link ChatRoom#addParticipants} is called with
    * only one participant.
    * 
    * @param encrypted Boolean value telling whether we should apply encryption or
    * not on chat messages sent and received on this room. 
    * 
    * @return The newly created client-side group chat room. 
    */
    public ChatRoom createClientGroupChatRoom(String subject, boolean fallback, boolean encrypted);

    /**
    * Create some default conference parameters for instanciating a a conference with
    * {@link Core#createConferenceWithParams}. 
    *
    * @return conference parameters. 
    */
    public ConferenceParams createConferenceParams();

    /**
    * Create a conference. 
    *
    * @param params Parameters of the conference. See {@link ConferenceParams}. 
    * 
    * @return A pointer on the freshly created conference. That object will be
    * automatically freed by the core after calling {@link Core#terminateConference}. 
    */
    public Conference createConferenceWithParams(ConferenceParams params);

    /**
    * Create a {@link Config} object from a user config file. 
    *
    * @param filename The filename of the config file to read to fill the
    * instantiated {@link Config} 
    */
    public Config createConfig(String filename);

    /**
    * Create a content with default values from Linphone core. 
    *
    * @return {@link Content} object with default values set 
    */
    public Content createContent();

    /**
    * Creates and returns the default chat room parameters. 
    *
    * @return LinphoneChatRoomParams 
    */
    public ChatRoomParams createDefaultChatRoomParams();

    /**
    * Create a default LinphoneFriend. 
    *
    * @return The created {@link Friend} object 
    */
    public Friend createFriend();

    /**
    * Create a new empty {@link FriendList} object. 
    *
    * @return A new {@link FriendList} object. 
    */
    public FriendList createFriendList();

    /**
    * Create a {@link Friend} from the given address. 
    *
    * @param address A string containing the address to create the {@link Friend}
    * from 
    * 
    * @return The created {@link Friend} object 
    */
    public Friend createFriendWithAddress(String address);

    /**
    * Creates an empty info message. 
    *
    * @return a new LinphoneInfoMessage.
    * 
    * The info message can later be filled with information using {@link
    * InfoMessage#addHeader} or {@link InfoMessage#setContent}, and finally sent with
    * linphone_core_send_info_message(). 
    */
    public InfoMessage createInfoMessage();

    /**
    * Create an independent media file player. 
    *
    * This player support WAVE and MATROSKA formats. 
    * 
    * @param windowId Id of the drawing window. Depend of video out 
    * 
    * @return A pointer on the new instance. null if faild. 
    */
    public Player createLocalPlayer(String soundCardName, String videoDisplayName, Object windowId);

    /**
    * Create a {@link MagicSearch} object. 
    *
    * @return The create {@link MagicSearch} object 
    */
    public MagicSearch createMagicSearch();

    /**
    * Create a new {@link NatPolicy} object with every policies being disabled. 
    *
    * @return A new {@link NatPolicy} object. 
    */
    public NatPolicy createNatPolicy();

    /**
    * Create a new {@link NatPolicy} by reading the config of a {@link Core}
    * according to the passed ref. 
    *
    * @param ref The reference of a NAT policy in the config of the {@link Core} 
    * 
    * @return A new {@link NatPolicy} object. 
    */
    public NatPolicy createNatPolicyFromConfig(String ref);

    /**
    * Create an out-of-dialog notification, specifying the destination resource, the
    * event name. 
    *
    * The notification can be send with {@link Event#notify}. 
    * 
    * @param event the event name 
    * 
    * @return a {@link Event} holding the context of the notification. 
    */
    public Event createNotify(Address resource, String event);

    /**
    * Create a publish context for a one-shot publish. 
    *
    * After being created, the publish must be sent using {@link Event#sendPublish}.
    * The {@link Event} is automatically terminated when the publish transaction is
    * finished, either with success or failure. The application must not call {@link
    * Event#terminate} for such one-shot publish. 
    * 
    * @param event the event name 
    * 
    * @return the {@link Event} holding the context of the publish. 
    */
    public Event createOneShotPublish(Address resource, String event);

    /**
    * Create a {@link PresenceActivity} with the given type and description. 
    *
    * @param description An additional description of the activity to set for the
    * activity. Can be null if no additional description is to be added. 
    * 
    * @return The created {@link PresenceActivity} object. 
    */
    public PresenceActivity createPresenceActivity(PresenceActivity.Type acttype, String description);

    /**
    * Create a default LinphonePresenceModel. 
    *
    * @return The created {@link PresenceModel} object. 
    */
    public PresenceModel createPresenceModel();

    /**
    * Create a {@link PresenceModel} with the given activity type and activity
    * description. 
    *
    * @param description An additional description of the activity to set for the
    * activity. Can be null if no additional description is to be added. 
    * 
    * @return The created {@link PresenceModel} object. 
    */
    public PresenceModel createPresenceModelWithActivity(PresenceActivity.Type acttype, String description);

    /**
    * Create a {@link PresenceModel} with the given activity type, activity
    * description, note content and note language. 
    *
    * @param lang The language of the note to be added to the created model. 
    * 
    * @return The created {@link PresenceModel} object. 
    */
    public PresenceModel createPresenceModelWithActivityAndNote(PresenceActivity.Type acttype, String description, String note, String lang);

    /**
    * Create a {@link PresenceNote} with the given content and language. 
    *
    * @param lang The language of the note to be created. 
    * 
    * @return The created {@link PresenceNote} object. 
    */
    public PresenceNote createPresenceNote(String content, String lang);

    /**
    * Create a {@link PresencePerson} with the given id. 
    *
    * @param id The id of the person to be created. 
    * 
    * @return The created {@link PresencePerson} object. 
    */
    public PresencePerson createPresencePerson(String id);

    /**
    * Create a {@link PresenceService} with the given id, basic status and contact. 
    *
    * @param contact A string containing a contact information corresponding to the
    * service to be created. 
    * 
    * @return The created {@link PresenceService} object. 
    */
    public PresenceService createPresenceService(String id, PresenceBasicStatus basicStatus, String contact);

    /**
    * Same as {@link Core#getPrimaryContact} but the result is a {@link Address}
    * object instead of const char *. 
    *
    */
    public Address createPrimaryContactParsed();

    /**
    * Create a proxy config with default values from Linphone core. 
    *
    * @return {@link ProxyConfig} with default values set 
    */
    public ProxyConfig createProxyConfig();

    /**
    * Create a publish context for an event state. 
    *
    * After being created, the publish must be sent using {@link Event#sendPublish}.
    * After expiry, the publication is refreshed unless it is terminated before. 
    * 
    * @param expires the lifetime of event being published, -1 if no associated
    * duration, in which case it will not be refreshed. 
    * 
    * @return the {@link Event} holding the context of the publish. 
    */
    public Event createPublish(Address resource, String event, int expires);

    /**
    * Create an outgoing subscription, specifying the destination resource, the event
    * name, and an optional content body. 
    *
    * If accepted, the subscription runs for a finite period, but is automatically
    * renewed if not terminated before. Unlike {@link Core#subscribe} the
    * subscription isn't sent immediately. It will be send when calling {@link
    * Event#sendSubscribe}. 
    * 
    * @param expires the whished duration of the subscription 
    * 
    * @return a {@link Event} holding the context of the created subcription. 
    */
    public Event createSubscribe(Address resource, ProxyConfig proxy, String event, int expires);

    /**
    * Create an outgoing subscription, specifying the destination resource, the event
    * name, and an optional content body. 
    *
    * If accepted, the subscription runs for a finite period, but is automatically
    * renewed if not terminated before. Unlike {@link Core#subscribe} the
    * subscription isn't sent immediately. It will be send when calling {@link
    * Event#sendSubscribe}. 
    * 
    * @param expires the whished duration of the subscription 
    * 
    * @return a {@link Event} holding the context of the created subcription. 
    */
    public Event createSubscribe(Address resource, String event, int expires);

    /**
    * Create a {@link XmlRpcSession} for a given url. 
    *
    * @param url The URL to the XML-RPC server. Must be NON null. 
    * 
    * @return The new {@link XmlRpcSession} object. 
    */
    public XmlRpcSession createXmlRpcSession(String url);

    /**
    * Decline a pending incoming call, with a reason. 
    *
    * @param reason The reason for rejecting the call: LinphoneReasonDeclined or
    * LinphoneReasonBusy 
    * 
    * @return 0 on success, -1 on failure
    * 
    * @deprecated Use {@link Call#decline} instead. Deprecated since 2017-02-13. 
    */
    @Deprecated
    public int declineCall(Call call, Reason reason);

    /**
    * When receiving a #LinphoneCallUpdatedByRemote state notification, prevent
    * {@link Core} from performing an automatic answer. 
    *
    * When receiving a #LinphoneCallUpdatedByRemote state notification (ie an
    * incoming reINVITE), the default behaviour of {@link Core} is defined by the
    * "defer_update_default" option of the "sip" section of the config. If this
    * option is 0 (the default) then the {@link Core} automatically answers the
    * reINIVTE with call parameters unchanged. However when for example when the
    * remote party updated the call to propose a video stream, it can be useful to
    * prompt the user before answering. This can be achieved by calling {@link
    * Core#deferCallUpdate} during the call state notification, to deactivate the
    * automatic answer that would just confirm the audio but reject the video. Then,
    * when the user responds to dialog prompt, it becomes possible to call {@link
    * Core#acceptCallUpdate} to answer the reINVITE, with eventually video enabled in
    * the {@link CallParams} argument.
    * 
    * The #LinphoneCallUpdatedByRemote notification can also arrive when receiving an
    * INVITE without SDP. In such case, an unchanged offer is made in the 200Ok, and
    * when the ACK containing the SDP answer is received,
    * #LinphoneCallUpdatedByRemote is triggered to notify the application of possible
    * changes in the media session. However in such case defering the update has no
    * meaning since we just generating an offer.
    * 
    * @param call The call for which to defer the update 
    * 
    * @return 0 if successful, -1 if the {@link Core#deferCallUpdate} was done
    * outside a valid #LinphoneCallUpdatedByRemote notification
    * 
    * @deprecated Use {@link Call#deferUpdate} instead 
    */
    @Deprecated
    public int deferCallUpdate(Call call);

    /**
    * Removes a chatroom including all message history from the LinphoneCore. 
    *
    * @param cr A {@link ChatRoom} object 
    */
    public void deleteChatRoom(ChatRoom cr);

    /**
    * Inconditionnaly disable incoming chat messages. 
    *
    * @param denyReason the deny reason (LinphoneReasonNone has no effect). 
    */
    public void disableChat(Reason denyReason);

    /**
    * Enable reception of incoming chat messages. 
    *
    * By default it is enabled but it can be disabled with {@link Core#disableChat}. 
    */
    public void enableChat();

    /**
    * Tells to {@link Core} to use Linphone Instant Messaging encryption. 
    *
    * @param val The new lime state
    * 
    * @deprecated Use linphone_core_enable_lime_x3dh instead. Depreacted since
    * 2019-02-04 
    */
    @Deprecated
    public void enableLime(LimeState val);

    /**
    * Call this method when you receive a push notification. 
    *
    * It will ensure the proxy configs are correctly registered to the proxy server,
    * so the call or the message will be correctly delivered. 
    */
    public void ensureRegistered();

    /**
    * This method is called by the application to notify the linphone core library
    * when it enters background mode. 
    *
    */
    public void enterBackground();

    /**
    * Join the local participant to the running conference. 
    *
    * @return 0 if succeeded. Negative number if failed 
    */
    public int enterConference();

    /**
    * This method is called by the application to notify the linphone core library
    * when it enters foreground mode. 
    *
    */
    public void enterForeground();

    /**
    * Returns whether a specific file format is supported. 
    *
    * See also: linphone_core_get_supported_file_formats 
    * 
    * @param fmt The format extension (wav, mkv). 
    */
    public boolean fileFormatSupported(String fmt);

    /**
    * Find authentication info matching realm, username, domain criteria. 
    *
    * First of all, (realm,username) pair are searched. If multiple results (which
    * should not happen because realm are supposed to be unique), then domain is
    * added to the search. 
    * 
    * @param sipDomain the SIP domain name (optional) 
    * 
    * @return a {@link AuthInfo} 
    */
    public AuthInfo findAuthInfo(String realm, String username, String sipDomain);

    /**
    * Search from the list of current calls if a remote address match uri. 
    *
    * @param uri which should match call remote uri 
    * 
    * @return {@link Call} or null is no match is found 
    */
    public Call findCallFromUri(String uri);

    /**
    * Get the call log matching the call id, or null if can't be found. 
    *
    * @param callId Call id of the call log to find 
    * 
    * @return {LinphoneCallLog} 
    */
    public CallLog findCallLogFromCallId(String callId);

    /**
    * Find a chat room. 
    *
    * No reference is transfered to the application. The {@link Core} keeps a
    * reference on the chat room. 
    * 
    * @param localAddr a linphone address. 
    * 
    * @return {@link ChatRoom} where messaging can take place. 
    */
    public ChatRoom findChatRoom(Address peerAddr, Address localAddr);

    /**
    * Retrieves a list of {@link Address} sort and filter. 
    *
    * @param sipOnly Only sip address or not 
    * 
    * @return A list of {@link Address} objects. LinphoneAddress  a list of filtered
    * {@link Address} + the {@link Address} created with the filter 
    */
    public Address[] findContactsByChar(String filter, boolean sipOnly);

    /**
    * Search a {@link Friend} by its address. 
    *
    * @param addr The address to use to search the friend. 
    * 
    * @return The {@link Friend} object corresponding to the given address. 
    */
    public Friend findFriend(Address addr);

    /**
    * Search all {@link Friend} matching an address. 
    *
    * @param addr The address to use to search the friends. 
    * 
    * @return A list of {@link Friend} objects. LinphoneFriend  a list of {@link
    * Friend} corresponding to the given address. 
    */
    public Friend[] findFriends(Address addr);

    /**
    * Find a one to one chat room. 
    *
    * No reference is transfered to the application. The {@link Core} keeps a
    * reference on the chat room. 
    * 
    * @param participantAddr a linphone address. 
    * 
    * @return {@link ChatRoom} where messaging can take place. 
    * 
    * @deprecated Use linphone_core_find_one_to_one_chat_room_2 instead 
    */
    @Deprecated
    public ChatRoom findOneToOneChatRoom(Address localAddr, Address participantAddr);

    /**
    * Find a one to one chat room. 
    *
    * No reference is transfered to the application. The {@link Core} keeps a
    * reference on the chat room. 
    * 
    * @param encrypted whether to look for an encrypted chat room or not 
    * 
    * @return {@link ChatRoom} where messaging can take place. 
    */
    public ChatRoom findOneToOneChatRoom(Address localAddr, Address participantAddr, boolean encrypted);

    /**
    * Get the call with the remote_address specified. 
    *
    * @param remoteAddress The remote address of the call that we want to get 
    * 
    * @return The call if it has been found, null otherwise    
    */
    @Nullable
    public Call getCallByRemoteAddress(String remoteAddress);

    /**
    * Get the call with the remote_address specified. 
    *
    * @param remoteAddress 
    * 
    * @return the {@link Call} of the call if found 
    */
    public Call getCallByRemoteAddress2(Address remoteAddress);

    /**
    * Get the list of call logs (past calls). 
    *
    * At the contrary of linphone_core_get_call_logs, it is your responsibility to
    * unref the logs and free this list once you are done using it. 
    * 
    * @param peerAddr A {@link Address} object. 
    * 
    * @return A list of {@link CallLog} objects. LinphoneCallLog  The objects inside
    * the list are freshly allocated with a reference counter equal to one, so they
    * need to be freed on list destruction with bctbx_list_free_with_data() for
    * instance.   
    */
    public CallLog[] getCallHistory(Address peerAddr, Address localAddr);

    /**
    * Get the list of call logs (past calls) that matches the given {@link Address}. 
    *
    * At the contrary of linphone_core_get_call_logs, it is your responsibility to
    * unref the logs and free this list once you are done using it. 
    * 
    * @param addr {@link Address} object 
    * 
    * @return A list of {@link CallLog} objects. LinphoneCallLog  The objects inside
    * the list are freshly allocated with a reference counter equal to one, so they
    * need to be freed on list destruction with bctbx_list_free_with_data() for
    * instance.   
    * 
    * @deprecated Use linphone_core_get_call_history_2 instead. Deprecated since
    * 2018-10-29. 
    */
    @Deprecated
    public CallLog[] getCallHistoryForAddress(Address addr);

    /**
    * Get a basic chat room. 
    *
    * If it does not exist yet, it will be created. No reference is transfered to the
    * application. The {@link Core} keeps a reference on the chat room. 
    * 
    * @param localAddr a linphone address. 
    * 
    * @return {@link ChatRoom} where messaging can take place. 
    */
    public ChatRoom getChatRoom(Address peerAddr, Address localAddr);

    /**
    * Get a basic chat room whose peer is the supplied address. 
    *
    * If it does not exist yet, it will be created. No reference is transfered to the
    * application. The {@link Core} keeps a reference on the chat room. 
    * 
    * @param addr a linphone address. 
    * 
    * @return {@link ChatRoom} where messaging can take place. 
    */
    public ChatRoom getChatRoom(Address addr);

    /**
    * Get a basic chat room for messaging from a sip uri like
    * sip:joe@sip.linphone.org. 
    *
    * If it does not exist yet, it will be created. No reference is transfered to the
    * application. The {@link Core} keeps a reference on the chat room. 
    * 
    * @param to The destination address for messages. 
    * 
    * @return {@link ChatRoom} where messaging can take place. 
    */
    public ChatRoom getChatRoomFromUri(String to);

    /**
    * Search a {@link Friend} by its reference key. 
    *
    * @param key The reference key to use to search the friend. 
    * 
    * @return The {@link Friend} object corresponding to the given reference key. 
    */
    public Friend getFriendByRefKey(String key);

    /**
    * Retrieves the list of {@link Friend} from the core that has the given display
    * name. 
    *
    * @param name the name of the list 
    * 
    * @return the first {@link FriendList} object or null 
    */
    public FriendList getFriendListByName(String name);

    /**
    * Get payload type from mime type and clock rate. 
    *
    * This function searches in audio and video codecs for the given payload type
    * name and clockrate. 
    * 
    * @param channels number of channels, can be
    * LINPHONE_FIND_PAYLOAD_IGNORE_CHANNELS 
    * 
    * @return Returns null if not found. If a {@link PayloadType} is returned, it
    * must be released with linphone_payload_type_unref after using it. 
    * 
    * Warning: The returned payload type is allocated as a floating reference i.e.
    * the reference counter is initialized to 0. 
    */
    public PayloadType getPayloadType(String type, int rate, int channels);

    /**
    *
    * @param idkey An arbitrary idkey string associated to a proxy configuration 
    * 
    * @return the proxy configuration for the given idkey value, or null if none
    * found 
    */
    public ProxyConfig getProxyConfigByIdkey(String idkey);

    /**
    * Return the unread chat message count for a given local address. 
    *
    * @param address {@link Address} object. 
    * 
    * @return The unread chat message count. 
    */
    public int getUnreadChatMessageCountFromLocal(Address address);

    /**
    * Get the zrtp sas validation status for a peer uri. 
    *
    * Once the SAS has been validated or rejected, the status will never return to
    * Unknown (unless you delete your cache) 
    * 
    * @param addr the peer uri
    * 
    * @return - LinphoneZrtpPeerStatusUnknown: this uri is not present in cache OR
    * during calls with the active device, SAS never was validated or rejected
    * 
    */
    public ZrtpPeerStatus getZrtpStatus(String addr);

    /**
    * Check whether the device has a hardware echo canceller. 
    *
    * @return  if it does, true otherwise 
    */
    public boolean hasBuiltinEchoCanceller();

    /**
    * Check whether the device is flagged has crappy opengl. 
    *
    * @return  if crappy opengl flag is set, true otherwise 
    */
    public boolean hasCrappyOpengl();

    /**
    * Tells whether there is a call running. 
    *
    * @return A boolean value telling whether a call is currently running or not 
    */
    public boolean inCall();

    /**
    * See linphone_proxy_config_normalize_sip_uri for documentation. 
    *
    * Default proxy config is used to parse the address. 
    */
    public Address interpretUrl(String url);

    /**
    * Initiates an outgoing call. 
    *
    * The application doesn't own a reference to the returned LinphoneCall object.
    * Use linphone_call_ref to safely keep the LinphoneCall pointer valid within your
    * application.
    * 
    * @param url The destination of the call (sip address, or phone number). 
    * 
    * @return A {@link Call} object or null in case of failure 
    */
    public Call invite(String url);

    /**
    * Initiates an outgoing call given a destination {@link Address} The {@link
    * Address} can be constructed directly using linphone_address_new, or created by
    * {@link Core#interpretUrl}. 
    *
    * The application doesn't own a reference to the returned {@link Call} object.
    * Use linphone_call_ref to safely keep the {@link Call} pointer valid within your
    * application. 
    * 
    * @param addr The destination of the call (sip address). 
    * 
    * @return A {@link Call} object or null in case of failure 
    */
    public Call inviteAddress(Address addr);

    /**
    * Initiates an outgoing call given a destination {@link Address} The {@link
    * Address} can be constructed directly using linphone_address_new, or created by
    * {@link Core#interpretUrl}. 
    *
    * The application doesn't own a reference to the returned {@link Call} object.
    * Use linphone_call_ref to safely keep the {@link Call} pointer valid within your
    * application. 
    * 
    * @param params Call parameters 
    * 
    * @return A {@link Call} object or null in case of failure 
    */
    public Call inviteAddressWithParams(Address addr, CallParams params);

    /**
    * Initiates an outgoing call according to supplied call parameters The
    * application doesn't own a reference to the returned {@link Call} object. 
    *
    * Use linphone_call_ref to safely keep the {@link Call} pointer valid within your
    * application. 
    * 
    * @param params Call parameters 
    * 
    * @return A {@link Call} object or null in case of failure 
    */
    public Call inviteWithParams(String url, CallParams params);

    /**
    * Checks if the given media filter is loaded and usable. 
    *
    * This is for advanced users of the library, mainly to expose mediastreamer video
    * filter status. 
    * 
    * @param filtername the filter name 
    * 
    * @return true if the filter is loaded and usable, false otherwise 
    */
    public boolean isMediaFilterSupported(String filtername);

    /**
    * Main loop function. 
    *
    * It is crucial that your application call it periodically.
    * 
    * {@link Core#iterate} performs various backgrounds tasks:
    * 
    */
    public void iterate();

    /**
    * Make the local participant leave the running conference. 
    *
    * @return 0 if succeeded. Negative number if failed 
    */
    public int leaveConference();

    /**
    * Tells if lime is available. 
    *
    * @deprecated Use linphone_core_lime_x3dh_available instead. Depreacted since
    * 2019-02-04 
    */
    @Deprecated
    public boolean limeAvailable();

    /**
    * Returns the lime state. 
    *
    * @return The current lime state
    * 
    * @deprecated Use linphone_core_lime_x3dh_enabled instead. Depreacted since
    * 2019-02-04 
    */
    @Deprecated
    public LimeState limeEnabled();

    /**
    * Tells if LIME X3DH is available. 
    *
    */
    public boolean limeX3DhAvailable();

    /**
    * Update current config with the content of a xml config file. 
    *
    * @param xmlUri the path to the xml file 
    */
    public void loadConfigFromXml(String xmlUri);

    /**
    * Check if a media encryption type is supported. 
    *
    * @param menc LinphoneMediaEncryption 
    * 
    * @return whether a media encryption scheme is supported by the {@link Core}
    * engine 
    */
    public boolean mediaEncryptionSupported(MediaEncryption menc);

    /**
    * Migrates the call logs from the linphonerc to the database if not done yet. 
    *
    */
    public void migrateLogsFromRcToDb();

    /**
    * Migrate configuration so that all SIP transports are enabled. 
    *
    * Versions of linphone < 3.7 did not support using multiple SIP transport
    * simultaneously. This function helps application to migrate the configuration so
    * that all transports are enabled. Existing proxy configuration are added a
    * transport parameter so that they continue using the unique transport that was
    * set previously. This function must be used just after creating the core, before
    * any call to {@link Core#iterate} 
    * 
    * @return 1 if migration was done, 0 if not done because unnecessary or already
    * done, -1 in case of error. 
    */
    public int migrateToMultiTransport();

    /**
    * Notify all friends that have subscribed. 
    *
    * @param presence {@link PresenceModel} to notify 
    */
    public void notifyAllFriends(PresenceModel presence);

    /**
    * Notifies the upper layer that a presence status has been received by calling
    * the appropriate callback if one has been set. 
    *
    * This method is for advanced usage, where customization of the liblinphone's
    * internal behavior is required. 
    * 
    * @param lf the {@link Friend} whose presence information has been received. 
    */
    public void notifyNotifyPresenceReceived(Friend lf);

    /**
    * Notifies the upper layer that a presence model change has been received for the
    * uri or telephone number given as a parameter, by calling the appropriate
    * callback if one has been set. 
    *
    * This method is for advanced usage, where customization of the liblinphone's
    * internal behavior is required. 
    * 
    * @param presenceModel the {@link PresenceModel} that has been modified 
    */
    public void notifyNotifyPresenceReceivedForUriOrTel(Friend lf, String uriOrTel, PresenceModel presenceModel);

    /**
    * Pause all currently running calls. 
    *
    * @return 0 
    */
    public int pauseAllCalls();

    /**
    * Pauses the call. 
    *
    * If a music file has been setup using {@link Core#setPlayFile}, this file will
    * be played to the remote user.
    * 
    * The only way to resume a paused call is to call {@link Core#resumeCall}. 
    * 
    * @param call The call to pause 
    * 
    * @return 0 on success, -1 on failure
    * 
    * See also: {@link Core#resumeCall} 
    * 
    * @deprecated Use {@link Call#pause} instead. Deprecated since 2017-02-13. 
    */
    @Deprecated
    public int pauseCall(Call call);

    /**
    * Plays a dtmf sound to the local user. 
    *
    * @param durationMs Duration in ms, -1 means play until next further call to
    * {@link Core#stopDtmf} 
    */
    public void playDtmf(char dtmf, int durationMs);

    /**
    * Plays an audio file to the local user. 
    *
    * This function works at any time, during calls, or when no calls are running. It
    * doesn't request the underlying audio system to support multiple playback
    * streams. 
    * 
    * @param audiofile The path to an audio file in wav PCM 16 bit format 
    * 
    * @return 0 on success, -1 on error 
    */
    public int playLocal(String audiofile);

    /**
    * Call generic OpenGL render for a given core. 
    *
    */
    public void previewOglRender();

    /**
    * Publish an event state. 
    *
    * This first create a {@link Event} with {@link Core#createPublish} and calls
    * {@link Event#sendPublish} to actually send it. After expiry, the publication is
    * refreshed unless it is terminated before. 
    * 
    * @param body the actual published data 
    * 
    * @return the {@link Event} holding the context of the publish. 
    */
    public Event publish(Address resource, String event, int expires, Content body);

    /**
    * Gets keep alive interval of real time text. 
    *
    * @return keep alive interval of real time text. 
    */
    public int realtimeTextGetKeepaliveInterval();

    /**
    * Set keep alive interval for real time text. 
    *
    * @param interval The keep alive interval of real time text, 25000 by default. 
    */
    public void realtimeTextSetKeepaliveInterval(int interval);

    /**
    * Redirect the specified call to the given redirect URI. 
    *
    * @param redirectUri The URI to redirect the call to 
    * 
    * @return 0 if successful, -1 on error.
    * 
    * @deprecated Use {@link Call#redirect} instead. Deprecated since 2017-02-13. 
    */
    @Deprecated
    public int redirectCall(Call call, String redirectUri);

    /**
    * force registration refresh to be initiated upon next iterate 
    *
    */
    public void refreshRegisters();

    /**
    * Black list a friend. 
    *
    * same as {@link Friend#setIncSubscribePolicy} with LinphoneSPDeny policy; 
    * 
    * @param lf {@link Friend} to add 
    */
    public void rejectSubscriber(Friend lf);

    /**
    * Reload mediastreamer2 plugins from specified directory. 
    *
    * @param path the path from where plugins are to be loaded, pass null to use
    * default (compile-time determined) plugin directory. 
    */
    public void reloadMsPlugins(String path);

    /**
    * Update detection of sound devices. 
    *
    * Use this function when the application is notified of USB plug events, so that
    * list of available hardwares for sound playback and capture is updated. 
    */
    public void reloadSoundDevices();

    /**
    * Update detection of camera devices. 
    *
    * Use this function when the application is notified of USB plug events, so that
    * list of available hardwares for video capture is updated. 
    */
    public void reloadVideoDevices();

    /**
    * Removes an authentication information object. 
    *
    * @param info The {@link AuthInfo} to remove. 
    */
    public void removeAuthInfo(AuthInfo info);

    /**
    * Remove a specific call log from call history list. 
    *
    * This function destroys the call log object. It must not be accessed anymore by
    * the application after calling this function. 
    * 
    * @param callLog {@link CallLog} object to remove. 
    */
    public void removeCallLog(CallLog callLog);

    /**
    * Removes a friend list. 
    *
    * @param list {@link FriendList} object 
    */
    public void removeFriendList(FriendList list);

    /**
    * Remove a call from the conference. 
    *
    * @param call a call that has been previously merged into the conference.
    * 
    * After removing the remote participant belonging to the supplied call, the call
    * becomes a normal call in paused state. If one single remote participant is left
    * alone together with the local user in the conference after the removal, then
    * the conference is automatically transformed into a simple call in
    * StreamsRunning state. The conference's resources are then automatically
    * destroyed.
    * 
    * In other words, unless {@link Core#leaveConference} is explicitly called, the
    * last remote participant of a conference is automatically put in a simple call
    * in running state.
    * 
    * @return 0 if successful, -1 otherwise. 
    */
    public int removeFromConference(Call call);

    /**
    * Remove the given linphone specs from the list of functionalities the linphone
    * client supports. 
    *
    * @param spec The spec to remove 
    */
    public void removeLinphoneSpec(String spec);

    /**
    * Removes a proxy configuration. 
    *
    * {@link Core} will then automatically unregister and place the proxy
    * configuration on a deleted list. For that reason, a removed proxy does NOT need
    * to be freed. 
    */
    public void removeProxyConfig(ProxyConfig config);

    /**
    * Remove a supported tag. 
    *
    * @param tag The tag to remove
    * 
    * See also: {@link Core#addSupportedTag} 
    */
    public void removeSupportedTag(String tag);

    /**
    * Reset the counter of missed calls. 
    *
    */
    public void resetMissedCallsCount();

    /**
    * Resumes a call. 
    *
    * The call needs to have been paused previously with {@link Core#pauseCall}. 
    * 
    * @param call The call to resume 
    * 
    * @return 0 on success, -1 on failure
    * 
    * See also: {@link Core#pauseCall} 
    * 
    * @deprecated Use {@link Call#resume} instead. Deprecated since 2017-02-13. 
    */
    @Deprecated
    public int resumeCall(Call call);

    /**
    * Sets the UDP port range from which to randomly select the port used for audio
    * streaming. 
    *
    * @param maxPort The upper bound of the audio port range to use 
    */
    public void setAudioPortRange(int minPort, int maxPort);

    /**
    * Assign an audio file to be played locally upon call failure, for a given
    * reason. 
    *
    * @param audiofile a wav file to be played when such call failure happens. 
    */
    public void setCallErrorTone(Reason reason, String audiofile);

    /**
    * Set the rectangle where the decoder will search a QRCode. 
    *
    * @param h height 
    */
    public void setQrcodeDecodeRect(int x, int y, int w, int h);

    /**
    * Sets the UDP port range from which to randomly select the port used for text
    * streaming. 
    *
    * @param maxPort The upper bound of the text port range to use 
    */
    public void setTextPortRange(int minPort, int maxPort);

    /**
    * Set the user agent string used in SIP messages. 
    *
    * Set the user agent string used in SIP messages as "[ua_name]/[version]". No
    * slash character will be printed if null is given to "version". If null is given
    * to "ua_name" and "version" both, the User-agent header will be empty.
    * 
    * This function should be called just after linphone_factory_create_core ideally. 
    * 
    * @param version Version of the user agent. 
    */
    public void setUserAgent(String uaName, String version);

    /**
    * Sets the UDP port range from which to randomly select the port used for video
    * streaming. 
    *
    * @param maxPort The upper bound of the video port range to use 
    */
    public void setVideoPortRange(int minPort, int maxPort);

    /**
    * Tells whether a specified sound device can capture sound. 
    *
    * @param device the device name as returned by linphone_core_get_sound_devices 
    * 
    * @return A boolean value telling whether the specified sound device can capture
    * sound 
    */
    public boolean soundDeviceCanCapture(String device);

    /**
    * Tells whether a specified sound device can play sound. 
    *
    * @param device the device name as returned by linphone_core_get_sound_devices 
    * 
    * @return A boolean value telling whether the specified sound device can play
    * sound 
    */
    public boolean soundDeviceCanPlayback(String device);

    /**
    * Check if a call will need the sound resources in near future (typically an
    * outgoing call that is awaiting response). 
    *
    * In liblinphone, it is not possible to have two independant calls using sound
    * device or camera at the same time. In order to prevent this situation, an
    * application can use {@link Core#soundResourcesLocked} to know whether it is
    * possible at a given time to start a new outgoing call. When the function
    * returns true, an application should not allow the user to start an outgoing
    * call. 
    * 
    * @return A boolean value telling whether a call will need the sound resources in
    * near future 
    */
    public boolean soundResourcesLocked();

    /**
    * Start a {@link Core} object after it has been instantiated and not
    * automatically started. 
    *
    * Also re-initialize a {@link Core} object that has been stopped using {@link
    * Core#stop}. Must be called only if LinphoneGlobalState is either Ready of Off.
    * State will changed to Startup, Configuring and then On.
    * 
    * @return 0: success, -1: global failure, -2: could not connect database 
    */
    public int start();

    /**
    * Start recording the running conference. 
    *
    * @param path Path to the file where the recording will be written 
    * 
    * @return 0 if succeeded. Negative number if failed 
    */
    public int startConferenceRecording(String path);

    /**
    * Special function to warm up dtmf feeback stream. 
    *
    * linphone_core_stop_dtmf_stream must() be called before entering FG mode 
    */
    public void startDtmfStream();

    /**
    * Starts an echo calibration of the sound devices, in order to find adequate
    * settings for the echo canceler automatically. 
    *
    * @return LinphoneStatus whether calibration has started or not. 
    */
    public int startEchoCancellerCalibration();

    /**
    * Start the simulation of call to test the latency with an external device. 
    *
    * @param rate Sound sample rate. 
    */
    public int startEchoTester(int rate);

    /**
    * Stop a {@link Core} object after it has been instantiated and started. 
    *
    * If stopped, it can be started again using {@link Core#start}. Must be called
    * only if LinphoneGlobalState is either On. State will changed to Shutdown and
    * then Off.
    */
    public void stop();

    /**
    * Stop asynchronously a {@link Core} object after it has been instantiated and
    * started. 
    *
    * State changes to Shutdown then {@link Core#iterate} must be called to allow the
    * Core to end asynchronous tasks (terminate call, etc.). When all tasks are
    * finished, State will change to Off. Must be called only if LinphoneGlobalState
    * is On. When LinphoneGlobalState is Off {@link Core} can be started again using
    * {@link Core#start}.
    */
    public void stopAsync();

    /**
    * Stop recording the running conference. 
    *
    * @return 0 if succeeded. Negative number if failed 
    */
    public int stopConferenceRecording();

    /**
    * Stops playing a dtmf started by {@link Core#playDtmf}. 
    *
    */
    public void stopDtmf();

    /**
    * Special function to stop dtmf feed back function. 
    *
    * Must be called before entering BG mode 
    */
    public void stopDtmfStream();

    /**
    * Stop the simulation of call. 
    *
    */
    public int stopEchoTester();

    /**
    * Whenever the liblinphone is playing a ring to advertise an incoming call or
    * ringback of an outgoing call, this function stops the ringing. 
    *
    * Typical use is to stop ringing when the user requests to ignore the call. 
    */
    public void stopRinging();

    /**
    * Create an outgoing subscription, specifying the destination resource, the event
    * name, and an optional content body. 
    *
    * If accepted, the subscription runs for a finite period, but is automatically
    * renewed if not terminated before. 
    * 
    * @param body an optional body, may be null.    
    * 
    * @return a {@link Event} holding the context of the created subcription. 
    */
    public Event subscribe(Address resource, String event, int expires, @Nullable Content body);

    /**
    * Take a photo of currently from capture device and write it into a jpeg file. 
    *
    * Note that the snapshot is asynchronous, an application shall not assume that
    * the file is created when the function returns.
    * 
    * @param file a path where to write the jpeg content. 
    * 
    * @return 0 if successfull, -1 otherwise (typically if jpeg format is not
    * supported). 
    */
    public int takePreviewSnapshot(String file);

    /**
    * Terminates all the calls. 
    *
    * @return 0 
    */
    public int terminateAllCalls();

    /**
    * Terminates a call. 
    *
    * @param call The LinphoneCall object representing the call to be terminated 
    * 
    * @return 0 on success, -1 on failure
    * 
    * @deprecated Use {@link Call#terminate} instead. Deprecated since 2017-02-13. 
    */
    @Deprecated
    public int terminateCall(Call call);

    /**
    * Terminate the running conference. 
    *
    * If it is a local conference, all calls inside it will become back separate
    * calls and will be put in #LinphoneCallPaused state. If it is a conference
    * involving a focus server, all calls inside the conference will be terminated. 
    * 
    * @return 0 if succeeded. Negative number if failed 
    */
    public int terminateConference();

    /**
    * Performs a simple call transfer to the specified destination. 
    *
    * The remote endpoint is expected to issue a new call to the specified
    * destination. The current call remains active and thus can be later paused or
    * terminated. It is possible to follow the progress of the transfer provided that
    * transferee sends notification about it. In this case, the
    * transfer_state_changed callback of the LinphoneCoreVTable is invoked to notify
    * of the state of the new call at the other party. The notified states are
    * #LinphoneCallOutgoingInit , #LinphoneCallOutgoingProgress,
    * #LinphoneCallOutgoingRinging and #LinphoneCallConnected. 
    * 
    * @param referTo The destination the call is to be refered to 
    * 
    * @return 0 on success, -1 on failure
    * 
    * @deprecated Use {@link Call#transfer} instead. Deprecated since 2017-02-13. 
    */
    @Deprecated
    public int transferCall(Call call, String referTo);

    /**
    * Transfers a call to destination of another running call. 
    *
    * This is used for "attended transfer" scenarios.
    * 
    * The transfered call is supposed to be in paused state, so that it is able to
    * accept the transfer immediately. The destination call is a call previously
    * established to introduce the transfered person. This method will send a
    * transfer request to the transfered person. The phone of the transfered is then
    * expected to automatically call to the destination of the transfer. The receiver
    * of the transfer will then automatically close the call with us (the 'dest'
    * call). It is possible to follow the progress of the transfer provided that
    * transferee sends notification about it. In this case, the
    * transfer_state_changed callback of the LinphoneCoreVTable is invoked to notify
    * of the state of the new call at the other party. The notified states are
    * #LinphoneCallOutgoingInit , #LinphoneCallOutgoingProgress,
    * #LinphoneCallOutgoingRinging and #LinphoneCallConnected. 
    * 
    * @param dest A running call whose remote person will receive the transfer 
    * 
    * @return 0 on success, -1 on failure
    * 
    * @deprecated Use {@link Call#transferToAnother} instead. Deprecated since
    * 2017-02-13. 
    */
    @Deprecated
    public int transferCallToAnother(Call call, Call dest);

    /**
    * Updates a running call according to supplied call parameters or parameters
    * changed in the LinphoneCore. 
    *
    * In this version this is limited to the following use cases:
    * 
    * 
    * @param params The new call parameters to use (may be null) 
    * 
    * @return 0 if successful, -1 otherwise.
    * 
    * @deprecated Use {@link Call#update} instead. Deprecated since 2017-02-13. 
    */
    @Deprecated
    public int updateCall(Call call, CallParams params);

    /**
    * Upload the log collection to the configured server url. 
    *
    */
    public void uploadLogCollection();

    /**
    * Tells the core to use a separate window for local camera preview video, instead
    * of inserting local view within the remote video window. 
    *
    * @param yesno  to use a separate window, true to insert the preview in the
    * remote video window. 
    */
    public void usePreviewWindow(boolean yesno);

    /**
    * Specify whether the tls server certificate must be verified when connecting to
    * a SIP/TLS server. 
    *
    * @param yesno A boolean value telling whether the tls server certificate must be
    * verified 
    */
    public void verifyServerCertificates(boolean yesno);

    /**
    * Specify whether the tls server certificate common name must be verified when
    * connecting to a SIP/TLS server. 
    *
    * @param yesno A boolean value telling whether the tls server certificate common
    * name must be verified 
    */
    public void verifyServerCn(boolean yesno);

    /**
    * Test if video is supported. 
    *
    */
    public boolean videoSupported();

    /**
    * Compress the log collection in a single file. 
    *
    * @return The path of the compressed log collection file (to be freed calling
    * ms_free()). 
    */
    public String compressLogCollection();

    /**
    * Enable the linphone core log collection to upload logs on a server. 
    *
    * @param state #LinphoneLogCollectionState value telling whether to enable log
    * collection or not. 
    */
    public void enableLogCollection(LogCollectionState state);

    /**
    * Get the max file size in bytes of the files used for log collection. 
    *
    * @return The max file size in bytes of the files used for log collection. 
    */
    public int getLogCollectionMaxFileSize();

    /**
    * Get the path where the log files will be written for log collection. 
    *
    * @return The path where the log files will be written. 
    */
    public String getLogCollectionPath();

    /**
    * Get the prefix of the filenames that will be used for log collection. 
    *
    * @return The prefix of the filenames used for log collection. 
    */
    public String getLogCollectionPrefix();

    /**
    * Get defined log level mask. 
    *
    * @return The loglevel parameter is a bitmask parameter. Therefore to enable only
    * warning and error messages, use ORTP_WARNING | ORTP_ERROR. To disable logs,
    * simply set loglevel to 0. 
    * 
    * @deprecated Use {@link LoggingService#getLogLevelMask} instead. Deprecated
    * since 2017-10-10. 
    */
    @Deprecated
    public int getLogLevelMask();

    /**
    * Returns liblinphone's version as a string. 
    *
    */
    public String getVersion();

    /**
    * Tells whether the linphone core log collection is enabled. 
    *
    * @return The state of the linphone core log collection. 
    */
    public LogCollectionState logCollectionEnabled();

    /**
    * Reset the log collection by removing the log files. 
    *
    */
    public void resetLogCollection();

    /**
    * Enable logs serialization (output logs from either the thread that creates the
    * linphone core or the thread that calls {@link Core#iterate}). 
    *
    * Must be called before creating the linphone core. 
    */
    public void serializeLogs();

    /**
    * Set the max file size in bytes of the files used for log collection. 
    *
    * Warning: this function should only not be used to change size dynamically but
    * instead only before calling See also: linphone_core_enable_log_collection. If
    * you increase max size on runtime, logs chronological order COULD be broken. 
    * 
    * @param size The max file size in bytes of the files used for log collection. 
    */
    public void setLogCollectionMaxFileSize(int size);

    /**
    * Set the path of a directory where the log files will be written for log
    * collection. 
    *
    * @param path The path where the log files will be written. 
    */
    public void setLogCollectionPath(String path);

    /**
    * Set the prefix of the filenames that will be used for log collection. 
    *
    * @param prefix The prefix to use for the filenames for log collection. 
    */
    public void setLogCollectionPrefix(String prefix);

    /**
    * Define the log level using mask. 
    *
    * The loglevel parameter is a bitmask parameter. Therefore to enable only warning
    * and error messages, use ORTP_WARNING | ORTP_ERROR. To disable logs, simply set
    * loglevel to 0.
    * 
    * @param mask A bitmask of the log levels to set. 
    * 
    * @deprecated Use {@link LoggingService#setLogLevel} instead. Deprecated since
    * 2017-10-10. 
    */
    @Deprecated
    public void setLogLevelMask(int mask);

    /**
    * True if tunnel support was compiled. 
    *
    */
    public boolean tunnelAvailable();

    /**
    * Return the availability of uPnP. 
    *
    * @return true if uPnP is available otherwise return false. 
    */
    public boolean upnpAvailable();

    /**
    * Tells whether VCARD support is builtin. 
    *
    * @return  if VCARD is supported, true otherwise. 
    */
    public boolean vcardSupported();

    /**
    */
    public void addListener(CoreListener listener);

    /**
    */
    public void removeListener(CoreListener listener);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();
}

class CoreImpl implements Core {

    protected long nativePtr = 0;
    protected Object userData = null;

    protected CoreImpl(long ptr) {
        nativePtr = ptr;
    }


    private native String getAdaptiveRateAlgorithm(long nativePtr);
    @Override
    synchronized public String getAdaptiveRateAlgorithm()  {
        
        return getAdaptiveRateAlgorithm(nativePtr);
    }

    private native void setAdaptiveRateAlgorithm(long nativePtr, String algorithm);
    @Override
    synchronized public void setAdaptiveRateAlgorithm(String algorithm)  {
        
        setAdaptiveRateAlgorithm(nativePtr, algorithm);
    }

    private native boolean adaptiveRateControlEnabled(long nativePtr);
    @Override
    synchronized public boolean adaptiveRateControlEnabled()  {
        
        return adaptiveRateControlEnabled(nativePtr);
    }

    private native void enableAdaptiveRateControl(long nativePtr, boolean enabled);
    @Override
    synchronized public void enableAdaptiveRateControl(boolean enabled)  {
        
        enableAdaptiveRateControl(nativePtr, enabled);
    }

    private native boolean audioAdaptiveJittcompEnabled(long nativePtr);
    @Override
    synchronized public boolean audioAdaptiveJittcompEnabled()  {
        
        return audioAdaptiveJittcompEnabled(nativePtr);
    }

    private native void enableAudioAdaptiveJittcomp(long nativePtr, boolean enable);
    @Override
    synchronized public void enableAudioAdaptiveJittcomp(boolean enable)  {
        
        enableAudioAdaptiveJittcomp(nativePtr, enable);
    }

    private native int getAudioDscp(long nativePtr);
    @Override
    synchronized public int getAudioDscp()  {
        
        return getAudioDscp(nativePtr);
    }

    private native void setAudioDscp(long nativePtr, int dscp);
    @Override
    synchronized public void setAudioDscp(int dscp)  {
        
        setAudioDscp(nativePtr, dscp);
    }

    private native int getAudioJittcomp(long nativePtr);
    @Override
    synchronized public int getAudioJittcomp()  {
        
        return getAudioJittcomp(nativePtr);
    }

    private native void setAudioJittcomp(long nativePtr, int milliseconds);
    @Override
    synchronized public void setAudioJittcomp(int milliseconds)  {
        
        setAudioJittcomp(nativePtr, milliseconds);
    }

    private native String getAudioMulticastAddr(long nativePtr);
    @Override
    synchronized public String getAudioMulticastAddr()  {
        
        return getAudioMulticastAddr(nativePtr);
    }

    private native int setAudioMulticastAddr(long nativePtr, String ip);
    @Override
    synchronized public int setAudioMulticastAddr(String ip)  {
        
        return setAudioMulticastAddr(nativePtr, ip);
    }

    private native boolean audioMulticastEnabled(long nativePtr);
    @Override
    synchronized public boolean audioMulticastEnabled()  {
        
        return audioMulticastEnabled(nativePtr);
    }

    private native void enableAudioMulticast(long nativePtr, boolean yesno);
    @Override
    synchronized public void enableAudioMulticast(boolean yesno)  {
        
        enableAudioMulticast(nativePtr, yesno);
    }

    private native int getAudioMulticastTtl(long nativePtr);
    @Override
    synchronized public int getAudioMulticastTtl()  {
        
        return getAudioMulticastTtl(nativePtr);
    }

    private native int setAudioMulticastTtl(long nativePtr, int ttl);
    @Override
    synchronized public int setAudioMulticastTtl(int ttl)  {
        
        return setAudioMulticastTtl(nativePtr, ttl);
    }

    private native PayloadType[] getAudioPayloadTypes(long nativePtr);
    @Override
    synchronized public PayloadType[] getAudioPayloadTypes()  {
        
        return getAudioPayloadTypes(nativePtr);
    }

    private native void setAudioPayloadTypes(long nativePtr, PayloadType[] payloadTypes);
    @Override
    synchronized public void setAudioPayloadTypes(PayloadType[] payloadTypes)  {
        
        setAudioPayloadTypes(nativePtr, payloadTypes);
    }

    private native int getAudioPort(long nativePtr);
    @Override
    synchronized public int getAudioPort()  {
        
        return getAudioPort(nativePtr);
    }

    private native void setAudioPort(long nativePtr, int port);
    @Override
    synchronized public void setAudioPort(int port)  {
        
        setAudioPort(nativePtr, port);
    }

    private native Range getAudioPortsRange(long nativePtr);
    @Override
    synchronized public Range getAudioPortsRange()  {
        
        return (Range)getAudioPortsRange(nativePtr);
    }

    private native AuthInfo[] getAuthInfoList(long nativePtr);
    @Override
    synchronized public AuthInfo[] getAuthInfoList()  {
        
        return getAuthInfoList(nativePtr);
    }

    private native int getAvpfMode(long nativePtr);
    @Override
    synchronized public AVPFMode getAvpfMode()  {
        
        return AVPFMode.fromInt(getAvpfMode(nativePtr));
    }

    private native void setAvpfMode(long nativePtr, int mode);
    @Override
    synchronized public void setAvpfMode(AVPFMode mode)  {
        
        setAvpfMode(nativePtr, mode.toInt());
    }

    private native int getAvpfRrInterval(long nativePtr);
    @Override
    synchronized public int getAvpfRrInterval()  {
        
        return getAvpfRrInterval(nativePtr);
    }

    private native void setAvpfRrInterval(long nativePtr, int interval);
    @Override
    synchronized public void setAvpfRrInterval(int interval)  {
        
        setAvpfRrInterval(nativePtr, interval);
    }

    private native CallLog[] getCallLogs(long nativePtr);
    @Override
    synchronized public CallLog[] getCallLogs()  {
        
        return getCallLogs(nativePtr);
    }

    private native String getCallLogsDatabasePath(long nativePtr);
    @Override
    synchronized public String getCallLogsDatabasePath()  {
        
        return getCallLogsDatabasePath(nativePtr);
    }

    private native void setCallLogsDatabasePath(long nativePtr, String path);
    @Override
    synchronized public void setCallLogsDatabasePath(String path)  {
        
        setCallLogsDatabasePath(nativePtr, path);
    }

    private native Call[] getCalls(long nativePtr);
    @Override
    synchronized public Call[] getCalls()  {
        
        return getCalls(nativePtr);
    }

    private native int getCallsNb(long nativePtr);
    @Override
    synchronized public int getCallsNb()  {
        
        return getCallsNb(nativePtr);
    }

    private native int getCameraSensorRotation(long nativePtr);
    @Override
    synchronized public int getCameraSensorRotation()  {
        
        return getCameraSensorRotation(nativePtr);
    }

    private native String getCaptureDevice(long nativePtr);
    @Override
    synchronized public String getCaptureDevice()  {
        
        return getCaptureDevice(nativePtr);
    }

    private native int setCaptureDevice(long nativePtr, String devid);
    @Override
    synchronized public int setCaptureDevice(String devid)  {
        
        return setCaptureDevice(nativePtr, devid);
    }

    private native String getChatDatabasePath(long nativePtr);
    @Override
    synchronized public String getChatDatabasePath()  {
        
        return getChatDatabasePath(nativePtr);
    }

    private native void setChatDatabasePath(long nativePtr, String path);
    @Override
    synchronized public void setChatDatabasePath(String path)  {
        
        setChatDatabasePath(nativePtr, path);
    }

    private native boolean chatEnabled(long nativePtr);
    @Override
    synchronized public boolean chatEnabled()  {
        
        return chatEnabled(nativePtr);
    }

    private native ChatRoom[] getChatRooms(long nativePtr);
    @Override
    synchronized public ChatRoom[] getChatRooms()  {
        
        return getChatRooms(nativePtr);
    }

    private native Conference getConference(long nativePtr);
    @Override
    synchronized public Conference getConference()  {
        
        return (Conference)getConference(nativePtr);
    }

    private native float getConferenceLocalInputVolume(long nativePtr);
    @Override
    synchronized public float getConferenceLocalInputVolume()  {
        
        return getConferenceLocalInputVolume(nativePtr);
    }

    private native boolean conferenceServerEnabled(long nativePtr);
    @Override
    synchronized public boolean conferenceServerEnabled()  {
        
        return conferenceServerEnabled(nativePtr);
    }

    private native void enableConferenceServer(long nativePtr, boolean enable);
    @Override
    synchronized public void enableConferenceServer(boolean enable)  {
        
        enableConferenceServer(nativePtr, enable);
    }

    private native int getConferenceSize(long nativePtr);
    @Override
    synchronized public int getConferenceSize()  {
        
        return getConferenceSize(nativePtr);
    }

    private native Config getConfig(long nativePtr);
    @Override
    synchronized public Config getConfig()  {
        
        return (Config)getConfig(nativePtr);
    }

    private native int getConsolidatedPresence(long nativePtr);
    @Override
    synchronized public ConsolidatedPresence getConsolidatedPresence()  {
        
        return ConsolidatedPresence.fromInt(getConsolidatedPresence(nativePtr));
    }

    private native void setConsolidatedPresence(long nativePtr, int presence);
    @Override
    synchronized public void setConsolidatedPresence(ConsolidatedPresence presence)  {
        
        setConsolidatedPresence(nativePtr, presence.toInt());
    }

    private native Call getCurrentCall(long nativePtr);
    @Override
    synchronized public Call getCurrentCall()  {
        
        return (Call)getCurrentCall(nativePtr);
    }

    private native Address getCurrentCallRemoteAddress(long nativePtr);
    @Override
    synchronized public Address getCurrentCallRemoteAddress()  {
        
        return (Address)getCurrentCallRemoteAddress(nativePtr);
    }

    private native VideoDefinition getCurrentPreviewVideoDefinition(long nativePtr);
    @Override
    synchronized public VideoDefinition getCurrentPreviewVideoDefinition()  {
        
        return (VideoDefinition)getCurrentPreviewVideoDefinition(nativePtr);
    }

    private native FriendList getDefaultFriendList(long nativePtr);
    @Override
    synchronized public FriendList getDefaultFriendList()  {
        
        return (FriendList)getDefaultFriendList(nativePtr);
    }

    private native ProxyConfig getDefaultProxyConfig(long nativePtr);
    @Override
    synchronized public ProxyConfig getDefaultProxyConfig()  {
        
        return (ProxyConfig)getDefaultProxyConfig(nativePtr);
    }

    private native void setDefaultProxyConfig(long nativePtr, ProxyConfig config);
    @Override
    synchronized public void setDefaultProxyConfig(ProxyConfig config)  {
        
        setDefaultProxyConfig(nativePtr, config);
    }

    private native String getDefaultVideoDisplayFilter(long nativePtr);
    @Override
    synchronized public String getDefaultVideoDisplayFilter()  {
        
        return getDefaultVideoDisplayFilter(nativePtr);
    }

    private native int getDelayedTimeout(long nativePtr);
    @Override
    synchronized public int getDelayedTimeout()  {
        
        return getDelayedTimeout(nativePtr);
    }

    private native void setDelayedTimeout(long nativePtr, int seconds);
    @Override
    synchronized public void setDelayedTimeout(int seconds)  {
        
        setDelayedTimeout(nativePtr, seconds);
    }

    private native int getDeviceRotation(long nativePtr);
    @Override
    synchronized public int getDeviceRotation()  {
        
        return getDeviceRotation(nativePtr);
    }

    private native void setDeviceRotation(long nativePtr, int rotation);
    @Override
    synchronized public void setDeviceRotation(int rotation)  {
        
        setDeviceRotation(nativePtr, rotation);
    }

    private native boolean dnsSearchEnabled(long nativePtr);
    @Override
    synchronized public boolean dnsSearchEnabled()  {
        
        return dnsSearchEnabled(nativePtr);
    }

    private native void enableDnsSearch(long nativePtr, boolean enable);
    @Override
    synchronized public void enableDnsSearch(boolean enable)  {
        
        enableDnsSearch(nativePtr, enable);
    }

    private native void setDnsServers(long nativePtr, String[] servers);
    @Override
    synchronized public void setDnsServers(String[] servers)  {
        
        setDnsServers(nativePtr, servers);
    }

    private native void setDnsServersApp(long nativePtr, String[] servers);
    @Override
    synchronized public void setDnsServersApp(String[] servers)  {
        
        setDnsServersApp(nativePtr, servers);
    }

    private native boolean getDnsSetByApp(long nativePtr);
    @Override
    synchronized public boolean getDnsSetByApp()  {
        
        return getDnsSetByApp(nativePtr);
    }

    private native boolean dnsSrvEnabled(long nativePtr);
    @Override
    synchronized public boolean dnsSrvEnabled()  {
        
        return dnsSrvEnabled(nativePtr);
    }

    private native void enableDnsSrv(long nativePtr, boolean enable);
    @Override
    synchronized public void enableDnsSrv(boolean enable)  {
        
        enableDnsSrv(nativePtr, enable);
    }

    private native int getDownloadBandwidth(long nativePtr);
    @Override
    synchronized public int getDownloadBandwidth()  {
        
        return getDownloadBandwidth(nativePtr);
    }

    private native void setDownloadBandwidth(long nativePtr, int bw);
    @Override
    synchronized public void setDownloadBandwidth(int bw)  {
        
        setDownloadBandwidth(nativePtr, bw);
    }

    private native int getDownloadPtime(long nativePtr);
    @Override
    synchronized public int getDownloadPtime()  {
        
        return getDownloadPtime(nativePtr);
    }

    private native void setDownloadPtime(long nativePtr, int ptime);
    @Override
    synchronized public void setDownloadPtime(int ptime)  {
        
        setDownloadPtime(nativePtr, ptime);
    }

    private native boolean echoCancellationEnabled(long nativePtr);
    @Override
    synchronized public boolean echoCancellationEnabled()  {
        
        return echoCancellationEnabled(nativePtr);
    }

    private native void enableEchoCancellation(long nativePtr, boolean val);
    @Override
    synchronized public void enableEchoCancellation(boolean val)  {
        
        enableEchoCancellation(nativePtr, val);
    }

    private native String getEchoCancellerFilterName(long nativePtr);
    @Override
    synchronized public String getEchoCancellerFilterName()  {
        
        return getEchoCancellerFilterName(nativePtr);
    }

    private native void setEchoCancellerFilterName(long nativePtr, String filtername);
    @Override
    synchronized public void setEchoCancellerFilterName(String filtername)  {
        
        setEchoCancellerFilterName(nativePtr, filtername);
    }

    private native boolean echoLimiterEnabled(long nativePtr);
    @Override
    synchronized public boolean echoLimiterEnabled()  {
        
        return echoLimiterEnabled(nativePtr);
    }

    private native void enableEchoLimiter(long nativePtr, boolean val);
    @Override
    synchronized public void enableEchoLimiter(boolean val)  {
        
        enableEchoLimiter(nativePtr, val);
    }

    private native void setEnableSipUpdate(long nativePtr, int value);
    @Override
    synchronized public void setEnableSipUpdate(int value)  {
        
        setEnableSipUpdate(nativePtr, value);
    }

    private native void setExpectedBandwidth(long nativePtr, int bw);
    @Override
    synchronized public void setExpectedBandwidth(int bw)  {
        
        setExpectedBandwidth(nativePtr, bw);
    }

    private native String getFileTransferServer(long nativePtr);
    @Override
    synchronized public String getFileTransferServer()  {
        
        return getFileTransferServer(nativePtr);
    }

    private native void setFileTransferServer(long nativePtr, String serverUrl);
    @Override
    synchronized public void setFileTransferServer(String serverUrl)  {
        
        setFileTransferServer(nativePtr, serverUrl);
    }

    private native void enableFriendListSubscription(long nativePtr, boolean enable);
    @Override
    synchronized public void enableFriendListSubscription(boolean enable)  {
        
        enableFriendListSubscription(nativePtr, enable);
    }

    private native String getFriendsDatabasePath(long nativePtr);
    @Override
    synchronized public String getFriendsDatabasePath()  {
        
        return getFriendsDatabasePath(nativePtr);
    }

    private native void setFriendsDatabasePath(long nativePtr, String path);
    @Override
    synchronized public void setFriendsDatabasePath(String path)  {
        
        setFriendsDatabasePath(nativePtr, path);
    }

    private native FriendList[] getFriendsLists(long nativePtr);
    @Override
    synchronized public FriendList[] getFriendsLists()  {
        
        return getFriendsLists(nativePtr);
    }

    private native boolean getGuessHostname(long nativePtr);
    @Override
    synchronized public boolean getGuessHostname()  {
        
        return getGuessHostname(nativePtr);
    }

    private native void setGuessHostname(long nativePtr, boolean val);
    @Override
    synchronized public void setGuessHostname(boolean val)  {
        
        setGuessHostname(nativePtr, val);
    }

    private native String getHttpProxyHost(long nativePtr);
    @Override
    synchronized public String getHttpProxyHost()  {
        
        return getHttpProxyHost(nativePtr);
    }

    private native void setHttpProxyHost(long nativePtr, String host);
    @Override
    synchronized public void setHttpProxyHost(String host)  {
        
        setHttpProxyHost(nativePtr, host);
    }

    private native int getHttpProxyPort(long nativePtr);
    @Override
    synchronized public int getHttpProxyPort()  {
        
        return getHttpProxyPort(nativePtr);
    }

    private native void setHttpProxyPort(long nativePtr, int port);
    @Override
    synchronized public void setHttpProxyPort(int port)  {
        
        setHttpProxyPort(nativePtr, port);
    }

    private native String getIdentity(long nativePtr);
    @Override
    synchronized public String getIdentity()  {
        
        return getIdentity(nativePtr);
    }

    private native ImNotifPolicy getImNotifPolicy(long nativePtr);
    @Override
    synchronized public ImNotifPolicy getImNotifPolicy()  {
        
        return (ImNotifPolicy)getImNotifPolicy(nativePtr);
    }

    private native int getInCallTimeout(long nativePtr);
    @Override
    synchronized public int getInCallTimeout()  {
        
        return getInCallTimeout(nativePtr);
    }

    private native void setInCallTimeout(long nativePtr, int seconds);
    @Override
    synchronized public void setInCallTimeout(int seconds)  {
        
        setInCallTimeout(nativePtr, seconds);
    }

    private native int getIncTimeout(long nativePtr);
    @Override
    synchronized public int getIncTimeout()  {
        
        return getIncTimeout(nativePtr);
    }

    private native void setIncTimeout(long nativePtr, int seconds);
    @Override
    synchronized public void setIncTimeout(int seconds)  {
        
        setIncTimeout(nativePtr, seconds);
    }

    private native boolean ipv6Enabled(long nativePtr);
    @Override
    synchronized public boolean ipv6Enabled()  {
        
        return ipv6Enabled(nativePtr);
    }

    private native void enableIpv66(long nativePtr, boolean val);
    @Override
    synchronized public void enableIpv6(boolean val)  {
        
        enableIpv66(nativePtr, val);
    }

    private native boolean isEchoCancellerCalibrationRequired(long nativePtr);
    @Override
    synchronized public boolean isEchoCancellerCalibrationRequired()  {
        
        return isEchoCancellerCalibrationRequired(nativePtr);
    }

    private native boolean isFriendListSubscriptionEnabled(long nativePtr);
    @Override
    synchronized public boolean isFriendListSubscriptionEnabled()  {
        
        return isFriendListSubscriptionEnabled(nativePtr);
    }

    private native boolean isInConference(long nativePtr);
    @Override
    synchronized public boolean isInConference()  {
        
        return isInConference(nativePtr);
    }

    private native boolean isIncomingInvitePending(long nativePtr);
    @Override
    synchronized public boolean isIncomingInvitePending()  {
        
        return isIncomingInvitePending(nativePtr);
    }

    private native boolean isMediaEncryptionMandatory(long nativePtr);
    @Override
    synchronized public boolean isMediaEncryptionMandatory()  {
        
        return isMediaEncryptionMandatory(nativePtr);
    }

    private native boolean isNetworkReachable(long nativePtr);
    @Override
    synchronized public boolean isNetworkReachable()  {
        
        return isNetworkReachable(nativePtr);
    }

    private native boolean isSenderNameHiddenInForwardMessage(long nativePtr);
    @Override
    synchronized public boolean isSenderNameHiddenInForwardMessage()  {
        
        return isSenderNameHiddenInForwardMessage(nativePtr);
    }

    private native boolean keepAliveEnabled(long nativePtr);
    @Override
    synchronized public boolean keepAliveEnabled()  {
        
        return keepAliveEnabled(nativePtr);
    }

    private native void enableKeepAlive(long nativePtr, boolean enable);
    @Override
    synchronized public void enableKeepAlive(boolean enable)  {
        
        enableKeepAlive(nativePtr, enable);
    }

    private native CallLog getLastOutgoingCallLog(long nativePtr);
    @Override
    synchronized public CallLog getLastOutgoingCallLog()  {
        
        return (CallLog)getLastOutgoingCallLog(nativePtr);
    }

    private native boolean limeX3DhEnabled(long nativePtr);
    @Override
    synchronized public boolean limeX3DhEnabled()  {
        
        return limeX3DhEnabled(nativePtr);
    }

    private native void enableLimeX3Dh(long nativePtr, boolean enable);
    @Override
    synchronized public void enableLimeX3Dh(boolean enable)  {
        
        enableLimeX3Dh(nativePtr, enable);
    }

    private native String getLimeX3DhServerUrl(long nativePtr);
    @Override
    synchronized public String getLimeX3DhServerUrl()  {
        
        return getLimeX3DhServerUrl(nativePtr);
    }

    private native void setLimeX3DhServerUrl(long nativePtr, String url);
    @Override
    synchronized public void setLimeX3DhServerUrl(String url)  {
        
        setLimeX3DhServerUrl(nativePtr, url);
    }

    private native String getLinphoneSpecs(long nativePtr);
    @Override
    synchronized public String getLinphoneSpecs()  {
        
        return getLinphoneSpecs(nativePtr);
    }

    private native void setLinphoneSpecs(long nativePtr, String specs);
    @Override
    synchronized public void setLinphoneSpecs(String specs)  {
        
        setLinphoneSpecs(nativePtr, specs);
    }

    private native String[] getLinphoneSpecsList(long nativePtr);
    @Override
    synchronized public String[] getLinphoneSpecsList()  {
        
        return getLinphoneSpecsList(nativePtr);
    }

    private native void setLinphoneSpecsList(long nativePtr, String[] specs);
    @Override
    synchronized public void setLinphoneSpecsList(String[] specs)  {
        
        setLinphoneSpecsList(nativePtr, specs);
    }

    private native String getLogCollectionUploadServerUrl(long nativePtr);
    @Override
    synchronized public String getLogCollectionUploadServerUrl()  {
        
        return getLogCollectionUploadServerUrl(nativePtr);
    }

    private native void setLogCollectionUploadServerUrl(long nativePtr, String serverUrl);
    @Override
    synchronized public void setLogCollectionUploadServerUrl(String serverUrl)  {
        
        setLogCollectionUploadServerUrl(nativePtr, serverUrl);
    }

    private native int getMaxCalls(long nativePtr);
    @Override
    synchronized public int getMaxCalls()  {
        
        return getMaxCalls(nativePtr);
    }

    private native void setMaxCalls(long nativePtr, int max);
    @Override
    synchronized public void setMaxCalls(int max)  {
        
        setMaxCalls(nativePtr, max);
    }

    private native int getMaxSizeForAutoDownloadIncomingFiles(long nativePtr);
    @Override
    synchronized public int getMaxSizeForAutoDownloadIncomingFiles()  {
        
        return getMaxSizeForAutoDownloadIncomingFiles(nativePtr);
    }

    private native void setMaxSizeForAutoDownloadIncomingFiles(long nativePtr, int size);
    @Override
    synchronized public void setMaxSizeForAutoDownloadIncomingFiles(int size)  {
        
        setMaxSizeForAutoDownloadIncomingFiles(nativePtr, size);
    }

    private native String getMediaDevice(long nativePtr);
    @Override
    synchronized public String getMediaDevice()  {
        
        return getMediaDevice(nativePtr);
    }

    private native int setMediaDevice(long nativePtr, String devid);
    @Override
    synchronized public int setMediaDevice(String devid)  {
        
        return setMediaDevice(nativePtr, devid);
    }

    private native int getMediaEncryption(long nativePtr);
    @Override
    synchronized public MediaEncryption getMediaEncryption()  {
        
        return MediaEncryption.fromInt(getMediaEncryption(nativePtr));
    }

    private native int setMediaEncryption(long nativePtr, int menc);
    @Override
    synchronized public int setMediaEncryption(MediaEncryption menc)  {
        
        return setMediaEncryption(nativePtr, menc.toInt());
    }

    private native void setMediaEncryptionMandatory(long nativePtr, boolean m);
    @Override
    synchronized public void setMediaEncryptionMandatory(boolean m)  {
        
        setMediaEncryptionMandatory(nativePtr, m);
    }

    private native void setMediaNetworkReachable(long nativePtr, boolean value);
    @Override
    synchronized public void setMediaNetworkReachable(boolean value)  {
        
        setMediaNetworkReachable(nativePtr, value);
    }

    private native boolean micEnabled(long nativePtr);
    @Override
    synchronized public boolean micEnabled()  {
        
        return micEnabled(nativePtr);
    }

    private native void enableMic(long nativePtr, boolean enable);
    @Override
    synchronized public void enableMic(boolean enable)  {
        
        enableMic(nativePtr, enable);
    }

    private native float getMicGainDb(long nativePtr);
    @Override
    synchronized public float getMicGainDb()  {
        
        return getMicGainDb(nativePtr);
    }

    private native void setMicGainDb(long nativePtr, float level);
    @Override
    synchronized public void setMicGainDb(float level)  {
        
        setMicGainDb(nativePtr, level);
    }

    private native int getMissedCallsCount(long nativePtr);
    @Override
    synchronized public int getMissedCallsCount()  {
        
        return getMissedCallsCount(nativePtr);
    }

    private native int getMtu(long nativePtr);
    @Override
    synchronized public int getMtu()  {
        
        return getMtu(nativePtr);
    }

    private native void setMtu(long nativePtr, int mtu);
    @Override
    synchronized public void setMtu(int mtu)  {
        
        setMtu(nativePtr, mtu);
    }

    private native String getNatAddress(long nativePtr);
    @Override
    synchronized public String getNatAddress()  {
        
        return getNatAddress(nativePtr);
    }

    private native void setNatAddress(long nativePtr, String addr);
    @Override
    synchronized public void setNatAddress(String addr)  {
        
        setNatAddress(nativePtr, addr);
    }

    private native NatPolicy getNatPolicy(long nativePtr);
    @Override
    synchronized public NatPolicy getNatPolicy()  {
        
        return (NatPolicy)getNatPolicy(nativePtr);
    }

    private native void setNatPolicy(long nativePtr, NatPolicy policy);
    @Override
    synchronized public void setNatPolicy(NatPolicy policy)  {
        
        setNatPolicy(nativePtr, policy);
    }

    private native Object getNativePreviewWindowId(long nativePtr);
    @Override
    synchronized public Object getNativePreviewWindowId()  {
        
        return getNativePreviewWindowId(nativePtr);
    }

    private native void setNativePreviewWindowId(long nativePtr, Object id);
    @Override
    synchronized public void setNativePreviewWindowId(Object id)  {
        
        setNativePreviewWindowId(nativePtr, id);
    }

    private native Object getNativeVideoWindowId(long nativePtr);
    @Override
    synchronized public Object getNativeVideoWindowId()  {
        
        return getNativeVideoWindowId(nativePtr);
    }

    private native void setNativeVideoWindowId(long nativePtr, Object id);
    @Override
    synchronized public void setNativeVideoWindowId(Object id)  {
        
        setNativeVideoWindowId(nativePtr, id);
    }

    private native void setNetworkReachable(long nativePtr, boolean value);
    @Override
    synchronized public void setNetworkReachable(boolean value)  {
        
        setNetworkReachable(nativePtr, value);
    }

    private native int getNortpTimeout(long nativePtr);
    @Override
    synchronized public int getNortpTimeout()  {
        
        return getNortpTimeout(nativePtr);
    }

    private native void setNortpTimeout(long nativePtr, int seconds);
    @Override
    synchronized public void setNortpTimeout(int seconds)  {
        
        setNortpTimeout(nativePtr, seconds);
    }

    private native String getPlayFile(long nativePtr);
    @Override
    synchronized public String getPlayFile()  {
        
        return getPlayFile(nativePtr);
    }

    private native void setPlayFile(long nativePtr, String file);
    @Override
    synchronized public void setPlayFile(String file)  {
        
        setPlayFile(nativePtr, file);
    }

    private native String getPlaybackDevice(long nativePtr);
    @Override
    synchronized public String getPlaybackDevice()  {
        
        return getPlaybackDevice(nativePtr);
    }

    private native int setPlaybackDevice(long nativePtr, String devid);
    @Override
    synchronized public int setPlaybackDevice(String devid)  {
        
        return setPlaybackDevice(nativePtr, devid);
    }

    private native float getPlaybackGainDb(long nativePtr);
    @Override
    synchronized public float getPlaybackGainDb()  {
        
        return getPlaybackGainDb(nativePtr);
    }

    private native void setPlaybackGainDb(long nativePtr, float level);
    @Override
    synchronized public void setPlaybackGainDb(float level)  {
        
        setPlaybackGainDb(nativePtr, level);
    }

    private native float getPreferredFramerate(long nativePtr);
    @Override
    synchronized public float getPreferredFramerate()  {
        
        return getPreferredFramerate(nativePtr);
    }

    private native void setPreferredFramerate(long nativePtr, float fps);
    @Override
    synchronized public void setPreferredFramerate(float fps)  {
        
        setPreferredFramerate(nativePtr, fps);
    }

    private native VideoDefinition getPreferredVideoDefinition(long nativePtr);
    @Override
    synchronized public VideoDefinition getPreferredVideoDefinition()  {
        
        return (VideoDefinition)getPreferredVideoDefinition(nativePtr);
    }

    private native void setPreferredVideoDefinition(long nativePtr, VideoDefinition vdef);
    @Override
    synchronized public void setPreferredVideoDefinition(VideoDefinition vdef)  {
        
        setPreferredVideoDefinition(nativePtr, vdef);
    }

    private native void setPreferredVideoSizeByName(long nativePtr, String name);
    @Override
    synchronized public void setPreferredVideoSizeByName(String name)  {
        
        setPreferredVideoSizeByName(nativePtr, name);
    }

    private native PresenceModel getPresenceModel(long nativePtr);
    @Override
    synchronized public PresenceModel getPresenceModel()  {
        
        return (PresenceModel)getPresenceModel(nativePtr);
    }

    private native void setPresenceModel(long nativePtr, PresenceModel presence);
    @Override
    synchronized public void setPresenceModel(PresenceModel presence)  {
        
        setPresenceModel(nativePtr, presence);
    }

    private native VideoDefinition getPreviewVideoDefinition(long nativePtr);
    @Override
    synchronized public VideoDefinition getPreviewVideoDefinition()  {
        
        return (VideoDefinition)getPreviewVideoDefinition(nativePtr);
    }

    private native void setPreviewVideoDefinition(long nativePtr, VideoDefinition vdef);
    @Override
    synchronized public void setPreviewVideoDefinition(VideoDefinition vdef)  {
        
        setPreviewVideoDefinition(nativePtr, vdef);
    }

    private native void setPreviewVideoSizeByName(long nativePtr, String name);
    @Override
    synchronized public void setPreviewVideoSizeByName(String name)  {
        
        setPreviewVideoSizeByName(nativePtr, name);
    }

    private native String getPrimaryContact(long nativePtr);
    @Override
    synchronized public String getPrimaryContact()  {
        
        return getPrimaryContact(nativePtr);
    }

    private native int setPrimaryContact(long nativePtr, String contact);
    @Override
    synchronized public int setPrimaryContact(String contact)  {
        
        return setPrimaryContact(nativePtr, contact);
    }

    private native Address getPrimaryContactParsed(long nativePtr);
    @Override
    synchronized public Address getPrimaryContactParsed()  {
        
        return (Address)getPrimaryContactParsed(nativePtr);
    }

    private native String getProvisioningUri(long nativePtr);
    @Override
    synchronized public String getProvisioningUri()  {
        
        return getProvisioningUri(nativePtr);
    }

    private native int setProvisioningUri(long nativePtr, String uri);
    @Override
    synchronized public int setProvisioningUri(String uri)  {
        
        return setProvisioningUri(nativePtr, uri);
    }

    private native ProxyConfig[] getProxyConfigList(long nativePtr);
    @Override
    synchronized public ProxyConfig[] getProxyConfigList()  {
        
        return getProxyConfigList(nativePtr);
    }

    private native boolean qrcodeVideoPreviewEnabled(long nativePtr);
    @Override
    synchronized public boolean qrcodeVideoPreviewEnabled()  {
        
        return qrcodeVideoPreviewEnabled(nativePtr);
    }

    private native void enableQrcodeVideoPreview(long nativePtr, boolean val);
    @Override
    synchronized public void enableQrcodeVideoPreview(boolean val)  {
        
        enableQrcodeVideoPreview(nativePtr, val);
    }

    private native boolean realtimeTextEnabled(long nativePtr);
    @Override
    synchronized public boolean realtimeTextEnabled()  {
        
        return realtimeTextEnabled(nativePtr);
    }

    private native String getRecordFile(long nativePtr);
    @Override
    synchronized public String getRecordFile()  {
        
        return getRecordFile(nativePtr);
    }

    private native void setRecordFile(long nativePtr, String file);
    @Override
    synchronized public void setRecordFile(String file)  {
        
        setRecordFile(nativePtr, file);
    }

    private native String getRemoteRingbackTone(long nativePtr);
    @Override
    synchronized public String getRemoteRingbackTone()  {
        
        return getRemoteRingbackTone(nativePtr);
    }

    private native void setRemoteRingbackTone(long nativePtr, String ring);
    @Override
    synchronized public void setRemoteRingbackTone(String ring)  {
        
        setRemoteRingbackTone(nativePtr, ring);
    }

    private native boolean retransmissionOnNackEnabled(long nativePtr);
    @Override
    synchronized public boolean retransmissionOnNackEnabled()  {
        
        return retransmissionOnNackEnabled(nativePtr);
    }

    private native void enableRetransmissionOnNack(long nativePtr, boolean val);
    @Override
    synchronized public void enableRetransmissionOnNack(boolean val)  {
        
        enableRetransmissionOnNack(nativePtr, val);
    }

    private native String getRing(long nativePtr);
    @Override
    synchronized public String getRing()  {
        
        return getRing(nativePtr);
    }

    private native void setRing(long nativePtr, String path);
    @Override
    synchronized public void setRing(String path)  {
        
        setRing(nativePtr, path);
    }

    private native boolean getRingDuringIncomingEarlyMedia(long nativePtr);
    @Override
    synchronized public boolean getRingDuringIncomingEarlyMedia()  {
        
        return getRingDuringIncomingEarlyMedia(nativePtr);
    }

    private native void setRingDuringIncomingEarlyMedia(long nativePtr, boolean enable);
    @Override
    synchronized public void setRingDuringIncomingEarlyMedia(boolean enable)  {
        
        setRingDuringIncomingEarlyMedia(nativePtr, enable);
    }

    private native String getRingback(long nativePtr);
    @Override
    synchronized public String getRingback()  {
        
        return getRingback(nativePtr);
    }

    private native void setRingback(long nativePtr, String path);
    @Override
    synchronized public void setRingback(String path)  {
        
        setRingback(nativePtr, path);
    }

    private native String getRingerDevice(long nativePtr);
    @Override
    synchronized public String getRingerDevice()  {
        
        return getRingerDevice(nativePtr);
    }

    private native int setRingerDevice(long nativePtr, String devid);
    @Override
    synchronized public int setRingerDevice(String devid)  {
        
        return setRingerDevice(nativePtr, devid);
    }

    private native String getRootCa(long nativePtr);
    @Override
    synchronized public String getRootCa()  {
        
        return getRootCa(nativePtr);
    }

    private native void setRootCa(long nativePtr, String path);
    @Override
    synchronized public void setRootCa(String path)  {
        
        setRootCa(nativePtr, path);
    }

    private native void setRootCaData(long nativePtr, String data);
    @Override
    synchronized public void setRootCaData(String data)  {
        
        setRootCaData(nativePtr, data);
    }

    private native boolean rtpBundleEnabled(long nativePtr);
    @Override
    synchronized public boolean rtpBundleEnabled()  {
        
        return rtpBundleEnabled(nativePtr);
    }

    private native void enableRtpBundle(long nativePtr, boolean value);
    @Override
    synchronized public void enableRtpBundle(boolean value)  {
        
        enableRtpBundle(nativePtr, value);
    }

    private native boolean sdp200AckEnabled(long nativePtr);
    @Override
    synchronized public boolean sdp200AckEnabled()  {
        
        return sdp200AckEnabled(nativePtr);
    }

    private native void enableSdp200Ack(long nativePtr, boolean enable);
    @Override
    synchronized public void enableSdp200Ack(boolean enable)  {
        
        enableSdp200Ack(nativePtr, enable);
    }

    private native boolean selfViewEnabled(long nativePtr);
    @Override
    synchronized public boolean selfViewEnabled()  {
        
        return selfViewEnabled(nativePtr);
    }

    private native void enableSelfView(long nativePtr, boolean val);
    @Override
    synchronized public void enableSelfView(boolean val)  {
        
        enableSelfView(nativePtr, val);
    }

    private native void enableSenderNameHiddenInForwardMessage(long nativePtr, boolean enable);
    @Override
    synchronized public void enableSenderNameHiddenInForwardMessage(boolean enable)  {
        
        enableSenderNameHiddenInForwardMessage(nativePtr, enable);
    }

    private native boolean getSessionExpiresEnabled(long nativePtr);
    @Override
    synchronized public boolean getSessionExpiresEnabled()  {
        
        return getSessionExpiresEnabled(nativePtr);
    }

    private native void setSessionExpiresEnabled(long nativePtr, boolean enabled);
    @Override
    synchronized public void setSessionExpiresEnabled(boolean enabled)  {
        
        setSessionExpiresEnabled(nativePtr, enabled);
    }

    private native int getSessionExpiresMinValue(long nativePtr);
    @Override
    synchronized public int getSessionExpiresMinValue()  {
        
        return getSessionExpiresMinValue(nativePtr);
    }

    private native void setSessionExpiresMinValue(long nativePtr, int min);
    @Override
    synchronized public void setSessionExpiresMinValue(int min)  {
        
        setSessionExpiresMinValue(nativePtr, min);
    }

    private native int getSessionExpiresRefresherValue(long nativePtr);
    @Override
    synchronized public SessionExpiresRefresher getSessionExpiresRefresherValue()  {
        
        return SessionExpiresRefresher.fromInt(getSessionExpiresRefresherValue(nativePtr));
    }

    private native void setSessionExpiresRefresherValue(long nativePtr, int refresher);
    @Override
    synchronized public void setSessionExpiresRefresherValue(SessionExpiresRefresher refresher)  {
        
        setSessionExpiresRefresherValue(nativePtr, refresher.toInt());
    }

    private native int getSessionExpiresValue(long nativePtr);
    @Override
    synchronized public int getSessionExpiresValue()  {
        
        return getSessionExpiresValue(nativePtr);
    }

    private native void setSessionExpiresValue(long nativePtr, int expires);
    @Override
    synchronized public void setSessionExpiresValue(int expires)  {
        
        setSessionExpiresValue(nativePtr, expires);
    }

    private native int getSipDscp(long nativePtr);
    @Override
    synchronized public int getSipDscp()  {
        
        return getSipDscp(nativePtr);
    }

    private native void setSipDscp(long nativePtr, int dscp);
    @Override
    synchronized public void setSipDscp(int dscp)  {
        
        setSipDscp(nativePtr, dscp);
    }

    private native void setSipNetworkReachable(long nativePtr, boolean value);
    @Override
    synchronized public void setSipNetworkReachable(boolean value)  {
        
        setSipNetworkReachable(nativePtr, value);
    }

    private native int getSipTransportTimeout(long nativePtr);
    @Override
    synchronized public int getSipTransportTimeout()  {
        
        return getSipTransportTimeout(nativePtr);
    }

    private native void setSipTransportTimeout(long nativePtr, int timeoutMs);
    @Override
    synchronized public void setSipTransportTimeout(int timeoutMs)  {
        
        setSipTransportTimeout(nativePtr, timeoutMs);
    }

    private native String[] getSoundDevicesList(long nativePtr);
    @Override
    synchronized public String[] getSoundDevicesList()  {
        
        return getSoundDevicesList(nativePtr);
    }

    private native String getStaticPicture(long nativePtr);
    @Override
    synchronized public String getStaticPicture()  {
        
        return getStaticPicture(nativePtr);
    }

    private native int setStaticPicture(long nativePtr, String path);
    @Override
    synchronized public int setStaticPicture(String path)  {
        
        return setStaticPicture(nativePtr, path);
    }

    private native float getStaticPictureFps(long nativePtr);
    @Override
    synchronized public float getStaticPictureFps()  {
        
        return getStaticPictureFps(nativePtr);
    }

    private native int setStaticPictureFps(long nativePtr, float fps);
    @Override
    synchronized public int setStaticPictureFps(float fps)  {
        
        return setStaticPictureFps(nativePtr, fps);
    }

    private native String getStunServer(long nativePtr);
    @Override
    synchronized public String getStunServer()  {
        
        return getStunServer(nativePtr);
    }

    private native void setStunServer(long nativePtr, String server);
    @Override
    synchronized public void setStunServer(String server)  {
        
        setStunServer(nativePtr, server);
    }

    private native String[] getSupportedFileFormatsList(long nativePtr);
    @Override
    synchronized public String[] getSupportedFileFormatsList()  {
        
        return getSupportedFileFormatsList(nativePtr);
    }

    private native PayloadType[] getTextPayloadTypes(long nativePtr);
    @Override
    synchronized public PayloadType[] getTextPayloadTypes()  {
        
        return getTextPayloadTypes(nativePtr);
    }

    private native void setTextPayloadTypes(long nativePtr, PayloadType[] payloadTypes);
    @Override
    synchronized public void setTextPayloadTypes(PayloadType[] payloadTypes)  {
        
        setTextPayloadTypes(nativePtr, payloadTypes);
    }

    private native int getTextPort(long nativePtr);
    @Override
    synchronized public int getTextPort()  {
        
        return getTextPort(nativePtr);
    }

    private native void setTextPort(long nativePtr, int port);
    @Override
    synchronized public void setTextPort(int port)  {
        
        setTextPort(nativePtr, port);
    }

    private native Range getTextPortsRange(long nativePtr);
    @Override
    synchronized public Range getTextPortsRange()  {
        
        return (Range)getTextPortsRange(nativePtr);
    }

    private native String getTlsCert(long nativePtr);
    @Override
    synchronized public String getTlsCert()  {
        
        return getTlsCert(nativePtr);
    }

    private native void setTlsCert(long nativePtr, String tlsCert);
    @Override
    synchronized public void setTlsCert(String tlsCert)  {
        
        setTlsCert(nativePtr, tlsCert);
    }

    private native String getTlsCertPath(long nativePtr);
    @Override
    synchronized public String getTlsCertPath()  {
        
        return getTlsCertPath(nativePtr);
    }

    private native void setTlsCertPath(long nativePtr, String tlsCertPath);
    @Override
    synchronized public void setTlsCertPath(String tlsCertPath)  {
        
        setTlsCertPath(nativePtr, tlsCertPath);
    }

    private native String getTlsKey(long nativePtr);
    @Override
    synchronized public String getTlsKey()  {
        
        return getTlsKey(nativePtr);
    }

    private native void setTlsKey(long nativePtr, String tlsKey);
    @Override
    synchronized public void setTlsKey(String tlsKey)  {
        
        setTlsKey(nativePtr, tlsKey);
    }

    private native String getTlsKeyPath(long nativePtr);
    @Override
    synchronized public String getTlsKeyPath()  {
        
        return getTlsKeyPath(nativePtr);
    }

    private native void setTlsKeyPath(long nativePtr, String tlsKeyPath);
    @Override
    synchronized public void setTlsKeyPath(String tlsKeyPath)  {
        
        setTlsKeyPath(nativePtr, tlsKeyPath);
    }

    private native Transports getTransports(long nativePtr);
    @Override
    synchronized public Transports getTransports()  {
        
        return (Transports)getTransports(nativePtr);
    }

    private native int setTransports(long nativePtr, Transports transports);
    @Override
    synchronized public int setTransports(Transports transports)  {
        
        return setTransports(nativePtr, transports);
    }

    private native Transports getTransportsUsed(long nativePtr);
    @Override
    synchronized public Transports getTransportsUsed()  {
        
        return (Transports)getTransportsUsed(nativePtr);
    }

    private native Tunnel getTunnel(long nativePtr);
    @Override
    synchronized public Tunnel getTunnel()  {
        
        return (Tunnel)getTunnel(nativePtr);
    }

    private native int getUnreadChatMessageCount(long nativePtr);
    @Override
    synchronized public int getUnreadChatMessageCount()  {
        
        return getUnreadChatMessageCount(nativePtr);
    }

    private native int getUnreadChatMessageCountFromActiveLocals(long nativePtr);
    @Override
    synchronized public int getUnreadChatMessageCountFromActiveLocals()  {
        
        return getUnreadChatMessageCountFromActiveLocals(nativePtr);
    }

    private native int getUploadBandwidth(long nativePtr);
    @Override
    synchronized public int getUploadBandwidth()  {
        
        return getUploadBandwidth(nativePtr);
    }

    private native void setUploadBandwidth(long nativePtr, int bw);
    @Override
    synchronized public void setUploadBandwidth(int bw)  {
        
        setUploadBandwidth(nativePtr, bw);
    }

    private native int getUploadPtime(long nativePtr);
    @Override
    synchronized public int getUploadPtime()  {
        
        return getUploadPtime(nativePtr);
    }

    private native void setUploadPtime(long nativePtr, int ptime);
    @Override
    synchronized public void setUploadPtime(int ptime)  {
        
        setUploadPtime(nativePtr, ptime);
    }

    private native String getUpnpExternalIpaddress(long nativePtr);
    @Override
    synchronized public String getUpnpExternalIpaddress()  {
        
        return getUpnpExternalIpaddress(nativePtr);
    }

    private native int getUpnpState(long nativePtr);
    @Override
    synchronized public UpnpState getUpnpState()  {
        
        return UpnpState.fromInt(getUpnpState(nativePtr));
    }

    private native boolean getUseFiles(long nativePtr);
    @Override
    synchronized public boolean getUseFiles()  {
        
        return getUseFiles(nativePtr);
    }

    private native void setUseFiles(long nativePtr, boolean yesno);
    @Override
    synchronized public void setUseFiles(boolean yesno)  {
        
        setUseFiles(nativePtr, yesno);
    }

    private native boolean getUseInfoForDtmf(long nativePtr);
    @Override
    synchronized public boolean getUseInfoForDtmf()  {
        
        return getUseInfoForDtmf(nativePtr);
    }

    private native void setUseInfoForDtmf(long nativePtr, boolean useInfo);
    @Override
    synchronized public void setUseInfoForDtmf(boolean useInfo)  {
        
        setUseInfoForDtmf(nativePtr, useInfo);
    }

    private native boolean getUseRfc2833ForDtmf(long nativePtr);
    @Override
    synchronized public boolean getUseRfc2833ForDtmf()  {
        
        return getUseRfc2833ForDtmf(nativePtr);
    }

    private native void setUseRfc2833ForDtmf(long nativePtr, boolean useRfc2833);
    @Override
    synchronized public void setUseRfc2833ForDtmf(boolean useRfc2833)  {
        
        setUseRfc2833ForDtmf(nativePtr, useRfc2833);
    }

    private native String getUserAgent(long nativePtr);
    @Override
    synchronized public String getUserAgent()  {
        
        return getUserAgent(nativePtr);
    }

    private native String getUserCertificatesPath(long nativePtr);
    @Override
    synchronized public String getUserCertificatesPath()  {
        
        return getUserCertificatesPath(nativePtr);
    }

    private native void setUserCertificatesPath(long nativePtr, String path);
    @Override
    synchronized public void setUserCertificatesPath(String path)  {
        
        setUserCertificatesPath(nativePtr, path);
    }

    private native VideoActivationPolicy getVideoActivationPolicy(long nativePtr);
    @Override
    synchronized public VideoActivationPolicy getVideoActivationPolicy()  {
        
        return (VideoActivationPolicy)getVideoActivationPolicy(nativePtr);
    }

    private native void setVideoActivationPolicy(long nativePtr, VideoActivationPolicy policy);
    @Override
    synchronized public void setVideoActivationPolicy(VideoActivationPolicy policy)  {
        
        setVideoActivationPolicy(nativePtr, policy);
    }

    private native boolean videoAdaptiveJittcompEnabled(long nativePtr);
    @Override
    synchronized public boolean videoAdaptiveJittcompEnabled()  {
        
        return videoAdaptiveJittcompEnabled(nativePtr);
    }

    private native void enableVideoAdaptiveJittcomp(long nativePtr, boolean enable);
    @Override
    synchronized public void enableVideoAdaptiveJittcomp(boolean enable)  {
        
        enableVideoAdaptiveJittcomp(nativePtr, enable);
    }

    private native boolean videoCaptureEnabled(long nativePtr);
    @Override
    synchronized public boolean videoCaptureEnabled()  {
        
        return videoCaptureEnabled(nativePtr);
    }

    private native void enableVideoCapture(long nativePtr, boolean enable);
    @Override
    synchronized public void enableVideoCapture(boolean enable)  {
        
        enableVideoCapture(nativePtr, enable);
    }

    private native String getVideoDevice(long nativePtr);
    @Override
    synchronized public String getVideoDevice()  {
        
        return getVideoDevice(nativePtr);
    }

    private native int setVideoDevice(long nativePtr, String id);
    @Override
    synchronized public int setVideoDevice(String id)  {
        
        return setVideoDevice(nativePtr, id);
    }

    private native String[] getVideoDevicesList(long nativePtr);
    @Override
    synchronized public String[] getVideoDevicesList()  {
        
        return getVideoDevicesList(nativePtr);
    }

    private native boolean videoDisplayEnabled(long nativePtr);
    @Override
    synchronized public boolean videoDisplayEnabled()  {
        
        return videoDisplayEnabled(nativePtr);
    }

    private native void enableVideoDisplay(long nativePtr, boolean enable);
    @Override
    synchronized public void enableVideoDisplay(boolean enable)  {
        
        enableVideoDisplay(nativePtr, enable);
    }

    private native String getVideoDisplayFilter(long nativePtr);
    @Override
    synchronized public String getVideoDisplayFilter()  {
        
        return getVideoDisplayFilter(nativePtr);
    }

    private native void setVideoDisplayFilter(long nativePtr, String filtername);
    @Override
    synchronized public void setVideoDisplayFilter(String filtername)  {
        
        setVideoDisplayFilter(nativePtr, filtername);
    }

    private native int getVideoDscp(long nativePtr);
    @Override
    synchronized public int getVideoDscp()  {
        
        return getVideoDscp(nativePtr);
    }

    private native void setVideoDscp(long nativePtr, int dscp);
    @Override
    synchronized public void setVideoDscp(int dscp)  {
        
        setVideoDscp(nativePtr, dscp);
    }

    private native boolean videoEnabled(long nativePtr);
    @Override
    synchronized public boolean videoEnabled()  {
        
        return videoEnabled(nativePtr);
    }

    private native int getVideoJittcomp(long nativePtr);
    @Override
    synchronized public int getVideoJittcomp()  {
        
        return getVideoJittcomp(nativePtr);
    }

    private native void setVideoJittcomp(long nativePtr, int milliseconds);
    @Override
    synchronized public void setVideoJittcomp(int milliseconds)  {
        
        setVideoJittcomp(nativePtr, milliseconds);
    }

    private native String getVideoMulticastAddr(long nativePtr);
    @Override
    synchronized public String getVideoMulticastAddr()  {
        
        return getVideoMulticastAddr(nativePtr);
    }

    private native int setVideoMulticastAddr(long nativePtr, String ip);
    @Override
    synchronized public int setVideoMulticastAddr(String ip)  {
        
        return setVideoMulticastAddr(nativePtr, ip);
    }

    private native boolean videoMulticastEnabled(long nativePtr);
    @Override
    synchronized public boolean videoMulticastEnabled()  {
        
        return videoMulticastEnabled(nativePtr);
    }

    private native void enableVideoMulticast(long nativePtr, boolean yesno);
    @Override
    synchronized public void enableVideoMulticast(boolean yesno)  {
        
        enableVideoMulticast(nativePtr, yesno);
    }

    private native int getVideoMulticastTtl(long nativePtr);
    @Override
    synchronized public int getVideoMulticastTtl()  {
        
        return getVideoMulticastTtl(nativePtr);
    }

    private native int setVideoMulticastTtl(long nativePtr, int ttl);
    @Override
    synchronized public int setVideoMulticastTtl(int ttl)  {
        
        return setVideoMulticastTtl(nativePtr, ttl);
    }

    private native PayloadType[] getVideoPayloadTypes(long nativePtr);
    @Override
    synchronized public PayloadType[] getVideoPayloadTypes()  {
        
        return getVideoPayloadTypes(nativePtr);
    }

    private native void setVideoPayloadTypes(long nativePtr, PayloadType[] payloadTypes);
    @Override
    synchronized public void setVideoPayloadTypes(PayloadType[] payloadTypes)  {
        
        setVideoPayloadTypes(nativePtr, payloadTypes);
    }

    private native int getVideoPort(long nativePtr);
    @Override
    synchronized public int getVideoPort()  {
        
        return getVideoPort(nativePtr);
    }

    private native void setVideoPort(long nativePtr, int port);
    @Override
    synchronized public void setVideoPort(int port)  {
        
        setVideoPort(nativePtr, port);
    }

    private native Range getVideoPortsRange(long nativePtr);
    @Override
    synchronized public Range getVideoPortsRange()  {
        
        return (Range)getVideoPortsRange(nativePtr);
    }

    private native String getVideoPreset(long nativePtr);
    @Override
    synchronized public String getVideoPreset()  {
        
        return getVideoPreset(nativePtr);
    }

    private native void setVideoPreset(long nativePtr, String preset);
    @Override
    synchronized public void setVideoPreset(String preset)  {
        
        setVideoPreset(nativePtr, preset);
    }

    private native boolean videoPreviewEnabled(long nativePtr);
    @Override
    synchronized public boolean videoPreviewEnabled()  {
        
        return videoPreviewEnabled(nativePtr);
    }

    private native void enableVideoPreview(long nativePtr, boolean val);
    @Override
    synchronized public void enableVideoPreview(boolean val)  {
        
        enableVideoPreview(nativePtr, val);
    }

    private native void enableVideoSourceReuse(long nativePtr, boolean enable);
    @Override
    synchronized public void enableVideoSourceReuse(boolean enable)  {
        
        enableVideoSourceReuse(nativePtr, enable);
    }

    private native boolean wifiOnlyEnabled(long nativePtr);
    @Override
    synchronized public boolean wifiOnlyEnabled()  {
        
        return wifiOnlyEnabled(nativePtr);
    }

    private native void enableWifiOnly(long nativePtr, boolean val);
    @Override
    synchronized public void enableWifiOnly(boolean val)  {
        
        enableWifiOnly(nativePtr, val);
    }

    private native String getZrtpSecretsFile(long nativePtr);
    @Override
    synchronized public String getZrtpSecretsFile()  {
        
        return getZrtpSecretsFile(nativePtr);
    }

    private native void setZrtpSecretsFile(long nativePtr, String file);
    @Override
    synchronized public void setZrtpSecretsFile(String file)  {
        
        setZrtpSecretsFile(nativePtr, file);
    }

    private native int acceptCall(long nativePtr, Call call);
    @Override
    synchronized public int acceptCall(Call call)  {
        
        return acceptCall(nativePtr, call);
    }

    private native int acceptCallUpdate(long nativePtr, Call call, CallParams params);
    @Override
    synchronized public int acceptCallUpdate(Call call, CallParams params)  {
        
        return acceptCallUpdate(nativePtr, call, params);
    }

    private native int acceptCallWithParams(long nativePtr, Call call, CallParams params);
    @Override
    synchronized public int acceptCallWithParams(Call call, CallParams params)  {
        
        return acceptCallWithParams(nativePtr, call, params);
    }

    private native int acceptEarlyMedia(long nativePtr, Call call);
    @Override
    synchronized public int acceptEarlyMedia(Call call)  {
        
        return acceptEarlyMedia(nativePtr, call);
    }

    private native int acceptEarlyMediaWithParams(long nativePtr, Call call, CallParams params);
    @Override
    synchronized public int acceptEarlyMediaWithParams(Call call, CallParams params)  {
        
        return acceptEarlyMediaWithParams(nativePtr, call, params);
    }

    private native int addAllToConference(long nativePtr);
    @Override
    synchronized public int addAllToConference()  {
        
        return addAllToConference(nativePtr);
    }

    private native void addAuthInfo(long nativePtr, AuthInfo info);
    @Override
    synchronized public void addAuthInfo(AuthInfo info)  {
        
        addAuthInfo(nativePtr, info);
    }

    private native void addFriend(long nativePtr, Friend fr);
    @Override
    synchronized public void addFriend(Friend fr)  {
        
        addFriend(nativePtr, fr);
    }

    private native void addFriendList(long nativePtr, FriendList list);
    @Override
    synchronized public void addFriendList(FriendList list)  {
        
        addFriendList(nativePtr, list);
    }

    private native void addLinphoneSpec(long nativePtr, String spec);
    @Override
    synchronized public void addLinphoneSpec(String spec)  {
        
        addLinphoneSpec(nativePtr, spec);
    }

    private native int addProxyConfig(long nativePtr, ProxyConfig config);
    @Override
    synchronized public int addProxyConfig(ProxyConfig config)  {
        
        return addProxyConfig(nativePtr, config);
    }

    private native void addSupportedTag(long nativePtr, String tag);
    @Override
    synchronized public void addSupportedTag(String tag)  {
        
        addSupportedTag(nativePtr, tag);
    }

    private native int addToConference(long nativePtr, Call call);
    @Override
    synchronized public int addToConference(Call call)  {
        
        return addToConference(nativePtr, call);
    }

    private native void audioSessionActivated(long nativePtr, boolean actived);
    @Override
    synchronized public void audioSessionActivated(boolean actived)  {
        
        audioSessionActivated(nativePtr, actived);
    }

    private native void callkitEnabled(long nativePtr, boolean enabled);
    @Override
    synchronized public void callkitEnabled(boolean enabled)  {
        
        callkitEnabled(nativePtr, enabled);
    }

    private native void checkForUpdate(long nativePtr, String currentVersion);
    @Override
    synchronized public void checkForUpdate(String currentVersion)  {
        
        checkForUpdate(nativePtr, currentVersion);
    }

    private native void clearAllAuthInfo(long nativePtr);
    @Override
    synchronized public void clearAllAuthInfo()  {
        
        clearAllAuthInfo(nativePtr);
    }

    private native void clearCallLogs(long nativePtr);
    @Override
    synchronized public void clearCallLogs()  {
        
        clearCallLogs(nativePtr);
    }

    private native void clearProxyConfig(long nativePtr);
    @Override
    synchronized public void clearProxyConfig()  {
        
        clearProxyConfig(nativePtr);
    }

    private native AccountCreator createAccountCreator(long nativePtr, String xmlrpcUrl);
    @Override
    synchronized public AccountCreator createAccountCreator(String xmlrpcUrl)  {
        
        return (AccountCreator)createAccountCreator(nativePtr, xmlrpcUrl);
    }

    private native Address createAddress(long nativePtr, String address);
    @Override
    synchronized public Address createAddress(String address)  {
        
        return (Address)createAddress(nativePtr, address);
    }

    private native AuthInfo createAuthInfo(long nativePtr, String username, String userid, String passwd, String ha1, String realm, String domain);
    @Override
    synchronized public AuthInfo createAuthInfo(String username, String userid, String passwd, String ha1, String realm, String domain)  {
        
        return (AuthInfo)createAuthInfo(nativePtr, username, userid, passwd, ha1, realm, domain);
    }

    private native CallLog createCallLog(long nativePtr, Address from, Address to, int dir, int duration, long startTime, long connectedTime, int status, boolean videoEnabled, float quality);
    @Override
    synchronized public CallLog createCallLog(Address from, Address to, Call.Dir dir, int duration, long startTime, long connectedTime, Call.Status status, boolean videoEnabled, float quality)  {
        
        return (CallLog)createCallLog(nativePtr, from, to, dir.toInt(), duration, startTime, connectedTime, status.toInt(), videoEnabled, quality);
    }

    private native CallParams createCallParams(long nativePtr, Call call);
    @Override
    synchronized public CallParams createCallParams(@Nullable Call call)  {
        
        return (CallParams)createCallParams(nativePtr, call);
    }

    private native ChatRoom createChatRoom2(long nativePtr, ChatRoomParams params, String subject, Address[] participants);
    @Override
    synchronized public ChatRoom createChatRoom(ChatRoomParams params, String subject, Address[] participants)  {
        
        return (ChatRoom)createChatRoom2(nativePtr, params, subject, participants);
    }

    private native ChatRoom createChatRoom3(long nativePtr, String subject, Address[] participants);
    @Override
    synchronized public ChatRoom createChatRoom(String subject, Address[] participants)  {
        
        return (ChatRoom)createChatRoom3(nativePtr, subject, participants);
    }

    private native ChatRoom createChatRoom4(long nativePtr, ChatRoomParams params, Address localAddr, Address participant);
    @Override
    synchronized public ChatRoom createChatRoom(ChatRoomParams params, Address localAddr, Address participant)  {
        
        return (ChatRoom)createChatRoom4(nativePtr, params, localAddr, participant);
    }

    private native ChatRoom createChatRoom5(long nativePtr, Address participant);
    @Override
    synchronized public ChatRoom createChatRoom(Address participant)  {
        
        return (ChatRoom)createChatRoom5(nativePtr, participant);
    }

    private native ChatRoom createChatRoom(long nativePtr, ChatRoomParams params, Address localAddr, String subject, Address[] participants);
    @Override
    synchronized public ChatRoom createChatRoom(ChatRoomParams params, Address localAddr, String subject, Address[] participants)  {
        
        return (ChatRoom)createChatRoom(nativePtr, params, localAddr, subject, participants);
    }

    private native ChatRoom createClientGroupChatRoom(long nativePtr, String subject, boolean fallback);
    @Override
    synchronized public ChatRoom createClientGroupChatRoom(String subject, boolean fallback)  {
        
        return (ChatRoom)createClientGroupChatRoom(nativePtr, subject, fallback);
    }

    private native ChatRoom createClientGroupChatRoom2(long nativePtr, String subject, boolean fallback, boolean encrypted);
    @Override
    synchronized public ChatRoom createClientGroupChatRoom(String subject, boolean fallback, boolean encrypted)  {
        
        return (ChatRoom)createClientGroupChatRoom2(nativePtr, subject, fallback, encrypted);
    }

    private native ConferenceParams createConferenceParams(long nativePtr);
    @Override
    synchronized public ConferenceParams createConferenceParams()  {
        
        return (ConferenceParams)createConferenceParams(nativePtr);
    }

    private native Conference createConferenceWithParams(long nativePtr, ConferenceParams params);
    @Override
    synchronized public Conference createConferenceWithParams(ConferenceParams params)  {
        
        return (Conference)createConferenceWithParams(nativePtr, params);
    }

    private native Config createConfig(long nativePtr, String filename);
    @Override
    synchronized public Config createConfig(String filename)  {
        
        return (Config)createConfig(nativePtr, filename);
    }

    private native Content createContent(long nativePtr);
    @Override
    synchronized public Content createContent()  {
        
        return (Content)createContent(nativePtr);
    }

    private native ChatRoomParams createDefaultChatRoomParams(long nativePtr);
    @Override
    synchronized public ChatRoomParams createDefaultChatRoomParams()  {
        
        return (ChatRoomParams)createDefaultChatRoomParams(nativePtr);
    }

    private native Friend createFriend(long nativePtr);
    @Override
    synchronized public Friend createFriend()  {
        
        return (Friend)createFriend(nativePtr);
    }

    private native FriendList createFriendList(long nativePtr);
    @Override
    synchronized public FriendList createFriendList()  {
        
        return (FriendList)createFriendList(nativePtr);
    }

    private native Friend createFriendWithAddress(long nativePtr, String address);
    @Override
    synchronized public Friend createFriendWithAddress(String address)  {
        
        return (Friend)createFriendWithAddress(nativePtr, address);
    }

    private native InfoMessage createInfoMessage(long nativePtr);
    @Override
    synchronized public InfoMessage createInfoMessage()  {
        
        return (InfoMessage)createInfoMessage(nativePtr);
    }

    private native Player createLocalPlayer(long nativePtr, String soundCardName, String videoDisplayName, Object windowId);
    @Override
    synchronized public Player createLocalPlayer(String soundCardName, String videoDisplayName, Object windowId)  {
        
        return (Player)createLocalPlayer(nativePtr, soundCardName, videoDisplayName, windowId);
    }

    private native MagicSearch createMagicSearch(long nativePtr);
    @Override
    synchronized public MagicSearch createMagicSearch()  {
        
        return (MagicSearch)createMagicSearch(nativePtr);
    }

    private native NatPolicy createNatPolicy(long nativePtr);
    @Override
    synchronized public NatPolicy createNatPolicy()  {
        
        return (NatPolicy)createNatPolicy(nativePtr);
    }

    private native NatPolicy createNatPolicyFromConfig(long nativePtr, String ref);
    @Override
    synchronized public NatPolicy createNatPolicyFromConfig(String ref)  {
        
        return (NatPolicy)createNatPolicyFromConfig(nativePtr, ref);
    }

    private native Event createNotify(long nativePtr, Address resource, String event);
    @Override
    synchronized public Event createNotify(Address resource, String event)  {
        
        return (Event)createNotify(nativePtr, resource, event);
    }

    private native Event createOneShotPublish(long nativePtr, Address resource, String event);
    @Override
    synchronized public Event createOneShotPublish(Address resource, String event)  {
        
        return (Event)createOneShotPublish(nativePtr, resource, event);
    }

    private native PresenceActivity createPresenceActivity(long nativePtr, int acttype, String description);
    @Override
    synchronized public PresenceActivity createPresenceActivity(PresenceActivity.Type acttype, String description)  {
        
        return (PresenceActivity)createPresenceActivity(nativePtr, acttype.toInt(), description);
    }

    private native PresenceModel createPresenceModel(long nativePtr);
    @Override
    synchronized public PresenceModel createPresenceModel()  {
        
        return (PresenceModel)createPresenceModel(nativePtr);
    }

    private native PresenceModel createPresenceModelWithActivity(long nativePtr, int acttype, String description);
    @Override
    synchronized public PresenceModel createPresenceModelWithActivity(PresenceActivity.Type acttype, String description)  {
        
        return (PresenceModel)createPresenceModelWithActivity(nativePtr, acttype.toInt(), description);
    }

    private native PresenceModel createPresenceModelWithActivityAndNote(long nativePtr, int acttype, String description, String note, String lang);
    @Override
    synchronized public PresenceModel createPresenceModelWithActivityAndNote(PresenceActivity.Type acttype, String description, String note, String lang)  {
        
        return (PresenceModel)createPresenceModelWithActivityAndNote(nativePtr, acttype.toInt(), description, note, lang);
    }

    private native PresenceNote createPresenceNote(long nativePtr, String content, String lang);
    @Override
    synchronized public PresenceNote createPresenceNote(String content, String lang)  {
        
        return (PresenceNote)createPresenceNote(nativePtr, content, lang);
    }

    private native PresencePerson createPresencePerson(long nativePtr, String id);
    @Override
    synchronized public PresencePerson createPresencePerson(String id)  {
        
        return (PresencePerson)createPresencePerson(nativePtr, id);
    }

    private native PresenceService createPresenceService(long nativePtr, String id, int basicStatus, String contact);
    @Override
    synchronized public PresenceService createPresenceService(String id, PresenceBasicStatus basicStatus, String contact)  {
        
        return (PresenceService)createPresenceService(nativePtr, id, basicStatus.toInt(), contact);
    }

    private native Address createPrimaryContactParsed(long nativePtr);
    @Override
    synchronized public Address createPrimaryContactParsed()  {
        
        return (Address)createPrimaryContactParsed(nativePtr);
    }

    private native ProxyConfig createProxyConfig(long nativePtr);
    @Override
    synchronized public ProxyConfig createProxyConfig()  {
        
        return (ProxyConfig)createProxyConfig(nativePtr);
    }

    private native Event createPublish(long nativePtr, Address resource, String event, int expires);
    @Override
    synchronized public Event createPublish(Address resource, String event, int expires)  {
        
        return (Event)createPublish(nativePtr, resource, event, expires);
    }

    private native Event createSubscribe2(long nativePtr, Address resource, ProxyConfig proxy, String event, int expires);
    @Override
    synchronized public Event createSubscribe(Address resource, ProxyConfig proxy, String event, int expires)  {
        
        return (Event)createSubscribe2(nativePtr, resource, proxy, event, expires);
    }

    private native Event createSubscribe(long nativePtr, Address resource, String event, int expires);
    @Override
    synchronized public Event createSubscribe(Address resource, String event, int expires)  {
        
        return (Event)createSubscribe(nativePtr, resource, event, expires);
    }

    private native XmlRpcSession createXmlRpcSession(long nativePtr, String url);
    @Override
    synchronized public XmlRpcSession createXmlRpcSession(String url)  {
        
        return (XmlRpcSession)createXmlRpcSession(nativePtr, url);
    }

    private native int declineCall(long nativePtr, Call call, int reason);
    @Override
    synchronized public int declineCall(Call call, Reason reason)  {
        
        return declineCall(nativePtr, call, reason.toInt());
    }

    private native int deferCallUpdate(long nativePtr, Call call);
    @Override
    synchronized public int deferCallUpdate(Call call)  {
        
        return deferCallUpdate(nativePtr, call);
    }

    private native void deleteChatRoom(long nativePtr, ChatRoom cr);
    @Override
    synchronized public void deleteChatRoom(ChatRoom cr)  {
        
        deleteChatRoom(nativePtr, cr);
    }

    private native void disableChat(long nativePtr, int denyReason);
    @Override
    synchronized public void disableChat(Reason denyReason)  {
        
        disableChat(nativePtr, denyReason.toInt());
    }

    private native void enableChat(long nativePtr);
    @Override
    synchronized public void enableChat()  {
        
        enableChat(nativePtr);
    }

    private native void enableLime(long nativePtr, int val);
    @Override
    synchronized public void enableLime(LimeState val)  {
        
        enableLime(nativePtr, val.toInt());
    }

    private native void ensureRegistered(long nativePtr);
    @Override
    synchronized public void ensureRegistered()  {
        
        ensureRegistered(nativePtr);
    }

    private native void enterBackground(long nativePtr);
    @Override
    synchronized public void enterBackground()  {
        
        enterBackground(nativePtr);
    }

    private native int enterConference(long nativePtr);
    @Override
    synchronized public int enterConference()  {
        
        return enterConference(nativePtr);
    }

    private native void enterForeground(long nativePtr);
    @Override
    synchronized public void enterForeground()  {
        
        enterForeground(nativePtr);
    }

    private native boolean fileFormatSupported(long nativePtr, String fmt);
    @Override
    synchronized public boolean fileFormatSupported(String fmt)  {
        
        return fileFormatSupported(nativePtr, fmt);
    }

    private native AuthInfo findAuthInfo(long nativePtr, String realm, String username, String sipDomain);
    @Override
    synchronized public AuthInfo findAuthInfo(String realm, String username, String sipDomain)  {
        
        return (AuthInfo)findAuthInfo(nativePtr, realm, username, sipDomain);
    }

    private native Call findCallFromUri(long nativePtr, String uri);
    @Override
    synchronized public Call findCallFromUri(String uri)  {
        
        return (Call)findCallFromUri(nativePtr, uri);
    }

    private native CallLog findCallLogFromCallId(long nativePtr, String callId);
    @Override
    synchronized public CallLog findCallLogFromCallId(String callId)  {
        
        return (CallLog)findCallLogFromCallId(nativePtr, callId);
    }

    private native ChatRoom findChatRoom(long nativePtr, Address peerAddr, Address localAddr);
    @Override
    synchronized public ChatRoom findChatRoom(Address peerAddr, Address localAddr)  {
        
        return (ChatRoom)findChatRoom(nativePtr, peerAddr, localAddr);
    }

    private native Address[] findContactsByChar(long nativePtr, String filter, boolean sipOnly);
    @Override
    synchronized public Address[] findContactsByChar(String filter, boolean sipOnly)  {
        
        return findContactsByChar(nativePtr, filter, sipOnly);
    }

    private native Friend findFriend(long nativePtr, Address addr);
    @Override
    synchronized public Friend findFriend(Address addr)  {
        
        return (Friend)findFriend(nativePtr, addr);
    }

    private native Friend[] findFriends(long nativePtr, Address addr);
    @Override
    synchronized public Friend[] findFriends(Address addr)  {
        
        return findFriends(nativePtr, addr);
    }

    private native ChatRoom findOneToOneChatRoom(long nativePtr, Address localAddr, Address participantAddr);
    @Override
    synchronized public ChatRoom findOneToOneChatRoom(Address localAddr, Address participantAddr)  {
        
        return (ChatRoom)findOneToOneChatRoom(nativePtr, localAddr, participantAddr);
    }

    private native ChatRoom findOneToOneChatRoom2(long nativePtr, Address localAddr, Address participantAddr, boolean encrypted);
    @Override
    synchronized public ChatRoom findOneToOneChatRoom(Address localAddr, Address participantAddr, boolean encrypted)  {
        
        return (ChatRoom)findOneToOneChatRoom2(nativePtr, localAddr, participantAddr, encrypted);
    }

    private native Call getCallByRemoteAddress(long nativePtr, String remoteAddress);
    @Override @Nullable
    synchronized public Call getCallByRemoteAddress(String remoteAddress)  {
        
        return (Call)getCallByRemoteAddress(nativePtr, remoteAddress);
    }

    private native Call getCallByRemoteAddress22(long nativePtr, Address remoteAddress);
    @Override
    synchronized public Call getCallByRemoteAddress2(Address remoteAddress)  {
        
        return (Call)getCallByRemoteAddress22(nativePtr, remoteAddress);
    }

    private native CallLog[] getCallHistory2(long nativePtr, Address peerAddr, Address localAddr);
    @Override
    synchronized public CallLog[] getCallHistory(Address peerAddr, Address localAddr)  {
        
        return getCallHistory2(nativePtr, peerAddr, localAddr);
    }

    private native CallLog[] getCallHistoryForAddress(long nativePtr, Address addr);
    @Override
    synchronized public CallLog[] getCallHistoryForAddress(Address addr)  {
        
        return getCallHistoryForAddress(nativePtr, addr);
    }

    private native ChatRoom getChatRoom2(long nativePtr, Address peerAddr, Address localAddr);
    @Override
    synchronized public ChatRoom getChatRoom(Address peerAddr, Address localAddr)  {
        
        return (ChatRoom)getChatRoom2(nativePtr, peerAddr, localAddr);
    }

    private native ChatRoom getChatRoom(long nativePtr, Address addr);
    @Override
    synchronized public ChatRoom getChatRoom(Address addr)  {
        
        return (ChatRoom)getChatRoom(nativePtr, addr);
    }

    private native ChatRoom getChatRoomFromUri(long nativePtr, String to);
    @Override
    synchronized public ChatRoom getChatRoomFromUri(String to)  {
        
        return (ChatRoom)getChatRoomFromUri(nativePtr, to);
    }

    private native Friend getFriendByRefKey(long nativePtr, String key);
    @Override
    synchronized public Friend getFriendByRefKey(String key)  {
        
        return (Friend)getFriendByRefKey(nativePtr, key);
    }

    private native FriendList getFriendListByName(long nativePtr, String name);
    @Override
    synchronized public FriendList getFriendListByName(String name)  {
        
        return (FriendList)getFriendListByName(nativePtr, name);
    }

    private native PayloadType getPayloadType(long nativePtr, String type, int rate, int channels);
    @Override
    synchronized public PayloadType getPayloadType(String type, int rate, int channels)  {
        
        return (PayloadType)getPayloadType(nativePtr, type, rate, channels);
    }

    private native ProxyConfig getProxyConfigByIdkey(long nativePtr, String idkey);
    @Override
    synchronized public ProxyConfig getProxyConfigByIdkey(String idkey)  {
        
        return (ProxyConfig)getProxyConfigByIdkey(nativePtr, idkey);
    }

    private native int getUnreadChatMessageCountFromLocal(long nativePtr, Address address);
    @Override
    synchronized public int getUnreadChatMessageCountFromLocal(Address address)  {
        
        return getUnreadChatMessageCountFromLocal(nativePtr, address);
    }

    private native int getZrtpStatus(long nativePtr, String addr);
    @Override
    synchronized public ZrtpPeerStatus getZrtpStatus(String addr)  {
        
        return ZrtpPeerStatus.fromInt(getZrtpStatus(nativePtr, addr));
    }

    private native boolean hasBuiltinEchoCanceller(long nativePtr);
    @Override
    synchronized public boolean hasBuiltinEchoCanceller()  {
        
        return hasBuiltinEchoCanceller(nativePtr);
    }

    private native boolean hasCrappyOpengl(long nativePtr);
    @Override
    synchronized public boolean hasCrappyOpengl()  {
        
        return hasCrappyOpengl(nativePtr);
    }

    private native boolean inCall(long nativePtr);
    @Override
    synchronized public boolean inCall()  {
        
        return inCall(nativePtr);
    }

    private native Address interpretUrl(long nativePtr, String url);
    @Override
    synchronized public Address interpretUrl(String url)  {
        
        return (Address)interpretUrl(nativePtr, url);
    }

    private native Call invite(long nativePtr, String url);
    @Override
    synchronized public Call invite(String url)  {
        
        return (Call)invite(nativePtr, url);
    }

    private native Call inviteAddress(long nativePtr, Address addr);
    @Override
    synchronized public Call inviteAddress(Address addr)  {
        
        return (Call)inviteAddress(nativePtr, addr);
    }

    private native Call inviteAddressWithParams(long nativePtr, Address addr, CallParams params);
    @Override
    synchronized public Call inviteAddressWithParams(Address addr, CallParams params)  {
        
        return (Call)inviteAddressWithParams(nativePtr, addr, params);
    }

    private native Call inviteWithParams(long nativePtr, String url, CallParams params);
    @Override
    synchronized public Call inviteWithParams(String url, CallParams params)  {
        
        return (Call)inviteWithParams(nativePtr, url, params);
    }

    private native boolean isMediaFilterSupported(long nativePtr, String filtername);
    @Override
    synchronized public boolean isMediaFilterSupported(String filtername)  {
        
        return isMediaFilterSupported(nativePtr, filtername);
    }

    private native void iterate(long nativePtr);
    @Override
    synchronized public void iterate()  {
        
        iterate(nativePtr);
    }

    private native int leaveConference(long nativePtr);
    @Override
    synchronized public int leaveConference()  {
        
        return leaveConference(nativePtr);
    }

    private native boolean limeAvailable(long nativePtr);
    @Override
    synchronized public boolean limeAvailable()  {
        
        return limeAvailable(nativePtr);
    }

    private native int limeEnabled(long nativePtr);
    @Override
    synchronized public LimeState limeEnabled()  {
        
        return LimeState.fromInt(limeEnabled(nativePtr));
    }

    private native boolean limeX3DhAvailable(long nativePtr);
    @Override
    synchronized public boolean limeX3DhAvailable()  {
        
        return limeX3DhAvailable(nativePtr);
    }

    private native void loadConfigFromXml(long nativePtr, String xmlUri);
    @Override
    synchronized public void loadConfigFromXml(String xmlUri)  {
        
        loadConfigFromXml(nativePtr, xmlUri);
    }

    private native boolean mediaEncryptionSupported(long nativePtr, int menc);
    @Override
    synchronized public boolean mediaEncryptionSupported(MediaEncryption menc)  {
        
        return mediaEncryptionSupported(nativePtr, menc.toInt());
    }

    private native void migrateLogsFromRcToDb(long nativePtr);
    @Override
    synchronized public void migrateLogsFromRcToDb()  {
        
        migrateLogsFromRcToDb(nativePtr);
    }

    private native int migrateToMultiTransport(long nativePtr);
    @Override
    synchronized public int migrateToMultiTransport()  {
        
        return migrateToMultiTransport(nativePtr);
    }

    private native void notifyAllFriends(long nativePtr, PresenceModel presence);
    @Override
    synchronized public void notifyAllFriends(PresenceModel presence)  {
        
        notifyAllFriends(nativePtr, presence);
    }

    private native void notifyNotifyPresenceReceived(long nativePtr, Friend lf);
    @Override
    synchronized public void notifyNotifyPresenceReceived(Friend lf)  {
        
        notifyNotifyPresenceReceived(nativePtr, lf);
    }

    private native void notifyNotifyPresenceReceivedForUriOrTel(long nativePtr, Friend lf, String uriOrTel, PresenceModel presenceModel);
    @Override
    synchronized public void notifyNotifyPresenceReceivedForUriOrTel(Friend lf, String uriOrTel, PresenceModel presenceModel)  {
        
        notifyNotifyPresenceReceivedForUriOrTel(nativePtr, lf, uriOrTel, presenceModel);
    }

    private native int pauseAllCalls(long nativePtr);
    @Override
    synchronized public int pauseAllCalls()  {
        
        return pauseAllCalls(nativePtr);
    }

    private native int pauseCall(long nativePtr, Call call);
    @Override
    synchronized public int pauseCall(Call call)  {
        
        return pauseCall(nativePtr, call);
    }

    private native void playDtmf(long nativePtr, char dtmf, int durationMs);
    @Override
    synchronized public void playDtmf(char dtmf, int durationMs)  {
        
        playDtmf(nativePtr, dtmf, durationMs);
    }

    private native int playLocal(long nativePtr, String audiofile);
    @Override
    synchronized public int playLocal(String audiofile)  {
        
        return playLocal(nativePtr, audiofile);
    }

    private native void previewOglRender(long nativePtr);
    @Override
    synchronized public void previewOglRender()  {
        
        previewOglRender(nativePtr);
    }

    private native Event publish(long nativePtr, Address resource, String event, int expires, Content body);
    @Override
    synchronized public Event publish(Address resource, String event, int expires, Content body)  {
        
        return (Event)publish(nativePtr, resource, event, expires, body);
    }

    private native int realtimeTextGetKeepaliveInterval(long nativePtr);
    @Override
    synchronized public int realtimeTextGetKeepaliveInterval()  {
        
        return realtimeTextGetKeepaliveInterval(nativePtr);
    }

    private native void realtimeTextSetKeepaliveInterval(long nativePtr, int interval);
    @Override
    synchronized public void realtimeTextSetKeepaliveInterval(int interval)  {
        
        realtimeTextSetKeepaliveInterval(nativePtr, interval);
    }

    private native int redirectCall(long nativePtr, Call call, String redirectUri);
    @Override
    synchronized public int redirectCall(Call call, String redirectUri)  {
        
        return redirectCall(nativePtr, call, redirectUri);
    }

    private native void refreshRegisters(long nativePtr);
    @Override
    synchronized public void refreshRegisters()  {
        
        refreshRegisters(nativePtr);
    }

    private native void rejectSubscriber(long nativePtr, Friend lf);
    @Override
    synchronized public void rejectSubscriber(Friend lf)  {
        
        rejectSubscriber(nativePtr, lf);
    }

    private native void reloadMsPlugins(long nativePtr, String path);
    @Override
    synchronized public void reloadMsPlugins(String path)  {
        
        reloadMsPlugins(nativePtr, path);
    }

    private native void reloadSoundDevices(long nativePtr);
    @Override
    synchronized public void reloadSoundDevices()  {
        
        reloadSoundDevices(nativePtr);
    }

    private native void reloadVideoDevices(long nativePtr);
    @Override
    synchronized public void reloadVideoDevices()  {
        
        reloadVideoDevices(nativePtr);
    }

    private native void removeAuthInfo(long nativePtr, AuthInfo info);
    @Override
    synchronized public void removeAuthInfo(AuthInfo info)  {
        
        removeAuthInfo(nativePtr, info);
    }

    private native void removeCallLog(long nativePtr, CallLog callLog);
    @Override
    synchronized public void removeCallLog(CallLog callLog)  {
        
        removeCallLog(nativePtr, callLog);
    }

    private native void removeFriendList(long nativePtr, FriendList list);
    @Override
    synchronized public void removeFriendList(FriendList list)  {
        
        removeFriendList(nativePtr, list);
    }

    private native int removeFromConference(long nativePtr, Call call);
    @Override
    synchronized public int removeFromConference(Call call)  {
        
        return removeFromConference(nativePtr, call);
    }

    private native void removeLinphoneSpec(long nativePtr, String spec);
    @Override
    synchronized public void removeLinphoneSpec(String spec)  {
        
        removeLinphoneSpec(nativePtr, spec);
    }

    private native void removeProxyConfig(long nativePtr, ProxyConfig config);
    @Override
    synchronized public void removeProxyConfig(ProxyConfig config)  {
        
        removeProxyConfig(nativePtr, config);
    }

    private native void removeSupportedTag(long nativePtr, String tag);
    @Override
    synchronized public void removeSupportedTag(String tag)  {
        
        removeSupportedTag(nativePtr, tag);
    }

    private native void resetMissedCallsCount(long nativePtr);
    @Override
    synchronized public void resetMissedCallsCount()  {
        
        resetMissedCallsCount(nativePtr);
    }

    private native int resumeCall(long nativePtr, Call call);
    @Override
    synchronized public int resumeCall(Call call)  {
        
        return resumeCall(nativePtr, call);
    }

    private native void setAudioPortRange(long nativePtr, int minPort, int maxPort);
    @Override
    synchronized public void setAudioPortRange(int minPort, int maxPort)  {
        
        setAudioPortRange(nativePtr, minPort, maxPort);
    }

    private native void setCallErrorTone(long nativePtr, int reason, String audiofile);
    @Override
    synchronized public void setCallErrorTone(Reason reason, String audiofile)  {
        
        setCallErrorTone(nativePtr, reason.toInt(), audiofile);
    }

    private native void setQrcodeDecodeRect(long nativePtr, int x, int y, int w, int h);
    @Override
    synchronized public void setQrcodeDecodeRect(int x, int y, int w, int h)  {
        
        setQrcodeDecodeRect(nativePtr, x, y, w, h);
    }

    private native void setTextPortRange(long nativePtr, int minPort, int maxPort);
    @Override
    synchronized public void setTextPortRange(int minPort, int maxPort)  {
        
        setTextPortRange(nativePtr, minPort, maxPort);
    }

    private native void setUserAgent(long nativePtr, String uaName, String version);
    @Override
    synchronized public void setUserAgent(String uaName, String version)  {
        
        setUserAgent(nativePtr, uaName, version);
    }

    private native void setVideoPortRange(long nativePtr, int minPort, int maxPort);
    @Override
    synchronized public void setVideoPortRange(int minPort, int maxPort)  {
        
        setVideoPortRange(nativePtr, minPort, maxPort);
    }

    private native boolean soundDeviceCanCapture(long nativePtr, String device);
    @Override
    synchronized public boolean soundDeviceCanCapture(String device)  {
        
        return soundDeviceCanCapture(nativePtr, device);
    }

    private native boolean soundDeviceCanPlayback(long nativePtr, String device);
    @Override
    synchronized public boolean soundDeviceCanPlayback(String device)  {
        
        return soundDeviceCanPlayback(nativePtr, device);
    }

    private native boolean soundResourcesLocked(long nativePtr);
    @Override
    synchronized public boolean soundResourcesLocked()  {
        
        return soundResourcesLocked(nativePtr);
    }

    private native int start(long nativePtr);
    @Override
    synchronized public int start()  {
        
        return start(nativePtr);
    }

    private native int startConferenceRecording(long nativePtr, String path);
    @Override
    synchronized public int startConferenceRecording(String path)  {
        
        return startConferenceRecording(nativePtr, path);
    }

    private native void startDtmfStream(long nativePtr);
    @Override
    synchronized public void startDtmfStream()  {
        
        startDtmfStream(nativePtr);
    }

    private native int startEchoCancellerCalibration(long nativePtr);
    @Override
    synchronized public int startEchoCancellerCalibration()  {
        
        return startEchoCancellerCalibration(nativePtr);
    }

    private native int startEchoTester(long nativePtr, int rate);
    @Override
    synchronized public int startEchoTester(int rate)  {
        
        return startEchoTester(nativePtr, rate);
    }

    private native void stop(long nativePtr);
    @Override
    synchronized public void stop()  {
        
        stop(nativePtr);
    }

    private native void stopAsync(long nativePtr);
    @Override
    synchronized public void stopAsync()  {
        
        stopAsync(nativePtr);
    }

    private native int stopConferenceRecording(long nativePtr);
    @Override
    synchronized public int stopConferenceRecording()  {
        
        return stopConferenceRecording(nativePtr);
    }

    private native void stopDtmf(long nativePtr);
    @Override
    synchronized public void stopDtmf()  {
        
        stopDtmf(nativePtr);
    }

    private native void stopDtmfStream(long nativePtr);
    @Override
    synchronized public void stopDtmfStream()  {
        
        stopDtmfStream(nativePtr);
    }

    private native int stopEchoTester(long nativePtr);
    @Override
    synchronized public int stopEchoTester()  {
        
        return stopEchoTester(nativePtr);
    }

    private native void stopRinging(long nativePtr);
    @Override
    synchronized public void stopRinging()  {
        
        stopRinging(nativePtr);
    }

    private native Event subscribe(long nativePtr, Address resource, String event, int expires, Content body);
    @Override
    synchronized public Event subscribe(Address resource, String event, int expires, @Nullable Content body)  {
        
        return (Event)subscribe(nativePtr, resource, event, expires, body);
    }

    private native int takePreviewSnapshot(long nativePtr, String file);
    @Override
    synchronized public int takePreviewSnapshot(String file)  {
        
        return takePreviewSnapshot(nativePtr, file);
    }

    private native int terminateAllCalls(long nativePtr);
    @Override
    synchronized public int terminateAllCalls()  {
        
        return terminateAllCalls(nativePtr);
    }

    private native int terminateCall(long nativePtr, Call call);
    @Override
    synchronized public int terminateCall(Call call)  {
        
        return terminateCall(nativePtr, call);
    }

    private native int terminateConference(long nativePtr);
    @Override
    synchronized public int terminateConference()  {
        
        return terminateConference(nativePtr);
    }

    private native int transferCall(long nativePtr, Call call, String referTo);
    @Override
    synchronized public int transferCall(Call call, String referTo)  {
        
        return transferCall(nativePtr, call, referTo);
    }

    private native int transferCallToAnother(long nativePtr, Call call, Call dest);
    @Override
    synchronized public int transferCallToAnother(Call call, Call dest)  {
        
        return transferCallToAnother(nativePtr, call, dest);
    }

    private native int updateCall(long nativePtr, Call call, CallParams params);
    @Override
    synchronized public int updateCall(Call call, CallParams params)  {
        
        return updateCall(nativePtr, call, params);
    }

    private native void uploadLogCollection(long nativePtr);
    @Override
    synchronized public void uploadLogCollection()  {
        
        uploadLogCollection(nativePtr);
    }

    private native void usePreviewWindow(long nativePtr, boolean yesno);
    @Override
    synchronized public void usePreviewWindow(boolean yesno)  {
        
        usePreviewWindow(nativePtr, yesno);
    }

    private native void verifyServerCertificates(long nativePtr, boolean yesno);
    @Override
    synchronized public void verifyServerCertificates(boolean yesno)  {
        
        verifyServerCertificates(nativePtr, yesno);
    }

    private native void verifyServerCn(long nativePtr, boolean yesno);
    @Override
    synchronized public void verifyServerCn(boolean yesno)  {
        
        verifyServerCn(nativePtr, yesno);
    }

    private native boolean videoSupported(long nativePtr);
    @Override
    synchronized public boolean videoSupported()  {
        
        return videoSupported(nativePtr);
    }

    private native String compressLogCollection(long nativePtr);
    @Override
    synchronized public String compressLogCollection()  {
        
        return compressLogCollection(nativePtr);
    }

    private native void enableLogCollection(long nativePtr, int state);
    @Override
    synchronized public void enableLogCollection(LogCollectionState state)  {
        
        enableLogCollection(nativePtr, state.toInt());
    }

    private native int getLogCollectionMaxFileSize(long nativePtr);
    @Override
    synchronized public int getLogCollectionMaxFileSize()  {
        
        return getLogCollectionMaxFileSize(nativePtr);
    }

    private native String getLogCollectionPath(long nativePtr);
    @Override
    synchronized public String getLogCollectionPath()  {
        
        return getLogCollectionPath(nativePtr);
    }

    private native String getLogCollectionPrefix(long nativePtr);
    @Override
    synchronized public String getLogCollectionPrefix()  {
        
        return getLogCollectionPrefix(nativePtr);
    }

    private native int getLogLevelMask(long nativePtr);
    @Override
    synchronized public int getLogLevelMask()  {
        
        return getLogLevelMask(nativePtr);
    }

    private native String getVersion(long nativePtr);
    @Override
    synchronized public String getVersion()  {
        
        return getVersion(nativePtr);
    }

    private native int logCollectionEnabled(long nativePtr);
    @Override
    synchronized public LogCollectionState logCollectionEnabled()  {
        
        return LogCollectionState.fromInt(logCollectionEnabled(nativePtr));
    }

    private native void resetLogCollection(long nativePtr);
    @Override
    synchronized public void resetLogCollection()  {
        
        resetLogCollection(nativePtr);
    }

    private native void serializeLogs(long nativePtr);
    @Override
    synchronized public void serializeLogs()  {
        
        serializeLogs(nativePtr);
    }

    private native void setLogCollectionMaxFileSize(long nativePtr, int size);
    @Override
    synchronized public void setLogCollectionMaxFileSize(int size)  {
        
        setLogCollectionMaxFileSize(nativePtr, size);
    }

    private native void setLogCollectionPath(long nativePtr, String path);
    @Override
    synchronized public void setLogCollectionPath(String path)  {
        
        setLogCollectionPath(nativePtr, path);
    }

    private native void setLogCollectionPrefix(long nativePtr, String prefix);
    @Override
    synchronized public void setLogCollectionPrefix(String prefix)  {
        
        setLogCollectionPrefix(nativePtr, prefix);
    }

    private native void setLogLevelMask(long nativePtr, int mask);
    @Override
    synchronized public void setLogLevelMask(int mask)  {
        
        setLogLevelMask(nativePtr, mask);
    }

    private native boolean tunnelAvailable(long nativePtr);
    @Override
    synchronized public boolean tunnelAvailable()  {
        
        return tunnelAvailable(nativePtr);
    }

    private native boolean upnpAvailable(long nativePtr);
    @Override
    synchronized public boolean upnpAvailable()  {
        
        return upnpAvailable(nativePtr);
    }

    private native boolean vcardSupported(long nativePtr);
    @Override
    synchronized public boolean vcardSupported()  {
        
        return vcardSupported(nativePtr);
    }

    private native void addListener(long nativePtr, CoreListener listener);
    @Override
    synchronized public void addListener(CoreListener listener)  {
        
        addListener(nativePtr, listener);
    }

    private native void removeListener(long nativePtr, CoreListener listener);
    @Override
    synchronized public void removeListener(CoreListener listener)  {
        
        removeListener(nativePtr, listener);
    }

    private native org.linphone.mediastream.Factory getMediastreamerFactory(long nativePtr);
    public org.linphone.mediastream.Factory getMediastreamerFactory() {
        return getMediastreamerFactory(nativePtr);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }
}
